#include "core.h"
#include <iostream>
#include <cassert>
#include <QCoreApplication>
#include <QSettings>
#include <QStringList>
#include <QTextStream>
#include <QDir>
#include <QThread>
#include <QMutex>
#include <gsl/gsl_errno.h>
#include "orbitan.h"
#include "orbitlib.h"
#include "schwarzschild.h"
#include "optimization.h"
#include "iodata.h"
#include "iosnapshot.h"
#include "iopotential.h"
#include "utils.h"

namespace smile{

/** default configuration parameters **/
const CConfigCore defaultConfigCore = {      // internal smile core/gui parameters (not related to any libsmile modules)
    50,   //double intTimeStepsPerPeriod;    // number of timesteps per unit time described above
    100,  //double intTimeInPeriods;         // orbit integration time measured in units of long-axis-orbit periods
    0,    //double intTimeMaxAdaptive;       // upper limits for adaptive integration time, expressed in energy-dependent orbital time periods
    1,    //double adaptiveTimeThreshold;    // parameter in Pfenniger's method for adaptive integration time
    false,//bool usePS;                      // whether to plot points on Poincare section
    // data for frequency map
    192,  //int fm_numOrbitsStationary;      // number of orbits in stationary start-space
    192,  //int fm_numOrbitsPrincipalPlane;  // number of orbits in principal-plane start-space
    0,    //int fm_numOrbitsYalpha;          // number of orbits in Y-alpha start space
    0,    //int fm_numOrbitsRandom;          // number of orbits in random(ergodic) start-space
    // data for Schwarzschild modelling
    1000, //int sm_numOrbits;
    24,   //int sm_numShells;                // grid size: number of radial shells
    2,    //int sm_linesPerSegment;          // grid size: number of lines dividing each of 3 segments of each shell into cells
    6,    //int sm_numAngularCoefs;          // order of angular spherical-harmonic expansion
    24,   //int sm_numRadialCoefs;           // number of radial coefs in BSE 
    0,    //double sm_Alpha;                 // Alpha-parameter defining family of basis functions in BSE model, 0=autodetect
    0,    //double sm_innerShellMass;        // fraction of mass in the innermost shell, 0 for default value of 1/(numShells+1)
    0,    //double sm_outerShellMass;        // fraction of mass enclosed by the outermost shell, 0 for default value of 1-1/(numShells+1)
    1e-3, //double chaoticMinFreqDiff
    0,    //double chaoticMinLambda
    0,    //double sm_chaoticPenalty
    0,    //int sm_numSamplingPoints;        // number of points sampled from trajectory, default NUM_SAMPLING_POINTS
    true, //bool sm_useBPMPD;                // whether to use BPMPD solver (if the executable is present), otherwise use GLPK
    CBasicSchwData::SD_DENS_CLASSIC,         // type of density constraints for Schwarzschild modelling
    0,    //sm_maxWeight;                    // upper bound on orbit weight
    0.01, //sm_regularization;               // parameter controlling the regularization of orbit weights in the quadratic optimization
    1,    //sm_constraintPenalty;            // penalty for constraint violation in the objective function in linear/quadratic optimization; 0 means infinity - must satisfy exactly
    false,//sm_constrainBeta;                // whether to constrain anisotropy
    0,0   //sm_betaIn, sm_betaOut;           // anisotropy coefficient requested (values for the inner and outer shells, linearly interpolated in between)
};

/** default parameters for orbit integration and classification **/
const CConfigOrbit defaultConfigOrbit = {
    1e-8,  // accuracyRel
    1e-8,  // accuracyAbs
    0.4,   // accTreecode
    false, // treecodeSymmetrizeTimestep
    1,     // calcLyapunovMethod
    true,  // freqStrictOrder
    0.98,  // minFreqX
    0.99   // minFreqYZ
};

/** default parameters for all possible potentials **/
const CConfigPotential defaultConfigPotential = {
    3,  // N_dim
    1,  // q
    1,  // p
    0,  // Mbh
    0,  // Rc
    1,  // Gamma
    1,  // Alpha
    0,  // Omega
    15, // Ncoefs_radial
    6,  // Ncoefs_angular
    -2, // nbpotEps
    0.5,// nbpotTol
    CDensity::PT_UNKNOWN, // PotentialType
    CDensity::PT_DEHNEN,  // DensityType
    CDensity::ST_TRIAXIAL,// SymmetryType
    1,  // splineSmoothFactor
    0,  // splineRadiusMin
    0   // splineRadiusMax
};

/// \cond INTERNAL_DOCS
QMutex mutexErrorInfo;  // mutex for ensuring that error line gets displayed at once
// change default GSL error handler to less nervous one
void gsl_verbose_error_handler_mutex (const char *reason, const char *file, int line, int gsl_errno)
{
    if(gsl_errno == GSL_ETOL || gsl_errno == GSL_EROUND/* || gsl_errno == GSL_EOVRFLW || gsl_errno == GSL_EUNDRFLW*/) return;  // do nothing
    mutexErrorInfo.lock();
    std::cerr << "GSL error " << gsl_errno << " in " << file << " at line " << line << ": " << reason << "\n";
    mutexErrorInfo.unlock();
} 
/// \endcond

CSmileCore::CSmileCore() :
  configCore(defaultConfigCore), 
  configPotential(defaultConfigPotential),
  configOrbit(defaultConfigOrbit)
{
    WorkDir = QDir::currentPath(); 
    AppDir  = getDirName(QCoreApplication::applicationFilePath());
    TempDir = "";
    potential=NULL;
    orbit=NULL;
    orbitlib=NULL;
    orbitIntRunning=false;
    orbitLibTimer=0;
    schwModel=NULL;
    useICe=false;
    gsl_set_error_handler(&gsl_verbose_error_handler_mutex);
    maxNumThread=std::max<int>(1, QThread::idealThreadCount());  // auto select
}

CSmileCore::~CSmileCore()
{   // commented out because the objects may be destroyed while running multi-threaded orbit library integration, 
    // which results in a crash. Probably there is a workaround but since the program is gonna close anyway, who cares.
#if 0
    if(orbit!=NULL)
    {
        delete orbit->getPotential();
        delete orbit;
    }
    if(model!=NULL)
        delete model;
    if(orbitlib!=NULL)
    {
        delete orbitlib->getPotential();
        delete orbitlib;
    }
    if(potential!=NULL)
        delete potential;
#endif
}

void CSmileCore::initPotential()
{
    if(potential) 
        delete potential;
    switch(configPotential.PotentialType)
    {
    case CDensity::PT_LOG:
        potential=new CPotentialLog(configPotential.N_dim, 
            configPotential.q, configPotential.p, configPotential.Mbh, configPotential.Rc);
        break;
    case CDensity::PT_HARMONIC:
        potential=new CPotentialHarmonic(configPotential.N_dim, 
            configPotential.q, configPotential.p, configPotential.Mbh);
        break;
    case CDensity::PT_DEHNEN:
        potential=new CPotentialDehnen(configPotential.N_dim, 
            configPotential.q, configPotential.p, configPotential.Mbh, configPotential.Gamma);
        break;
    case CDensity::PT_SCALEFREE:
        potential=new CPotentialScaleFree(configPotential.N_dim, 
            configPotential.q, configPotential.p, configPotential.Mbh, configPotential.Gamma);
        break;
    case CDensity::PT_SCALEFREESH:
        if(configPotential.DensityType==CDensity::PT_COEFS)
            initPotentialNB();   // load coefs from a text file
        else
            potential=new CPotentialScaleFreeSH(configPotential.N_dim, 
                configPotential.q, configPotential.p, configPotential.Mbh, configPotential.Gamma, 
                configPotential.Ncoefs_angular);
        break;
    case CDensity::PT_NB:
        initPotentialNB();
        break;
    case CDensity::PT_SPLINE:
    case CDensity::PT_BSE:   
        // two possible variants: either init from analytic density model or from Nbody/coef/massmodel file
        if( configPotential.DensityType==CDensity::PT_NB || 
            configPotential.DensityType==CDensity::PT_ELLIPSOIDAL || 
            configPotential.DensityType==CDensity::PT_MGE || 
            configPotential.DensityType==CDensity::PT_COEFS )
            initPotentialNB();
        else {   
            // create temporary instance of analytic density profile, 
            // pass it to the constructor to compute expansion coefficients
            CDensity* densModel=NULL;
            switch(configPotential.DensityType) 
            {
            case CDensity::PT_DEHNEN: 
                densModel = new CPotentialDehnen(3, configPotential.q, configPotential.p, 0, configPotential.Gamma); 
                break;
            case CDensity::PT_PLUMMER:
                densModel = new CDensityPlummer(configPotential.q, configPotential.p); 
                break;
            case CDensity::PT_ISOCHRONE:
                densModel = new CDensityIsochrone(configPotential.q, configPotential.p);
                break;
            case CDensity::PT_PERFECTELLIPSOID:
                densModel = new CDensityPerfectEllipsoid(configPotential.q, configPotential.p); 
                break;
            case CDensity::PT_NFW: {
                double concentration = std::max<double>(configPotential.Rc, 1.0);
                densModel = new CDensityNFW(configPotential.q, configPotential.p, concentration);
                break; }
            default: densModel=NULL;
            }
            if(densModel==NULL)  
            {  // error? init in some default way
                if(configPotential.PotentialType==CDensity::PT_BSE) 
                    potential = new CPotentialBSE(configPotential.N_dim, configPotential.Mbh, 1, 1, 1, CPointMassSet<double>()); 
                else potential = new CPotentialSpline(configPotential.N_dim, configPotential.Mbh, 1, 1, CPointMassSet<double>()); 
            }
            else
            {
                if(configPotential.PotentialType==CPotential::PT_BSE)
                    potential = new CPotentialBSE(configPotential.N_dim, configPotential.Mbh, 
                        configPotential.Alpha, configPotential.Ncoefs_radial, configPotential.Ncoefs_angular, densModel);
                else {
                    if(configPotential.splineRadiusMin>0 && configPotential.splineRadiusMax>0)
                    {
                        vectord grid;
                        createNonuniformGrid(configPotential.Ncoefs_radial+1, 
                            configPotential.splineRadiusMin, configPotential.splineRadiusMax, true, &grid);
                        potential = new CPotentialSpline(configPotential.N_dim, configPotential.Mbh, 
                            configPotential.Ncoefs_radial, configPotential.Ncoefs_angular, densModel, &grid);
                    } else  // no user-supplied grid extent -- will be assigned by defaule
                        potential = new CPotentialSpline(configPotential.N_dim, configPotential.Mbh, 
                            configPotential.Ncoefs_radial, configPotential.Ncoefs_angular, densModel);
                }
                delete densModel;
            }
        }
        break;
    default:
        initPotentialDefault();
    }
    configPotential.PotentialType = potential->PotentialType();
    configPotential.SymmetryType  = potential->symmetry();
    if(configPotential.DensityType==CPotential::PT_UNKNOWN)
        configPotential.DensityType=CPotential::PT_DEHNEN;
#ifdef DEBUGPRINT
    std::cerr << "Gamma="<<potential->getGamma() << ", total mass=" << potential->totalMass() << "\n";
#endif
}

void CSmileCore::initPotentialNB()
{
    if(NbodyFile=="") {
        my_message("Empty file name in initPotentialNB");
        initPotentialDefault();
        return;
    }
    QFileInfo NbodyFileName(QDir(WorkDir), NbodyFile);
    potential=readPotential(NbodyFileName.absoluteFilePath().toStdString(), &configPotential);
    if(!potential)
        initPotentialDefault();
}

void CSmileCore::initPotentialDefault()
{
    potential=new CPotentialLog(configPotential.N_dim, configPotential.q, configPotential.p, configPotential.Mbh, configPotential.Rc);
}

/* Export/import operations */
bool CSmileCore::exportPotential(const QString& fileName)
{
    if(potential==NULL) return false;
    vectord points[N_DIM];
    // export potential and its derivatives in a set of points: along three principal axes, three diagonals in each plane, and a x=y=z line
    const int Npoints=100;
    const double rmin=0.001, rmax=1000;
    vectord radii;
    createNonuniformGrid(Npoints, rmin, rmax, false, &radii);
    double axiscoefs[7][3] = { {1,0,0}, {0,1,0}, {0,0,1}, {0.7,0.7,0}, {0,0.7,0.7}, {0.7,0,0.7}, {0.58,0.58,0.58} };
    for(int d=0; d<7; d++)
    {
        if((potential->symmetry() & CPotential::ST_PLANESYM) == 0)   // no triaxial symmetry -- output both negative and positive radii
        {
            for(int i=Npoints-1; i>=0; i--)
            {
                double r=-radii[i];
                for(unsigned int c=0; c<N_DIM; c++)
                    points[c].push_back(r*axiscoefs[d][c]);
            }
        }
        for(int i=0; i<Npoints; i++)
        {
            double r=radii[i];
            for(unsigned int c=0; c<N_DIM; c++)
                points[c].push_back(r*axiscoefs[d][c]);
        }
    }
    if(!exportPotentialAtPoints(fileName, points, static_cast<int>(points[0].size())/7))
        return false;
    // export potential, forces and density at points defined as initial conditions for orbit library (if it exists)
    if(orbitlib && orbitlib->size()>0)
    {
        size_t count=orbitlib->size();
        for(unsigned int c=0; c<N_DIM; c++)
            points[c].assign(count, 0);
        for(size_t i=0; i<count; i++)
        {
            const COrbitDesc* od=orbitlib->getOrbitDesc(i);
            COrbitInitData<float> initData(od->getInitData());
            for(unsigned int c=0; c<N_DIM; c++)
                points[c][i]=initData.initCond.Pos[c];
        }
        if(!exportPotentialAtPoints(fileName+".points", points))
            return false;
    }
    // export coefficients if it makes sense
    if(potential->PotentialType()==CDensity::PT_BSE || potential->PotentialType()==CDensity::PT_SPLINE || 
        potential->PotentialType()==CDensity::PT_SCALEFREESH)
    {
        return writePotential((fileName+".coef"+
           (potential->PotentialType()==CDensity::PT_BSE?"_bse":
            potential->PotentialType()==CDensity::PT_SPLINE?"_spl":"")).toStdString(), potential);
    }
    else
        return true;
}

bool CSmileCore::exportPotentialAtPoints(const QString& fileName, const vectord points[N_DIM], int blockSize)
{
    QFile file(fileName);
    if (!file.open(QFile::WriteOnly | QFile::Text)) 
        return false;
    QTextStream out(&file);
    out.setRealNumberPrecision(8);
    out << "x\ty\tz\tPhi\tFx\tFy\tFz\tRho\n";
    for(size_t p=0; p<points[0].size(); p++)
    {
        double v[6]={points[0][p], points[1][p], points[2][p], 0, 0, 0};
        double f[8];
        double Phi = potential->Phi(v[0],v[1],v[2]);
        potential->DiffEq(N_DIM*2, 0, v, f);
        out << v[0] << "\t" << v[1] << "\t" << v[2] << "\t" << Phi << "\t" 
            << -f[3] << "\t" << -f[4] << "\t" << -f[5] << "\t" << potential->Rho(v[0],v[1],v[2]) << "\n";
        if(blockSize>0 && (p+1)%blockSize==0) out << "\n";  // separating line
    }
    return true;
}

bool CSmileCore::exportOrbit(const QString& fileName)
{
    if(orbit==NULL || orbit->getIntTime()==0) return false;
    CFileOutputText out(fileName.toStdString());
    out.writeOrbit(orbit);
    return out.ok();
}

bool CSmileCore::importOrbit(const QString &fileName)
{
    if(isOrbitIntRunning()) return false;
    CFileInputText in(fileName.toStdString());
    if(!in.ok()) return false;
    CPotential* newPot=potential->clone();
    COrbit* newOrb=in.readOrbit(&configOrbit, newPot);
    if(!newOrb)
    {
        delete newPot;
        return false;
    }
    if(orbit!=NULL)
    {
        delete orbit->getPotential();  // manually delete, since it was created by cloning externally to the object
        delete orbit;
    }
    if(newOrb->getIntTime()<=0 || newOrb->getInitData().timeStep<=0 || newOrb->getInitData().timeUnit<=0)  // something wrong with the data
    {
        delete newOrb;
        delete newPot;
        return false;
    }
    orbit=newOrb;
    InitData = orbit->getInitData();
    initCondE = potential->totalEnergy(InitData.initCond);
    configCore.intTimeInPeriods=orbit->getIntTime()/InitData.timeUnit; 
    configCore.intTimeStepsPerPeriod=InitData.timeUnit/InitData.timeStep;
    return in.ok();
}

bool CSmileCore::exportOrbitLib(const QString &fileName, bool withModelData)
{
    if(orbitlib==NULL || orbitlib->size()==0) return false;
    CFileOutputText out(fileName.toStdString());
    out.writeOrbitLibrary(orbitlib, withModelData);
    return out.ok();
}

bool CSmileCore::importOrbitLib(const QString &fileName, bool withModelData)
{
    assert(potential);
    CFileInputText in(fileName.toStdString());
    if(!in.ok()) return false;
    CPotential* newPot=potential->clone();
    COrbitLibrary* newOrbLib=in.readOrbitLibrary(&configOrbit, newPot, withModelData);
    if(!newOrbLib)
    {
        delete newPot;
        return false;
    }
    if(orbitlib!=NULL)
    {
        delete orbitlib->getPotential();  // manually delete, since it was created by cloning externally to the object
        delete orbitlib;
    }
    orbitlib=newOrbLib;
    if(withModelData) {
        initSchwData();
        CSchwModelQuadOpt model(orbitlib, schwData, NULL, NULL, 0, 0, 0, 0, 0, 0);
        emit signalInfo(QString::number(orbitlib->size())+" orbits loaded\n" + 
            QString::fromStdString(model.getStatistics()));
    }
    else emit signalInfo(QString::number(orbitlib->size())+" orbits loaded");
    return true;
}

bool CSmileCore::exportSchwModelStats(const QString& fileName)
{
    if(orbitlib==NULL || schwData.size()==0)
        return false;
    std::ofstream strm(fileName.toStdString().c_str(), std::ios::out);
    if(!strm) return false;
    CSchwModelQuadOpt model(orbitlib, schwData, NULL, NULL, 0, 0, 0, 0, 0, 0);    // create temp.model for computing constraint deviations
    size_t ind_constr=0;
    for(vectorSchwData::iterator iter=schwData.begin(); iter!=schwData.end(); iter++)
    {
        strm << "#Data_type: " << (*iter)->dataName() << "\n";
        const CSchwDataDensClassic* model_cl=NULL;
        const CSchwDataDensSH*      model_sph=NULL;
        const CSchwDataKinemShell*  model_kin=NULL;
        switch((*iter)->dataType()) {
            case CBasicSchwData::SD_DENS_CLASSIC:
                model_cl=static_cast<const CSchwDataDensClassic*>(*iter);
                strm << "#index\tRequiredCoef\tActualCoef\tDifference\tNormFactor\tradius\tx\ty\tz\n";
                break;
            case CBasicSchwData::SD_DENS_SHGRID:
                model_sph=static_cast<const CSchwDataDensSH*>(*iter); 
                strm << "#index\tRequiredCoef\tActualCoef\tDifference\tNormFactor\tradius\tl\tm\n";
                break;
            case CBasicSchwData::SD_DENS_SHBSE:
                model_sph=static_cast<const CSchwDataDensSH*>(*iter); 
                strm << "#index\tRequiredCoef\tActualCoef\tDifference\tNormFactor\tn\tl\tm\n";
                break;
            case CBasicSchwData::SD_KINEM_SHELL:
                strm << "#index\tradius\tM(r)\tsigma_r^2\tsigma_t^2\n";
                model_kin=static_cast<const CSchwDataKinemShell*>(*iter); 
                break;
            default: ;
        }
        if(!model_kin) {
            for(size_t i=0; i<(*iter)->numConstraints(); i++)
            {
                double coefVal =(*iter)->getConstraint(i);
                double coefNorm=(*iter)->getConstraintNormFactor(i);
                double coefDev =model.isCorrect() ? model.calcCoefDeviation(i+ind_constr)*coefNorm : gsl_nan();
                strm << i << "\t" << coefVal << "\t" << (coefDev+coefVal) << "\t" << coefDev << "\t" << coefNorm;
                if(model_cl)
                {
                    double X, Y, Z;
                    model_cl->cellCenter(i, &X, &Y, &Z);
                    strm << "\t" << sqrt(X*X+Y*Y+Z*Z) << "\t" << X << "\t" << Y << "\t" << Z;
                }
                else if(model_sph)
                {
                    double n;
                    int l, m;
                    model_sph->getCoefDecomposition(i, &n, &l, &m);
                    strm << "\t" << n << "\t" << l << "\t" << m;
                }
                strm << "\n";
            }
        }
        else //model_kin
        {
            size_t numShellsKin=model_kin->numConstraints();
            vectord 
                shellVr(numShellsKin, 0), 
                shellVt(numShellsKin, 0), 
                shellTime(numShellsKin, 0);

            if(model.isCorrect())  for(size_t o=0; o<orbitlib->size(); o++)
            {
                // load kinematic information data from orbit desc
                for(size_t inf=0; inf<orbitlib->getOrbitDesc(o)->getInfoNum(); inf++)
                {
                    const CBasicInformation* info=orbitlib->getOrbitDesc(o)->getInfo(inf);
                    if(info->infoType()==CBasicInformation::IT_SCHW &&
                        static_cast<const CSchwInformation*>(info)->schwDataType()==CBasicSchwData::SD_KINEM_SHELL)
                    {
                        const CSchwInformation* infoKin=static_cast<const CSchwInformation*>(info);
                        double weight=orbitlib->getOrbitDesc(o)->getWeight();
                        for(size_t s=0; s<numShellsKin; s++)
                        {
                            double shTime = infoKin->getSchwData(s) * weight;
                            shellVr[s] += shTime * infoKin->getSchwData(s+1*numShellsKin);
                            shellVt[s] += shTime * infoKin->getSchwData(s+2*numShellsKin);
                            shellTime[s] += shTime;
                        }
                    }
                }
            }
            for(size_t s=0; s<numShellsKin; s++)
            {
                strm << s << "\t" << model_kin->getShellRadius(s) << "\t"
                    << orbitlib->getPotential()->Mass(model_kin->getShellRadius(s)) << "\t" 
                    << (shellVr[s]/shellTime[s]) << "\t" << (shellVt[s]/shellTime[s]) << "\n";
            }
        }
        strm << "\n";
    }
    return static_cast<bool>(strm);
}

/* ----------- orbit integration ------------ */
void CSmileCore::initIC()
{
    if(useICe)   // assign from initCondE
    {
        InitData.initCond=CPosVelPoint<double>();
        InitData.initCond.Pos[0]=potential->longaxisradius(initCondE);
    }
    else
    {
        initCondE = potential->totalEnergy(InitData.initCond);
    }
    // calculate T_orb (for x-axis orbit)
    InitData.timeUnit = potential->longaxisperiod(initCondE);
}

void CSmileCore::startOrbit()
{
    if(InitData.timeUnit<=0 || configCore.intTimeInPeriods<=0 || configCore.intTimeStepsPerPeriod<=0) return;  // error
    if(isOrbitIntRunning()) return;   // can't invoke this while another orbit is running
    if(orbit!=NULL)  // this one left from the previous orbitintegration (is not destroyed after it finishes)
    {
        delete orbit->getPotential();  // manually delete, since it was created by cloning externally to the object
        delete orbit;
    }
    vectorRuntimeFncCreators crts;   // this array of generators is different from core->creatorsLib -- 
    // the latter is for all orbits in orbit library, the former is for a single orbit independent of the library
    crts.push_back(new COrbitRuntimeTrajectoryCreator());
    if(configCore.usePS)
        crts.push_back(new COrbitRuntimePoincareCreator(0, 1));
    crts.push_back(new COrbitRuntimePericenterCreator());
    orbit = new COrbit(&configOrbit, COrbitInitData<double>(
        potential->clone(), InitData.timeUnit/configCore.intTimeStepsPerPeriod, 
        InitData.timeUnit, InitData.initCond, InitData.calcLyapunov), &crts);
    // create a worker object for doing orbit integration
    COrbitIntWorker *wkr = new COrbitIntWorker(orbit, configCore.intTimeInPeriods*InitData.timeUnit);
    // create a separate thread to perform the job
    QThread* thr = new QThread();
    // move the worker object to the new thread
    wkr->moveToThread(thr);
    // when thread starts, launch the worker object's doIntegration method
    connect(thr, SIGNAL(started()), wkr, SLOT(doIntegration()));
    // when orbit integration finishes, notify about it
    connect(wkr, SIGNAL(calcFinished()), this, SLOT(coreOrbitFinished()));
    // when worker object finishes, the thread that hosted it should be deleted
    connect(wkr, SIGNAL(calcFinished()), thr, SLOT(quit()));
    // when the thread has finished, it will be deleted shortly thereafter
    connect(thr, SIGNAL(finished()), thr, SLOT(deleteLater()));
    // now start the thread, no need to care about its proper deletion hereafter
    thr->start();
    orbitIntRunning=true;
}

void CSmileCore::coreOrbitFinished()
{
    orbitIntRunning=false;
    emit signalOrbitFinished();
}

/* ----------- Frequency map / Schwarzschild modelling ------------- */

void CSmileCore::startOrbitLibrary(const char* finishSignal)
{
    for(int t=0; t<maxNumThread; t++)
    {
        // create a worker object for doing orbit integration in parallel
        COrbitLibWorker *wkr = new COrbitLibWorker(orbitlib);
        // create a separate thread to perform the job
        QThread* thr = new QThread();
        // move the worker object to the new thread
        wkr->moveToThread(thr);
        // when thread starts, launch the worker object's run method
        connect(thr, SIGNAL(started()), wkr, SLOT(run()));
        // if termination of orbit library integration is requested, inform the worker object about it
        connect(this, SIGNAL(signalStopOrbitLibrary()), wkr, SLOT(halt()), Qt::DirectConnection);
        // when the integration of entire orbit library finishes, notify about it with the provided signal
        connect(wkr, SIGNAL(calcFinished()), this, finishSignal);
        // when worker object finishes, the thread that hosted it should be deleted
        connect(wkr, SIGNAL(calcFinished()), thr, SLOT(quit()));
        // when the thread has finished, it will be deleted shortly thereafter
        connect(thr, SIGNAL(finished()), thr, SLOT(deleteLater()));
        // now start the thread, no need to care about its proper deletion hereafter
        thr->start(QThread::LowPriority);
    }
    emit signalInfo("Starting integration..");
    orbitLibTimer = startTimer(2000);  // to refresh progress and to mark that orbit library integration is in progress
}

void CSmileCore::coreOrbitLibraryFinished()
{
    killTimer(orbitLibTimer);
    orbitLibTimer=0;  // stop showing progress indicator
    // delete non-used orbits (if integration threads were terminated prematurely, there exist some)
    orbitlib->removeUnused();
    // finally, send signal to the owner object
    emit signalOrbitLibraryFinished();
}

void CSmileCore::timerEvent(QTimerEvent *)
{
    if(orbitLibTimer==0)
        return;           // shouldn't happen - this routine is invoked only during orbit library integration
    emit signalTimer();   // triggered regularly while orbit library integration is running
}

void CSmileCore::buildFreqMapIC(double energy, bool useExistingIC)
{
    assert(potential);
    schwData.clear();   // no schw model is used in frequency map
    for(vectorRuntimeFncCreators::const_iterator iter=creatorsLib.begin(); iter!=creatorsLib.end(); ++iter)
        delete (*iter);
    creatorsLib.clear();
    creatorsLib.push_back(new COrbitRuntimeTrajectoryCreator());
    creatorsLib.push_back(new COrbitRuntimePericenterCreator());
    if(useExistingIC && orbitlib!=NULL && orbitlib->size()>0)
    {
        delete orbitlib->getPotential();  // manually delete, since it was created by cloning externally to the object
        orbitlib->modifyInitialConditions(potential->clone(), &creatorsLib, 
            configCore.intTimeStepsPerPeriod, configCore.intTimeInPeriods, 
            configCore.intTimeMaxAdaptive, InitData.calcLyapunov, true);
    }
    else
    {
        if(orbitlib)
        {
            delete orbitlib->getPotential();  // manually delete, since it was created by cloning externally to the object
            delete orbitlib;
        }
        orbitlib=new COrbitLibrary(&configOrbit, potential->clone(), &creatorsLib);
        emit signalInfo("Preparing initial conditions..");
        orbitlib->prepareInitialConditionsE(
            configCore.fm_numOrbitsStationary, configCore.fm_numOrbitsPrincipalPlane, 
            configCore.fm_numOrbitsYalpha, configCore.fm_numOrbitsRandom, 
            energy, configCore.intTimeStepsPerPeriod, configCore.intTimeInPeriods, 
            configCore.intTimeMaxAdaptive, InitData.calcLyapunov);
    }
}

void CSmileCore::initSchwData()
{
    for(vectorSchwData::iterator iter=schwData.begin(); iter!=schwData.end(); iter++)
        delete (*iter);
    schwData.clear();
    assert(potential->N_dim==N_DIM);
    emit signalInfo("Creating Schwarzschild model..");
    CBasicSchwData* dataDens=NULL;
    switch(configCore.sm_densDataType) {
        case CBasicSchwData::SD_DENS_CLASSIC:
            dataDens=new CSchwDataDensClassic(potential, configCore.sm_numShells, 
                configCore.sm_innerShellMass, configCore.sm_outerShellMass, configCore.sm_linesPerSegment);
            break;
        case CBasicSchwData::SD_DENS_SHGRID:
            dataDens=new CSchwDataDensSHGrid(potential, configCore.sm_numShells, 
                configCore.sm_innerShellMass, configCore.sm_outerShellMass, configCore.sm_numAngularCoefs);
            break;
        case CBasicSchwData::SD_DENS_SHBSE:
            dataDens=new CSchwDataDensSHBSE(potential, 
                configCore.sm_numAngularCoefs, configCore.sm_numRadialCoefs, configCore.sm_Alpha);
            break;
        default: my_error("Unrecognized type of Schwarzschild model"); dataDens=NULL;
    }
    if(dataDens!=NULL) schwData.push_back(dataDens);
    schwData.push_back(new CSchwDataKinemShell(potential, configCore.sm_numShells, 
        configCore.sm_innerShellMass, configCore.sm_outerShellMass));
}

void CSmileCore::buildSchwIC(bool useExistingIC, CPointCountSetFloat* orbitsForReintegrate)
{
    assert(potential);
    if(potential->N_dim!=3) {
        my_error("Schwarzschild modelling is only possible in three dimensions");
        return;
    }
    // Prepare initial conditions
    if(useExistingIC && orbitlib!=NULL && orbitlib->size()>0)
    {
        delete orbitlib->getPotential();  // manually delete, since it was created by cloning externally to the object
        orbitlib->modifyInitialConditions(potential->clone(), &creatorsLib, 
            configCore.intTimeStepsPerPeriod, configCore.intTimeInPeriods, 
            configCore.intTimeMaxAdaptive, InitData.calcLyapunov, orbitsForReintegrate==NULL);
    }
    else
    {
        if(orbitlib)
        {
            delete orbitlib->getPotential();  // manually delete, since it was created by cloning externally to the object
            delete orbitlib;
        }
        orbitlib=new COrbitLibrary(&configOrbit, potential->clone(), &creatorsLib);
        emit signalInfo("Preparing initial conditions..");
        //double maxRadiusMass=configCore.sm_outerShellMass>0 && configCore.sm_outerShellMass<1 ? 1-0.25*(1-configCore.sm_outerShellMass) : 1-0.25/configCore.sm_numShells;
        orbitlib->prepareInitialConditions(
            configCore.sm_numOrbits, 
            0/*potential->getRadiusByMass(maxRadiusMass*potential->totalMass())*/, 
            configCore.intTimeStepsPerPeriod, 
            configCore.intTimeInPeriods, 
            configCore.intTimeMaxAdaptive, 
            InitData.calcLyapunov);
    }
    if(orbitsForReintegrate==NULL || schwData.size()==0)
    {   // need to create data objects 
        initSchwData();
    }
    // fill the list of runtime function creators
    for(vectorRuntimeFncCreators::const_iterator iter=creatorsLib.begin(); iter!=creatorsLib.end(); ++iter)
        delete (*iter);
    creatorsLib.clear();
    creatorsLib.push_back(new COrbitRuntimeTrajectoryCreator());
    creatorsLib.push_back(new COrbitRuntimePericenterCreator());
    if(configCore.sm_numSamplingPoints>0 || orbitsForReintegrate!=NULL) 
    {
        creatorsLib.push_back(new COrbitRuntimeTrajSampleCreator(
            configCore.sm_numSamplingPoints, configCore.intTimeInPeriods, orbitsForReintegrate));
    }
    for(vectorSchwData::iterator iter=schwData.begin(); iter!=schwData.end(); iter++)
    {
        const CBasicOrbitRuntimeFncCreator* crt=createSchwCreator(*iter);
        if(crt!=NULL) 
            creatorsLib.push_back(crt);
    }
}

void CSmileCore::schwStartOptimization(bool quadratic)
{
    if(isSchwOptRunning())
    {
        my_error("Schwarzschild optimization is already running");
        return;      // do not allow to run two threads at once
    }
    if(!orbitlib || orbitlib->size()==0 || schwData.size()==0)
    {
        my_error("No Schwarzschild model defined");
        return;
    }
    // create quad.solver object and orbit penalty fnc if necessary
    const CBasicQuadraticOptimizationSolver<smNumType>* solver = NULL;
    if(configCore.sm_useBPMPD) 
        solver=new COptimizationSolverBPMPD<smNumType>(AppDir.toStdString()+"bpmpd", TempDir.toStdString()); 
    else {
#ifdef HAVE_GLPK
        if(!quadratic)
            solver=new COptimizationSolverGLPK<smNumType>();
#endif
#ifdef HAVE_CVXOPT
        if(solver==NULL)
            solver=new COptimizationSolverCVXOPT<smNumType>();
#endif
    }
    if(solver==NULL) {
        my_error(std::string(quadratic ? "Quadratic" : "Linear") + " optimization solver not available!");
        return;
    }
    const CBasicOrbitFilteringFnc* orbitPenaltyFnc = 
        configCore.sm_chaoticPenalty!=0 ? 
        new CChaosOrbitFilteringFnc(configCore.sm_chaoticPenalty, configCore.chaoticMinLambda, configCore.chaoticMinFreqDiff) : NULL;
    // create the Schw.quad.opt.model class which will be run within the context of a separate thread
    schwModel=new CSchwModelQuadOpt(
        orbitlib, schwData, solver, orbitPenaltyFnc, 
        quadratic ? configCore.sm_regularization : 0, configCore.sm_constraintPenalty, 
        configCore.sm_maxWeight, configCore.sm_constrainBeta, configCore.sm_betaIn, configCore.sm_betaOut);
    // create a worker object for doing optimization
    CSchwarzschildWorker* wkr = new CSchwarzschildWorker(schwModel);
    // create a separate thread to perform the job
    QThread* thr = new QThread();
    // move the worker object to the new thread
    wkr->moveToThread(thr);
    // when thread starts, launch the worker object's doWork method
    connect(thr, SIGNAL(started()), wkr, SLOT(doWork()));
    // when optimization finishes, display results
    connect(wkr, SIGNAL(message(const QString&)), this, SLOT(coreSchwOptimizationFinished(const QString&)));
    // when worker object finishes, the thread that hosted it should be deleted
    connect(wkr, SIGNAL(workDone()), thr, SLOT(quit()));
    // when the thread has finished, it will be deleted shortly thereafter
    connect(thr, SIGNAL(finished()), thr, SLOT(deleteLater()));
    // now start the thread, yeah!!
    thr->start();
}

void CSmileCore::coreSchwOptimizationFinished(const QString& result)
{
    if(schwModel!=NULL) delete schwModel; // associated solver and penalty fnc will be deleted automatically
    schwModel=NULL;
    emit signalSchwOptimizationFinished(result);
}

void CSmileCore::schwExportNbody(int numPoints, const QString &fileName, const QString &fileFormat, int numBinsRefine)
{
    if(orbitlib==NULL || orbitlib->size()==0)
        return;  // nothing to share...
    if(fileName.isEmpty()) {
        my_error("Filename cannot be empty!");
        return;
    }
    if(isExportNbodyRunning()) {
        my_error("Error, export to Nbody is already in progress");
        return;
    }
    numPointsNbodyExport=numPoints;
    fileNameNbodyExport=fileName;     // non-empty filename means that export is in progress
    fileFormatNbodyExport=fileFormat;
    numBinsRefineNbodyExport=numBinsRefine;
    firstCallNbodyExport=true;
    coreNbodyExport();
}

void CSmileCore::coreNbodyExport()
{
    if(fileNameNbodyExport.isEmpty()) 
        return;  // incorrect call, shouldn't happen
    if(orbitLibTimer!=0) {  // called after reintegration was finished (or terminated by user)
        killTimer(orbitLibTimer); // to stop showing up progress bar after reintegration is complete
        orbitLibTimer=0;
        emit signalNbodyReintegrationFinished();
    }
    const CEnergyOrbitFilteringFnc rankFnc;
    const CMassRefinementFnc massRefineFnc(orbitlib, numBinsRefineNbodyExport, &rankFnc);
    CPointMassSetFloat* points = NULL;
    CPointCountSetFloat* orbitsForReintegrate = NULL;
    std::string result=orbitlib->exportNbody(
        numPointsNbodyExport, 
        numBinsRefineNbodyExport ? &massRefineFnc : NULL, 
        &points, &orbitsForReintegrate);
    QString info = QString::fromStdString(result);
    if(points==NULL)
    {
        coreNbodyExportFinished("Error in creating Nbody model: "+info);
        return;
    }
    if(orbitsForReintegrate!=NULL && !orbitsForReintegrate->empty())
    {   // requires reintegration
        if(!firstCallNbodyExport)
        {   // for some reason the reintegration still did not create enough sampling points
            coreNbodyExportFinished("Export of Nbody model cancelled or was unsuccessful");
            delete points;
            delete orbitsForReintegrate;
            return;
        }
        firstCallNbodyExport=false;
        buildSchwIC(true, orbitsForReintegrate);
        startOrbitLibrary(SLOT(coreNbodyExport()));   // will be called once again
        emit signalInfo("Reintegrating "+
            QString::number(static_cast<unsigned int>(orbitsForReintegrate->size()))+" orbits..");
        delete points;
        delete orbitsForReintegrate;  // a working copy was made by TrajSampleCreator
        return;
    }
    emit signalInfo("Exporting Nbody model..         ");
    bool fileExportResult=true;
    CBasicIOSnapshot* snap = createIOSnapshotWrite(fileFormatNbodyExport.toStdString(), fileNameNbodyExport.toStdString(), 
        "SMILE version="VERSION" potential="+potential->PotentialName()+
        " schwmodel="+getSchwDataNameByType(configCore.sm_densDataType)+
        " numorbits="+QString::number(static_cast<unsigned int>(orbitlib->size())).toStdString());
    if(snap) { 
        fileExportResult &= snap->writeSnapshot(points);
        delete snap;
    } else {
        my_error("Invalid export file format: "+fileFormatNbodyExport.toStdString());
        fileExportResult=false;
    }
    if(!fileExportResult) {
        coreNbodyExportFinished("Error writing file "+fileNameNbodyExport);
        delete points;
        return;
    }
    // start a helper thread with a worker object to compute virial ratio
    emit signalInfo(info+"\nComputing virial ratio...");
    // points will be deleted when the worker object is destroyed
    CNbodyExportStatWorker* wkr = new CNbodyExportStatWorker(points, info);
    // create a separate thread to perform the job
    QThread* thr = new QThread();
    // move the worker object to the new thread
    wkr->moveToThread(thr);
    // when thread starts, launch the worker object's doWork method
    connect(thr, SIGNAL(started()), wkr, SLOT(doWork()));
    // to terminate the thread upon request
    connect(this, SIGNAL(signalStopExportNbody()), wkr, SLOT(halt()), Qt::DirectConnection);
    // when computation is done, display results and reset the filename
    connect(wkr, SIGNAL(exportFinished(const QString&)), this, SLOT(coreNbodyExportFinished(const QString&)));
    // when worker object finishes, the thread that hosted it should be deleted
    connect(wkr, SIGNAL(workDone()), thr, SLOT(quit()));
    // when the thread has finished, it will be deleted shortly thereafter
    connect(thr, SIGNAL(finished()), thr, SLOT(deleteLater()));
    // now start the thread
    thr->start();
}

void CSmileCore::coreNbodyExportFinished(const QString &result)
{
    fileNameNbodyExport.clear();   // clear internal flag that means the export is in progress
    emit signalNbodyExportFinished(result);
}

/* ----------- load/save settings ------------- */

bool CSmileCore::loadSettings(QString fileName)
{
    if(fileName=="") 
        fileName=AppDir+DEFAULT_SETTINGS_FILE_NAME;
    else if(!QFile::exists(fileName))
        return false;
    QSettings settings(fileName, QSettings::IniFormat);
    QString WorkDir1 = settings.value("Common/WorkDir", "").toString();
    if(WorkDir1!="")
    {
        QFileInfo fi(WorkDir1);
        if(fi.isRelative()) WorkDir = AppDir+WorkDir1; else WorkDir = WorkDir1;
    }
    QString TempDir1 = settings.value("Common/TempDir", "").toString();
    if(TempDir1!="")
    {
        QFileInfo fi(TempDir1);
        if(fi.isRelative()) TempDir = AppDir+TempDir1; else TempDir = TempDir1;
    } 
    else if(TempDir=="") TempDir=WorkDir;
    if(!WorkDir.endsWith('/')) WorkDir += '/';
    if(!TempDir.endsWith('/')) TempDir += '/';
    if(!QFile::exists(WorkDir)) { 
        my_error("Warning, working directory "+WorkDir.toStdString()+" doesn't exist!");
        WorkDir=AppDir;
    }
    if(!QFile::exists(TempDir)) { 
        my_error("Warning, temp directory "+TempDir.toStdString()+" doesn't exist!");
        TempDir=AppDir;
    }
    int newMaxNumThread=settings.value("Common/MaxThreads", "0").toInt();
    if(newMaxNumThread>=1) maxNumThread=newMaxNumThread;   // if no value exists in ini file, leave current
    configCore.sm_useBPMPD = settings.value("Common/useBPMPD", configCore.sm_useBPMPD).toBool();
    if(!QFile::exists(AppDir+"bpmpd.exe")) configCore.sm_useBPMPD=false;  // if BPMPD.exe does not exist, resort to GLPK or CVXOPT
    // potential section
    configPotential.PotentialType = getPotentialTypeByName(settings.value("Potential/Type", "").toString().toStdString());
    configPotential.DensityType = getDensityTypeByName(settings.value("Potential/DensityModel", "").toString().toStdString());
    configPotential.N_dim = std::max<int>(2, std::min<int>(3, settings.value("Potential/N_dim", N_DIM).toInt()));
    configPotential.q = std::max<double>(0, std::min<double>(1, settings.value("Potential/q_YtoX", configPotential.q).toDouble())); 
    configPotential.p = std::max<double>(0, std::min<double>(1, settings.value("Potential/p_ZtoX", configPotential.p).toDouble())); 
    configPotential.Mbh = std::max<double>(0, settings.value("Potential/Mbh", configPotential.Mbh).toDouble());
    configPotential.Rc = std::max<double>(0, settings.value("Potential/Rc", configPotential.Rc).toDouble());
    configPotential.Gamma = std::max<double>(0, std::min<double>(2, settings.value("Potential/Gamma", configPotential.Gamma).toDouble()));
    // allowed alpha range is [0.5 - infinity], alpha=0 means try to autodetect..
    configPotential.Alpha = std::max<double>(0, settings.value("Potential/Alpha", configPotential.Alpha).toDouble());
    configPotential.Omega = settings.value("Potential/Omega", configPotential.Omega).toDouble();
    if(configPotential.Omega<0) configPotential.Omega=-configPotential.Omega;
    configPotential.Ncoefs_radial = std::min<size_t>(MAX_NCOEFS_RADIAL, settings.value("Potential/Ncoefs_radial", 
        static_cast<unsigned int>(configPotential.Ncoefs_radial)).toInt());
    configPotential.Ncoefs_angular = std::min<size_t>(MAX_NCOEFS_ANGULAR, settings.value("Potential/Ncoefs_angular", 
        static_cast<unsigned int>(configPotential.Ncoefs_angular)).toInt());
    NbodyFile  = settings.value("Potential/NbodyFile").toString();
    configPotential.SymmetryType = getSymmetryTypeByName(settings.value("Potential/Symmetry", "").toString().toStdString());
    configPotential.nbpotEps   = settings.value("Potential/treecodeEps", configPotential.nbpotEps).toDouble();
    configPotential.nbpotTol   = std::min<double>(1, std::max<double>(0, settings.value("Potential/treecodeTheta", configPotential.nbpotTol).toDouble()));
    configPotential.splineSmoothFactor = std::max<double>(0, settings.value("Potential/splineSmoothFactor", configPotential.splineSmoothFactor).toDouble());
    configPotential.splineRadiusMin = settings.value("Potential/splineRadiusMin", 0).toDouble();
    configPotential.splineRadiusMax = settings.value("Potential/splineRadiusMax", 0).toDouble();
    initPotential();
    // orbit integration section
    configOrbit.accuracyRel= settings.value("Potential/accuracyRelative", configOrbit.accuracyRel).toDouble();
    configOrbit.accuracyAbs= settings.value("Potential/accuracyAbsolute", configOrbit.accuracyAbs).toDouble();
    configOrbit.accTreecode= settings.value("Potential/accuracyTreecode", configOrbit.accTreecode).toDouble();
    configOrbit.treecodeSymmetrizeTimestep= settings.value("Potential/treecodeSymmetrizeTimestep", configOrbit.treecodeSymmetrizeTimestep).toBool();
    double icX = settings.value("Orbit/x", 1).toDouble();
    double icY = settings.value("Orbit/y", 0).toDouble();
    double icZ = settings.value("Orbit/z", 0).toDouble();
    double icVx = settings.value("Orbit/vx", 0).toDouble();
    double icVy = settings.value("Orbit/vy", 0).toDouble();
    double icVz = settings.value("Orbit/vz", 0).toDouble();
    initCondE = settings.value("Orbit/E", 0).toDouble();
    useICe = settings.value("Orbit/useE", false).toBool();
    InitData.initCond=CPosVelPoint<double>(icX, icY, icZ, icVx, icVy, icVz);
    InitData.calcLyapunov = settings.value("Orbit/calcLyapunov", false).toBool();
    configOrbit.calcLyapunovMethod = settings.value("Orbit/calcLyapunovMethod", configOrbit.calcLyapunovMethod).toInt();
    configCore.intTimeInPeriods = settings.value("Orbit/intTime", configCore.intTimeInPeriods).toDouble();
    configCore.intTimeStepsPerPeriod = settings.value("Orbit/stepsPerPeriod", configCore.intTimeStepsPerPeriod).toDouble();
    configCore.intTimeMaxAdaptive = settings.value("Orbit/intTimeMax", configCore.intTimeMaxAdaptive).toDouble();
    configCore.adaptiveTimeThreshold = settings.value("Orbit/adaptiveTimeThreshold", configCore.adaptiveTimeThreshold).toDouble();
    configCore.usePS = settings.value("Orbit/usePS", configCore.usePS).toBool();
    // frequency map
    configCore.fm_numOrbitsStationary = settings.value("Frequency_map/numOrbitsStationary", configCore.fm_numOrbitsStationary).toInt();
    configCore.fm_numOrbitsPrincipalPlane = settings.value("Frequency_map/numOrbitsPrincipalPlane", configCore.fm_numOrbitsPrincipalPlane).toInt();
    configCore.fm_numOrbitsYalpha = settings.value("Frequency_map/numOrbitsYalpha", configCore.fm_numOrbitsYalpha).toInt();
    configCore.fm_numOrbitsRandom = settings.value("Frequency_map/numOrbitsRandom", configCore.fm_numOrbitsRandom).toInt();
    // schwarzschild modelling
    configCore.sm_densDataType = getSchwDataTypeByName(settings.value("Schwarzschild_model/densityModelType", "").toString().toStdString());
    configCore.sm_numOrbits = settings.value("Schwarzschild_model/numOrbits", configCore.sm_numOrbits).toInt();
    configCore.sm_linesPerSegment = settings.value("Schwarzschild_model/linesPerSegment", configCore.sm_linesPerSegment).toInt();
    configCore.sm_numAngularCoefs = settings.value("Schwarzschild_model/numAngularCoefs", configCore.sm_numAngularCoefs).toInt();
    configCore.sm_numRadialCoefs = settings.value("Schwarzschild_model/numRadialCoefs", configCore.sm_numRadialCoefs).toInt();
    configCore.sm_Alpha = settings.value("Schwarzschild_model/smAlpha", configCore.sm_Alpha).toDouble();
    configCore.sm_numShells = settings.value("Schwarzschild_model/numShells", configCore.sm_numShells).toInt();
    configCore.sm_innerShellMass = settings.value("Schwarzschild_model/innerShellMass", configCore.sm_innerShellMass).toDouble();
    configCore.sm_outerShellMass = settings.value("Schwarzschild_model/outerShellMass", configCore.sm_outerShellMass).toDouble();
    configCore.sm_numSamplingPoints = settings.value("Schwarzschild_model/numSamplingPoints", configCore.sm_numSamplingPoints).toInt();
    configCore.chaoticMinFreqDiff = settings.value("Schwarzschild_model/chaoticMinFreqDiff", configCore.chaoticMinFreqDiff).toDouble();
    configCore.chaoticMinLambda = settings.value("Schwarzschild_model/chaoticMinLambda", configCore.chaoticMinLambda).toDouble();
    configCore.sm_chaoticPenalty = settings.value("Schwarzschild_model/chaoticWeightFactor", configCore.sm_chaoticPenalty).toDouble();
    configCore.sm_maxWeight = settings.value("Schwarzschild_model/maxWeight", configCore.sm_maxWeight).toDouble();
    configCore.sm_regularization = settings.value("Schwarzschild_model/regularization", configCore.sm_regularization).toDouble();
    configCore.sm_constraintPenalty = settings.value("Schwarzschild_model/constraintPenalty", configCore.sm_constraintPenalty).toDouble();
    configCore.sm_constrainBeta = settings.value("Schwarzschild_model/constrainBeta", configCore.sm_constrainBeta).toBool();
    configCore.sm_betaIn = settings.value("Schwarzschild_model/betaIn", configCore.sm_betaIn).toDouble();
    configCore.sm_betaOut= settings.value("Schwarzschild_model/betaOut",configCore.sm_betaOut).toDouble();

    return true;
}

void CSmileCore::saveSettings(QString fileName)
{
    if(fileName=="") fileName=AppDir+DEFAULT_SETTINGS_FILE_NAME;
    QSettings settings(fileName, QSettings::IniFormat);
    // potential section
    settings.setValue("Potential/Type", QString::fromStdString(PotentialNames[configPotential.PotentialType]));
    settings.setValue("Potential/DensityModel", QString::fromStdString(DensityNames[configPotential.DensityType]));
    settings.setValue("Potential/N_dim", configPotential.N_dim);
    settings.setValue("Potential/q_YtoX", configPotential.q);
    settings.setValue("Potential/p_ZtoX", configPotential.p);
    settings.setValue("Potential/Mbh", configPotential.Mbh);
    settings.setValue("Potential/Rc", configPotential.Rc);
    settings.setValue("Potential/Gamma", configPotential.Gamma);
    settings.setValue("Potential/Alpha", configPotential.Alpha);
    settings.setValue("Potential/Omega", configPotential.Omega);
    settings.setValue("Potential/Ncoefs_radial",  static_cast<unsigned int>(configPotential.Ncoefs_radial));
    settings.setValue("Potential/Ncoefs_angular", static_cast<unsigned int>(configPotential.Ncoefs_angular));
    settings.setValue("Potential/Symmetry", QString::fromStdString(SymmetryNames[configPotential.SymmetryType]));
    settings.setValue("Potential/NbodyFile", NbodyFile);
    settings.setValue("Potential/treecodeEps", configPotential.nbpotEps);
    settings.setValue("Potential/treecodeTheta", configPotential.nbpotTol);
    settings.setValue("Potential/accuracyRelative", configOrbit.accuracyRel);
    settings.setValue("Potential/accuracyAbsolute", configOrbit.accuracyAbs);
    settings.setValue("Potential/accuracyTreecode", configOrbit.accTreecode);
    settings.setValue("Potential/treecodeSymmetrizeTimestep", configOrbit.treecodeSymmetrizeTimestep);
    settings.setValue("Potential/splineSmoothFactor", configPotential.splineSmoothFactor);
    settings.setValue("Potential/splineRadiusMin", configPotential.splineRadiusMin);
    settings.setValue("Potential/splineRadiusMax", configPotential.splineRadiusMax);
    // orbit integration section
    settings.setValue("Orbit/x", InitData.initCond.Pos[0]);
    settings.setValue("Orbit/y", InitData.initCond.Pos[1]);
    settings.setValue("Orbit/z", InitData.initCond.Pos[2]);
    settings.setValue("Orbit/vx", InitData.initCond.Vel[0]);
    settings.setValue("Orbit/vy", InitData.initCond.Vel[1]);
    settings.setValue("Orbit/vz", InitData.initCond.Vel[2]);
    settings.setValue("Orbit/E", initCondE);
    settings.setValue("Orbit/useE", useICe);
    settings.setValue("Orbit/intTime", configCore.intTimeInPeriods);
    settings.setValue("Orbit/stepsPerPeriod", configCore.intTimeStepsPerPeriod);
    settings.setValue("Orbit/intTimeMax", configCore.intTimeMaxAdaptive);
    settings.setValue("Orbit/adaptiveTimeThreshold", configCore.adaptiveTimeThreshold);
    settings.setValue("Orbit/calcLyapunov", InitData.calcLyapunov);
    settings.setValue("Orbit/calcLyapunovMethod", configOrbit.calcLyapunovMethod);
    settings.setValue("Orbit/usePS", configCore.usePS);
    // frequency map
    settings.setValue("Frequency_map/numOrbitsStationary", configCore.fm_numOrbitsStationary);
    settings.setValue("Frequency_map/numOrbitsPrincipalPlane", configCore.fm_numOrbitsPrincipalPlane);
    settings.setValue("Frequency_map/numOrbitsYalpha", configCore.fm_numOrbitsYalpha);
    settings.setValue("Frequency_map/numOrbitsRandom", configCore.fm_numOrbitsRandom);
    // schwarzschild modelling
    settings.setValue("Schwarzschild_model/numOrbits", configCore.sm_numOrbits);
    settings.setValue("Schwarzschild_model/densityModelType", QString::fromStdString(getSchwDataNameByType(configCore.sm_densDataType)));
    settings.setValue("Schwarzschild_model/linesPerSegment", configCore.sm_linesPerSegment);
    settings.setValue("Schwarzschild_model/numAngularCoefs", configCore.sm_numAngularCoefs);
    settings.setValue("Schwarzschild_model/numRadialCoefs", configCore.sm_numRadialCoefs);
    settings.setValue("Schwarzschild_model/smAlpha", configCore.sm_Alpha);
    settings.setValue("Schwarzschild_model/numShells", configCore.sm_numShells);
    settings.setValue("Schwarzschild_model/innerShellMass", configCore.sm_innerShellMass);
    settings.setValue("Schwarzschild_model/outerShellMass", configCore.sm_outerShellMass);
    settings.setValue("Schwarzschild_model/numSamplingPoints", configCore.sm_numSamplingPoints);
    settings.setValue("Schwarzschild_model/chaoticMinFreqDiff", configCore.chaoticMinFreqDiff);
    settings.setValue("Schwarzschild_model/chaoticMinLambda", configCore.chaoticMinLambda);
    settings.setValue("Schwarzschild_model/chaoticWeightFactor", configCore.sm_chaoticPenalty);
    settings.setValue("Schwarzschild_model/maxWeight", configCore.sm_maxWeight);
    settings.setValue("Schwarzschild_model/regularization", configCore.sm_regularization);
    settings.setValue("Schwarzschild_model/constraintPenalty", configCore.sm_constraintPenalty);
    settings.setValue("Schwarzschild_model/constrainBeta", configCore.sm_constrainBeta);
    settings.setValue("Schwarzschild_model/betaIn",  configCore.sm_betaIn);
    settings.setValue("Schwarzschild_model/betaOut", configCore.sm_betaOut);
}

QString getDirName(const QString &fileName)  // return file path ending with "/"
{
    QFileInfo fi(fileName);
    QString fn = fi.absolutePath();
    if(!fn.endsWith('/')) fn += '/';
    return fn;
}

/* ----------- Class for performing console/interactive scripting ------------ */ 

void CSmileConsole::runScript(QString scriptFileName)
{
    std::cout << "SMILE console version "VERSION"  build "__DATE__ << std::endl;
    core->loadSettings();
    core->initIC();
    if(scriptFileName!="")
    {
        input.setFileName(scriptFileName);
        isConsoleInput=false;
        if(!input.open(QIODevice::ReadOnly | QIODevice::Text))
        {
            std::cout << "Error! cannot read script file " << scriptFileName.toStdString() << std::endl;
            emit scriptDone();
        }
    }
    else
    {
        input.open(stdin, QIODevice::ReadOnly | QIODevice::Text);
        isConsoleInput=true;
    }
    connect(this, SIGNAL(scriptNextLine()), SLOT(scriptProcessLine()));
    connect(core, SIGNAL(signalOrbitLibraryFinished()), SLOT(scriptOrbitLibraryFinished()));
    connect(core, SIGNAL(signalSchwOptimizationFinished(const QString&)), SLOT(scriptTaskFinished(const QString&)));
    connect(core, SIGNAL(signalNbodyExportFinished(const QString&)), SLOT(scriptTaskFinished(const QString&)));
    connect(core, SIGNAL(signalInfo(const QString&)), SLOT(scriptInfo(const QString&)));
    connect(core, SIGNAL(signalTimer()), SLOT(scriptTimerEvent()));
    timeStartup=QDateTime::currentDateTime();
    emit scriptNextLine();
}

void CSmileConsole::scriptInfo(const QString& message)
{
    std::cout << message.toStdString() << std::endl;
}

void CSmileConsole::scriptTimerEvent()
{
    if(!core->orbitlib) return;
    std::cout << core->orbitlib->numComplete() << " out of " << core->orbitlib->size() << " orbits done\015" << std::flush;
}

void CSmileConsole::scriptOrbitLibraryFinished()
{
    int timeElapsed = timeOrbitLibStarted.msecsTo(QDateTime::currentDateTime());
    std::cout << core->orbitlib->numComplete() << " orbits total, " << 
        QString::number(1000.0*core->orbitlib->numComplete()/timeElapsed,'g',3).toStdString() << " orbits/s" << std::endl;
    emit scriptNextLine();
}

void CSmileConsole::scriptTaskFinished(const QString& message)
{
    emit scriptInfo(message);
    emit scriptNextLine();
}

void CSmileConsole::scriptProcessLine()
{
    QString Line;
    std::cout << "> ";
    if(!input.atEnd())
        Line = input.readLine().trimmed();
    else
        Line = "Exit";
    if(Line=="" || Line.startsWith("#"))
    {
        emit scriptNextLine();
        return;
    }
    if(Line.startsWith("Exit", Qt::CaseInsensitive) || Line.startsWith("Quit", Qt::CaseInsensitive) || Line.startsWith("Bye", Qt::CaseInsensitive))
    {
        std::cout << "Done!  Total runtime: " << 
            timeStartup.secsTo(QDateTime::currentDateTime()) << " seconds" << std::endl;
        emit scriptDone();
        return;
    }
    if(!isConsoleInput) 
        std::cout << (Line.toStdString()) << std::endl;
    
    // parse input string and perform desired action
    if(Line.startsWith("ReadIni", Qt::CaseInsensitive) || Line.startsWith("ReadConfig", Qt::CaseInsensitive) || 
        Line.startsWith("LoadIni", Qt::CaseInsensitive) || Line.startsWith("LoadConfig", Qt::CaseInsensitive))
    {
        int i1=Line.indexOf("(""");
        int i2=Line.indexOf(""")");
        QString FileName = (i1>=2 && i2>i1+3) ? Line.mid(i1+2, i2-i1-3) : (i1==-1&&i2==-1 ? DEFAULT_SETTINGS_FILE_NAME : "");
        if(!FileName.isEmpty())
        {
            if(core->loadSettings(FileName))
            {
                std::cout << "Loaded config from " << FileName.toStdString() << std::endl;
                core->initIC();
            }
            else
                std::cout << "Error! Cannot load config from " << FileName.toStdString() << std::endl;
        }
        else std::cout << "Error! No valid filename specified" << std::endl;
    }
    else if(Line.startsWith("ExportPotential", Qt::CaseInsensitive))
    {
        int i1=Line.indexOf("(""");
        int i2=Line.indexOf(""")");
        QString FileName = Line.mid(i1+2, i2-i1-3);
        if(i1>0 && i2>i1)
        {
            if(core->exportPotential(FileName))
                std::cout << "Potential exported to " << FileName.toStdString() << std::endl;
            else
                std::cout << "Error! Cannot export potential to " << FileName.toStdString() << std::endl;
        }
        else std::cout << "Error! No filename specified" << std::endl;
    }
    else if(Line.startsWith("BuildFreq", Qt::CaseInsensitive) || 
            Line.startsWith("BuildSchw", Qt::CaseInsensitive) || Line.startsWith("BuildModel", Qt::CaseInsensitive))
    {
        if(Line.startsWith("BuildFreq", Qt::CaseInsensitive))
            core->buildFreqMapIC(core->initCondE, Line.contains("exist", Qt::CaseInsensitive));
        else
            core->buildSchwIC(Line.contains("exist", Qt::CaseInsensitive));
        core->startOrbitLibrary(SIGNAL(signalOrbitLibraryFinished()));
        timeOrbitLibStarted=QDateTime::currentDateTime();
        return;
    }
    else if(Line.startsWith("SchwLinear", Qt::CaseInsensitive) || Line.startsWith("SchwQuad", Qt::CaseInsensitive))
    {
        if(core->schwData.size()>0 && core->orbitlib!=NULL && core->orbitlib->size()>0)
        {
            core->schwStartOptimization(Line.startsWith("SchwQuad", Qt::CaseInsensitive));
            return;
        }
        else
            std::cout << "No Schwarzschild model or orbit library defined" << std::endl;
    }
    else if(Line.startsWith("ExportOrbits", Qt::CaseInsensitive) || Line.startsWith("SaveOrbits", Qt::CaseInsensitive) || 
            Line.startsWith("ExportModel", Qt::CaseInsensitive)  || Line.startsWith("SaveModel", Qt::CaseInsensitive))
    {
        if(core->orbitlib!=NULL)
        {
            int i1=Line.indexOf("(""");
            int i2=Line.indexOf(""")");
            QString FileName = Line.mid(i1+2, i2-i1-3);
            if(i1>0 && i2>i1)
            {
                bool withmodel=Line.startsWith("ExportModel", Qt::CaseInsensitive) || Line.startsWith("SaveModel", Qt::CaseInsensitive);
                if(!core->exportOrbitLib(FileName, withmodel))
                    std::cout << "Error writing file: " << FileName.toStdString() << std::endl;
            }
            else std::cout << "No filename specified" << std::endl;
        }
        else
            std::cout << "No orbit library defined" << std::endl;
    }
    else if(Line.startsWith("ImportOrbits", Qt::CaseInsensitive) || Line.startsWith("LoadOrbits", Qt::CaseInsensitive) || 
            Line.startsWith("ImportModel", Qt::CaseInsensitive)  || Line.startsWith("LoadModel", Qt::CaseInsensitive))
    {
        int i1=Line.indexOf("(""");
        int i2=Line.indexOf(""")");
        QString FileName = Line.mid(i1+2, i2-i1-3);
        if(i1>0 && i2>i1)
        {
            bool withmodel=Line.startsWith("ImportModel", Qt::CaseInsensitive) || Line.startsWith("LoadModel", Qt::CaseInsensitive);
            if(QFile::exists(FileName+".ini") && core->loadSettings(FileName+".ini"))
            {
                std::cout << "Loaded config from " << FileName.toStdString() << ".ini" << std::endl;
                core->initIC();
            }
            if(!core->importOrbitLib(FileName, withmodel))
                std::cout << "Error reading file: " << FileName.toStdString() << std::endl;
            else {
                std::cout << core->orbitlib->numComplete() << " orbits loaded" << std::endl;
            }
        }
        else std::cout << "No filename specified" << std::endl;
    }
    else if(Line.startsWith("ExportNbody", Qt::CaseInsensitive))
    {
        if(core->orbitlib)
        {
            int i1=Line.indexOf("(");
            int i2=Line.lastIndexOf(")");
            QString paramstr = Line.mid(i1+1, i2-i1-1);
            QStringList params=paramstr.split(QRegExp("[,\\s]+"));
            if(params.count()<2)
                std::cout << "Syntax: ExportNbody(NumBodies, \"FileName\" [, \"FileFormat\"=\"" << 
                formatsIOSnapshot.front() << "\" [, RefineFactor=0]])" << std::endl;
            else{
                int NPoints=static_cast<int>(params[0].toDouble());
                i1=params[1].indexOf("\"");
                i2=params[1].lastIndexOf("\"");
                QString fileName=params[1].mid(i1+1, i2-i1-1);
                std::string fileFormat=formatsIOSnapshot.front();
                if(params.count()>=3) {
                    i1=params[2].indexOf("\"");
                    i2=params[2].lastIndexOf("\"");
                    fileFormat=params[2].mid(i1+1, i2-i1-1).toStdString();
                    bool found=false;
                    for(size_t i=0; i<formatsIOSnapshot.size(); i++)
                        if(fileFormat==formatsIOSnapshot[i]) found=true;
                    if(!found) {
                        fileFormat=formatsIOSnapshot.front();
                        std::cout << "Warning, incorrect specification for file format, using \""<<fileFormat<<"\" instead" << std::endl;
                    }
                }
                int refineFactor=0;
                if(params.count()>=4)
                    refineFactor=params[3].toInt();
                if(NPoints==0)
                    std::cout << "Incorrect first parameter: NumBodies" << std::endl;
                else
                if(fileName!="")
                {
                    core->schwExportNbody(NPoints, fileName, QString(fileFormat.c_str()), refineFactor);
                    return;
                }
                else std::cout << "No filename specified" << std::endl;
            }
        }
        else
            std::cout << "No orbit library defined" << std::endl;
    }
    else 
        std::cout << "Invalid command" << std::endl;
    emit scriptNextLine();
}

/* -------- Helper threads that run various tasks in parallel to main program ------ */

// run a single orbit integration
void COrbitIntWorker::doIntegration()
{
    orbit->integrateToTime(time);
    emit calcFinished();
    deleteLater();
}

// run orbit library integration in parallel threads
volatile int nthreads=0;
QMutex mutexL;
QMutex mutexC;

void COrbitLibWorker::run()
{
    mutexC.lock();
    ++nthreads;
    mutexC.unlock();
    finished=false;
    while(!finished)
    {
        mutexL.lock();
        int no = orbits->findUnstartedOrbit();
        mutexL.unlock();
        if(no>=0) {
            orbits->runOrbit(no);
            const COrbitDesc* last=orbits->getOrbitDesc(no);
            if(last!=NULL && last->getInfoNum()>=1 && 
                last->getInfo(0)->infoType()==CBasicInformation::IT_TRAJ_ANALYSIS &&
                static_cast<const COrbitInformation<float>* >(last->getInfo(0))->getDescription()=="Failed")   // indicates no memory
                    halt();   // need to stop entire orbit library integration
        }
        else finished=true;
    }
    mutexC.lock();
    --nthreads;
    if(nthreads==0)
        emit calcFinished();   // only once
    mutexC.unlock();
    deleteLater();
}

void CSchwarzschildWorker::doWork()
{
    if(model==NULL)
        emit message( "Error, no Schw.model object is present!" );
    std::string result = model->performModelling();
    emit message(QString::fromStdString(result));
    emit workDone();
    deleteLater();
};

void CNbodyExportStatWorker::doWork()
{
    handler = new CPointMassSetHandler<float>(points);
    double T=0, W=0;
    handler->computeVirialRatio(&T, &W);
    delete handler;
    handler=NULL;
    delete points;   // deleting an object which was created externally and passed 
    // as a parameter to the constructor, but is no longer needed
    emit exportFinished(info + 
        "\nT="+QString::number(T, 'g', 5) + 
        ", W="+QString::number(W, 'g', 5) + 
        "\nVirial ratio="+QString::number(2*T/W, 'g', 5));
    emit workDone();
    deleteLater();
}

}