/** \file    core.h
    \brief   Main business logic for both GUI and console versions, and the console scripting class
    \author  EV
    \date    2009-2013

    This file defines smile::CSmileCore - the driver class for both console and GUI versions of SMILE, 
    which handles configuration data in the ini file, 
    starts and manages computation threads, saves/loads data.
    A number of helper thread classes for performing various tasks are also defined here.
    Finally, the smile::CSmileConsole class manages console scripting (interactive or from a file).
*/
#pragma once
#include "common.h"
#include "potential.h"
#include "orbitan.h"
#include "orbitlib.h"
#include "schwarzschild.h"
#include <QObject>
#include <QTime>
#include <QFile>

namespace smile{

/* -------- Worker objects that run various tasks in separate threads parallel to the main program -------- */

/// Worker object that runs single orbit integration in a separate thread
class COrbitIntWorker : public QObject
{
    Q_OBJECT
public:
    COrbitIntWorker(COrbit* _orbit, double _time): orbit(_orbit), time(_time) {};
public slots:
    void doIntegration();  ///< call orbit->integrateToTime()
signals:
    void calcFinished();   ///< inform that integration is finished
private:
    COrbit *orbit;         ///< pointer to the orbit
    double time;           ///< required integration time
};

/// Worker object that runs orbit library integration (for use in multiple parallel threads).
/// TODO: reimplement the functionality using higher-level API such as QRunnable/QThreadPool or QtConcurrent.
class COrbitLibWorker : public QObject
{
    Q_OBJECT
public:
    COrbitLibWorker(COrbitLibrary* _orbits): orbits(_orbits) {};
signals:
    void calcFinished();  ///< emitted only once, by the last object that finishes its work
public slots:
    /// loop over orbit library to find the next orbit to integrate it, until no more orbits left
    void run();
    /// terminate the integration of the current orbit and the entire library
    void halt() { orbits->halt(); finished=true; }
private:
    COrbitLibrary* orbits;
    bool finished;
};

/** Worker object that runs Schwarzschild model optimization solver.
    It is created by CSmileCore and put into a separate thread, so that the main program continues.
    When the optimization is done, emit a signal with text representation of the result from model 
    or the error from the solver.
**/
class CSchwarzschildWorker : public QObject
{
    Q_OBJECT
public:
    CSchwarzschildWorker(CBasicSchwModel* _model): model(_model) {}; 
public slots:
    void doWork();                         ///< call model->performModelling()
signals:
    void message(const QString& result);   ///< communicate the modelling result
    void workDone();                       ///< tell that it is safe to delete the helper thread
private:
    CBasicSchwModel* model;                ///< pointer to an existing schwarzschild model object
};

/** Worker object that calculates statistics after Nbody export (currently, computes virial ratio).
    Created by CSmileCore and put into a separate thread.
    Initialized with a pointer to point mass set (Nbody model), and an info string.
    Point mass set, passed as a parameter to the constructor, is destroyed upon thread completion!
    Returns info string + virial ratio computed.
**/
class CNbodyExportStatWorker : public QObject
{
    Q_OBJECT
public:
    CNbodyExportStatWorker(const CPointMassSetFloat* _points, const QString& _info) : 
        points(_points), handler(NULL), info(_info) {};
signals:
    void exportFinished(const QString& result);  ///< communicate the virial ratio computed and the previous info string
    void workDone();                             ///< signal that the helper thread may be deleted
public slots:
    void doWork();                               ///< compute the virial ratio, then delete points (which were const, huh?)
    void halt()                                  ///  terminate computing the virial ratio immediately
    { if(handler!=NULL) handler->halt(); };
private:
    const CPointMassSetFloat* points;           ///< the point masses of the snapshot
    CPointMassSetHandler<float>* handler;       ///< actual computation is done by this object
    QString info;                               ///< previous info string which is appended to the output of exportFinished()
};

/// Common configuration parameters for various tasks in CSmileCore
struct CConfigCore
{
    /// \name orbit integration parameters
    ///@{
    double intTimeStepsPerPeriod;    ///< number of timesteps per unit time described above
    double intTimeInPeriods;         ///< orbit integration time measured in units of long-axis-orbit periods
    double intTimeMaxAdaptive;       ///< upper limits for adaptive integration time, expressed in energy-dependent orbital time periods
    double adaptiveTimeThreshold;    ///< (unused) parameter in Pfenniger's method for adaptive integration time
    bool usePS;                      ///< whether to plot points on Poincare section
    ///@}
    /// \name frequency map parameters
    ///@{
    int fm_numOrbitsStationary;      ///< number of orbits in stationary start-space
    int fm_numOrbitsPrincipalPlane;  ///< number of orbits in principal-plane start-space
    int fm_numOrbitsYalpha;          ///< number of orbits in Y-alpha start space
    int fm_numOrbitsRandom;          ///< number of orbits in random(ergodic) start-space
    ///@}
    /// \name Schwarzschild modelling parameters
    ///@{
    int sm_numOrbits;                ///< number of orbits with random initial conditions in the Schwarzschild model
    int sm_numShells;                ///< grid size: number of radial shells
    int sm_linesPerSegment;          ///< grid size: number of lines dividing each of 3 segments of each shell into cells
    int sm_numAngularCoefs;          ///< order of angular spherical-harmonic expansion
    int sm_numRadialCoefs;           ///< number of radial coefs in BSE / SHGrid Schwarzschild model
    double sm_Alpha;                 ///< Alpha-parameter defining family of basis functions in BSE model
    double sm_innerShellMass;        ///< fraction of mass in the innermost shell, 0 for default value of 1/(numShells+1)
    double sm_outerShellMass;        ///< fraction of mass enclosed by the outermost shell, 0 for default value of 1-1/(numShells+1)
    double chaoticMinFreqDiff;       ///< minimal value of frequency diffusion rate to consider orbit as chaotic
    double chaoticMinLambda;         ///< minimal value of Lyapunov indicator to consider orbit as chaotic (in OR relation with previous criterion)
    double sm_chaoticPenalty;        ///< weight factor for chaotic (super)orbit in objective function: +1 - disfavor them, -1 - favor, 0 - treat equally as regular
    int sm_numSamplingPoints;        ///< number of points sampled from trajectory, default NUM_SAMPLING_POINTS
    bool sm_useBPMPD;                ///< whether to use BPMPD solver (if the executable is present), otherwise use GLPK
    CBasicSchwData::SCHWDATATYPE sm_densDataType;  ///< variant of Schwarzschild density model
    double sm_maxWeight;             ///< upper bound on orbit weight
    double sm_regularization;        ///< parameter controlling the regularization of orbit weights in the quadratic optimization
    double sm_constraintPenalty;     ///< penalty for constraint violation in the objective function in linear/quadratic optimization; 0 means infinity - must satisfy exactly
    bool sm_constrainBeta;           ///< whether to constrain anisotropy
    double sm_betaIn, sm_betaOut;    ///< anisotropy coefficient requested (values for the inner and outer shells, linearly interpolated in between)
    ///@}
};

/** Class that provides all core computational functionality.
    It contains configuration parameters and routines to load/save them in an ini file; 
    routines for managing orbit library integration, Schwarzschild modelling and export of 
    a model to an Nbody snapshot; manages import/export of data to text/binary files.
    This class is used in the console version - by CSmileConsole, and in the GUI - by CSmileGUI.
    Most computational tasks are performed in separate threads by worker objects; 
    there are four of them:
    - integration of a single orbit (COrbitIntWorker);
    - integration of the orbit library (COrbitLibWorker) in several concurrent threads;
    - performing Schwarzschild model optimization (CSchwarzschildWorker);
    - export of Schw.model to an Nbody snapshot (CNbodyExportStatWorker), 
      the latter operation is partly performed within the context of main thread.
    The first task is independent of the other three and both may run in parallel; 
    however all operations with orbit library should not overlap (it must be ensured by the 
    upper-level object starting these tasks). 
*/
class CSmileCore: public QObject
{
    Q_OBJECT

public:
    /* --- configuration data --- */
    CConfigCore configCore;                ///< global parameters of the program
    CConfigPotential configPotential;      ///< parameters related to the potential
    CConfigOrbit configOrbit;              ///< parameters related to orbit integration
    COrbitInitData<double> InitData;       ///< initial conditions for a single orbit
    double initCondE;                      ///< energy corresponding to initial conditions
    bool useICe;                           ///< whether to use energy as an initial condition instead of position/velocity (applies to freq.map)
    QString WorkDir;                       ///< current working directory (for GUI file dialogs)
    QString TempDir;                       ///< directory for exchanging data with external programs (qdelaunay, bpmpd)
    QString AppDir;                        ///< directory where the application executable is located and external programs are searched for
    QString NbodyFile;                     ///< name of file containing discrete points for tree-code or BSE/spline expansion potentials
    /* --- end of configuration data --- */

    const CPotential* potential;           ///< instance of potential used in all calculations, created by initPotential()
    COrbit* orbit;                         ///< current orbit
    COrbitLibrary* orbitlib;               ///< library of orbits (for frequency map and Schwarzschild model)
    vectorSchwData schwData;               ///< Schwarzschild data objects (density and kinematic constraints)

private:
    int maxNumThread;                      ///< number of concurrent threads in orbit integration (set in ini file or automatically)
    int orbitLibTimer;                     ///< timer ID for displaying progress in orbit library integration; set to 0 when no integration is in progress
    bool orbitIntRunning;                  ///< set to true for the duration of single orbit integration
    CBasicSchwModel* schwModel;            ///< instance of Schw model object internally created for the duration of optimization
    vectorRuntimeFncCreators creatorsLib;  ///< array of runtime function generators for orbit library
    /* --- internally stored parameters for Nbody export --- */
    int numPointsNbodyExport;              ///< number of points in Nbody model
    QString fileNameNbodyExport;           ///< Nbody snapshot filename. If assigned, it means that the export is in progress.
    QString fileFormatNbodyExport;         ///< format selector for Nbody snapshot
    int numBinsRefineNbodyExport;          ///< if mass refinement in Nbody model is used, this is the number of mass bins
    bool firstCallNbodyExport;             ///< flag signalling if the coreNbodyExport() was called for the first time or after reintegration

public:
    CSmileCore();      ///< construct the object and set default values for internal state variables
    ~CSmileCore();     ///< destroy the object

    /* --- configuration file --- */
    /// load configuration data (at start - from predefined file smile.ini, later - when loading orbit library)
    bool loadSettings(QString fileName="");
    
    /// store configuration data (when storing orbit library, or at exit, if demanded in GUI)
    void saveSettings(QString fileName="");

    /* --- potential --- */
    /// create a new instance of potential specified by configPotential
    void initPotential();
    
    /* --- orbit integration --- */
    /// set up initial conditions for a single orbit integration (calc E from X,Y,Z or vice-versa, calc timeUnit, etc)
    void initIC();
    /// start thread that runs single orbit integration
    void startOrbit();
    
    /* --- Frequency map and Schwarzschild modelling --- */
    /// create initial conditions for orbit library for frequency map at given energy
    void buildFreqMapIC(double energy, bool useExistingIC);
    
    /// create initial conditions for orbit library for Schwarzschild model
    void buildSchwIC(bool useExistingIC, CPointCountSetFloat* orbitsForReintegrate=NULL);
    
    /// start several threads to integrate orbits in the orbit library
    void startOrbitLibrary(const char* finishSignal);
    
    /// start thread that solves Schwarzschild optimization problem
    void schwStartOptimization(bool quadratic);
    
    /// export Schw model to Nbody model
    void schwExportNbody(int numPoints, const QString &fileName, const QString &fileFormat, int numBinsRefine);
    
    /* --- import/export functions --- */
    /// read trajectory from a text file and assign it to the current orbit
    bool importOrbit(const QString &fileName);

    /// export current orbit into a text file
    bool exportOrbit(const QString &fileName);
    
    /// export potential file (values of potential and forces in certain points of space, for statistics)
    bool exportPotential(const QString &fileName);
    
    /// load orbit library from a text file, possibly with Schwarzschild data from an accompanying binary file
    bool importOrbitLib(const QString &fileName, bool withModelData);
    
    /// save orbit library to a text file, possibly with Schwarzschild data
    bool exportOrbitLib(const QString &fileName, bool withModelData);
    
    /// export statistical information about the model
    bool exportSchwModelStats(const QString &fileName);

    /* --- information about currently running tasks --- */
    /// inform whether a single orbit integration is running (in a separate thread)
    bool isOrbitIntRunning() const { return orbitIntRunning; };

    /// inform whether the orbit library integration is in progress
    bool isOrbitLibRunning() const { return orbitLibTimer!=0; };

    /// inform whether the optimization is in progress (in a helper thread)
    bool isSchwOptRunning()  const { return schwModel!=NULL; };

    /// inform whether export to Nbody model is in progress (at any stage)
    bool isExportNbodyRunning() const { return !fileNameNbodyExport.isEmpty(); }

private:
    void initSchwData();                       ///< create a set of schwData objects specified by config->modeltype
    void initPotentialNB();                    ///< load Nbody coords file, use direct N-body potential or calculate density and potential in basis set / spline expansion
    void initPotentialDefault();               ///< called when an error occurs during initPotential, since one cannot leave this object uninitialized
    bool exportPotentialAtPoints               ///  actually do export potential, forces and density at given set of points
        (const QString &fileName, const vectord points[N_DIM], int blockSize=0);
    virtual void timerEvent(QTimerEvent*);     ///< to display progress indicator in orbit library integration

signals:
    void signalInfo(const QString &info);      ///< display various information (GUI - in message area, console - print out string)
    void signalTimer();                        ///< emitted regularly during orbit library integration, to update progress info
    void signalOrbitFinished();                ///< emitted when orbit integration finished
    void signalOrbitLibraryFinished();         ///< emitted when the orbit library integration is finished (for the frequency map or Schwarzschild orbit library)
    void signalSchwOptimizationFinished(const QString& result);  ///< emitted when Schwarzschild optimization finished
    void signalNbodyReintegrationFinished();   ///< emitted upon finish of reintegrating orbits for nbody export (if it was necessary to get more sampling points)
    void signalNbodyExportFinished(const QString& result);       ///< emitted when export to nbody model is finished
    void signalStopOrbitLibrary();             ///< emitted when the user cancels orbit library integration; transmitted to the worker thread objects
    void signalStopExportNbody();              ///< emitted when the user stops export to Nbody model
private slots:
    /// called when single orbit integration is done; emits signalOrbitFinished()
    void coreOrbitFinished();

    /// called upon finish of last thread in building orbit library for frequency map or Schw model; 
    /// stops timers, clears unused orbits and emits signalOrbitLibraryFinished()
    void coreOrbitLibraryFinished();
    
    /// called upon finish of Schwarzschild optimization; 
    /// destroys the model object and emits signalSchwOptimizationFinished()
    void coreSchwOptimizationFinished(const QString& result);
    
    /// called when export to nbody model is finished; 
    /// clears the filename and emits signalNbodyExportFinished()
    void coreNbodyExportFinished(const QString& result);
    
    /** perform actual work on export to nbody file. 
        The export consists of three stages: 
        - First the Nbody model is created by call to COrbitLibrary::exportNbody();
        this procedure creates the requested number of particles but may produce duplicate points 
        if the number of sampling points was not sufficient for some orbits.
        In this case, these orbits are marked for reintegration and a call to startOrbitLibrary() 
        launches the integration in parallel threads. 
        Upon finishing the reintegration signalNbodyReintegrationFinished() is emitted, 
        and coreNbodyExport() is called once again. 
        - Then the Nbody snapshot is written to a file by calling createIOSnapshotWrite().
        - Finally, the virial ratio is computed in a separate thread by a worker object 
        CNbodyExportStatWorker; upon finishing signalNbodyExportFinished() is emitted.
    */
    void coreNbodyExport();                    
};

/** The console version of application is implemented in this class.
    It may be directed by either interactive input from the user, or a script file.
    Each task is processed in turn, even if performed in a separate thread;
    unlike the GUI version, there is no way to stop a running operation. 
    Single orbit integration is not used in the console version; only tasks related 
    to the orbit library and Schwarzschild modelling are implemented.
    For a full list of available commands see readme.pdf
*/
class CSmileConsole: public QObject
{
    Q_OBJECT
public:
    /// create the instance of CSmileCore object
    CSmileConsole() { core=new CSmileCore(); }
    /// destroy the core object
    ~CSmileConsole() { delete core; }
    /// run a script from a file or process data from console if the parameter is empty
    void runScript(QString scriptFileName="");
private:
    CSmileCore* core;              ///< the instance of core object
    QFile input;                   ///< input file for scripting (or console)
    bool isConsoleInput;           ///< set when running console version; if interactive, set ConsoleInput=true, if processing script, set to false
    QDateTime timeStartup;         ///< set to current time at startup
    QDateTime timeOrbitLibStarted; ///< measure time for orbit library integration
signals:
    void scriptNextLine();         ///< emitted upon having processed a line in console script
    void scriptDone();             ///< emitted when finished processing console script

private slots:
    /// the main command parser: process a line in the script file or from the console 
    /// and start the appropriate operation
    void scriptProcessLine();

    /// called after finishing frequency map or Schwarzschild orbit library;
    /// display some statistics and call scriptProcessLine()
    void scriptOrbitLibraryFinished();

    /// called after finishing optimization or Nbody export; 
    /// display the text message and call scriptProcessLine()
    void scriptTaskFinished(const QString& message);

    /// display progress in orbit library integration (connected to CSmilecore::signalTimer() )
    void scriptTimerEvent();

    /// displays message in console mode (connected to CSmileCore::signalInfo() )
    void scriptInfo(const QString& message);
};

QString getDirName(const QString &fileName);   ///< convenience function, return file path ending with "/"

}