#include "gui.h"

#include <cmath>
#include <vector>
#include <algorithm>
#include <cassert>
#include <fstream>
#include <iostream>

#include <QKeyEvent>
#include <QFileDialog>
#include <QMessageBox>
#include <QPrinter>
#include <QImage>
#include <QMap>
#include <QInputDialog>
#include <QMetaObject>

#include <qwt_plot.h>
#include <qwt_plot_curve.h>
#include <qwt_text.h>
#include <qwt_symbol.h>
#include <qwt_plot_picker.h>
#include <qwt_plot_panner.h>
#include <qwt_plot_zoomer.h>
#include <qwt_scale_engine.h>
#include <qwt_plot_marker.h>

#include "core.h"
#include "potential.h"
#include "orbitan.h"
#include "orbitlib.h"
#include "schwarzschild.h"
#include "utils.h"
#include "iosnapshot.h"

#ifdef USE3DFACETS
// for triangulation thread
#include <QVector>
#include <QProcess>
#include <QTextStream>
#endif

namespace smile{

/// The global variable for the instance of GUI object is used to redirect text messages to it
CSmileGUI* GUI;

/// display the message string in the bottom-right text area of the main window
void my_message_gui(const std::string &message)
{
#ifndef DEBUGPRINT
    std::cerr << message <<"\n";
#endif
    QMetaObject::invokeMethod(GUI, "info", Q_ARG(QString, QString::fromStdString(message)));
}

/// pop up a message box with the error string
void my_error_gui(const std::string &message)
{
    std::cerr << message <<"\n";
    QMetaObject::invokeMethod(GUI, "error", Q_ARG(QString, QString::fromStdString(message)));
}

void CSmileGUI::info(const QString &message)
{
#ifdef DEBUGPRINT
    std::cerr << message.toStdString() <<"\n";
#endif
    ui.label_info->setPlainText(message);
    ui.label_info->repaint();  // to ensure immediate displaying
}

void CSmileGUI::error(const QString &message)
{
    QMessageBox::warning(this, "Error", message);
}

CSmileGUI::CSmileGUI(QWidget *parent, Qt::WFlags flags)
    : QMainWindow(parent, flags)
{
    GUI=this;
    ui.setupUi(this);

    core=new CSmileCore();  // initialize the instance of "business logic" object
    my_error_ptr=&my_error_gui;
    my_message_ptr=&my_message_gui;
    // connect signals to the core
    connect(core, SIGNAL(signalOrbitFinished()), SLOT(CalcFinished()));
    connect(core, SIGNAL(signalOrbitLibraryFinished()), this, SLOT(OrbitLibraryFinished()));
    connect(core, SIGNAL(signalSchwOptimizationFinished(const QString&)), SLOT(SchwOptimizationFinished(const QString&)));
    connect(core, SIGNAL(signalNbodyReintegrationFinished()), SLOT(NbodyReintegrationFinished()));
    connect(core, SIGNAL(signalNbodyExportFinished(const QString&)), SLOT(NbodyExportFinished(const QString&)));
    connect(core, SIGNAL(signalInfo(const QString&)), SLOT(info(const QString&)));
    connect(core, SIGNAL(signalTimer()), SLOT(timerEventCore()));
    connect(this, SIGNAL(stopOrbitLibrary()), core, SIGNAL(signalStopOrbitLibrary()));
    connect(this, SIGNAL(stopExportNbody()),  core, SIGNAL(signalStopExportNbody()));
    
    Eprev=0;
    // fill potential, density, symmetry and SM types dropdown list
    ui.selectPotentialType->clear();
    for(PotentialNameMapType::iterator iterp=PotentialNames.begin(); iterp!=PotentialNames.end(); iterp++)
        ui.selectPotentialType->insertItem(100, QString::fromStdString(iterp->second));
    ui.selectDensityType->clear();
    for(DensityNameMapType::iterator iterd=DensityNames.begin(); iterd!=DensityNames.end(); iterd++)
        ui.selectDensityType->insertItem(100, QString::fromStdString(iterd->second));
    ui.selectSymmetryType->clear();
    for(SymmetryNameMapType::iterator iters=SymmetryNames.begin(); iters!=SymmetryNames.end(); iters++)
        ui.selectSymmetryType->insertItem(100, QString::fromStdString(iters->second));
    ui.select_smType->clear();
    ui.select_smType->insertItem(100, QString::fromStdString(CSchwDataDensClassic::myName()));
    ui.select_smType->insertItem(100, QString::fromStdString(CSchwDataDensSHGrid::myName()));
    ui.select_smType->insertItem(100, QString::fromStdString(CSchwDataDensSHBSE::myName()));
    
    // orbit plot
    PlotOrbit = new QwtPlot(QwtText("Orbit"), ui.frame_orbitarea);
    CurveOrbit= new QwtPlotCurve("");
    CurveOrbit->setPen(QColor(Qt::blue));
    CurveOrbit->attach(PlotOrbit);
    CurveEquipotential= new QwtPlotCurve("");
    CurveEquipotential->setPen(QColor(Qt::red));
    CurveEquipotential->attach(PlotOrbit);
    PlotOrbit->show();
    QVBoxLayout* LPlotOrbit = new QVBoxLayout(ui.frame_orbitarea);
    LPlotOrbit->addWidget(PlotOrbit);
    ZoomerOrbit= new QwtPlotZoomer2( QwtPlot::xBottom, QwtPlot::yLeft, 
        QwtPicker::DragSelection | QwtPicker::CornerToCorner, QwtPicker::ActiveOnly, PlotOrbit->canvas(), true);
    ZoomerOrbit->setMousePattern(QwtEventPattern::MouseSelect3, Qt::RightButton);
    QwtPlotPanner* PannerOrbit = new QwtPlotPanner(PlotOrbit->canvas());
    PannerOrbit->setMouseButton(Qt::MidButton);

    PlotOrbitParam = new QwtPlot(QwtText("Distribution of orbital parameters"), ui.frame_orbitarea);
    CurveOrbitParam= new QwtPlotCurve("");
    CurveOrbitParam->attach(PlotOrbitParam);
    PlotOrbitParam->hide();
    LPlotOrbit->addWidget(PlotOrbitParam);
    ZoomerOrbitParam= new QwtPlotZoomer2( QwtPlot::xBottom, QwtPlot::yLeft, 
        QwtPicker::DragSelection | QwtPicker::CornerToCorner, QwtPicker::ActiveOnly, PlotOrbitParam->canvas(), true);
    ZoomerOrbitParam->setMousePattern(QwtEventPattern::MouseSelect3, Qt::RightButton);
    QwtPlotPanner* PannerOrbitParam = new QwtPlotPanner(PlotOrbitParam->canvas());
    PannerOrbitParam->setMouseButton(Qt::MidButton);

#ifdef HAVE_QWT3D
    // 3d orbit
    Plot3d=new Qwt3D::SurfacePlot(ui.frame_orbitarea);
    Plot3d->setCoordinateStyle(Qwt3D::FRAME);
    Plot3d->setRotation(90,0,0);
    QColor bkg=ui.tabWidget->palette().color(QPalette::Background);
    Plot3d->setBackgroundColor(Qwt3D::RGBA(bkg.redF(), bkg.greenF(), bkg.blueF()));
    Plot3d->setMeshColor(Qwt3D::RGBA(0.2,0,0.4));
    for (unsigned i=0; i!=Plot3d->coordinates()->axes.size(); ++i)
    {
      Plot3d->coordinates()->axes[i].setMajors(7);
      Plot3d->coordinates()->axes[i].setMinors(4);
    }
    LPlotOrbit->addWidget(Plot3d);
    ui.progressBarOrbit->hide();
#ifdef USE3DFACETS
    TriangThr = new CTriangThread(core->AppDir+"qdelaunay");
    connect(TriangThr, SIGNAL(finished(bool)), SLOT(redrawOrbit(bool)));
    Plot3d->setLightComponent(GL_DIFFUSE, 0.3);
    Plot3d->setMaterialComponent(GL_SPECULAR, 0.0);
    Plot3d->enableLighting();
    Plot3d->illuminate();
    Plot3d->updateGL();
#endif
#endif

    // poincare plot
    PlotPoincare = new QwtPlot(QwtText("Poincare section"), ui.frame_poincarearea);
    PlotPoincare->show();
    QVBoxLayout * LPlotPoincare = new QVBoxLayout(ui.frame_poincarearea);
    LPlotPoincare->addWidget(PlotPoincare);
    QwtPlotPicker *PickerPoincare = new QwtPlotPicker2(QwtPlot::xBottom, QwtPlot::yLeft,
        QwtPicker::PointSelection, QwtPlotPicker::NoRubberBand, QwtPicker::AlwaysOn, PlotPoincare->canvas());
    PickerPoincare->setMousePattern(QwtEventPattern::MouseSelect1, Qt::RightButton);
    connect(PickerPoincare, SIGNAL(selected(const QwtDoublePoint &)), SLOT(PSclicked(const QwtDoublePoint &)));
    ZoomerPoincare = new QwtPlotZoomer2( QwtPlot::xBottom, QwtPlot::yLeft,
        QwtPicker::DragSelection | QwtPicker::CornerToCorner, QwtPicker::AlwaysOn, PlotPoincare->canvas(), true);
    ZoomerPoincare->setMousePattern(QwtEventPattern::MouseSelect3, Qt::RightButton, Qt::ControlModifier);
    ZoomerPoincare->setMousePattern(QwtEventPattern::MouseSelect2, Qt::RightButton, Qt::ControlModifier);
    ZoomerPoincare->setKeyPattern(QwtEventPattern::KeySelect1, Qt::Key_Space);
    QwtPlotPanner* PannerPoincare = new QwtPlotPanner(PlotPoincare->canvas());
    PannerPoincare->setMouseButton(Qt::MidButton);

    // frequency plot
    PlotFreq = new QwtPlot(QwtText("Frequencies"), ui.frame_frequencies);
    CurveFreqX= new QwtPlotCurve("X");
    CurveFreqX->setPen(QColor(Qt::blue));
    CurveFreqX->attach(PlotFreq);
    CurveFreqY= new QwtPlotCurve("Y");
    CurveFreqY->setPen(QColor(Qt::darkGreen));
    CurveFreqY->attach(PlotFreq);
    CurveFreqZ= new QwtPlotCurve("Z");
    CurveFreqZ->setPen(QColor(Qt::darkRed));
    CurveFreqZ->attach(PlotFreq);
    CurveFreqPeaks= new QwtPlotCurve();
    CurveFreqPeaks->setStyle(QwtPlotCurve::Sticks);
    CurveFreqPeaks->attach(PlotFreq);
    CurveFreqPeaksP= new QwtPlotCurve();
    CurveFreqPeaksP->setStyle(QwtPlotCurve::Sticks);
    CurveFreqPeaksP->setPen(QColor(Qt::white));
    CurveFreqPeaksP->attach(PlotFreq);
    PlotFreq->show();
    QVBoxLayout* LPlotFreq = new QVBoxLayout(ui.frame_frequencies);
    LPlotFreq->addWidget(PlotFreq);
    ZoomerFreq = new QwtPlotZoomer2( QwtPlot::xBottom, QwtPlot::yLeft,
        QwtPicker::DragSelection | QwtPicker::CornerToCorner, QwtPicker::AlwaysOn, PlotFreq->canvas(), true);
    ZoomerFreq->setMousePattern(QwtEventPattern::MouseSelect3, Qt::RightButton);
    QwtPlotPanner* PannerFreq = new QwtPlotPanner(PlotFreq->canvas());
    PannerFreq->setMouseButton(Qt::MidButton);

    // Lyapunov plot
    PlotLyapunov = new QwtPlot(QwtText("Lyapunov number"), ui.tab_l);
    PlotLyapunov->setAxisScaleEngine(QwtPlot::xBottom, new QwtLog10ScaleEngine);
    PlotLyapunov->setAxisScaleEngine(QwtPlot::yLeft, new QwtLog10ScaleEngine);
    PlotLyapunov->setAxisScaleEngine(QwtPlot::yRight, new QwtLog10ScaleEngine);
    CurveDeviation = new QwtPlotCurve("");
    CurveDeviation->setPen(QColor(Qt::red));
    CurveDeviation->attach(PlotLyapunov);
    CurveDeviation->setYAxis(QwtPlot::yRight);
    PlotLyapunov->enableAxis(QwtPlot::yRight);
    CurveLyapunov = new QwtPlotCurve("");
    CurveLyapunov->setPen(QColor(Qt::blue));
    CurveLyapunov->attach(PlotLyapunov);
    PlotLyapunov->show();
    /*QwtPlotPicker *PickerLyapunov = */new QwtPlotPicker2(QwtPlot::xBottom, QwtPlot::yLeft,
        QwtPicker::PointSelection, QwtPlotPicker::NoRubberBand, QwtPicker::AlwaysOn, PlotLyapunov->canvas());
    QVBoxLayout* LPlotLyapunov = new QVBoxLayout(ui.tab_l);
    LPlotLyapunov->addWidget(PlotLyapunov);

    // Frequency Map plot
    PlotFreqMap = new QwtPlot(QwtText("Frequency map"), ui.frame_fmarea);
    QwtSymbol sym;
    sym.setStyle(QwtSymbol::Rect);
    sym.setSize(2);
    sym.setPen(Qt::NoPen);
    sym.setBrush(QColor(Qt::red));
    CurveFreqMapC = new QwtPlotCurve("");
    CurveFreqMapC->setSymbol(sym);
    CurveFreqMapC->setStyle(QwtPlotCurve::Dots);
    CurveFreqMapC->attach(PlotFreqMap);
    sym.setBrush(QColor(Qt::blue));
    CurveFreqMapR = new QwtPlotCurve("");
    CurveFreqMapR->setSymbol(sym);
    CurveFreqMapR->setStyle(QwtPlotCurve::Dots);
    CurveFreqMapR->attach(PlotFreqMap);
    drawFreqMapResonances();
    PlotFreqMap->show();
    QVBoxLayout* LPlotFreqMap = new QVBoxLayout(ui.frame_fmarea);
    LPlotFreqMap->addWidget(PlotFreqMap);
    QwtPlotPicker *PickerFreqMap = new QwtPlotPicker2(QwtPlot::xBottom, QwtPlot::yLeft,
        QwtPicker::PointSelection, QwtPlotPicker::NoRubberBand, QwtPicker::AlwaysOn, PlotFreqMap->canvas());
    PickerFreqMap->setMousePattern(QwtEventPattern::MouseSelect1, Qt::RightButton);
    connect(PickerFreqMap, SIGNAL(selected(const QwtDoublePoint &)), SLOT(FMclicked(const QwtDoublePoint &)));
    ZoomerFreqMap = new QwtPlotZoomer2( QwtPlot::xBottom, QwtPlot::yLeft,
        QwtPicker::DragSelection | QwtPicker::CornerToCorner, QwtPicker::AlwaysOn, PlotFreqMap->canvas(), true);
    ZoomerFreqMap->setMousePattern(QwtEventPattern::MouseSelect3, Qt::RightButton, Qt::ControlModifier);
    ZoomerFreqMap->setMousePattern(QwtEventPattern::MouseSelect2, Qt::RightButton, Qt::ControlModifier);
    ZoomerFreqMap->setKeyPattern(QwtEventPattern::KeySelect1, Qt::Key_Space);
    QwtPlotPanner* PannerFreqMap = new QwtPlotPanner(PlotFreqMap->canvas());
    PannerFreqMap->setMouseButton(Qt::MidButton);

    PlotFreqMapHisto = new QwtPlot(QwtText("Histograms"), ui.frame_fmarea);
    PlotFreqMapHisto->setAxisScaleEngine(QwtPlot::xBottom, new QwtLog10ScaleEngine);
    CurveFreqMapHisto = new QwtPlotCurve("");
    CurveFreqMapHisto->setStyle(QwtPlotCurve::Steps);
    CurveFreqMapHisto->attach(PlotFreqMapHisto);
    LPlotFreqMap->addWidget(PlotFreqMapHisto);
    /*QwtPlotPicker *PickerFreqMapHisto = */new QwtPlotPicker2(QwtPlot::xBottom, QwtPlot::yLeft,
        QwtPicker::PointSelection, QwtPlotPicker::NoRubberBand, QwtPicker::AlwaysOn, PlotFreqMapHisto->canvas());
    PlotFreqMapHisto->hide();

    PlotFreqMapSS = new QwtPlot(QwtText("Start-space"), ui.frame_fmarea);
    sym.setStyle(QwtSymbol::Rect);
    sym.setSize(4);
    sym.setBrush(QColor(Qt::blue));
    CurveFreqMapSSR = new QwtPlotCurve("");
    CurveFreqMapSSR->setSymbol(sym);
    CurveFreqMapSSR->setStyle(QwtPlotCurve::Dots);
    CurveFreqMapSSR->attach(PlotFreqMapSS);
    sym.setBrush(QColor(Qt::red));
    CurveFreqMapSSC = new QwtPlotCurve("");
    CurveFreqMapSSC->setSymbol(sym);
    CurveFreqMapSSC->setStyle(QwtPlotCurve::Dots);
    CurveFreqMapSSC->attach(PlotFreqMapSS);
    QwtPlotPicker *PickerFreqMapSS = new QwtPlotPicker2(QwtPlot::xBottom, QwtPlot::yLeft,
        QwtPicker::PointSelection, QwtPlotPicker::NoRubberBand, QwtPicker::ActiveOnly, PlotFreqMapSS->canvas());
    PickerFreqMapSS->setMousePattern(QwtEventPattern::MouseSelect1, Qt::RightButton);
    connect(PickerFreqMapSS, SIGNAL(selected(const QwtDoublePoint &)), SLOT(FMclicked(const QwtDoublePoint &)));
    LPlotFreqMap->addWidget(PlotFreqMapSS);
    PlotFreqMapSS->hide();
    ui.progressBarFM->hide();

    // Schwarzschild plot
    QVBoxLayout* LPlotSchw = new QVBoxLayout(ui.frame_smarea);
    PlotSchw = new QwtPlot(QwtText("Schwarzschild model grid"), ui.frame_smarea);
    sym.setStyle(QwtSymbol::Diamond);
    sym.setSize(8);
    sym.setBrush(QColor(Qt::blue));
    CurveSchwCellF = new QwtPlotCurve("");
    CurveSchwCellF->setSymbol(sym);
    CurveSchwCellF->setStyle(QwtPlotCurve::Dots);
    CurveSchwCellF->attach(PlotSchw);
    sym.setBrush(QColor(Qt::red));
    CurveSchwCellI= new QwtPlotCurve("");
    CurveSchwCellI->setSymbol(sym);
    CurveSchwCellI->setStyle(QwtPlotCurve::Dots);
    CurveSchwCellI->attach(PlotSchw);
    QwtPlotPicker *PickerFreqMapSM = new QwtPlotPicker2(QwtPlot::xBottom, QwtPlot::yLeft,
        QwtPicker::PointSelection, QwtPlotPicker::NoRubberBand, QwtPicker::ActiveOnly, PlotSchw->canvas());
    PickerFreqMapSM->setMousePattern(QwtEventPattern::MouseSelect1, Qt::RightButton);
    connect(PickerFreqMapSM, SIGNAL(selected(const QwtDoublePoint &)), SLOT(SMclicked(const QwtDoublePoint &)));
    PlotSchw->show();
    LPlotSchw->addWidget(PlotSchw);
    ui.progressBarSM->hide();
    PlotSchwHisto = new QwtPlot(QwtText("Schwarzschild model orbit weights"), ui.frame_smarea);
    sym.setSize(2);
    sym.setStyle(QwtSymbol::Rect);
    sym.setBrush(QColor(Qt::blue));
    CurveSchwHistoR = new QwtPlotCurve("");
    CurveSchwHistoR->setStyle(QwtPlotCurve::Dots);
    CurveSchwHistoR->setSymbol(sym);
    CurveSchwHistoR->attach(PlotSchwHisto);
    sym.setBrush(QColor(Qt::red));
    CurveSchwHistoC = new QwtPlotCurve("");
    CurveSchwHistoC->setStyle(QwtPlotCurve::Dots);
    CurveSchwHistoC->setSymbol(sym);
    CurveSchwHistoC->attach(PlotSchwHisto);
    LPlotSchw->addWidget(PlotSchwHisto);
    QwtPlotPicker *PickerSchwHisto = new QwtPlotPicker2(QwtPlot::xBottom, QwtPlot::yLeft,
        QwtPicker::PointSelection, QwtPlotPicker::NoRubberBand, QwtPicker::AlwaysOn, PlotSchwHisto->canvas());
    PickerSchwHisto->setMousePattern(QwtEventPattern::MouseSelect1, Qt::RightButton);
    connect(PickerSchwHisto, SIGNAL(selected(const QwtDoublePoint &)), SLOT(SMHistoClicked(const QwtDoublePoint &)));
    PlotSchwHisto->hide();

    // connect various events to change corresponding parameters in core
    connect(ui.lineEdit_x, SIGNAL(textChanged(const QString&)), SLOT(ICchanged()));
    connect(ui.lineEdit_y, SIGNAL(textChanged(const QString&)), SLOT(ICchanged()));
    connect(ui.lineEdit_z, SIGNAL(textChanged(const QString&)), SLOT(ICchanged()));
    connect(ui.lineEdit_vx, SIGNAL(textChanged(const QString&)), SLOT(ICchanged()));
    connect(ui.lineEdit_vy, SIGNAL(textChanged(const QString&)), SLOT(ICchanged()));
    connect(ui.lineEdit_vz, SIGNAL(textChanged(const QString&)), SLOT(ICchanged()));
    connect(ui.lineEdit_E, SIGNAL(textChanged(const QString&)), SLOT(ICchanged()));
    connect(ui.lineEdit_q, SIGNAL(textChanged(const QString&)), SLOT(potentialChanged()));
    connect(ui.lineEdit_p, SIGNAL(textChanged(const QString&)), SLOT(potentialChanged()));
    connect(ui.lineEdit_Mbh, SIGNAL(textChanged(const QString&)), SLOT(potentialChanged()));
    connect(ui.lineEdit_Rc, SIGNAL(textChanged(const QString&)), SLOT(potentialChanged()));
    connect(ui.lineEdit_Gamma, SIGNAL(textChanged(const QString&)), SLOT(potentialChanged()));
    connect(ui.lineEdit_Omega, SIGNAL(textChanged(const QString&)), SLOT(potentialChanged()));
    connect(ui.lineEdit_Ncoefs_radial, SIGNAL(textChanged(const QString&)), SLOT(potentialChanged()));
    connect(ui.lineEdit_Ncoefs_angular, SIGNAL(textChanged(const QString&)), SLOT(potentialChanged()));
    connect(ui.lineEdit_Alpha, SIGNAL(textChanged(const QString&)), SLOT(potentialChanged()));
    connect(ui.lineEdit_eps, SIGNAL(textChanged(const QString&)), SLOT(potentialChanged()));
    connect(ui.lineEdit_tol, SIGNAL(textChanged(const QString&)), SLOT(potentialChanged()));
    connect(ui.selectPotentialType, SIGNAL(currentIndexChanged(int)), SLOT(potentialChanged()));
    connect(ui.selectDensityType, SIGNAL(currentIndexChanged(int)), SLOT(potentialChanged()));
    connect(ui.selectSymmetryType, SIGNAL(currentIndexChanged(int)), SLOT(potentialChanged()));
    connect(ui.buttonChooseNbodyFile, SIGNAL(clicked()), SLOT(chooseNbodyFile()));
    connect(ui.lineEdit_NbodyFile, SIGNAL(returnPressed()), SLOT(typeinNbodyFile()));
    connect(ui.radioButton_2dorbit, SIGNAL(toggled(bool)), SLOT(redrawOrbit(bool)));
    connect(ui.radioButton_3dline, SIGNAL(toggled(bool)), SLOT(redrawOrbit(bool)));
    connect(ui.radioButton_3dmesh, SIGNAL(toggled(bool)), SLOT(redrawOrbit(bool)));
    connect(ui.radioButton_opPericenter, SIGNAL(toggled(bool)), SLOT(redrawOrbit(bool)));
    connect(ui.radioButton_opLPeri, SIGNAL(toggled(bool)), SLOT(redrawOrbit(bool)));
    connect(ui.radioButton_opL, SIGNAL(toggled(bool)), SLOT(redrawOrbit(bool)));
    connect(ui.radioButton_2d_xy, SIGNAL(toggled(bool)), SLOT(redrawOrbit(bool)));
    connect(ui.radioButton_2d_xz, SIGNAL(toggled(bool)), SLOT(redrawOrbit(bool)));
    connect(ui.radioButton_2d_yz, SIGNAL(toggled(bool)), SLOT(redrawOrbit(bool)));
    connect(ui.radioButton_fmmapdiff, SIGNAL(toggled(bool)), SLOT(redrawFreqMap(bool)));
    connect(ui.radioButton_fmmaplyapunov, SIGNAL(toggled(bool)), SLOT(redrawFreqMap(bool)));
    connect(ui.radioButton_fmhistodiff, SIGNAL(toggled(bool)), SLOT(redrawFreqMap(bool)));
    connect(ui.radioButton_fmhistolyapunov, SIGNAL(toggled(bool)), SLOT(redrawFreqMap(bool)));
    connect(ui.radioButton_fmssdiff, SIGNAL(toggled(bool)), SLOT(redrawFreqMap(bool)));
    connect(ui.radioButton_fmsslyapunov, SIGNAL(toggled(bool)), SLOT(redrawFreqMap(bool)));
    connect(ui.radioButton_fmspdiff, SIGNAL(toggled(bool)), SLOT(redrawFreqMap(bool)));
    connect(ui.radioButton_fmsplyapunov, SIGNAL(toggled(bool)), SLOT(redrawFreqMap(bool)));
    connect(ui.radioButton_fmsydiff, SIGNAL(toggled(bool)), SLOT(redrawFreqMap(bool)));
    connect(ui.radioButton_fmsylyapunov, SIGNAL(toggled(bool)), SLOT(redrawFreqMap(bool)));
    connect(ui.radioButton_smviewgrid, SIGNAL(toggled(bool)), SLOT(redrawSchwModel(bool)));
    connect(ui.radioButton_smviewweights, SIGNAL(toggled(bool)), SLOT(redrawSchwModel(bool)));
    connect(ui.radioButton_smviewhisto, SIGNAL(toggled(bool)), SLOT(redrawSchwModel(bool)));
    connect(ui.lineEdit_intTime, SIGNAL(textChanged(const QString&)), SLOT(intTime_changed()));
    connect(ui.lineEdit_timeStep,SIGNAL(textChanged(const QString&)), SLOT(intTime_changed()));
    connect(ui.lineEdit_fms, SIGNAL(textChanged(const QString&)), SLOT(fmparams_changed()));
    connect(ui.lineEdit_fmp, SIGNAL(textChanged(const QString&)), SLOT(fmparams_changed()));
    connect(ui.lineEdit_fmy, SIGNAL(textChanged(const QString&)), SLOT(fmparams_changed()));
    connect(ui.lineEdit_fmr, SIGNAL(textChanged(const QString&)), SLOT(fmparams_changed()));
    connect(ui.lineEdit_smNumOrbits, SIGNAL(textChanged(const QString&)), SLOT(smparams_changed()));
    connect(ui.select_smType, SIGNAL(currentIndexChanged(int)), SLOT(smparams_changed()));
    connect(ui.lineEdit_smNumLines, SIGNAL(textChanged(const QString&)), SLOT(smparams_changed()));
    connect(ui.lineEdit_smNumAngCoefs, SIGNAL(textChanged(const QString&)), SLOT(smparams_changed()));
    connect(ui.lineEdit_smNumRadCoefs, SIGNAL(textChanged(const QString&)), SLOT(smparams_changed()));
    connect(ui.lineEdit_smAlpha, SIGNAL(textChanged(const QString&)), SLOT(smparams_changed()));
    connect(ui.lineEdit_smNumShells, SIGNAL(textChanged(const QString&)), SLOT(smparams_changed()));
    connect(ui.lineEdit_smInnerShellMass, SIGNAL(textChanged(const QString&)), SLOT(smparams_changed()));
    connect(ui.lineEdit_smOuterShellMass, SIGNAL(textChanged(const QString&)), SLOT(smparams_changed()));
    connect(ui.lineEdit_smNumSamplingPoints, SIGNAL(textChanged(const QString&)), SLOT(smparams_changed()));
    connect(ui.lineEdit_smConstraintPenalty, SIGNAL(textChanged(const QString&)), SLOT(smparams_changed()));
    connect(ui.lineEdit_smQuadraticPenalty, SIGNAL(textChanged(const QString&)), SLOT(smparams_changed()));
    connect(ui.lineEdit_smChaoticWeight, SIGNAL(textChanged(const QString&)), SLOT(smparams_changed()));
    connect(ui.checkBox_smBeta, SIGNAL(toggled(bool)), SLOT(smparams_changed()));
    connect(ui.lineEdit_smBetaIn, SIGNAL(textChanged(const QString&)), SLOT(smparams_changed()));
    connect(ui.lineEdit_smBetaOut, SIGNAL(textChanged(const QString&)), SLOT(smparams_changed()));
    connect(ui.lineEdit_smMaxOrbitWeight, SIGNAL(textChanged(const QString&)), SLOT(smparams_changed()));
    connect(ui.lineEdit_MinFreqDiff, SIGNAL(textChanged(const QString&)), SLOT(ccparams_changed()));
    connect(ui.lineEdit_MinLambda, SIGNAL(textChanged(const QString&)), SLOT(ccparams_changed()));
    ui.lineEdit_x->installEventFilter(this);
    ui.lineEdit_y->installEventFilter(this);
    ui.lineEdit_z->installEventFilter(this);
    ui.lineEdit_vx->installEventFilter(this);
    ui.lineEdit_vy->installEventFilter(this);
    ui.lineEdit_vz->installEventFilter(this);
    ui.lineEdit_q->installEventFilter(this);
    ui.lineEdit_p->installEventFilter(this);
    ui.lineEdit_Mbh->installEventFilter(this);
    ui.lineEdit_Rc->installEventFilter(this);
    ui.lineEdit_intTime->installEventFilter(this);
    ui.lineEdit_timeStep->installEventFilter(this);
    ui.checkBox_Lyapunov->installEventFilter(this);
    ui.radioButton_2d->installEventFilter(this);
    ui.radioButton_3d->installEventFilter(this);
    ui.radioButton_2dorbit->installEventFilter(this);
    ui.radioButton_3dline->installEventFilter(this);
    ui.radioButton_3dmesh->installEventFilter(this);
    ui.ButtonClearPS->installEventFilter(this);
    ui.ButtonRandom->installEventFilter(this);
    ui.tab_f->installEventFilter(this);
    ui.tab_p->installEventFilter(this);
    ui.tab_o->installEventFilter(this);
    ui.tabWidget->installEventFilter(this);
    PlotOrbit->installEventFilter(this);
    PlotPoincare->installEventFilter(this);
    PlotFreq->installEventFilter(this);
    PlotFreqMap->installEventFilter(this);

    // load and apply settings
    loadApplySettings();
}

CSmileGUI::~CSmileGUI()
{
    if(ui.checkBox_saveSettings->isChecked())
        core->saveSettings();
    delete core;
}

bool CSmileGUI::loadApplySettings(QString fileName)
{
    ui.label_Omega->setVisible(false);  // !!! no full support for rotating potential yet...
    ui.lineEdit_Omega->setVisible(false);

    bool result=core->loadSettings(fileName);
    InternalRecalc=true;
    // find potential name by its value
    ui.selectPotentialType->setCurrentIndex(ui.selectPotentialType->findText(QString::fromStdString(PotentialNames[core->configPotential.PotentialType])));
    ui.selectDensityType  ->setCurrentIndex(ui.selectDensityType  ->findText(QString::fromStdString(DensityNames  [core->configPotential.DensityType])));
    ui.selectSymmetryType ->setCurrentIndex(ui.selectSymmetryType ->findText(QString::fromStdString(SymmetryNames [core->configPotential.SymmetryType])));
    ui.select_smType->setCurrentIndex(ui.select_smType->findText(QString::fromStdString(getSchwDataNameByType(core->configCore.sm_densDataType))));
    ui.lineEdit_q->setText(QString::number(core->configPotential.q));
    ui.lineEdit_p->setText(QString::number(core->configPotential.p));
    ui.lineEdit_Mbh->setText(QString::number(core->configPotential.Mbh));
    ui.lineEdit_Rc->setText(QString::number(core->configPotential.Rc));
    ui.lineEdit_Gamma->setText(QString::number(core->configPotential.Gamma));
    ui.lineEdit_Omega->setText(QString::number(core->configPotential.Omega));
    ui.lineEdit_Ncoefs_radial->setText(QString::number(core->configPotential.Ncoefs_radial));
    ui.lineEdit_Ncoefs_angular->setText(QString::number(core->configPotential.Ncoefs_angular));
    ui.lineEdit_Alpha->setText(QString::number(core->configPotential.Alpha));
    ui.lineEdit_NbodyFile->setText(core->NbodyFile);
    ui.lineEdit_eps->setText(QString::number(core->configPotential.nbpotEps));
    ui.lineEdit_tol->setText(QString::number(core->configPotential.nbpotTol));
    ui.lineEdit_x->setText(QString::number(core->InitData.initCond.Pos[0]));
    ui.lineEdit_y->setText(QString::number(core->InitData.initCond.Pos[1]));
    ui.lineEdit_z->setText(QString::number(core->InitData.initCond.Pos[2]));
    ui.lineEdit_vx->setText(QString::number(core->InitData.initCond.Vel[0]));
    ui.lineEdit_vy->setText(QString::number(core->InitData.initCond.Vel[1]));
    ui.lineEdit_vz->setText(QString::number(core->InitData.initCond.Vel[2]));
    ui.lineEdit_E->setText(QString::number(core->initCondE));
    if(core->useICe)
    {
        ui.lineEdit_E->setEnabled(true);
        ui.groupBox_ICx->setEnabled(false);
    }
    else
    {
        ui.lineEdit_E->setEnabled(false);
        ui.groupBox_ICx->setEnabled(true);
    }
    ui.lineEdit_intTime->setText(QString::number(core->configCore.intTimeInPeriods));
    ui.lineEdit_timeStep->setText(QString::number(core->configCore.intTimeStepsPerPeriod));
    ui.checkBox_Lyapunov->setChecked(core->InitData.calcLyapunov);
    ui.checkBox_usePS->setChecked(core->configCore.usePS);
    ui.lineEdit_fms->setText(QString::number(core->configCore.fm_numOrbitsStationary));
    ui.lineEdit_fmp->setText(QString::number(core->configCore.fm_numOrbitsPrincipalPlane));
    ui.lineEdit_fmy->setText(QString::number(core->configCore.fm_numOrbitsYalpha));
    ui.lineEdit_fmr->setText(QString::number(core->configCore.fm_numOrbitsRandom));
    ui.lineEdit_smNumOrbits->setText(QString::number(core->configCore.sm_numOrbits));
    ui.lineEdit_smNumLines->setText(QString::number(core->configCore.sm_linesPerSegment));
    ui.lineEdit_smNumAngCoefs->setText(QString::number(core->configCore.sm_numAngularCoefs));
    ui.lineEdit_smNumRadCoefs->setText(QString::number(core->configCore.sm_numRadialCoefs));
    ui.lineEdit_smAlpha->setText(QString::number(core->configCore.sm_Alpha));
    ui.lineEdit_smNumShells->setText(QString::number(core->configCore.sm_numShells));
    ui.lineEdit_smInnerShellMass->setText(QString::number(core->configCore.sm_innerShellMass));
    ui.lineEdit_smOuterShellMass->setText(QString::number(core->configCore.sm_outerShellMass));
    ui.lineEdit_smNumSamplingPoints->setText(QString::number(core->configCore.sm_numSamplingPoints));
    ui.selectEnergyLevel_sm->setMaximum(core->configCore.sm_numShells);
    ui.lineEdit_MinFreqDiff->setText(QString::number(core->configCore.chaoticMinFreqDiff));
    ui.lineEdit_MinLambda->setText(QString::number(core->configCore.chaoticMinLambda));
    ui.lineEdit_smConstraintPenalty->setText(QString::number(core->configCore.sm_constraintPenalty));
    ui.lineEdit_smQuadraticPenalty->setText(QString::number(core->configCore.sm_regularization));
    ui.lineEdit_smChaoticWeight->setText(QString::number(core->configCore.sm_chaoticPenalty));
    ui.lineEdit_smMaxOrbitWeight->setText(QString::number(core->configCore.sm_maxWeight));
    ui.lineEdit_smBetaIn->setText(QString::number(core->configCore.sm_betaIn));
    ui.lineEdit_smBetaOut->setText(QString::number(core->configCore.sm_betaOut));
    ui.checkBox_smBeta->setChecked(core->configCore.sm_constrainBeta);
    ui.lineEdit_smBetaIn->setEnabled(core->configCore.sm_constrainBeta);
    ui.lineEdit_smBetaOut->setEnabled(core->configCore.sm_constrainBeta);
#ifndef HAVE_CVXOPT
    ui.ButtonSchwQP->setVisible(core->configCore.sm_useBPMPD);  // disable QP button if we are not using BPMPD and the program was compiled without CVXOPT support
#endif
    if(core->useICe) ui.radioButton_ICe->setChecked(true); else ui.radioButton_ICe->setChecked(false);
    if(core->potential->N_dim==2) ui.radioButton_2d->setChecked(true); else ui.radioButton_3d->setChecked(true);
    InternalRecalc=false;
    potentialChanged();
    ui.buttonInitPotential->hide();   // this button will be shown only when parameters changed
    smparams_changed();
    ICchanged();
    return result;
}

bool CSmileGUI::eventFilter(QObject* target, QEvent* Event)
{
    if(Event->type() == QEvent::KeyPress) 
    {  
        QKeyEvent* keyEvent = static_cast<QKeyEvent*>(Event);
        if((keyEvent->key() == Qt::Key_Return) || (keyEvent->key() == Qt::Key_Enter))
        {
            on_ButtonStart_clicked();
            return true;
        }
    }
    return QMainWindow::eventFilter(target, Event);
}

void CSmileGUI::timerEventCore()
{
    if(core->orbitlib==NULL) return;  // shouldn't happen
    // update both progress bars; only the one for the currently active operation is visible
    ui.progressBarFM->setValue(static_cast<int>(core->orbitlib->numComplete()));
    ui.progressBarSM->setValue(static_cast<int>(core->orbitlib->numComplete()));
    // also repaint frequency map if it is visible
    if(ui.tabWidget->currentWidget()==ui.tab_fm)
        redrawFreqMap(true);
}

void CSmileGUI::timerEvent(QTimerEvent * Event)
{
    if(Event->timerId()==myTimerOrbit && core->isOrbitIntRunning())
    {
        ui.progressBarOrbit->setValue(core->orbit->getIntTime()*100 / (core->configCore.intTimeInPeriods*core->InitData.timeUnit));
        ui.progressBarOrbit->show();
        redrawOrbit(true);
    }
}

/* ----------------------------  single orbit integration --------- */
void CSmileGUI::on_ButtonStart_clicked()
{
    if(core->isOrbitIntRunning()) { 
        core->orbit->halt();    // stop current computation
        return; 
    }
    if(core->InitData.timeUnit<=0 || core->configCore.intTimeInPeriods<=0 || core->configCore.intTimeStepsPerPeriod<=0) 
    {
        error("Total energy is outside allowed range");
        return;  // error
    }
    ui.groupBox_N_dim->setEnabled(false);
    ui.groupBox_potential->setEnabled(false);
    ui.ButtonStart->setText("Stop!");
    ui.ButtonRandom->setEnabled(false);
    core->startOrbit();
    timeOI.start();
    myTimerOrbit = startTimer(2000);  // to refresh orbit plot
}

void CSmileGUI::CalcFinished()    // called from core->coreOrbitFinished()
{
    if(!core->isOrbitLibRunning() && !core->isExportNbodyRunning() && !core->isSchwOptRunning()) {
        ui.groupBox_N_dim->setEnabled(true);
        ui.groupBox_potential->setEnabled(true);
    }
    ui.ButtonStart->setText("Start");
    ui.ButtonRandom->setEnabled(true);
    timeOIelapsed=timeOI.elapsed();  // store to re-use in subsequent "refresh spectrum" events
    killTimer(myTimerOrbit);
    ui.progressBarOrbit->hide();
    const COrbitRuntimeTrajectory* trj = static_cast<const COrbitRuntimeTrajectory*>(core->orbit->getRuntimeFnc(0));   // 0th runtime function is the orbit trajectory analyzer
    assert(trj->FncType()==COrbitRuntimeTrajectory::FT_TRAJ_ANALYSIS);
    const COrbitInitData<double> InitData=core->orbit->getInitData();

    // Poincare section display (if it was used)
    const COrbitRuntimePoincare* tps=NULL;
    for(size_t rf=0; rf<core->orbit->getInfoNum(); rf++)
        if(core->orbit->getRuntimeFnc(rf)->FncType() == CBasicOrbitRuntimeFnc::FT_POINCARE)
            tps = static_cast<const COrbitRuntimePoincare* >(core->orbit->getRuntimeFnc(rf));
    if(tps!=NULL && tps->size()>0)
    {
        QwtPlotCurve* CurvePS = new QwtPlotCurve();
        QwtSymbol sym;
        sym.setStyle(QwtSymbol::Rect);
        sym.setPen(Qt::NoPen);
        sym.setBrush(QColor(rand()%224, rand()%192, rand()%224));
        sym.setSize(2);
        CurvePS->setSymbol(sym);
        CurvePS->setStyle(QwtPlotCurve::Dots);
        CurvePS->attach(PlotPoincare);
        unsigned int nPS=tps->size();
        vectord psx(nPS), psvx(nPS);
        for(unsigned int i=0; i<nPS; i++)
        {
            psx[i]=tps->at(i).first;
            psvx[i]=tps->at(i).second;
        }
        // remove closest
        bool actiontaken;
        if(psx.size()>100)
        {
            double maxdistx=0, maxdistvx=0;
            for(unsigned int k1=1; k1<psx.size(); k1++)
            {
                double x1=psx[k1], y1=psvx[k1];
                for(unsigned int k2=0; k2<k1; k2++)
                {
                    double d=fabs(psx[k2]-x1);
                    if(d>maxdistx)
                        maxdistx=d;
                    d=fabs(psvx[k2]-y1);
                    if(d>maxdistvx) maxdistvx=d;
                }
            }
          do{
            double mindist=1000;
            unsigned int minp1=0, minp2=0;
            for(unsigned int k1=1; k1<psx.size(); k1++)
            {
                double x1=psx[k1], y1=psvx[k1];
                for(unsigned int k2=0; k2<k1; k2++)
                {
                    double d=fabs(psx[k2]-x1)/maxdistx + fabs(psvx[k2]-y1)/maxdistvx;
                    if(d<mindist)
                    {
                        mindist=d;
                        minp1=k1;
                        minp2=k2;
                    }
                }
            }
            if(mindist<1e-3)
            {
                psx.erase(psx.begin()+minp1);
                psvx.erase(psvx.begin()+minp1);
                actiontaken=true;
            }
            else
                actiontaken=false;
          }while(actiontaken);
        }
        CurvePS->setData(&(psx.front()), &(psvx.front()), static_cast<int>(psx.size()));
        PlotPoincare->replot();
        CurvesPoincare<<(CurvePS);
    }

#ifdef HAVE_QWT3D
#ifdef USE3DFACETS
    if(core->potential->N_dim==3)
    {
        // calc max distance between adjacent points (for triangulation purposes)
        maxdist=trj->getMaxDist() * 1.01;
        ui.radioButton_3dmesh_full->setChecked(true);
        ui.lineEdit_3dmesh_maxdist->setText(QString::number(maxdist, 'g', 3));
        on_pushButton_refresh3dmesh_clicked();
    }
#endif
#endif

    // perform frequency graphs and orbit classification (also set label text and redraw orbit)
    if(ui.spinBox_intervalCount->value()!=1)
        ui.spinBox_intervalCount->setValue(1);  // refreshes frequency graph and calls for orbit classification
    else
        on_spinBox_intervalCount_valueChanged(1);
    // lyapunov
    if(InitData.calcLyapunov)
    {
      size_t nActual=trj->getTrajSize();
      if(nActual<100)
      {
        CurveLyapunov->setData(NULL, NULL, 0);
        CurveDeviation->setData(NULL, NULL, 0);
        PlotLyapunov->replot();
      }
      else
      {
        vectord timesteps(nActual);
        vectord lambdatmp(nActual);
        double devvecmin=1e4;
        for(size_t s=100; s<nActual; s++) 
        {
            timesteps[s] = s*InitData.timeStep/InitData.timeUnit;
            double lnw=trj->getLnDevVec(s);
            lambdatmp[s] = std::max<double>(lnw/timesteps[s], 1e-4);
            devvecmin = std::min<double>(lnw - log(timesteps[s]), devvecmin);
        }
        double devvecadd=(devvecmin<0.1?0.1-devvecmin:0);
        CurveLyapunov->setData(&(timesteps.at(100)), &(lambdatmp.at(100)), static_cast<int>(nActual)-100);
        for(size_t s=100; s<nActual; s++) 
            lambdatmp[s] = trj->getLnDevVec(s) - log(timesteps[s]) + devvecadd;
        CurveDeviation->setData(&(timesteps.at(100)), &(lambdatmp.at(100)), static_cast<int>(nActual)-100);
        //for(int i=0; i<PlotLyapunov->axisCnt; i++) PlotLyapunov->setAxisAutoScale(i); 
        PlotLyapunov->setAxisScale(QwtPlot::yLeft, 1e-4, 10);
        PlotLyapunov->setAxisScale(QwtPlot::yRight, 0.1, 1e4);
        PlotLyapunov->setAxisScale(QwtPlot::xBottom, 1, timesteps.back()*1.05);
        PlotLyapunov->replot();
      }
    }
}

/* ----------------------------  orbit rendering --------- */
void CSmileGUI::redrawOrbit(bool show)
{
    if(!show || !core->orbit || !core->orbit->getIntTime()) return;

    const COrbitRuntimeTrajectory* trj = static_cast<const COrbitRuntimeTrajectory*>(core->orbit->getRuntimeFnc(0));   // 0th runtime function is the orbit trajectory analyzer
    assert(trj->FncType()==COrbitRuntimeTrajectory::FT_TRAJ_ANALYSIS);
    size_t nActual=trj->getTrajSize();
    const COrbitInitData<double> InitData=core->orbit->getInitData();

    size_t intNum = ui.spinBox_IntervalNum->value()-1;
    size_t intCount = ui.spinBox_intervalCount->value();
    size_t start=nActual * intNum / intCount;
    size_t count=nActual / intCount; 
    if(start+count>nActual) count=nActual-start;
    ui.groupBox_2dplane->setEnabled(true);
    if(ui.radioButton_2dorbit->isChecked())
    {   // draw 2d orbit projection
        ui.groupBox_3dmesh->setEnabled(false);
        vectord hor(count), ver(count);
        size_t indh=0, indv=0;
        if(InitData.potential->N_dim==2)
        {
            indh=0; indv=1;
        }
        else
        {
            if(ui.radioButton_2d_yz->isChecked()) { indh=1; indv=2; }
            if(ui.radioButton_2d_xz->isChecked()) { indh=0; indv=2; }
            if(ui.radioButton_2d_xy->isChecked()) { indh=0; indv=1; }
        }
        for(size_t i=0; i<count; i++)
        {
            const CPosVelPoint<double>& pt = trj->getTraj(i+start);
            hor[i]=pt.Pos[indh];
            ver[i]=pt.Pos[indv];
        }
        if(count>0)
            CurveOrbit->setData(&(hor.front()), &(ver.front()), static_cast<int>(count));
        else
            CurveOrbit->setData(NULL, NULL, 0);
        // equipotential surface
        double eqh[NUM_POINTS_PLOT_EQUIPOTENTIAL*4], eqv[NUM_POINTS_PLOT_EQUIPOTENTIAL*4];
        double E = core->initCondE;
        int num_points_plot_equipotential = (InitData.potential->PotentialType()==CDensity::PT_NB)?NUM_POINTS_PLOT_EQUIPOTENTIAL/5:NUM_POINTS_PLOT_EQUIPOTENTIAL;  
        CDensity::SYMMETRYTYPE sym=InitData.potential->symmetry();
        for(int i=0; i<num_points_plot_equipotential; i++)
        {
            double angle=(i+0.5)*M_PI/2/num_points_plot_equipotential;
            double scrcoords[4][3] = { // four directions(quadrants) and three coords: horizontal, vertical and unused
                { cos(angle), sin(angle), 0}, 
                {-cos(angle), sin(angle), 0}, 
                {-cos(angle),-sin(angle), 0}, 
                { cos(angle),-sin(angle), 0} 
            };
            int coordindex[3]={0,0,0};  // X_k = scrcoords[quadrant][coordindex[k]];
            if(ui.radioButton_2d_xy->isChecked()) { coordindex[0]=0; coordindex[1]=1; coordindex[2]=2; }
            if(ui.radioButton_2d_yz->isChecked()) { coordindex[0]=2; coordindex[1]=0; coordindex[2]=1; }
            if(ui.radioButton_2d_xz->isChecked()) { coordindex[0]=0; coordindex[1]=2; coordindex[2]=1; }
            double r0=InitData.potential->findintersection(E, scrcoords[0][coordindex[0]], scrcoords[0][coordindex[1]], scrcoords[0][coordindex[2]]);
            double r1=r0, r2=r0, r3=r0;
            if((sym & CDensity::ST_TRIAXIAL) == 0)  // no symmetry w.r.t. changing sign of any coordinate
            {
                r1=InitData.potential->findintersection(E, scrcoords[1][coordindex[0]], scrcoords[1][coordindex[1]], scrcoords[1][coordindex[2]]);
                if((sym & CDensity::ST_REFLECTION) == 0)  // no symmetry w.r.t. reflection of all three coords
                {
                    r2=InitData.potential->findintersection(E, scrcoords[2][coordindex[0]], scrcoords[2][coordindex[1]], scrcoords[2][coordindex[2]]);
                    r3=InitData.potential->findintersection(E, scrcoords[3][coordindex[0]], scrcoords[3][coordindex[1]], scrcoords[3][coordindex[2]]);
                }
                else {r2=r0; r3=r1; }
            }
            eqh[i]=scrcoords[0][0]*r0; 
            eqv[i]=scrcoords[0][1]*r0;
            eqh[num_points_plot_equipotential*2-1-i]=scrcoords[1][0]*r1;
            eqv[num_points_plot_equipotential*2-1-i]=scrcoords[1][1]*r1;
            eqh[num_points_plot_equipotential*2+i]  =scrcoords[2][0]*r2;
            eqv[num_points_plot_equipotential*2+i]  =scrcoords[2][1]*r2;
            eqh[num_points_plot_equipotential*4-1-i]=scrcoords[3][0]*r3;
            eqv[num_points_plot_equipotential*4-1-i]=scrcoords[3][1]*r3;
        }
        CurveEquipotential->setData(eqh, eqv, 4*num_points_plot_equipotential);
        for(int i=0; i<PlotOrbit->axisCnt; i++) PlotOrbit->setAxisAutoScale(i); 
        PlotOrbit->replot();
        ZoomerOrbit->setZoomBase(false);
#ifdef HAVE_QWT3D
        Plot3d->hide();
#endif
        PlotOrbitParam->hide();
        PlotOrbit->show();
    }
    else if(ui.radioButton_3dline->isChecked() || ui.radioButton_3dmesh->isChecked())
    {   // draw 3d orbit
        ui.groupBox_2dplane->setEnabled(false);
        ui.groupBox_3dmesh->setEnabled(false);
#ifdef HAVE_QWT3D 
        if(ui.radioButton_3dline->isChecked())
        {
            Qwt3D::Cell trjcell;
            trjpos.clear();
            trjcells.clear();
            trjcell.clear();
            for(size_t i=0; i<count; i++)
            {
                const CPosVelPoint<double>& pt = trj->getTraj(i+start);
                trjpos.push_back(Qwt3D::Triple(pt.Pos[0], pt.Pos[1], pt.Pos[2]));
                trjcell.push_back(static_cast<unsigned int>(i));
            }
            trjcells.push_back(trjcell);
            // draw trajectory
            Plot3d->loadFromData(trjpos, trjcells);
            Plot3d->setPlotStyle(Qwt3D::WIREFRAME);
        }
#ifdef USE3DFACETS
        if(ui.radioButton_3dmesh->isChecked())
        {
            // draw mesh
            if(TriangThr->restart)   // triangulation is in process, so do not display anything
                Plot3d->loadFromData(Qwt3D::TripleField(), Qwt3D::CellField());
            else
                Plot3d->loadFromData(TriangThr->trjpos, TriangThr->facets);
            Plot3d->setPlotStyle(Qwt3D::FILLEDMESH);
            ui.groupBox_3dmesh->setEnabled(true);
        }
#endif
        Plot3d->updateData();
        Plot3d->updateGL();
        Plot3d->show();
        PlotOrbitParam->hide();
        PlotOrbit->hide();
#endif
    }
    else
    {   // orbit params distribution
        ui.groupBox_2dplane->setEnabled(false);
        ui.groupBox_3dmesh->setEnabled(false);
        vectord data;
        const COrbitRuntimePericenter* tpc=NULL;
        for(size_t rf=0; rf<core->orbit->getInfoNum(); rf++)
            if(core->orbit->getRuntimeFnc(rf)->FncType() == CBasicOrbitRuntimeFnc::FT_PERICENTER)
                tpc = static_cast<const COrbitRuntimePericenter* >(core->orbit->getRuntimeFnc(rf));
        if(ui.radioButton_opLPeri->isChecked())
        {
            if(tpc!=NULL)
                for(size_t i=0; i<tpc->size(); i++)
                    data.push_back(tpc->getl2(i));
        }
        else if(ui.radioButton_opPericenter->isChecked())
        {
            if(tpc!=NULL)
                for(size_t i=0; i<tpc->size(); i++)
                    data.push_back(tpc->getr(i));
        }
        else if(ui.radioButton_opL->isChecked())
        {
            data.resize(nActual);
            for(size_t i=0; i<nActual; i++)
            {
                const CPosVelPoint<double>& pt = trj->getTraj(i);
                data[i] = pow_2(pt.Pos[1]*pt.Vel[2] - pt.Pos[2]*pt.Vel[1]) +
                        + pow_2(pt.Pos[2]*pt.Vel[0] - pt.Pos[0]*pt.Vel[2]) +
                        + pow_2(pt.Pos[0]*pt.Vel[1] - pt.Pos[1]*pt.Vel[0]);
            }
        }
        size_t nump=data.size();
        if(nump>0)
        {
            std::sort(data.begin(), data.end());
            vectord frac(nump);
            for(size_t i=0; i<nump; i++)
                frac[i]=i*1.0/nump;
            CurveOrbitParam->setData(&(frac.front()), &(data.front()), static_cast<int>(nump));
        }
        else CurveOrbitParam->setData(NULL, NULL, 0);
        for(int i=0; i<PlotOrbitParam->axisCnt; i++) PlotOrbitParam->setAxisAutoScale(i); 
        //PlotOrbitParam->replot();
        ZoomerOrbitParam->setZoomBase(true);
#ifdef HAVE_QWT3D
        Plot3d->hide();
#endif
        PlotOrbit->hide();
        PlotOrbitParam->show();
    }
}

void CSmileGUI::on_pushButton_refresh3dmesh_clicked()
{
#ifdef USE3DFACETS
    bool ok;
    double md = ui.lineEdit_3dmesh_maxdist->text().toDouble(&ok);   
    if(ok) maxdist=md;
    int meshstyle=0;
    if(ui.radioButton_3dmesh_x->isChecked()) meshstyle=1; else
    if(ui.radioButton_3dmesh_y->isChecked()) meshstyle=2; else
    if(ui.radioButton_3dmesh_z->isChecked()) meshstyle=3; 
    std::vector<CPosPoint<double> > trajp;
    assert(core->orbit);
    const COrbitRuntimeTrajectory* trj = static_cast<const COrbitRuntimeTrajectory*>(core->orbit->getRuntimeFnc(0));   // 0th runtime function is the orbit trajectory analyzer
    assert(trj->FncType()==COrbitRuntimeTrajectory::FT_TRAJ_ANALYSIS);
    size_t nActual=trj->getTrajSize();
    trajp.reserve(nActual);
    for(size_t i=0; i<nActual; i++)
        trajp.push_back(trj->getTraj(i));
    TriangThr->startTriangulation(trajp, maxdist, meshstyle);
#endif
}

/* ----------------------------  spectrum rendering --------- */
void CSmileGUI::on_spinBox_intervalCount_valueChanged(int count)
{
    if(InternalRecalc) return;
    bool doNotRefresh = (ui.spinBox_IntervalNum->value() > count);
    ui.spinBox_IntervalNum->setMaximum(count);
    if(doNotRefresh) return;   // setting maximum more than current value led to change of value of IntervalNum and hence the refresh function was already called
    if(ui.spinBox_IntervalNum->value()!=1)
        ui.spinBox_IntervalNum->setValue(1);  // refreshes frequency graph and calls for orbit classification
    else
        refreshOrbitAndSpectrum();
}

void CSmileGUI::on_spinBox_IntervalNum_valueChanged(int)
{
    refreshOrbitAndSpectrum();
}

void CSmileGUI::refreshOrbitAndSpectrum() 
{
    const COrbitRuntimeTrajectory* trj = static_cast<const COrbitRuntimeTrajectory*>(core->orbit->getRuntimeFnc(0));   // 0th runtime function is the orbit trajectory analyzer
    assert(trj->FncType()==COrbitRuntimeTrajectory::FT_TRAJ_ANALYSIS);
    size_t nActual=trj->getTrajSize();
    const COrbitInitData<double> InitData=core->orbit->getInitData();

    size_t intNum = ui.spinBox_IntervalNum->value() - 1;
    size_t intCount = ui.spinBox_intervalCount->value();
    if (intNum>=intCount) return;
    size_t start=nActual * intNum / intCount;
    size_t count=nActual / intCount; 
    if(start+count>nActual) count=nActual-start;

    timeOI.start();
    // obtain spectrum for given interval
    CSpectrum sl[N_DIM];
    std::vector<double> fa[N_DIM];
    COrbitInformation<double>* infoOrbit = trj->analyzeOrbit<double>(start, count, sl, fa);
    int nPointsSpectrum = static_cast<int>(fa[0].size()); 
    if(nPointsSpectrum<=0) return;  // ??? shouldn't happen but who knows..
    vectord freq(nPointsSpectrum);
    for(int i=0; i<nPointsSpectrum; i++)
        freq[i]=i/core->orbit->getIntTime()*InitData.timeUnit * intCount;
    CurveFreqX->setData(&freq.front(), &fa[0].front(), nPointsSpectrum*std::min<double>(2*PLOT_FREQ_MAX_F/ui.lineEdit_timeStep->text().toDouble(),1));
    CurveFreqY->setData(&freq.front(), &fa[1].front(), nPointsSpectrum*std::min<double>(2*PLOT_FREQ_MAX_F/ui.lineEdit_timeStep->text().toDouble(),1));
    if(core->potential->N_dim==3)
        CurveFreqZ->setData(&freq.front(), &fa[2].front(), nPointsSpectrum*std::min<double>(2*PLOT_FREQ_MAX_F/ui.lineEdit_timeStep->text().toDouble(),1));
    else
        CurveFreqZ->setData(NULL, NULL, 0);
    double sf[MAX_SPECTRAL_LINES*N_DIM], sa[MAX_SPECTRAL_LINES*N_DIM];
    double sfp[MAX_SPECTRAL_LINES*N_DIM], sap[MAX_SPECTRAL_LINES*N_DIM];
    int si=0, sip=0;
    for(unsigned int d=0; d<core->potential->N_dim; d++)
      for(size_t s=0; s<sl[d].size() && s<MAX_SPECTRAL_LINES*N_DIM; s++) 
          if(sl[d][s].freq<5)
          {
            if(sl[d][s].method>0)
            {
                sfp[sip]=sl[d][s].freq;
                sap[sip]=sl[d][s].ampl;
                sip++;
            }
            else
            {
                sf[si]=sl[d][s].freq;
                sa[si]=sl[d][s].ampl;
                si++;
            }
          }
    CurveFreqPeaks->setData(sf, sa, si);
    CurveFreqPeaksP->setData(sfp, sap, sip);
    for(int i=0; i<PlotFreq->axisCnt; i++) PlotFreq->setAxisAutoScale(i); 
    PlotFreq->replot();
    ZoomerFreq->setZoomBase(true);

    // set label text
    std::string strInfo = infoOrbit->toString();   // obtain information from orbit analysis - not for entire trajectory but for a segment of it.
    delete infoOrbit;
    for(size_t rf=1; rf<core->orbit->getInfoNum(); rf++)    // obtain the rest (starting from 1 since 0th runtime function is the orbit analysis which was called manually above)
    {
        const CBasicInformation* info=core->orbit->getNewInfo(rf);
        strInfo += info->toString();
        delete info;
    }
    strInfo+=core->orbit->toString(false);
    info(QString::fromStdString(strInfo)
        + "Time elapsed="+QString::number((timeOIelapsed+timeOI.elapsed())*0.001)+" s");
    redrawOrbit(true);  // redraw trajectory for given interval start--count
}

void CSmileGUI::on_checkBox_freqLines_stateChanged(int)
{
    if(ui.checkBox_freqLines->isChecked())
    {
        CurveFreqPeaksP->setStyle(QwtPlotCurve::Sticks);
        CurveFreqPeaks->setStyle(QwtPlotCurve::Sticks);
    }
    else
    {
        CurveFreqPeaksP->setStyle(QwtPlotCurve::NoCurve);
        CurveFreqPeaks->setStyle(QwtPlotCurve::NoCurve);
    }
    PlotFreq->replot();   
}

/* ----------------------------  update initial conditions --------- */
void CSmileGUI::potentialChanged()
{
    if(InternalRecalc) return;
    InternalRecalc=true;    // temporary set flag so that changing options below does not call this function again
    CDensity::POTENTIALTYPE PotentialType = getPotentialTypeByName(ui.selectPotentialType->currentText().toStdString());
    CDensity::POTENTIALTYPE DensityType   = getDensityTypeByName  (ui.selectDensityType  ->currentText().toStdString());
    CDensity::SYMMETRYTYPE SymmetryType   = getSymmetryTypeByName (ui.selectSymmetryType ->currentText().toStdString());

    if(PotentialType==CDensity::PT_BSE || PotentialType==CDensity::PT_SPLINE)
    {
        ui.selectDensityType->show();
        ui.label_densityType->show();
        ui.lineEdit_Ncoefs_radial->show();
        ui.label_Ncoefs_radial->show();
        ui.lineEdit_Ncoefs_radial->setEnabled(DensityType!=CDensity::PT_COEFS);
    }
    else
    {
        ui.selectDensityType->hide();
        ui.label_densityType->hide();
        ui.lineEdit_Ncoefs_radial->hide();
        ui.label_Ncoefs_radial->hide();
    }
    if(PotentialType==CDensity::PT_BSE)
    {
        ui.lineEdit_Alpha->show();
        ui.label_Alpha->show();
        ui.lineEdit_Alpha->setEnabled(DensityType!=CDensity::PT_COEFS);  // if coefs are loaded from file, it is assigned from the same file
    }
    else
    {
        ui.lineEdit_Alpha->hide();
        ui.label_Alpha->hide();
    }
    if(PotentialType==CDensity::PT_NB || 
        ( (PotentialType==CDensity::PT_BSE || PotentialType==CDensity::PT_SPLINE) && 
        (DensityType==CDensity::PT_NB || DensityType==CDensity::PT_ELLIPSOIDAL || 
        DensityType==CDensity::PT_MGE || DensityType==CDensity::PT_COEFS) ) )
    {
        ui.lineEdit_NbodyFile->show();
        ui.buttonChooseNbodyFile->show();
        ui.lineEdit_p->hide();
        ui.lineEdit_q->hide();
        ui.label_p->hide();
        ui.label_q->hide();
        if(PotentialType!=CDensity::PT_NB && DensityType==CDensity::PT_NB) {
            ui.selectSymmetryType->show(); 
            ui.label_symmetryType->show();
        }else{
            ui.selectSymmetryType->hide();
            ui.label_symmetryType->hide();
        }
    }
    else
    {
        ui.lineEdit_NbodyFile->hide();
        ui.buttonChooseNbodyFile->hide();
        ui.lineEdit_p->show();
        ui.lineEdit_q->show();
        ui.label_p->show();
        ui.label_q->show();
        ui.selectSymmetryType->hide();
        ui.label_symmetryType->hide();
    }
    if(PotentialType==CDensity::PT_BSE || PotentialType==CDensity::PT_SPLINE || PotentialType==CDensity::PT_SCALEFREESH)
    {
        ui.lineEdit_Ncoefs_angular->show();
        ui.label_Ncoefs_angular->show();
        ui.lineEdit_Ncoefs_angular->setEnabled(DensityType!=CDensity::PT_COEFS || PotentialType==CDensity::PT_SCALEFREESH);
    }
    else
    {
        ui.lineEdit_Ncoefs_angular->hide();
        ui.label_Ncoefs_angular->hide();
    }
    if(PotentialType==CDensity::PT_LOG || 
        ((PotentialType==CDensity::PT_BSE || PotentialType==CDensity::PT_SPLINE) && DensityType==CDensity::PT_NFW) )
    {
        ui.lineEdit_Rc->show();
        ui.label_core->show();
    }
    else
    {
        ui.lineEdit_Rc->hide();
        ui.label_core->hide();
    }
    if(PotentialType==CDensity::PT_DEHNEN || PotentialType==CDensity::PT_SCALEFREE || PotentialType==CDensity::PT_SCALEFREESH || 
      ((PotentialType==CDensity::PT_BSE || PotentialType==CDensity::PT_SPLINE) && DensityType==CDensity::PT_DEHNEN))
    {
        ui.lineEdit_Gamma->show();
        ui.label_cusp->show();
    }
    else
    {
        ui.lineEdit_Gamma->hide();
        ui.label_cusp->hide();
    }
    if(PotentialType==CDensity::PT_NB)
    {
        ui.label_eps->show();
        ui.label_tol->show();
        ui.lineEdit_eps->show();
        ui.lineEdit_tol->show();
        // disable Lyapunov exponent
        core->InitData.calcLyapunov=false;
        ui.checkBox_Lyapunov->setChecked(false);
        ui.checkBox_Lyapunov->setEnabled(false);
    }
    else
    {
        ui.label_eps->hide();
        ui.label_tol->hide();
        ui.lineEdit_eps->hide();
        ui.lineEdit_tol->hide();
        ui.groupBox_2d3d->setEnabled(true);
        ui.checkBox_Lyapunov->setEnabled(true);
    }
    if(PotentialType==CDensity::PT_BSE || PotentialType==CDensity::PT_SPLINE || PotentialType==CDensity::PT_NB)
    {
        // set N_dim=3
        core->configPotential.N_dim=3;
        on_radioButton_2d_toggled(false);
        ui.radioButton_3d->setChecked(true);
        ui.groupBox_2d3d->setEnabled(false);
    }
    if(PotentialType==CDensity::PT_SCALEFREESH)
        DensityType=CDensity::PT_SCALEFREE;    // reset density type to scale-free (another possibility is PT_COEFS but it is not handled in GUI)
    InternalRecalc=false;
    bool ok=true;
    double newq=(ui.lineEdit_q->text().toDouble(&ok));  if(!ok || (newq==0)) return;
    double newp=(ui.lineEdit_p->text().toDouble(&ok));  if(!ok || (newp==0)) return;
    double newRc=ui.lineEdit_Rc->text().toDouble(&ok);  if (!ok) return;
    double newGamma=ui.lineEdit_Gamma->text().toDouble(&ok);  if (!ok) return;
    double newOmega=ui.lineEdit_Omega->text().toDouble(&ok);  if (!ok) return;
    double newAlpha=ui.lineEdit_Alpha->text().toDouble(&ok);  if (!ok) return;
    double newMbh=ui.lineEdit_Mbh->text().toDouble(&ok);  if(!ok) return;
    unsigned int newNcoefs_radial=ui.lineEdit_Ncoefs_radial->text().toInt(&ok);  if(!ok) return;
    unsigned int newNcoefs_angular=ui.lineEdit_Ncoefs_angular->text().toInt(&ok);  if(!ok) return;
    double newnbpotEps=ui.lineEdit_eps->text().toDouble(&ok); if(!ok) return;
    double newnbpotTol=ui.lineEdit_tol->text().toDouble(&ok); if(!ok) return;
    if(PotentialType!=core->configPotential.PotentialType || 
        DensityType!=core->configPotential.DensityType || 
        SymmetryType!=core->configPotential.SymmetryType || 
        newq!=core->configPotential.q || newp!=core->configPotential.p || 
        newRc!=core->configPotential.Rc || newMbh!=core->configPotential.Mbh ||
        newGamma!=core->configPotential.Gamma || newAlpha!=core->configPotential.Alpha ||  
        newNcoefs_radial!=core->configPotential.Ncoefs_radial || 
        newNcoefs_angular!=core->configPotential.Ncoefs_angular || 
        newOmega!=core->configPotential.Omega || 
        newnbpotEps!=core->configPotential.nbpotEps || 
        newnbpotTol!=core->configPotential.nbpotTol)
    {
        core->configPotential.q=newq; core->configPotential.p=newp; 
        core->configPotential.Rc=newRc; 
        core->configPotential.Gamma=newGamma; 
        core->configPotential.Mbh=newMbh; 
        core->configPotential.Omega=newOmega;
        core->configPotential.Ncoefs_radial=newNcoefs_radial; 
        core->configPotential.Ncoefs_angular=newNcoefs_angular; 
        core->configPotential.Alpha=newAlpha;
        core->configPotential.nbpotEps=newnbpotEps; 
        core->configPotential.nbpotTol=newnbpotTol;
        core->configPotential.PotentialType=PotentialType;
        core->configPotential.DensityType=DensityType;
        core->configPotential.SymmetryType=SymmetryType;
        if(PotentialType!=CPotential::PT_BSE && PotentialType!=CPotential::PT_SPLINE && PotentialType!=CPotential::PT_NB)
            on_buttonInitPotential_clicked();   // do initialization for 'lightweight' potentials only; should manually press this button for 'heavy-weight' initialization of general-purpose potentials listed above
        else
            ui.buttonInitPotential->show();   // let the user press this button manually
    }
    ui.frame->repaint();
}

void CSmileGUI::on_buttonInitPotential_clicked()
{
    core->initPotential();
    ICchanged();
    if(core->configPotential.PotentialType==CPotential::PT_BSE || core->configPotential.PotentialType==CDensity::PT_SPLINE)
    {   // the symmetry type and parameters may have changed by potential constructor, especially if it loaded coefficients from file rather than initialized from Nbody file
        InternalRecalc=true;
        ui.selectDensityType-> setCurrentIndex(ui.selectDensityType-> findText(QString::fromStdString(DensityNames [core->configPotential.DensityType])));
        ui.selectSymmetryType->setCurrentIndex(ui.selectSymmetryType->findText(QString::fromStdString(SymmetryNames[core->configPotential.SymmetryType])));
        ui.lineEdit_Ncoefs_radial->setEnabled(core->configPotential.DensityType!=CDensity::PT_COEFS);
        ui.lineEdit_Ncoefs_angular->setEnabled(core->configPotential.DensityType!=CDensity::PT_COEFS);
        ui.lineEdit_Alpha->setEnabled(core->configPotential.DensityType!=CDensity::PT_COEFS);
        // ensure that N_radial, N_angular and Alpha input lines are consistent with corresponding variables (which may have been updated after loading a coef file)
        if(core->configPotential.Ncoefs_radial != static_cast<unsigned int>(ui.lineEdit_Ncoefs_radial->text().toInt()))
            ui.lineEdit_Ncoefs_radial->setText(QString::number(core->configPotential.Ncoefs_radial));
        if(core->configPotential.Ncoefs_angular != static_cast<unsigned int>(ui.lineEdit_Ncoefs_angular->text().toInt())) 
            ui.lineEdit_Ncoefs_angular->setText(QString::number(core->configPotential.Ncoefs_angular));
        if(core->configPotential.Alpha != ui.lineEdit_Alpha->text().toDouble())
            ui.lineEdit_Alpha->setText(QString::number(core->configPotential.Alpha));
        if(core->configPotential.Mbh != ui.lineEdit_Mbh->text().toDouble())
            ui.lineEdit_Mbh->setText(QString::number(core->configPotential.Mbh));
        InternalRecalc=false;
    }
    ui.buttonInitPotential->hide();   // this button will be shown only when parameters changed
}

void CSmileGUI::chooseNbodyFile()
{
    assignNbodyFile(QFileDialog::getOpenFileName(this, "Open Nbody file", core->WorkDir));
}

void CSmileGUI::typeinNbodyFile()
{
    QString fileName = ui.lineEdit_NbodyFile->text();
    if(!QFile::exists(fileName))
        fileName = QFileDialog::getOpenFileName(this, "Open Nbody file", core->WorkDir);
    assignNbodyFile(fileName);
}

void CSmileGUI::assignNbodyFile(QString fileName)
{
    if(fileName.isEmpty() || !QFile::exists(fileName)) return;
    core->WorkDir = getDirName(fileName);
    core->NbodyFile=fileName;
    ui.lineEdit_NbodyFile->setText(fileName);
    core->configPotential.PotentialType = getPotentialTypeByName(ui.selectPotentialType->currentText().toStdString());   // set explicitly because it may have been reset to default if previous load was unsuccessful
    on_buttonInitPotential_clicked();  // do initialization
}

void CSmileGUI::ICchanged()
{
    if(InternalRecalc) return;
    bool ok=true;
    double newE=ui.lineEdit_E->text().toDouble(&ok);  if(!ok) return;
    double X=ui.lineEdit_x->text().toDouble(&ok);  if(!ok) return;
    double Y=ui.lineEdit_y->text().toDouble(&ok);  if(!ok) return;
    double Z=(core->potential->N_dim==3)?ui.lineEdit_z->text().toDouble(&ok):0;  if(!ok) return;
    double Vx=ui.lineEdit_vx->text().toDouble(&ok);  if(!ok) return;
    double Vy=ui.lineEdit_vy->text().toDouble(&ok);  if(!ok) return;
    double Vz=(core->potential->N_dim==3)?ui.lineEdit_vz->text().toDouble(&ok):0;  if(!ok) return;
    if(core->useICe)
    {
        core->initCondE=newE;
    }
    else
    {
        core->InitData.initCond=CPosVelPoint<double>(X, Y, Z, Vx, Vy, Vz);
    }
    InternalRecalc=true;
    core->initIC();
    if(core->useICe)
    {
        ui.lineEdit_x->setText(QString::number(core->InitData.initCond.Pos[0]));
        ui.lineEdit_y->setText(QString::number(core->InitData.initCond.Pos[1]));
        ui.lineEdit_z->setText(QString::number(core->InitData.initCond.Pos[2]));
        ui.lineEdit_vx->setText(QString::number(core->InitData.initCond.Vel[0]));
        ui.lineEdit_vy->setText(QString::number(core->InitData.initCond.Vel[1]));
        ui.lineEdit_vz->setText(QString::number(core->InitData.initCond.Vel[2]));
    }
    else
    {
        ui.lineEdit_E->setText(QString::number(core->initCondE));
    }
    ui.lineEdit_Torb->setText(QString::number(core->InitData.timeUnit));
    InternalRecalc=false;
    if(fabs(core->initCondE-Eprev)> 1e-4)
        on_ButtonClearPS_clicked();
}

/* ----------------------------  modification of various parameters by GUI lineEdits ----- */
void CSmileGUI::intTime_changed()
{
    if(InternalRecalc) return;
    bool ok;
    double t = ui.lineEdit_intTime->text().toDouble(&ok);
    if(ok)  core->configCore.intTimeInPeriods=t;
    t = ui.lineEdit_timeStep->text().toDouble(&ok);
    if(ok)  core->configCore.intTimeStepsPerPeriod=t;
}

void CSmileGUI::on_checkBox_fmExistingSS_stateChanged(int Checked)
{
    ui.lineEdit_fms->setEnabled(Checked==0);
    ui.lineEdit_fmp->setEnabled(Checked==0);
    ui.lineEdit_fmr->setEnabled(Checked==0);
    ui.lineEdit_fmy->setEnabled(Checked==0);
}

void CSmileGUI::on_checkBox_smExistingSS_stateChanged(int Checked)
{
    ui.lineEdit_smNumOrbits->setEnabled(Checked==0);
}

void CSmileGUI::fmparams_changed()
{
    if(InternalRecalc) return;
    bool ok; int t;
    t = ui.lineEdit_fms->text().toInt(&ok);
    if(ok)  core->configCore.fm_numOrbitsStationary=t;
    t = ui.lineEdit_fmp->text().toInt(&ok);
    if(ok)  core->configCore.fm_numOrbitsPrincipalPlane=t;
    t = ui.lineEdit_fmy->text().toInt(&ok);
    if(ok)  core->configCore.fm_numOrbitsYalpha=t;
    t = ui.lineEdit_fmr->text().toInt(&ok);
    if(ok)  core->configCore.fm_numOrbitsRandom=t;
}

void CSmileGUI::smparams_changed()
{
    if(InternalRecalc) return;
    bool ok; int t; double d;
    // orbit library tab
    t = ui.lineEdit_smNumOrbits->text().toInt(&ok);
    if(ok)  core->configCore.sm_numOrbits=t;
    t = ui.lineEdit_smNumSamplingPoints->text().toInt(&ok);
    if(ok) core->configCore.sm_numSamplingPoints=t;
    // model params tab
    CBasicSchwData::SCHWDATATYPE densDataType = getSchwDataTypeByName(ui.select_smType->currentText().toStdString());
    if(densDataType==CBasicSchwData::SD_DENS_CLASSIC)
    {
        ui.lineEdit_smNumLines->show();
        ui.label_smNumLines->show();
        ui.lineEdit_smNumAngCoefs->hide();
        ui.label_smNumAngCoefs->hide();
    }
    else
    {
        ui.lineEdit_smNumLines->hide();
        ui.label_smNumLines->hide();
        ui.lineEdit_smNumAngCoefs->show();
        ui.label_smNumAngCoefs->show();
    }
    if(densDataType==CBasicSchwData::SD_DENS_SHBSE)
    {
        ui.lineEdit_smNumRadCoefs->show();
        ui.lineEdit_smAlpha->show();
        ui.label_smNumRadCoefs->show();
        ui.label_smAlpha->show();
    }
    else
    {
        ui.lineEdit_smNumRadCoefs->hide();
        ui.lineEdit_smAlpha->hide();
        ui.label_smNumRadCoefs->hide();
        ui.label_smAlpha->hide();
    }
    if(densDataType!=core->configCore.sm_densDataType)
    {
        ui.groupBox_smOptim->setEnabled(false);
        core->configCore.sm_densDataType=densDataType;
    }
    t = ui.lineEdit_smNumAngCoefs->text().toInt(&ok);
    if(ok&&(core->configCore.sm_numAngularCoefs!=t))
    {
        ui.groupBox_smOptim->setEnabled(false);
        core->configCore.sm_numAngularCoefs=t;
    }
    t = ui.lineEdit_smNumRadCoefs->text().toInt(&ok);
    if(ok&&(core->configCore.sm_numRadialCoefs!=t))
    {
        ui.groupBox_smOptim->setEnabled(false);
        core->configCore.sm_numRadialCoefs=t;
    }
    d = ui.lineEdit_smAlpha->text().toDouble(&ok);
    if(ok&&(core->configCore.sm_Alpha!=d))
    {
        ui.groupBox_smOptim->setEnabled(false);
        core->configCore.sm_Alpha=d;
    }
    t = ui.lineEdit_smNumLines->text().toInt(&ok);
    if(ok&&(core->configCore.sm_linesPerSegment!=t))
    {
        ui.groupBox_smOptim->setEnabled(false);
        core->configCore.sm_linesPerSegment=t;
    }
    t = ui.lineEdit_smNumShells->text().toInt(&ok);
    if(ok)
    {
        if(core->configCore.sm_numShells!=t) ui.groupBox_smOptim->setEnabled(false);
        ui.selectEnergyLevel_sm->setMaximum(t);
        ui.selectEnergyLevel_fm->setMaximum(t);
        core->configCore.sm_numShells=t;
    }
    d = ui.lineEdit_smInnerShellMass->text().toDouble(&ok);
    if(ok&&(core->configCore.sm_innerShellMass!=d))
    {
        ui.groupBox_smOptim->setEnabled(false);
        core->configCore.sm_innerShellMass=d;
    }
    d = ui.lineEdit_smOuterShellMass->text().toDouble(&ok);
    if(ok&&(core->configCore.sm_outerShellMass!=d))
    {
        ui.groupBox_smOptim->setEnabled(false);
        core->configCore.sm_outerShellMass=d;
    }
    // optimization tab
    d = ui.lineEdit_smConstraintPenalty->text().toDouble(&ok);
    if(ok) core->configCore.sm_constraintPenalty = d;
    d = ui.lineEdit_smQuadraticPenalty->text().toDouble(&ok);
    if(ok) core->configCore.sm_regularization = d;
    d = ui.lineEdit_smChaoticWeight->text().toDouble(&ok);
    if(ok) core->configCore.sm_chaoticPenalty = d;
    d = ui.lineEdit_smMaxOrbitWeight->text().toDouble(&ok);
    if(ok) core->configCore.sm_maxWeight = d;
    core->configCore.sm_constrainBeta=ui.checkBox_smBeta->isChecked();
    ui.lineEdit_smBetaIn->setEnabled(core->configCore.sm_constrainBeta);
    ui.lineEdit_smBetaOut->setEnabled(core->configCore.sm_constrainBeta);
    d = ui.lineEdit_smBetaIn->text().toDouble(&ok);
    if(ok)  core->configCore.sm_betaIn = d;
    d = ui.lineEdit_smBetaOut->text().toDouble(&ok);
    if(ok)  core->configCore.sm_betaOut = d;
    // compute # of coefs
    int numCoefs=0;
    switch(densDataType)
    {
    case CBasicSchwData::SD_DENS_CLASSIC: 
        numCoefs=3*core->configCore.sm_linesPerSegment*core->configCore.sm_linesPerSegment*core->configCore.sm_numShells;
        break;
    case CBasicSchwData::SD_DENS_SHGRID: 
        numCoefs=1+core->configCore.sm_numShells*(core->configCore.sm_numAngularCoefs/2+1)*(core->configCore.sm_numAngularCoefs/2+2)/2;
        break;
    case CBasicSchwData::SD_DENS_SHBSE: 
        numCoefs=(1+core->configCore.sm_numRadialCoefs)*(core->configCore.sm_numAngularCoefs/2+1)*(core->configCore.sm_numAngularCoefs/2+2)/2;
        break;
    default: info("Invalid type of Schwarzschild model");
    }
    ui.lineEdit_smNumCoefs->setText(QString::number(numCoefs));
}

void CSmileGUI::ccparams_changed()
{
    if(InternalRecalc) return;
    bool ok;
    double d = ui.lineEdit_MinFreqDiff->text().toDouble(&ok);
    if(ok)  core->configCore.chaoticMinFreqDiff=d;
    d = ui.lineEdit_MinLambda->text().toDouble(&ok);
    if(ok)  core->configCore.chaoticMinLambda=d;
    redrawFreqMap(true);
}

void CSmileGUI::on_checkBox_Lyapunov_stateChanged(int Checked)
{
    if(InternalRecalc) return;
    core->InitData.calcLyapunov = (Checked>0);
}

void CSmileGUI::on_checkBox_usePS_stateChanged(int Checked)
{
    if(InternalRecalc) return;
    core->configCore.usePS = (Checked>0);
}

void CSmileGUI::on_radioButton_2d_toggled(bool n2d)
{
    if(n2d)
    {
        ui.radioButton_2d_xy->setChecked(true);
        ui.radioButton_2d_xz->setEnabled(false);
        ui.radioButton_2d_yz->setEnabled(false);
        ui.radioButton_3dline->setEnabled(false);
        ui.radioButton_3dmesh->setEnabled(false);
    }
    else
    {
        ui.radioButton_2d_xz->setEnabled(true);
        ui.radioButton_2d_yz->setEnabled(true);
#ifdef HAVE_QWT3D
        ui.radioButton_3dline->setEnabled(true);
#ifdef USE3DFACETS
        ui.radioButton_3dmesh->setEnabled(true);
#endif
#endif
    }
    if(InternalRecalc) return;
    core->configPotential.N_dim=n2d?2:3;
    core->initPotential();
    ICchanged();
}

void CSmileGUI::on_radioButton_ICe_toggled(bool useICe)
{
    if(InternalRecalc) return;
    core->useICe=useICe;
    if(useICe)
    {
        ui.lineEdit_E->setEnabled(true);
        ui.groupBox_ICx->setEnabled(false);
    }
    else
    {
        ui.lineEdit_E->setEnabled(false);
        ui.groupBox_ICx->setEnabled(true);
    }
}

/* ----------------------------  Random initial conditions --------- */
void CSmileGUI::on_ButtonRandom_clicked()
{
    double Xmax=core->potential->longaxisradius(core->initCondE);
    if(Xmax<0)
    {
        error("Total energy is outside allowed range");
        return;  // error
    }
    const int npsampling=1000;
    std::vector<double> Xs(npsampling), Ys(npsampling), Zs(npsampling), Ws(npsampling);
    double X, Y, Z, Et, totalWeight=0;
    for(int s=0; s<npsampling; s++)
    {   // sample accessible configuration space with appropriate weight
      do{
        X = (rand()*2.0/RAND_MAX-1) * Xmax;
        Y = (rand()*2.0/RAND_MAX-1) * Xmax;
        Z = core->potential->N_dim==3 ? (rand()*2.0/RAND_MAX-1) * Xmax : 0;
        Et = core->potential->Phi(X, Y, Z);
      }
      while(Et>core->initCondE);
      double W=sqrt(2*(core->initCondE-Et));
      Xs[s]=X; Ys[s]=Y; Zs[s]=Z; Ws[s]=W; totalWeight+=W;
    }
    // now select point from sampled
    double indw=rand() * totalWeight / RAND_MAX;
    int i=0;
    while(i<npsampling-1 && indw>=Ws[i]) 
    {
        indw-=Ws[i];
        i++;
    }
    X=Xs[i]; Y=Ys[i]; Z=Zs[i]; 
    double V=Ws[i];
    double costh=(core->potential->N_dim==3)? (rand()*2.0/RAND_MAX-1) : 0;
    double sinth=sqrt(1-pow_2(costh));
    double phi=rand()*2*M_PI/RAND_MAX;
    InternalRecalc=true;

    core->InitData.initCond=CPosVelPoint<double>(X, Y, Z, V*sinth*cos(phi), V*sinth*sin(phi), V*costh);
    //core->initIC();
    ui.lineEdit_x->setText(QString::number(core->InitData.initCond.Pos[0]));
    ui.lineEdit_y->setText(QString::number(core->InitData.initCond.Pos[1]));
    ui.lineEdit_z->setText(QString::number(core->InitData.initCond.Pos[2]));
    ui.lineEdit_vx->setText(QString::number(core->InitData.initCond.Vel[0]));
    ui.lineEdit_vy->setText(QString::number(core->InitData.initCond.Vel[1]));
    ui.lineEdit_vz->setText(QString::number(core->InitData.initCond.Vel[2]));
    InternalRecalc=false;
    on_ButtonStart_clicked();
}

/* ----------------------------  Poincare section operations --------- */
void CSmileGUI::PSclicked(const QPointF &pos)
{
    double newx=pos.x();
    double newvx=pos.y();
    double newvy2=2*(core->initCondE - core->potential->Phi(newx, 0, 0)) - pow_2(newvx);
    if(newvy2>=0)
    {
        InternalRecalc=true;
        ui.lineEdit_x->setText(QString::number(newx, 'f', 7));
        ui.lineEdit_vx->setText(QString::number(newvx, 'f', 7));
        ui.lineEdit_vy->setText(QString::number(sqrt(newvy2), 'f', 7));
        ui.lineEdit_y->setText("0");
        ui.lineEdit_z->setText("0");
        ui.lineEdit_vz->setText("0");
        InternalRecalc=false;
        ICchanged();
    }
}

void CSmileGUI::on_ButtonClearPS_clicked()
{
    while(CurvesPoincare.size()>0)
    {
        delete CurvesPoincare[CurvesPoincare.size()-1];
        CurvesPoincare.remove(CurvesPoincare.size()-1);
    }
    double xm[NUM_POINTS_PLOT_POINCARE*4], vxm[NUM_POINTS_PLOT_POINCARE*4];
    double Xmax=core->potential->longaxisradius(core->initCondE);
    if(Xmax>0) 
    {
        for(unsigned int i=0; i<NUM_POINTS_PLOT_POINCARE; i++)
        {
            xm[i]=Xmax*(i+0.5)/NUM_POINTS_PLOT_POINCARE;
            double vx2=2*(core->initCondE - core->potential->Phi(xm[i], 0, 0));
            if(vx2>0) vxm[i]=sqrt(vx2); else vxm[i]=0;
            xm[NUM_POINTS_PLOT_POINCARE*2-1-i]=xm[i];
            vxm[NUM_POINTS_PLOT_POINCARE*2-1-i]=-vxm[i];
            xm[NUM_POINTS_PLOT_POINCARE*2+i]=-xm[i];
            vxm[NUM_POINTS_PLOT_POINCARE*2+i]=-vxm[i];
            xm[NUM_POINTS_PLOT_POINCARE*4-1-i]=-xm[i];
            vxm[NUM_POINTS_PLOT_POINCARE*4-1-i]=vxm[i];
        }
        QwtPlotCurve* CurveBoundary = new QwtPlotCurve();
        CurveBoundary->setPen(QColor(Qt::red));
        CurveBoundary->attach(PlotPoincare);
        CurveBoundary->setData(xm, vxm, 4*NUM_POINTS_PLOT_POINCARE);
        CurvesPoincare<<(CurveBoundary);
    }
    for(int i=0; i<PlotPoincare->axisCnt; i++) PlotPoincare->setAxisAutoScale(i); 
    PlotPoincare->replot();
    ZoomerPoincare->setZoomBase(true);
    Eprev=core->initCondE;
}

/* ----------------------------  Frequency map operations --------- */

// disable several buttons for the duration of orbit library integration, optimization, or export to Nbody
void CSmileGUI::enableOrbitLibControls(bool enable)
{
    ui.groupBox_N_dim->setEnabled(enable && !core->isOrbitIntRunning());
    ui.groupBox_potential->setEnabled(enable && !core->isOrbitIntRunning());
    ui.groupBox_smOptim->setEnabled(enable);
    ui.groupBox_smGrid->setEnabled(enable);
    ui.ButtonImportFM->setEnabled(enable);
    ui.ButtonImportSM->setEnabled(enable);
}

void CSmileGUI::on_ButtonFreqMap_clicked()
{
    if(core->isExportNbodyRunning() || core->isSchwOptRunning()) return;
    if(core->isOrbitLibRunning())
    {
        emit stopOrbitLibrary();
        return;
    }
    CurveFreqMapR->setData(NULL, NULL, 0);
    CurveFreqMapC->setData(NULL, NULL, 0);
    // start building freq map
    core->buildFreqMapIC(core->initCondE, ui.checkBox_fmExistingSS->isChecked());
    core->startOrbitLibrary(SLOT(coreOrbitLibraryFinished()));
    ui.progressBarFM->setRange(0, static_cast<int>(core->orbitlib->size()));
    ui.progressBarFM->setValue(0);
    ui.progressBarFM->show();
    timeFM.start();
    ui.ButtonFreqMap->setText("Stop");
    ui.tabWidget_sm->setEnabled(false);   // can't do Schw.modelling while running freq.map
    enableOrbitLibControls(false);
    ui.ButtonSchwNbody->setEnabled(false);
}

void CSmileGUI::OrbitLibraryFinished()
{   // called when either freq.map or schw.orbit library is finished
    enableOrbitLibControls(true);
    ui.ButtonFreqMap->setEnabled(true);
    ui.ButtonFreqMap->setText("Start");
    ui.ButtonSchwOrbitLib->setText("Start");
    ui.tabWidget_sm->setEnabled(true);
    if(core->schwData.size()==0)   // was not running in the context of creation of Schw.model orbit library
        ui.groupBox_smOptim->setEnabled(false);
    ui.progressBarFM->hide();
    ui.progressBarSM->hide();
    redrawFreqMap(true);
    info(QString::number(core->orbitlib->numComplete())+" orbits, " +
        QString::number(1000.0/timeFM.elapsed()*core->orbitlib->numComplete(),'g',3)+" orbits/second\n" );
}

bool comparePaird(paird elem1, paird elem2)
{
    return elem1.second < elem2.second;
}

// return an orbit filtering function which selects orbits belonging to a given energy shell, 
// if a Schw.data object with energy shells was initialized and numShell>0
// may also return shell energy if the last parameter is not NULL
CShellOrbitFilteringFnc shellFilter(const vectorSchwData &schwData, int numShell, double* shellEnergy=NULL)
{
    const CSchwDataKinemShell* model_sh=NULL;
    for(vectorSchwData::const_iterator iter=schwData.begin(); iter!=schwData.end(); ++iter)
        if((*iter)->dataType() == CBasicSchwData::SD_KINEM_SHELL)
        {
            model_sh=static_cast<const CSchwDataKinemShell*>(*iter);
            if(numShell>0 && shellEnergy!=NULL) *shellEnergy=model_sh->getShellEnergy(numShell-1);
        }
    return CShellOrbitFilteringFnc(model_sh, numShell);
}

void CSmileGUI::redrawFreqMap(bool show)
{
    if(!show || core->orbitlib==NULL) return;
    size_t nFMpoints = core->orbitlib->size();
    if(nFMpoints==0) return;
    int numShell = ui.selectEnergyLevel_fm->value();
    double shellEnergy=12345;
    const CShellOrbitFilteringFnc filter=shellFilter(core->schwData, numShell, &shellEnergy);
    //const CChaosOrbitFilteringFnc chaos(1, core->configCore.chaoticMinLambda, core->configCore.chaoticMinFreqDiff);
    if(numShell>0 && shellEnergy!=12345)
        ui.lineEdit_E->setText(QString::number(shellEnergy));
    bool useNonzeroOnly = ui.checkBox_fmConfigSpace->isChecked();
    if(ui.radioButton_fmmapdiff->isChecked() || ui.radioButton_fmmaplyapunov->isChecked())
    {   // draw frequency map itself
        vectord w1w3r(nFMpoints),  w2w3r(nFMpoints),  w1w3c(nFMpoints),  w2w3c(nFMpoints);
        int nr=0, nc=0;
        for(unsigned int i=0; i<nFMpoints; i++)
        {
            bool useOrbit=true;
            useOrbit = filter.eval(core->orbitlib->getOrbitDesc(i))>=0.5 && (!useNonzeroOnly || core->orbitlib->getOrbitDesc(i)->getWeight()>0);
            if(core->orbitlib->getOrbitDesc(i)->getState()==COrbitDesc::OS_DONE && useOrbit)
            {
                const COrbitInformation<float>* info=static_cast<const COrbitInformation<float>*>(core->orbitlib->getOrbitDesc(i)->getInfoByType(CBasicInformation::IT_TRAJ_ANALYSIS));
                if(info==NULL) continue;
                bool criterion;
                if(ui.radioButton_fmmapdiff->isChecked()) 
                    criterion = info->getlfccdiff() <= core->configCore.chaoticMinFreqDiff;
                else 
                    criterion = info->getlambda() <= core->configCore.chaoticMinLambda;
                double denom = (core->potential->N_dim==3)?info->getlf(2):1;
                if(denom==0) continue;
                if(criterion)
                {
                    w1w3r[nr] = info->getlf(0)/denom;
                    w2w3r[nr] = info->getlf(1)/denom;
                    nr++;
                }
                else
                {
                    w1w3c[nc] = info->getlf(0)/denom;
                    w2w3c[nc] = info->getlf(1)/denom;
                    nc++;
                }
            }
        }
        CurveFreqMapR->setData(&(w1w3r.front()), &(w2w3r.front()), nr);
        CurveFreqMapC->setData(&(w1w3c.front()), &(w2w3c.front()), nc);
        if(core->potential->N_dim==3)
        {
            PlotFreqMap->setAxisScale(QwtPlot::yLeft, 0.5, 1.1);
            PlotFreqMap->setAxisScale(QwtPlot::xBottom, 0.3, 1.1);
        }
        PlotFreqMap->replot();
        PlotFreqMapHisto->hide();
        PlotFreqMapSS->hide();
        PlotFreqMap->show();

        if(!core->isOrbitLibRunning())
        {
            info(getOrbitPopulation(numShell, useNonzeroOnly));
        }
    }
    else if(ui.radioButton_fmhistodiff->isChecked() || ui.radioButton_fmhistolyapunov->isChecked())
    {   // draw histogram of frequency diffusion rates or Lyapunov exponents
        std::vector<paird> values;
        double totalWeight=0;
        for(unsigned int i=0; i<nFMpoints; i++)
            if(core->orbitlib->getOrbitDesc(i)->getState()==COrbitDesc::OS_DONE && filter.eval(core->orbitlib->getOrbitDesc(i))>=0.5)
            {
                const COrbitInformation<float>* info=static_cast<const COrbitInformation<float>*>(core->orbitlib->getOrbitDesc(i)->getInfoByType(CBasicInformation::IT_TRAJ_ANALYSIS));
                if(info==NULL) continue;
                double weight=(useNonzeroOnly ? core->orbitlib->getOrbitDesc(i)->getWeight() : 0);
                totalWeight+=weight;
                values.push_back(paird(weight, ui.radioButton_fmhistodiff->isChecked() ? info->getlfccdiff() : info->getlambda()));
            }
        std::sort(values.begin(), values.end(), comparePaird);
        int npoints=static_cast<int>(values.size());
        if(npoints>0)
        {
            vectord val(npoints), np(npoints);
            double minval=1, tw=0;
            for(int i=0; i<npoints; i++)
            {
                val[i]=values[i].second;
                if(totalWeight)
                {
                    np[i]=tw/totalWeight;
                    tw+=values[i].first;
                }
                else
                    np[i]=i*1.0/npoints;
                if((val[i]>0) && (val[i]<minval)) minval=val[i];
            }
            CurveFreqMapHisto->setData(&(val.front()), &(np.front()), npoints);
            PlotFreqMapHisto->setAxisScale(QwtPlot::xBottom, minval/1.5, val[npoints-1]*1.1);
            PlotFreqMapHisto->replot();
        }
        PlotFreqMap->hide();
        PlotFreqMapSS->hide();
        PlotFreqMapHisto->show();
    }
    else if(ui.radioButton_fmssdiff->isChecked() || ui.radioButton_fmsslyapunov->isChecked() ||
        ui.radioButton_fmspdiff->isChecked() || ui.radioButton_fmsplyapunov->isChecked() ||
        ui.radioButton_fmsydiff->isChecked() || ui.radioButton_fmsylyapunov->isChecked())
    {   // draw start spaces
        vectord Xr(nFMpoints), Yr(nFMpoints), Xc(nFMpoints), Yc(nFMpoints);
        int nr=0, nc=0;
        double rad=0;
        COrbitInitData<float> InitData;
        for(unsigned int i=0; i<nFMpoints; i++)
        {
            if(core->orbitlib->getOrbitDesc(i)->getState()==COrbitDesc::OS_DONE && filter.eval(core->orbitlib->getOrbitDesc(i))>=0.5)
            {
                bool criterion;
                const COrbitInformation<float>* info=static_cast<const COrbitInformation<float>*>(core->orbitlib->getOrbitDesc(i)->getInfoByType(CBasicInformation::IT_TRAJ_ANALYSIS));
                if(info==NULL) continue;
                if(ui.radioButton_fmssdiff->isChecked() || ui.radioButton_fmspdiff->isChecked() || ui.radioButton_fmsydiff->isChecked()) 
                    criterion = info->getlfccdiff() <= core->configCore.chaoticMinFreqDiff;
                else
                    criterion = info->getlambda() <= core->configCore.chaoticMinLambda;
                const COrbitInitData<float> InitData=core->orbitlib->getOrbitDesc(i)->getInitData();
                double X=InitData.initCond.Pos[0];
                double Y=InitData.initCond.Pos[1];
                double Z=InitData.initCond.Pos[2];
                double Xp, Yp;
                if(ui.radioButton_fmsydiff->isChecked() || ui.radioButton_fmsylyapunov->isChecked())
                {
                    Xp=Y;
                    Yp=atan2(InitData.initCond.Vel[0], InitData.initCond.Vel[2]);
                }
                else
                    XYZtoXY(X,Y,Z,&Xp,&Yp);
                if(pow_2(Xp)+pow_2(Yp)>pow_2(rad)) rad=sqrt(pow_2(Xp)+pow_2(Yp));
                if(ui.radioButton_fmssdiff->isChecked() || ui.radioButton_fmsslyapunov->isChecked())
                {
                    if(pow_2(InitData.initCond.Vel[0])+pow_2(InitData.initCond.Vel[1])+pow_2(InitData.initCond.Vel[2])>0) continue;   // this is not a stationary orbit
                }
                else if(ui.radioButton_fmspdiff->isChecked() || ui.radioButton_fmsplyapunov->isChecked())
                {
                    if(X*Y*Z!=0 || (X==0 && Z==0)) continue;   // this is not a principal-plane orbit
                }
                else
                {
                    if(X!=0 || Z!=0) continue;  // this point is not on Y axis
                }
                if(criterion)
                {
                    Xr[nr] = Xp;
                    Yr[nr] = Yp;
                    nr++;
                }
                else
                {
                    Xc[nc] = Xp;
                    Yc[nc] = Yp;
                    nc++;
                }
            }
        }
        CurveFreqMapSSR->setData(&(Xr.front()), &(Yr.front()), nr);
        CurveFreqMapSSC->setData(&(Xc.front()), &(Yc.front()), nc);
        PlotFreqMapSS->setAxisScale(QwtPlot::yLeft, -rad, rad);
        PlotFreqMapSS->setAxisScale(QwtPlot::xBottom, -rad, rad);
        PlotFreqMapSS->replot();
        PlotFreqMap->hide();
        PlotFreqMapHisto->hide();
        PlotFreqMapSS->show();
    }
}

void CSmileGUI::on_selectEnergyLevel_fm_valueChanged(int /*numShell*/)
{
    redrawFreqMap(true);
    //if(numShell>0 && core->model!=NULL && (core->model->ModelType() & CBasicSchwData::SD_DENS_SHELL))  // redundant
    //    ui.lineEdit_E->setText(QString::number(static_cast<CBasicShellSchwModel*>(core->model)->getShellEnergy(numShell-1)));
}

void CSmileGUI::on_checkBox_fmConfigSpace_stateChanged(int)
{
    redrawFreqMap(true);
}

QString CSmileGUI::getOrbitPopulation(int numShell, bool useWeights) const
{
    if(core->orbitlib==NULL) return "Error!";
    const CShellOrbitFilteringFnc filter=shellFilter(core->schwData, numShell);
    const CChaosOrbitFilteringFnc chaos(1, core->configCore.chaoticMinLambda, core->configCore.chaoticMinFreqDiff);
    return QString::fromStdString(core->orbitlib->getOrbitPopulation(&filter, &chaos, useWeights));
}

void CSmileGUI::drawFreqMapResonances()
{
    const int NR=18;   // number of resonances depicted on the map
    double res[NR][5] = {
        {0,1,-1 ,0.80,1},
        {1,-1,0 ,0.55,0.55},
        {2,-1,0 ,0.45,0.90},
        {3,-2,0 ,0.34,0.51},
        {4,-3,0 ,0.39,0.52},
        {2,0,-1 ,0.50,1.08},
        {3,0,-2 ,0.66,0.52},
        {1,-3,2 ,0.34,0.78},
        {1,-2,1 ,0.80,0.90},
        {2,-3,1 ,0.33,0.55},
        {4,-2,-1,0.78,1.08},
        {3,-1,-1,0.69,1.08},
        {3,1,-3 ,0.63,1.08},
        {2,1,-2 ,0.74,0.52},
        {2,2,-3 ,0.80,0.70},
        {1,2,-2 ,0.74,0.63},
        {1,1,-1 ,0.47,0.53} };
    double xmin=0.3, xmax=1., ymin=0.4, ymax=1.1;
    double boundrect[4][3] = { {1, 0, -xmin}, {1, 0, -xmax}, {0, 1, -ymin}, {0, 1, -ymax} };
    //QString S;
    for(int i=0; i<NR; i++)
    {
        double x[2], y[2];
        if(res[i][1]==0)
        {
            x[0]=x[1]=-res[i][2]/res[i][0];
            y[0]=ymin;
            y[1]=ymax;
        }
        else if(res[i][0]==0)
        {
            x[0]=xmin;
            x[1]=xmax;
            y[0]=y[1]=-res[i][2]/res[i][1];
        }
        else
        {   // intersect with bounding rectangle
            std::vector<double> ax(0), ay(0);
            for(int c=0; c<4; c++)
            {
                double D = res[i][0]*boundrect[c][1] - res[i][1]*boundrect[c][0];
                ax.push_back( (-res[i][2]*boundrect[c][1] + res[i][1]*boundrect[c][2]) / D);
                ay.push_back(-(-res[i][2]*boundrect[c][0] + res[i][0]*boundrect[c][2]) / D);
            }
            std::sort(ax.begin(), ax.end());
            std::sort(ay.begin(), ay.end());
            int sgn = (res[i][0]*res[i][1]>0)?1:0;
            x[0]=ax[1]; x[1]=ax[2];
            y[0]=ay[1+sgn]; y[1]=ay[2-sgn];
        }
        QwtPlotCurve* c = new QwtPlotCurve();
        c->setPen(QPen(Qt::darkGray, 1, Qt::DotLine));
        c->setData(x, y, 2);
        c->attach(PlotFreqMap);
        QwtPlotMarker* m = new QwtPlotMarker();
        m->setValue(res[i][3], res[i][4]);
        m->setLabel(QString::number(res[i][0])+","+QString::number(res[i][1])+","+QString::number(res[i][2]));
        m->attach(PlotFreqMap);
    }
}

void CSmileGUI::FMclicked(const QPointF &pos)
{
    if(core->orbitlib==NULL || core->orbitlib->size()==0)
        return;
    int mode=-1;
    if(ui.radioButton_fmmapdiff->isChecked() || ui.radioButton_fmmaplyapunov->isChecked()) mode=0;
    else if(ui.radioButton_fmssdiff->isChecked() || ui.radioButton_fmsslyapunov->isChecked()) mode=1;
    else if(ui.radioButton_fmspdiff->isChecked() || ui.radioButton_fmsplyapunov->isChecked()) mode=2;
    if(mode<0) return;
    double px=pos.x();
    double py=pos.y();
    double mindist=1; 
    int np=0;
    int numShell = ui.selectEnergyLevel_fm->value();
    //const CBasicShellSchwModel* model_sh = core->model!=NULL && (core->model->ModelType() & CBasicSchwData::SD_DENS_SHELL) ? static_cast<const CBasicShellSchwModel*>(core->model) : NULL;
    const CShellOrbitFilteringFnc filter=shellFilter(core->schwData, numShell);
    for(unsigned int i=0; i<core->orbitlib->size(); i++)
    {
        const COrbitInformation<float>* info=static_cast<const COrbitInformation<float>*>(core->orbitlib->getOrbitDesc(i)->getInfoByType(CBasicInformation::IT_TRAJ_ANALYSIS));
        if(info==NULL) continue;
        double dist=1;
        if(mode==0)
        {
            double lf2 = (core->potential->N_dim==3)?info->getlf(2):1;
            double w1w3o = info->getlf(0)/lf2;
            double w2w3o = info->getlf(1)/lf2;
            dist = pow_2(w1w3o-px) + pow_2(w2w3o-py);
        }
        else
        {
            const COrbitInitData<float> InitData=core->orbitlib->getOrbitDesc(i)->getInitData();
            double X=InitData.initCond.Pos[0];
            double Y=InitData.initCond.Pos[1];
            double Z=InitData.initCond.Pos[2];
            double Xp, Yp;
            XYZtoXY(X,Y,Z,&Xp,&Yp);
            if((mode==1 && X*Y*Z==0) || (mode==2 && X*Y*Z!=0)) continue; 
            dist = pow_2(Xp-px) + pow_2(Yp-py);
        }
        if(dist<mindist && filter.eval(core->orbitlib->getOrbitDesc(i))>=0.5)
        {
            np=i;
            mindist=dist;
        }
    }
    if(mindist>=1) return;  // too far
    // load data from np'th orbit
    bool tmpLyapunov=core->InitData.calcLyapunov;
    core->InitData=core->orbitlib->getOrbitDesc(np)->getInitData();
    core->InitData.calcLyapunov=tmpLyapunov;
    // set exact values
    bool tmpUseE=core->useICe;
    core->useICe=false;  // need to put in values for IC, not for E
    core->initIC();
    core->useICe=tmpUseE;
    InternalRecalc=true;
    ui.lineEdit_x->setText(QString::number(core->InitData.initCond.Pos[0]));
    ui.lineEdit_y->setText(QString::number(core->InitData.initCond.Pos[1]));
    ui.lineEdit_z->setText(QString::number(core->InitData.initCond.Pos[2]));
    ui.lineEdit_vx->setText(QString::number(core->InitData.initCond.Vel[0]));
    ui.lineEdit_vy->setText(QString::number(core->InitData.initCond.Vel[1]));
    ui.lineEdit_vz->setText(QString::number(core->InitData.initCond.Vel[2]));
    ui.lineEdit_E->setText(QString::number(core->initCondE));
    ui.lineEdit_Torb->setText(QString::number(core->InitData.timeUnit));
    InternalRecalc=false;
    info(QString::fromStdString(core->orbitlib->getOrbitDesc(np)->toString()));
}

void CSmileGUI::XYZtoXY(const double X, const double Y, const double Z, double* x, double* y)
{
    *x = (Y-X)/sqrt(2.0);
    *y = (2*Z-X-Y)/sqrt(6.0);
}

/* ----------------------------  Schwarzschild modelling --------- */
void CSmileGUI::on_ButtonSchwOrbitLib_clicked()
{
    if(core->isExportNbodyRunning())  // don't interfere with exporting to Nbody
        return;
    if(core->isSchwOptRunning())  // stop? noooo!!
    {
        error("Cannot stop optimization routine");
        return;
    }
    if(core->isOrbitLibRunning())
    {
        emit stopOrbitLibrary();
        return;
    }
    if(core->potential->N_dim!=3)
    {
        error("Schwarzschild modelling is only possible in 3 dimensions");
        return;
    }

    // start building library
    core->buildSchwIC(ui.checkBox_smExistingSS->isChecked());
    core->startOrbitLibrary(SLOT(coreOrbitLibraryFinished()));
    enableOrbitLibControls(false);
    ui.ButtonFreqMap->setEnabled(false);  // can't run freq.map while integrating orbits for Schw/model
    ui.ButtonSchwOrbitLib->setText("Stop");
    timeFM.start();
    ui.progressBarSM->setRange(0, static_cast<int>(core->orbitlib->size()));
    ui.progressBarSM->setValue(0);
    ui.progressBarSM->show();
}

void CSmileGUI::NbodyReintegrationFinished()    // slot called when core->signalNbodyReintegrationFinished() emitted
{
    ui.progressBarSM->hide();
    info("Reintegration finished, exporting Nbody file");
}

void CSmileGUI::NbodyExportFinished(const QString& result)    // slot called when core->signalschwExportNbodyFinished() emitted
{
    ui.ButtonSchwNbody->setText("Create Nbody IC");
    enableOrbitLibControls(true);
    ui.progressBarSM->hide();
    info(result);
}

void CSmileGUI::on_ButtonSchwLP_clicked()
{
    if(core->isExportNbodyRunning() || core->isOrbitLibRunning() || 
        core->isSchwOptRunning() || core->schwData.size()==0) return;
    enableOrbitLibControls(false);
    core->schwStartOptimization(false);
}

void CSmileGUI::on_ButtonSchwQP_clicked()
{
    if(core->isExportNbodyRunning() || core->isOrbitLibRunning() || 
        core->isSchwOptRunning() || core->schwData.size()==0) return;
    enableOrbitLibControls(false);
    core->schwStartOptimization(true);
}

void CSmileGUI::SchwOptimizationFinished(const QString& message)
{
    enableOrbitLibControls(true);
    ui.ButtonSchwNbody->setEnabled(true);
    SchwModelResult = message;
    redrawSchwModel(true);
}

void CSmileGUI::on_selectEnergyLevel_sm_valueChanged(int)
{
    redrawSchwModel(true);
}

void CSmileGUI::redrawSchwModel(bool updateInfo)
{
    if(core->schwData.size()==0 || !core->orbitlib)
        return;
    int numShell=ui.selectEnergyLevel_sm->value();
    size_t nstart=0, nend=0;      // range of cells to display
    vectorSchwData tmpSchwData;   // temporary vector for keeping only known Schw.density data objects
    // init one or more pointers to specific classes of Schw.data objects
    const CSchwDataKinemShell*  model_sh=NULL;
    const CSchwDataDensClassic* model_cl=NULL;
    const CSchwDataDensSH*      model_sph=NULL;
    for(vectorSchwData::iterator iter=core->schwData.begin(); iter!=core->schwData.end(); iter++)
        switch((*iter)->dataType()) {
            case CBasicSchwData::SD_KINEM_SHELL:
                model_sh=static_cast<const CSchwDataKinemShell*>(*iter); 
                break;
            case CBasicSchwData::SD_DENS_CLASSIC:
                model_cl=static_cast<const CSchwDataDensClassic*>(*iter);
                nstart=numShell==0 ? 0 : (numShell-1)*model_cl->getCellsPerShell();
                nend=numShell==0 ? (*iter)->numConstraints() : numShell*model_cl->getCellsPerShell();
                tmpSchwData.push_back(model_cl);
                break;
            case CBasicSchwData::SD_DENS_SHGRID:
            case CBasicSchwData::SD_DENS_SHBSE:
                model_sph=static_cast<const CSchwDataDensSH*>(*iter); 
                nend=(*iter)->numConstraints();
                tmpSchwData.push_back(model_sph);
                break;
            default: ;
        }
    if(model_cl) {
        ui.selectEnergyLevel_sm->setEnabled(true);
    } else {
        ui.selectEnergyLevel_sm->setEnabled(true);
        numShell=0;
    }
    if(numShell>0 && model_sh!=NULL)
        ui.lineEdit_E->setText(QString::number(model_sh->getShellEnergy(numShell-1)));
    if(ui.radioButton_smviewgrid->isChecked() && model_sh!=NULL)
    {
        vectord CellsFx, CellsFy, CellsIx, CellsIy;
        // create a temporary instance of Schw Model object to compute statistics
        CSchwModelQuadOpt model(core->orbitlib, tmpSchwData, NULL, NULL, 0, 0, 0, 0, 0, 0);
        double Xmin=0, Xmax=0, Ymin=0, Ymax=0;
        int nn=-1; double nnprev=-1;
        for(size_t c=nstart; c<nend; c++)
        {
            double Xp=0, Yp=0;
            if(model_cl)
            {
                double X=0,Y=0,Z=0;
                model_cl->cellCenter(c, &X, &Y, &Z);
                XYZtoXY(X,Y,Z,&Xp,&Yp);
                if(fabs(Xp)>Xmax) { Xmax=fabs(Xp); Xmin=-Xmax; }
                if(Yp>Ymax) { Ymax=Yp; }
                if(Yp<Ymin) { Ymin=Yp; }
            } 
            else if(model_sph)
            {
                double n;
                int l, m;
                model_sph->getCoefDecomposition(c, &n, &l, &m);
                l/=2; m/=2;
                if(n!=nnprev) { nnprev=n; nn++; }
                Xp=l*(l+1)/2 + m*(1.-1./(l+4));  // put the points in row but make somewhat larger spacing between l and l+1 series
                Yp=nn;
                Xmax=std::max<double>(Xmax,Xp);
                Ymax=std::max<double>(Ymax,Yp);
            }
            if(model.calcCoefDeviation(c)!=0)
            {
                CellsIx.push_back(Xp);
                CellsIy.push_back(Yp);
            }
            else
            {
                CellsFx.push_back(Xp);
                CellsFy.push_back(Yp);
            }
        }
        if(CellsFx.empty())
            CurveSchwCellF->setData(NULL, NULL, 0);
        else
            CurveSchwCellF->setData(&(CellsFx.front()), &(CellsFy.front()), static_cast<int>(CellsFx.size()));
        if(CellsIx.empty())
            CurveSchwCellI->setData(NULL, NULL, 0);
        else
            CurveSchwCellI->setData(&(CellsIx.front()), &(CellsIy.front()), static_cast<int>(CellsIx.size()));
        PlotSchw->setAxisScale(QwtPlot::xBottom, Xmin, Xmax);
        PlotSchw->setAxisScale(QwtPlot::yLeft, Ymin, Ymax);
        PlotSchw->replot();
        PlotSchwHisto->hide();
        PlotSchw->show();
        if(updateInfo)  info(SchwModelResult + "\n" + 
            getOrbitPopulation(numShell, true));
    }
    else if(ui.radioButton_smviewbeta->isChecked())
    {   // plot anisotropy coefficient beta
        if(model_sh)
        {
            size_t numKinemShells=model_sh->numConstraints();
            vectord val(nend), np(nend);
            for(size_t s=0; s<numKinemShells; s++)
            {
                double vr2=0, vt2=0;
                for(size_t o=0; o<core->orbitlib->size(); o++)
                {
                    const COrbitDesc* orb=core->orbitlib->getOrbitDesc(o);
                    if(orb->getState()==COrbitDesc::OS_DONE && orb->getWeight()>0)
                    {
                        // load kinematic information data from orbit desc
                        const CSchwInformation* infoSchwKinem=NULL;
                        for(size_t inf=0; inf<orb->getInfoNum(); inf++)
                        {
                            const CBasicInformation* info=orb->getInfo(inf);
                            if(info->infoType()==CBasicInformation::IT_SCHW && 
                                static_cast<const CSchwInformation*>(info)->schwDataType() == CBasicSchwData::SD_KINEM_SHELL) 
                                infoSchwKinem=static_cast<const CSchwInformation*>(info);
                        }
                        if(infoSchwKinem!=NULL && infoSchwKinem->getSchwData().size()==numKinemShells*3)
                        {
                            vr2 += infoSchwKinem->getSchwData(s + numKinemShells) * infoSchwKinem->getSchwData(s) * orb->getWeight();
                            vt2 += infoSchwKinem->getSchwData(s+2*numKinemShells) * infoSchwKinem->getSchwData(s) * orb->getWeight();
                        }
                    }
                }
                if(vr2>0)
                    val[s] = 1-vt2/vr2/2;
                else
                    val[s] = 1;
                np[s]=s*1.0;
            }
            if(numKinemShells>0)
                CurveSchwHistoR->setData(&(np.front()), &(val.front()), static_cast<int>(numKinemShells));
            else
                CurveSchwHistoR->setData(NULL, NULL, 0);
            CurveSchwHistoR->setStyle(QwtPlotCurve::Steps);
            CurveSchwHistoC->setData(NULL, NULL, 0);
        } else {
            CurveSchwHistoR->setData(NULL, NULL, 0);
            CurveSchwHistoC->setData(NULL, NULL, 0);
        }
        PlotSchwHisto->setAxisScale(QwtPlot::yLeft, -1, 1);
        PlotSchwHisto->setAxisAutoScale(QwtPlot::xBottom);
        PlotSchwHisto->setAxisScaleEngine(QwtPlot::yLeft, new QwtLinearScaleEngine);
        PlotSchwHisto->replot();
        PlotSchw->hide();
        PlotSchwHisto->show();
        if(updateInfo)  info(SchwModelResult + "\n" + 
            getOrbitPopulation(numShell, true));
    }
    else if(ui.radioButton_smviewweights->isChecked() || ui.radioButton_smviewhisto->isChecked())
    {   // plot histogram of orbit weights
        vectord valuesR, valuesC, npR, npC;
        double minval=1, maxval=0;
        const CShellOrbitFilteringFnc filter(model_sh, numShell);
        const CChaosOrbitFilteringFnc chaos(1, core->configCore.chaoticMinLambda, core->configCore.chaoticMinFreqDiff);
        for(size_t o=0; o<core->orbitlib->size(); o++)
        {
            const COrbitDesc* orb=core->orbitlib->getOrbitDesc(o);
            if(orb->getState()==COrbitDesc::OS_DONE && filter.eval(orb)>=0.5)
            {
                if(chaos.eval(orb)>=0.5 && !ui.radioButton_smviewhisto->isChecked())
                {
                    valuesC.push_back(orb->getWeight());
                    npC.push_back(o*1.0);
                }
                else
                {
                    valuesR.push_back(orb->getWeight());
                    npR.push_back(o*1.0);
                }
                if(orb->getWeight()>0) minval = std::min<double>(minval, orb->getWeight());
                maxval = std::max<double>(maxval, orb->getWeight());
            }
        }
        if(ui.radioButton_smviewhisto->isChecked()) 
        {
            std::sort(valuesR.begin(), valuesR.end());
            for(size_t i=1; i<valuesR.size(); i++) valuesR[i]+=valuesR[i-1];
            for(size_t i=0; i<valuesR.size(); i++) valuesR[i]/=valuesR.back();
            minval=0; maxval=1;
            CurveSchwHistoC->setData(NULL, NULL, 0);
            PlotSchwHisto->setAxisScaleEngine(QwtPlot::yLeft, new QwtLinearScaleEngine);
        }
        else
        {
            if(npC.size()==0)
                CurveSchwHistoC->setData(NULL, NULL, 0);
            else
                CurveSchwHistoC->setData(&(npC.front()), &(valuesC.front()), static_cast<int>(valuesC.size()));
            PlotSchwHisto->setAxisScaleEngine(QwtPlot::yLeft, new QwtLog10ScaleEngine);
        }
        if(npR.size()==0)
            CurveSchwHistoR->setData(NULL, NULL, 0);
        else
            CurveSchwHistoR->setData(&(npR.front()), &(valuesR.front()), static_cast<int>(valuesR.size()));
        CurveSchwHistoR->setStyle(ui.radioButton_smviewhisto->isChecked() ? QwtPlotCurve::Steps : QwtPlotCurve::Dots);
        PlotSchwHisto->setAxisScale(QwtPlot::yLeft, minval, maxval);
        //PlotSchwHisto->setAxisAutoScale(QwtPlot::xBottom);
        PlotSchwHisto->setAxisScale(QwtPlot::xBottom, 0, core->orbitlib->size());
        PlotSchwHisto->replot();
        PlotSchw->hide();
        PlotSchwHisto->show();
        if(updateInfo)  info(SchwModelResult + "\n" + 
            getOrbitPopulation(numShell, true));
    }
}

void CSmileGUI::SMclicked(const QPointF& pos)
{
    if(core->orbitlib==NULL || core->orbitlib->size()==0 || core->schwData.size()==0)
        return;
    double px=pos.x();
    double py=pos.y();
    double mindist=10000; 
    size_t numCoefs=0;
    // init one or more pointers to specific classes of Schw.data objects
    const CSchwDataDensClassic* model_cl=NULL;
    const CSchwDataDensSH*      model_sph=NULL;
    vectorSchwData tmpSchwData;         // temporary vector for keeping only known Schw.density data objects
    for(vectorSchwData::iterator iter=core->schwData.begin(); iter!=core->schwData.end(); iter++)
        switch((*iter)->dataType()) {
            case CBasicSchwData::SD_DENS_CLASSIC:
                model_cl=static_cast<const CSchwDataDensClassic*>(*iter);
                numCoefs=model_cl->numConstraints();
                tmpSchwData.push_back(model_cl);
                break;
            case CBasicSchwData::SD_DENS_SHGRID:
            case CBasicSchwData::SD_DENS_SHBSE:
                model_sph=static_cast<const CSchwDataDensSH*>(*iter); 
                numCoefs=model_sph->numConstraints();
                tmpSchwData.push_back(model_sph);
                break;
            default: ;
        }
    if(tmpSchwData.size()!=1) return;   // must have one and only one known density data type object
    int nc=-1, nn=-1; double nnprev=-1;
    for(size_t c=0; c<numCoefs; c++)
    {
        double Xp=0, Yp=0;
        if(model_cl)
        {
            double X=0,Y=0,Z=0;
            model_cl->cellCenter(c, &X, &Y, &Z);
            XYZtoXY(X,Y,Z,&Xp,&Yp);
        } 
        else if(model_sph)
        {
            double n;
            int l, m;
            model_sph->getCoefDecomposition(c, &n, &l, &m);
            l/=2; m/=2;
            if(n!=nnprev) { nnprev=n; nn++; }
            Xp=l*(l+1)/2 + m*(1.-1./(l+4));  // put the points in row but make somewhat larger spacing between l and l+1 series
            Yp=nn;
        }
        double dist = pow_2(Xp-px) + pow_2(Yp-py);
        if(dist<mindist)
        {
            nc=static_cast<int>(c);
            mindist=dist;
        }
    }
    if(nc<0) return;  // nc= index of cell/coefficient that was clicked
    // create a temporary instance of Schw Model object to compute penalty (deviation of a coef from required value)
    CSchwModelQuadOpt model(core->orbitlib, tmpSchwData, NULL, NULL, 0, 0, 0, 0, 0, 0);
    model.calcCoefDeviation(nc);
    double coefValue  = tmpSchwData[0]->getConstraint(nc);
    double normFactor = tmpSchwData[0]->getConstraintNormFactor(nc);
    double coefDeviation= model.calcCoefDeviation(nc)*normFactor;
    if(coefDeviation!=0) coefDeviation=coefDeviation/coefValue*100;  // relative deviation in percents
    info("Constraint index="+QString::number(nc) + "\nRequired value="+QString::number(coefValue) + 
        "\nDeviation="+QString::number(coefDeviation) + "%\nNormalization="+QString::number(normFactor));
}

void CSmileGUI::SMHistoClicked(const QPointF &pos)
{
    if(core->orbitlib==NULL || core->orbitlib->size()==0 || !ui.radioButton_smviewweights->isChecked())
        return;
    int numShell = ui.selectEnergyLevel_fm->value();
    const CShellOrbitFilteringFnc filter=shellFilter(core->schwData, numShell);
    double px=pos.x();
    double py=pos.y();
    // scale clicked value also to [0..1] interval
    double minyval=1, maxyval=0;
    size_t maxxval=core->orbitlib->size(), minxval=0;
    for(size_t i=0; i<core->orbitlib->size(); i++)
        if(core->orbitlib->getOrbitDesc(i)->getState()==COrbitDesc::OS_DONE && filter.eval(core->orbitlib->getOrbitDesc(i))>=0.5)
        {
            //if(i<minxval) minxval=i;
            //if(i>maxxval) maxxval=i;
            float weight=core->orbitlib->getOrbitDesc(i)->getWeight();
            if(weight>0)
            {
                minyval=std::min<double>(minyval, weight);
                maxyval=std::max<double>(minyval, weight);
            }
        }
    if(maxxval==0 || maxyval==0) return;
    minyval=log10(minyval);
    maxyval=log10(maxyval);
    px = (px-minxval)*1.0/(maxxval-minxval);
    py = (log10(py)-minyval)/(maxyval-minyval);
    // find nearest point
    double mindist=1; 
    size_t np=0;
    for(size_t i=0; i<core->orbitlib->size(); i++)
    {
        double dist=1;
        if(core->orbitlib->getOrbitDesc(i)->getWeight()>0 && core->orbitlib->getOrbitDesc(i)->getState()==COrbitDesc::OS_DONE && filter.eval(core->orbitlib->getOrbitDesc(i))>=0.5)
        {
            double ox=(i-minxval)*1.0/(maxxval-minxval);
            double oy=(log10(core->orbitlib->getOrbitDesc(i)->getWeight())-minyval)/(maxyval-minyval);
            dist = pow_2(ox-px) + pow_2(oy-py);
        }
        if(dist<mindist)
        {
            np=i;
            mindist=dist;
        }
    }
    if(mindist>=1) return;  // too far
    // load data from np'th orbit
    bool tmpLyapunov=core->InitData.calcLyapunov;
    core->InitData=core->orbitlib->getOrbitDesc(np)->getInitData();
    core->InitData.calcLyapunov=tmpLyapunov;
    // set exact values
    bool tmpUseE=core->useICe;
    core->useICe=false;  // need to put in values for IC, not for E
    core->initIC();
    core->useICe=tmpUseE;
    InternalRecalc=true;
    ui.lineEdit_x->setText(QString::number(core->InitData.initCond.Pos[0]));
    ui.lineEdit_y->setText(QString::number(core->InitData.initCond.Pos[1]));
    ui.lineEdit_z->setText(QString::number(core->InitData.initCond.Pos[2]));
    ui.lineEdit_vx->setText(QString::number(core->InitData.initCond.Vel[0]));
    ui.lineEdit_vy->setText(QString::number(core->InitData.initCond.Vel[1]));
    ui.lineEdit_vz->setText(QString::number(core->InitData.initCond.Vel[2]));
    ui.lineEdit_E->setText(QString::number(core->initCondE));
    ui.lineEdit_Torb->setText(QString::number(core->InitData.timeUnit));
    InternalRecalc=false;
    info(QString::fromStdString(core->orbitlib->getOrbitDesc(np)->toString()));
}

/*   -------------  export-import ------------- */
void CSmileGUI::on_ButtonSchwNbody_clicked()
{
    if(core->orbitlib==NULL || core->orbitlib->size()==0)
        return;
    if(core->isExportNbodyRunning())
    {   // stop export (determine which stage we are at now)
        if(core->isOrbitLibRunning()) 
        {   // reintegration stage - abort
            emit stopOrbitLibrary();
            return;
        } else {  // perhaps still computing virial ratio?
            emit stopExportNbody();  // if not, this will be discarded
            return;
        }
    }
    if(core->isOrbitLibRunning() || core->isSchwOptRunning())  // can't export while performing regular orbit library integration
        return;
    bool ok;
    int NPoints=QInputDialog::getInteger(this, "Export to N-body", "Enter number of points", 100000, 0, 1e8, 1, &ok);
    if(!ok) return;
    QString fileFormats, selectedFormat;
    for(size_t i=0; i<formatsIOSnapshot.size(); i++) {
        fileFormats += (fileFormats.length()>0 ? ";;" : "") + QString(formatsIOSnapshot[i].c_str()) + " (*)";
        if(i==0) selectedFormat=fileFormats;
    }
    QString fileName = QFileDialog::getSaveFileName(this, "Export to N-body", 
        core->WorkDir, fileFormats, &selectedFormat, QFileDialog::HideNameFilterDetails | QFileDialog::DontUseNativeDialog);
    selectedFormat.replace(" (*)", "");
    if(fileName.isEmpty())
        return;
    core->WorkDir = getDirName(fileName);
    int refineFactor=QInputDialog::getInteger(this, "Export to N-body", "Mass refine factor (2^N)", 0, 0, 6, 1, &ok);
    if(!ok) refineFactor=0;

    ui.ButtonSchwNbody->setText("Cancel");
    enableOrbitLibControls(false);
    ui.progressBarSM->setValue(0);
    ui.progressBarSM->setRange(0, static_cast<int>(core->orbitlib->size()));
    ui.progressBarSM->show();   // to display progress in orbit library reintegration, if it is necessary
    core->schwExportNbody(NPoints, fileName, selectedFormat, refineFactor);
}

void CSmileGUI::on_ButtonExportFM_clicked()
{
    if(core->orbitlib==NULL || core->orbitlib->size()==0)
        return;
    QString fileName = QFileDialog::getSaveFileName(this, "Export Frequency map", core->WorkDir);
    if (fileName.isEmpty())
        return;
    core->WorkDir = getDirName(fileName);

    if(!core->exportOrbitLib(fileName, false))
        error("Cannot write file "+fileName);
    core->saveSettings(fileName+".ini");
}

void CSmileGUI::on_ButtonImportFM_clicked()
{
    QString fileName = QFileDialog::getOpenFileName(this, "Import Frequency map", core->WorkDir);
    if (fileName.isEmpty())
        return;
    core->WorkDir = getDirName(fileName);
    loadApplySettings(fileName+".ini");   // try loading saved potential and other data
    if (!core->importOrbitLib(fileName, false)) {
        error("Cannot read file "+fileName);
        return;
    }
    redrawFreqMap(true);
}

void CSmileGUI::on_ButtonExportSM_clicked()
{
    if(core->orbitlib==NULL || core->orbitlib->size()==0 || core->schwData.size()==0)
        return;
    QString fileName = QFileDialog::getSaveFileName(this, "Export Schwarzschild model", core->WorkDir);
    if (fileName.isEmpty())
        return;
    core->WorkDir = getDirName(fileName);

    if(!core->exportOrbitLib(fileName, true))
        error("Cannot write file "+fileName);
    core->saveSettings(fileName+".ini");
}

void CSmileGUI::on_ButtonImportSM_clicked()
{
    QString fileName = QFileDialog::getOpenFileName(this, "Import Schwarzschild model", core->WorkDir);
    if (fileName.isEmpty())
        return;
    core->WorkDir = getDirName(fileName);
    loadApplySettings(fileName+".ini");   // try loading saved potential and other data
    if(core->potential->N_dim!=N_DIM)
    {
        error("N_dim should be equal to 3 for Schwarzschild modelling");
        return;
    }
    if(!core->importOrbitLib(fileName, true)) 
    {
        error("Cannot read file "+fileName);
        return;
    }
    ui.groupBox_smGrid->setEnabled(true);
    ui.groupBox_smOptim->setEnabled(true);
    ui.ButtonSchwNbody->setEnabled(true);
    SchwModelResult=ui.label_info->toPlainText();
    redrawFreqMap(true);
    on_selectEnergyLevel_sm_valueChanged(ui.selectEnergyLevel_sm->value());  // redraw grid
}

void CSmileGUI::on_ButtonExportOrbit_clicked()
{
    if(!core->orbit || !core->orbit->getIntTime())
        return;
#ifdef USE3DFACETS
    if(ui.radioButton_3dmesh->isChecked())
    {   // export mesh, not the trajectory
        QString fileName = QFileDialog::getSaveFileName(this, "Export 3d orbit shape to Wavefront OBJ file", core->WorkDir, "*.obj");
        if (fileName.isEmpty())
            return;
        core->WorkDir = getDirName(fileName);
        // export mesh to Wavefront OBJ text file, to render in an external application (e.g. Blender or ParaView)
        std::ofstream strm(fileName.toStdString().c_str(), std::ios_base::out);
        for(size_t t=0; t<TriangThr->trjpos.size(); t++)
            strm << "v " << TriangThr->trjpos[t].x << " " << TriangThr->trjpos[t].y << " " << TriangThr->trjpos[t].z << std::endl;
        for(size_t c=0; c<TriangThr->facets.size(); c++)
            strm << "f " << TriangThr->facets[c][0] <<" " << TriangThr->facets[c][1] <<" " << TriangThr->facets[c][2]<< std::endl; 
        if(!strm) error("Cannot write file "+fileName);
        return;
    }
#endif

    QString fileName = QFileDialog::getSaveFileName(this, "Export orbit", core->WorkDir);
    if (fileName.isEmpty())
        return;
    core->WorkDir = getDirName(fileName);
    if(!core->exportOrbit(fileName))
        error("Cannot write file "+fileName);
}

void CSmileGUI::on_ButtonImportOrbit_clicked()
{
    QString fileName = QFileDialog::getOpenFileName(this, "Import orbit", core->WorkDir);
    if (fileName.isEmpty())
        return;
    core->WorkDir = getDirName(fileName);
    if(!core->importOrbit(fileName)) {
        error("Cannot read file "+fileName);
        return;
    }
    InternalRecalc=true;
    ui.lineEdit_x->setText(QString::number(core->InitData.initCond.Pos[0]));
    ui.lineEdit_y->setText(QString::number(core->InitData.initCond.Pos[1]));
    ui.lineEdit_z->setText(QString::number(core->InitData.initCond.Pos[2]));
    ui.lineEdit_vx->setText(QString::number(core->InitData.initCond.Vel[0]));
    ui.lineEdit_vy->setText(QString::number(core->InitData.initCond.Vel[1]));
    ui.lineEdit_vz->setText(QString::number(core->InitData.initCond.Vel[2]));
    ui.lineEdit_E->setText(QString::number(core->initCondE));
    ui.lineEdit_Torb->setText(QString::number(core->InitData.timeUnit));
    ui.lineEdit_intTime->setText(QString::number(core->configCore.intTimeInPeriods));
    ui.lineEdit_timeStep->setText(QString::number(core->configCore.intTimeStepsPerPeriod));
    ui.spinBox_intervalCount->setValue(1);
    InternalRecalc=false;
    refreshOrbitAndSpectrum();
}

void CSmileGUI::on_ButtonExportModelStats_clicked()
{
    if(core->orbitlib==NULL || core->orbitlib->size()==0 || core->schwData.size()==0)
        return;
    QString fileName = QFileDialog::getSaveFileName(this, "Export Schwarzschild model statistics", core->WorkDir);
    if (fileName.isEmpty())
        return;
    core->WorkDir = getDirName(fileName);
    if(!core->exportSchwModelStats(fileName))
        error("Cannot write file "+fileName);
}

void CSmileGUI::on_ButtonExportPS_clicked()
{
    if(CurvesPoincare.size()<=1) return;
    QString fileName = QFileDialog::getSaveFileName(this, "Export Poincare surface of section", core->WorkDir);
    if (fileName.isEmpty())
        return;
    core->WorkDir = getDirName(fileName);
    std::ofstream strm(fileName.toStdString().c_str(), std::ios::out);
    if(!strm) { error("Cannot write file "+fileName); return; }
    std::vector< paird > curve;
    for(int c=1; c<CurvesPoincare.size(); c++)
    {
        curve.resize(CurvesPoincare[c]->data().size());
        for(size_t i=0; i<CurvesPoincare[c]->data().size(); i++)
        {
            curve[i].first =CurvesPoincare[c]->data().x(i);
            curve[i].second=CurvesPoincare[c]->data().y(i);
        }
        // sort points in curve in nearest-neighbour order
        for(size_t i=0; i<curve.size()-1; i++)
        {
            double distmin=1e10; 
            size_t indmin=0;
            // find next unused point closest to i-th 
            for(size_t j=i+1; j<curve.size(); j++)
            {
                double dist=pow_2(curve[i].first-curve[j].first)+pow_2(curve[i].second-curve[j].second);
                if(dist<distmin) { distmin=dist; indmin=j; }
            }
            if(indmin>0)
            {
                paird tmp(curve[indmin]);
                curve[indmin]=curve[i+1];
                curve[i+1]=tmp;
            }
        }
        for(size_t i=0; i<curve.size()-1; i++)
            strm << curve[i].first << "\t" << curve[i].second << "\n";
        strm << "\n";
    }
}

void CSmileGUI::on_ButtonPrint_clicked()
{
    QwtPlot* fig = NULL;  
    if(PlotOrbit->isVisible())
        fig = PlotOrbit;
    if(PlotPoincare->isVisible())
        fig = PlotPoincare;
    if(PlotFreq->isVisible())
        fig = PlotFreq;
    if(PlotLyapunov->isVisible())
        fig = PlotLyapunov;
    if(PlotFreqMap->isVisible())
        fig = PlotFreqMap;
    if(PlotFreqMapHisto->isVisible())
        fig = PlotFreqMapHisto;
    if(PlotFreqMapSS->isVisible())
        fig = PlotFreqMapSS;
    if(PlotSchw->isVisible())
        fig = PlotSchw;
    if(PlotSchwHisto->isVisible())
        fig = PlotSchwHisto;
#ifdef HAVE_QWT3D
    bool Plot3dVisible=Plot3d->isVisible();
    // 3d plot is not inherited from QwtPlot, so treat it differently
#else
    bool Plot3dVisible=false;
#endif
    if(fig==NULL && !Plot3dVisible)
        return;
    QString fileName = QFileDialog::getSaveFileName(this, "Print figure to file", core->WorkDir, "PostScript (*.ps);;PDF (*.pdf)");
    if (fileName.isEmpty())
        return;
    core->WorkDir = getDirName(fileName);
    if(fig)
    {
        QPrinter P;
        P.setOutputFileName(fileName);
        fig->print(P);
    }
#ifdef HAVE_QWT3D
    else if(Plot3dVisible)
    {
        QString Format=fileName.endsWith(".pdf", Qt::CaseInsensitive) ? "PDF" : "PS";
        Qwt3D::VectorWriter* printer = (Qwt3D::VectorWriter*)Qwt3D::IO::outputHandler(Format);
        printer->setTextMode(Qwt3D::VectorWriter::NATIVE);
        printer->setSortMode(Qwt3D::VectorWriter::BSPSORT);
        printer->setLandscape(Qwt3D::VectorWriter::OFF);
        Qwt3D::IO::save(Plot3d, fileName, Format);
    }
#endif
}

/* ------------- Triangulation thread helper class -------------- */
#ifdef USE3DFACETS
CTriangThread::CTriangThread(const QString& _appPath) : appPath(_appPath)
{
    finish = restart = false;
}

CTriangThread::~CTriangThread()
{
    mutex.lock();
    finish=restart=true;
    condition.wakeOne();
    mutex.unlock();
}
 
void CTriangThread::run()
{
    while(!finish)
    {
        createTriangulation();
        mutex.lock();
        if(!restart)
            condition.wait(&mutex);
        restart=false;
        mutex.unlock();
    }
}

void CTriangThread::startTriangulation(const std::vector<CPosPoint<double> > &_trajp, double _maxdist, int _slice)
{
    QMutexLocker locker(&mutex);
    // store data 
    nPointsMy = static_cast<int>(_trajp.size());
    trajpMy=_trajp;
    sliceMy = _slice;
    maxdistMy = _maxdist;
    // start or restart thread
    if(!isRunning())
        start(QThread::LowPriority);
    else
    {    
        restart=true;
        condition.wakeOne();
    }
}

bool CTriangThread::toolong(int v1, int v2)
{
    return (pow_2(trjpos[v1].x-trjpos[v2].x) + pow_2(trjpos[v1].y-trjpos[v2].y) + pow_2(trjpos[v1].z-trjpos[v2].z) > pow_2(maxdist));
}
void CTriangThread::addface(int v1, int v2, int v3)
{
    CTriangle T(v1, v2, v3);
    if(Triangles.remove(T)) 
        T.dup=true;
    Triangles << T;
}

void CTriangThread::createTriangulation()
{
    {
        Triangles.clear();
        // before proceeding further, copy data to local variables
        mutex.lock();
        facets.clear();
        trjpos.clear();
        maxdist=maxdistMy;
        int coord = sliceMy-1;
        int nPoints=0;
        for(int i=0; i<nPointsMy; i++)
        {
            if((sliceMy==0) || (trajpMy[i].Pos[coord]>=0))
            {
                trjpos.push_back(Qwt3D::Triple(trajpMy[i].Pos[0], trajpMy[i].Pos[1], trajpMy[i].Pos[2]));
                nPoints++;
            }
        }
        mutex.unlock();

        QProcess process;
        process.setReadChannel(QProcess::StandardOutput);
        process.start(appPath, QStringList("i"));
        if(!process.waitForStarted()) return;
        QTextStream strmI(&process);
        strmI << "3" << endl << nPoints << endl;
        for(int i=0; i<nPoints; i++)
        {
            strmI << trjpos[i].x << " " << trjpos[i].y << " " << trjpos[i].z << endl;
            if(restart) return;
        }
        process.closeWriteChannel();
        int nttot=0, nt=0;
        if(process.waitForFinished())
        {
            QTextStream strmO(&process);
            QString ln = strmO.readLine();
            nttot = ln.toInt();
            while(!strmO.atEnd())
            {
                ln = strmO.readLine();
                QStringList abcd=ln.split(" ");
                if(abcd.size() >= 4)
                {
                    int v1=abcd[0].toInt();
                    int v2=abcd[1].toInt();
                    int v3=abcd[2].toInt();
                    int v4=abcd[3].toInt();
                    if((v1>=nPoints) || (v2>=nPoints) || (v3>=nPoints) || (v4>=nPoints)) continue;
                    if(!(toolong(v1,v2) || toolong(v1,v3) || toolong(v1,v4) ||
                         toolong(v2,v3) || toolong(v2,v4) || toolong(v3,v4)))
                    {
                        addface(v1, v2, v3);
                        addface(v1, v2, v4);
                        addface(v1, v3, v4);
                        addface(v2, v3, v4);
                        nt++;
                    }
                }
                if(restart) return;
            }
        }
        // update faces
        int ntot=0, nout=0;
        QList<CTriangle> TList = Triangles.toList();
        Qwt3D::Cell tricell(3);
        foreach (CTriangle T, TList)
        {
            ntot++;
            if(!T.dup)
            {
                tricell[0] = T.c1;
                tricell[1] = T.c2;
                tricell[2] = T.c3;
                facets.push_back(tricell);
                nout++;
            }
            if(restart) return;
        }
        emit finished(true);
    }
}

#endif

QwtText QwtPlotPicker2::trackerText( const QPointF &pos ) const
{
    QString text;

    switch ( rubberBand() )
    {
        case HLineRubberBand:
            text.sprintf( "%.4g", pos.y() );
            break;
        case VLineRubberBand:
            text.sprintf( "%.4g", pos.x() );
            break;
        default:
            text.sprintf( "%.4g, %.4g", pos.x(), pos.y() );
    }
    return QwtText( text );
}

QwtText QwtPlotZoomer2::trackerText( const QPointF &pos ) const
{
    QString text;

    switch ( rubberBand() )
    {
        case HLineRubberBand:
            text.sprintf( "%.4g", pos.y() );
            break;
        case VLineRubberBand:
            text.sprintf( "%.4g", pos.x() );
            break;
        default:
            text.sprintf( "%.4g, %.4g", pos.x(), pos.y() );
    }
    return QwtText( text );
}

}  // namespace