#include "iodata.h"
#include "potential.h"
#include "orbitan.h"
#include "orbitlib.h"
#include "utils.h"

namespace smile{

/* ----- Input from text file (and possibly additional binary files with bulky orbit data) ---- */

void CFileInputText::open(const std::string &_baseFileName)
{
    baseFileName=_baseFileName;
    if(isopen) 
    {
        isopen=false;
        if(strmText.is_open()) strmText.close();
#ifndef HAVE_HDF5
        if(strmSchwData.is_open()) strmSchwData.close();
        if(strmTrajSample.is_open()) strmTrajSample.close();
#else
        if(h5_fileSchwData) H5Fclose(h5_fileSchwData);
        if(h5_datasetSchwData) H5Dclose(h5_datasetSchwData);
        if(h5_fileTrajSample) H5Fclose(h5_fileTrajSample);
        if(h5_datasetTrajSample) H5Dclose(h5_datasetTrajSample);
#endif
    }
#ifdef HAVE_HDF5
    // these variables will be initialized when readOrbitLibrary is called
    h5_fileSchwData=0;
    h5_datasetSchwData=0;
    h5_fileTrajSample=0;
    h5_datasetTrajSample=0;
#endif
    status=true;
    strmText.open(baseFileName.c_str(), std::ios::in);
    if(!strmText) { status=false; return; }
    isopen=status;
}

CFileInputText::~CFileInputText()
{
    if(strmText.is_open()) strmText.close();
#ifndef HAVE_HDF5
    if(strmSchwData.is_open()) strmSchwData.close();
    if(strmTrajSample.is_open()) strmTrajSample.close();
#else
    if(h5_fileSchwData) H5Fclose(h5_fileSchwData);
    if(h5_datasetSchwData) H5Dclose(h5_datasetSchwData);
    if(h5_fileTrajSample) H5Fclose(h5_fileTrajSample);
    if(h5_datasetTrajSample) H5Dclose(h5_datasetTrajSample);
#endif
}

COrbit* CFileInputText::readOrbit(const CConfigOrbit* configOrbit, const CPotential* potential)
{
    if(!isopen || !strmText) return NULL;
    CPosVelDataDouble trj;
    double time=0;
    int countlines=0;
    std::string buffer;
    std::vector<std::string> fields;
    while(std::getline(strmText, buffer) && !strmText.eof())
    {
        if(buffer.size()>0 && ((buffer[0]>='0' && buffer[0]<='9') || buffer[0]=='-' || buffer[0]=='+'))
        {
            splitString(buffer, " \t", &fields);
            if(fields.size()<7)
                return NULL;
            time=convertTo<double>(fields[0]);
            trj.push_back(CPosVelPoint<double>(   // don't check for errors in conversion, assume value=0 instead
                convertTo<double>(fields[1]), convertTo<double>(fields[2]), convertTo<double>(fields[3]),
                convertTo<double>(fields[4]), convertTo<double>(fields[5]), convertTo<double>(fields[6]) ));
            countlines++;
        }
    }
    if(countlines>1)
        return new COrbit(configOrbit, potential, time/(countlines-1) /*timeStep*/, 0 /*timeUnit=autodetect*/, trj);
    else return NULL;
}

COrbitDesc* CFileInputText::readOrbitDesc(const CConfigOrbit* configOrbit, const CPotential* potential, bool withModelData)
{
    if(!isopen || !strmText) return NULL;
    std::string buffer;
    std::vector<std::string> fields;
    if(std::getline(strmText, buffer) && buffer.size()>0 && ((buffer[0]>='0' && buffer[0]<='9') || buffer[0]=='-' || buffer[0]=='+'))
    {
        try{
            std::vector<std::string> fields;
            splitString(buffer, " \t", &fields);
            size_t nf=fields.size();
            if(nf<3)
                return NULL;
            vectorInformation info;
            COrbitInitData<float> InitData(potential, 
                nf>8 ? convertTo<float>(fields[8]) : 0, 
                nf>8 ? convertTo<float>(fields[7]) : 0,
                CPosVelPoint<float>( convertTo<float>(fields[0]), convertTo<float>(fields[1]), convertTo<float>(fields[2]), 
                nf>5 ? convertTo<float>(fields[3]) : 0, nf>5 ? convertTo<float>(fields[4]) : 0, nf>5 ? convertTo<float>(fields[5]) : 0), 
                nf>17 ? convertTo<float>(fields[17])>=0 /* calcLyapunov: if Lyapunov exp.!=-1, then it was at least calculated */ : 0);
            float weight= nf>6 ? convertTo<float>(fields[6]) : 0;
            float intTime=nf>9 ? convertTo<float>(fields[9]) : 0;
            float maxTime=nf>10? convertTo<float>(fields[10]) : 0;
            float Einit  =nf>11? convertTo<float>(fields[11]) : 0;
            float Ediff  =nf>12? convertTo<float>(fields[12]) : 0;
            if(InitData.timeUnit==0)
            {
                if(Einit==0) Einit=static_cast<float>(potential->totalEnergy(InitData.initCond));
                InitData.timeUnit =static_cast<float>(potential->longaxisperiod(Einit));
            }
            float lf[N_DIM]={0,0,0};
            if(nf>15) {
                lf[0]=convertTo<float>(fields[13]);
                lf[1]=convertTo<float>(fields[14]);
                lf[2]=convertTo<float>(fields[15]);
            }
            float lfdiff =nf>16? convertTo<float>(fields[16]) : 0;
            float lambda =nf>17? convertTo<float>(fields[17]) : -1;
            std::string Description=nf>18 ? fields[18] : "empty";
            std::string::size_type ind;
            while((ind = Description.find("_")) != std::string::npos)
                Description=Description.replace(ind, 1, " ");
            float inert[N_DIM]={0,0,0};
            if(nf>21) {
                inert[0]=convertTo<float>(fields[19]);
                inert[1]=convertTo<float>(fields[20]);
                inert[2]=convertTo<float>(fields[21]);
            }
            float Lavg[N_DIM]={0,0,0};
            float Lvar[N_DIM]={0,0,0};
            if(nf>27) {
                Lavg[0]=convertTo<float>(fields[22]);
                Lvar[0]=convertTo<float>(fields[23]);
                Lavg[1]=convertTo<float>(fields[24]);
                Lvar[1]=convertTo<float>(fields[25]);
                Lavg[2]=convertTo<float>(fields[26]);
                Lvar[2]=convertTo<float>(fields[27]);
            }
            info.push_back(new COrbitInformation<float>(Description, Einit, Ediff, lf, lfdiff, lambda, inert, Lavg, Lvar));
            if(nf>32) {
                float fitIntercept=convertTo<float>(fields[28]);
                float fitSlope=convertTo<float>(fields[29]);
                float fitScatter=convertTo<float>(fields[30]);
                float fitSignificance=convertTo<float>(fields[31]);
                float Lcirc2=convertTo<float>(fields[32]);
                info.push_back(new CPericenterInformation<float>(fitIntercept, fitSlope, fitScatter, fitSignificance, Lcirc2));
            }
#ifdef HAVE_HDF5
            if(withModelData && h5_fileSchwData)
            {
                // obtain the file dataspace and datatype
                hid_t h5_dataspaceSchwData = H5Dget_space(h5_datasetSchwData);
                hid_t h5_typedefSchwData = H5Dget_type(h5_datasetSchwData);
                // create a memory dataspace of one element
                hid_t h5_memspaceSchwData = H5Screate(H5S_SCALAR);
                // select an element in the file dataset to read the data from it
                hsize_t dsetstart[1] = { orbitIndex };
                hsize_t dsetcount[1] = { 1 };
                H5Sselect_hyperslab(h5_dataspaceSchwData, H5S_SELECT_SET, dsetstart, NULL, dsetcount, NULL);
                // reserve space for the entire collection of Schw.data arrays for this orbit
                size_t schwsize=0;
                for(size_t i=0; i<schwsizes.size(); i++) schwsize+=schwsizes[i];
                std::vector<sdNumType> schwDataArray(schwsize);
                // read all Schw.data elements for the orbit into the temp.array
                herr_t h5_status = H5Dread(h5_datasetSchwData, h5_typedefSchwData, h5_memspaceSchwData, 
                                           h5_dataspaceSchwData, H5P_DEFAULT, &(schwDataArray.front()));
                if(h5_status<0) status=false;
                else {
                    // de-serialize Schw.data arrays from the temp.array
                    size_t offset=0;
                    for(size_t inddata=0; inddata<schwtypes.size(); inddata++) {
                        std::vector<sdNumType> schwdata;
                        schwdata.assign(schwDataArray.begin()+offset, schwDataArray.begin()+offset+schwsizes[inddata]);
                        info.push_back(new CSchwInformation(schwtypes[inddata], schwdata));
                        offset+=schwsizes[inddata];
                    }
                }
                // close the dataspaces and typedef
                H5Sclose(h5_dataspaceSchwData);
                H5Sclose(h5_memspaceSchwData);
                H5Tclose(h5_typedefSchwData);
            }
            
            if(withModelData && h5_fileTrajSample)
            {
                // obtain the file dataspace and datatype
                hid_t h5_dataspaceTrajSample = H5Dget_space(h5_datasetTrajSample);
                hid_t h5_typedefTrajSample = H5Dget_type(h5_datasetTrajSample);
                // create a memory dataspace of one element
                hid_t h5_memspaceTrajSample = H5Screate(H5S_SCALAR);
                // select an element in the file dataset to read the data from it
                hsize_t dsetstart[1] = { orbitIndex };
                hsize_t dsetcount[1] = { 1 };
                H5Sselect_hyperslab(h5_dataspaceTrajSample, H5S_SELECT_SET, dsetstart, NULL, dsetcount, NULL);
                // create hdf5 variable-length data element
                hvl_t h5_varlengthdata;
                // read one element from the file
                herr_t h5_status = H5Dread(h5_datasetTrajSample, h5_typedefTrajSample, h5_memspaceTrajSample, 
                                           h5_dataspaceTrajSample, H5P_DEFAULT, &h5_varlengthdata);
                size_t nump=h5_varlengthdata.len;
                if(h5_status<0) status=false;
                else if(nump>0) {
                    // copy data from temp var-len array to a new TrajSample
                    std::vector< CPosVelPoint<float> > data(nump);
                    float* srcdata = static_cast<float*>(h5_varlengthdata.p);
                    for(size_t i=0; i<nump; i++) {
                        data[i].Pos[0] = srcdata[i*6+0];
                        data[i].Pos[1] = srcdata[i*6+1];
                        data[i].Pos[2] = srcdata[i*6+2];
                        data[i].Vel[0] = srcdata[i*6+3];
                        data[i].Vel[1] = srcdata[i*6+4];
                        data[i].Vel[2] = srcdata[i*6+5];
                    }
                    float x=data.back().Pos[0]; 
                    if(x==0) data.pop_back();
                    info.push_back( new CTrajSampleInformation<float>(data) );
                }
                // free memory allocated for the hdf5 variable-length array
                H5Dvlen_reclaim(h5_typedefTrajSample, h5_memspaceTrajSample, H5P_DEFAULT, &h5_varlengthdata);
                // close the dataspaces and typedef
                H5Sclose(h5_dataspaceTrajSample);
                H5Sclose(h5_memspaceTrajSample);
                H5Tclose(h5_typedefTrajSample);
            }
#else
            if(withModelData && strmSchwData.is_open() && strmSchwData.good())
            {
                unsigned int numdata;
                strmSchwData.read( reinterpret_cast<char*>(&numdata), sizeof(numdata) );
                for(size_t infoind=0; infoind<numdata; infoind++)
                {
                    // for each item of Schwarzschild data of the given orbit, store the datatype, data length and bulk data array
                    unsigned int datasize=0;
                    CBasicSchwData::SCHWDATATYPE datatype;
                    strmSchwData.read( reinterpret_cast<char*>(&datatype), sizeof(datatype) );
                    strmSchwData.read( reinterpret_cast<char*>(&datasize), sizeof(datasize) );
                    std::vector<sdNumType> schwdata(datasize);
                    if(datasize>0) strmSchwData.read( reinterpret_cast<char*>(&(schwdata.front()) ), datasize*sizeof(sdNumType));
                    info.push_back(new CSchwInformation(datatype, schwdata));
                }
                if(!strmSchwData) status=false;
            }
            if(withModelData && strmTrajSample.is_open() && strmTrajSample.good())
            {
                unsigned int nump=0;
                strmTrajSample.read( reinterpret_cast<char*>(&nump), sizeof(unsigned int) );
                if(nump>0)
                {
                    std::vector< CPosVelPoint<float> > data(nump);
                    if(nump>0) strmTrajSample.read( reinterpret_cast<char*>( &(data.front()) ), nump * sizeof(CPosVelPoint<float>));
                    info.push_back( new CTrajSampleInformation<float>(data) );
                }
                if(!strmTrajSample) status=false;
            }
#endif
            orbitIndex++;
            if(!strmText || !status) {
                status=false; 
                for(size_t ind=0; ind<info.size(); ind++) delete info[ind];
                return NULL;
            }
            return new COrbitDesc(configOrbit, InitData, intTime, maxTime, weight, info);
        }
        catch(const std::bad_alloc&) {
            return NULL;
        }
    }
    else return NULL;
}
    
COrbitLibrary* CFileInputText::readOrbitLibrary(const CConfigOrbit* configOrbit, const CPotential* potential, bool withModelData)
{
    if(!status || !isopen) return NULL;
    COrbitLibrary* orbitlib=NULL;
    orbitIndex=0;
    if(withModelData)
    {
#ifdef HAVE_HDF5
        bool success=true;   // flag telling that the HDF5 file exists and is of appropriate format
        // open the HDF5 file for Schwarzschild data
        h5_fileSchwData = H5Fopen((baseFileName+".schw.h5").c_str(), H5F_ACC_RDONLY, H5P_DEFAULT);
        if(h5_fileSchwData<0) success=false;
        if(success) {
            // open the dataset in the file
            h5_datasetSchwData = H5Dopen2(h5_fileSchwData, H5DATASETNAME_SCHW, H5P_DEFAULT);
            if(h5_datasetSchwData<0) success=false;
        }
        hid_t h5_typedefSchwData;
        if(success) {
            // obtain the data type stored in the file, it describes the number and type of Schwarzschild data records for each orbit
            h5_typedefSchwData = H5Dget_type(h5_datasetSchwData);
            if(H5Tget_class(h5_typedefSchwData) != H5T_COMPOUND) success=false;
        }
        if(success) {
            int numdata = H5Tget_nmembers(h5_typedefSchwData);
            for(int inddata=0; success && inddata<numdata; inddata++) {
                hid_t h5_typedefSchwDataElement = H5Tget_member_type(h5_typedefSchwData, inddata);
                if(H5Tget_class(h5_typedefSchwDataElement) != H5T_ARRAY) success=false;
                std::string dataname(H5Tget_member_name(h5_typedefSchwData, inddata));
                CBasicSchwData::SCHWDATATYPE datatype=getSchwDataTypeByName(dataname);
                if(getSchwDataNameByType(datatype) != dataname) success=false;
                hsize_t datasize[1];
                if(H5Tget_array_ndims(h5_typedefSchwDataElement)==1) 
                    H5Tget_array_dims(h5_typedefSchwDataElement, datasize);
                else success=false;
                // atomic type used in the data storage (float or double)
                hid_t h5_typedefNumType = sizeof(sdNumType)==sizeof(float) ? H5T_NATIVE_FLOAT : sizeof(sdNumType)==sizeof(double) ? H5T_NATIVE_DOUBLE : H5I_INVALID_HID;
                if(H5Tequal(H5Tget_super(h5_typedefSchwDataElement), h5_typedefNumType)<=0) success=false;  // in fact this is not required; one could imagine converting e.g. from float to double
                if(success) {
                    schwtypes.push_back(datatype);
                    schwsizes.push_back(static_cast<size_t>(datasize[0]));
                }
            }
        }
        if(success) {
            // obtain the dimensions of the dataset and ensure that it is a 1d array
            hid_t h5_dataspaceSchwData = H5Dget_space(h5_datasetSchwData);
            if(H5Sget_simple_extent_ndims(h5_dataspaceSchwData) != 1) success=false;
            H5Sclose(h5_dataspaceSchwData);
        }
        if(!success) {
            if(h5_datasetSchwData) H5Dclose(h5_datasetSchwData);
            if(h5_fileSchwData) H5Fclose(h5_fileSchwData);
            h5_fileSchwData=0;
            h5_datasetSchwData=0;
            status=false;
        }
        
        // open the HDF5 file for trajectory sample
        h5_fileTrajSample = H5Fopen((baseFileName+".smpl.h5").c_str(), H5F_ACC_RDONLY, H5P_DEFAULT);
        if(h5_fileTrajSample<0) success=false;
        if(success) {
            // open the dataset in the file
            h5_datasetTrajSample = H5Dopen2(h5_fileTrajSample, H5DATASETNAME_TRAJ, H5P_DEFAULT);
            if(h5_datasetTrajSample<0) success=false;
        }
        if(success) {
            // obtain the data type stored in the file and check that it is a variable-length array of 3+3 floats
            hsize_t dims3[1] = { 3 };
            hid_t h5_typedefTripleFloat = H5Tarray_create(H5T_NATIVE_FLOAT, 1, dims3);
            hid_t h5_typedefPosVelPoint = H5Tcreate(H5T_COMPOUND, sizeof(CPosVelPoint<float>));
            H5Tinsert(h5_typedefPosVelPoint, "Pos", /*HOFFSET(CPosVelPoint<float>, Pos)*/ 0*sizeof(float), h5_typedefTripleFloat);
            H5Tinsert(h5_typedefPosVelPoint, "Vel", /*HOFFSET(CPosVelPoint<float>, Vel)*/ 3*sizeof(float), h5_typedefTripleFloat);
            hid_t h5_typedefTrajSample = H5Tvlen_create(h5_typedefPosVelPoint);
            if(H5Tget_class(h5_typedefTrajSample) != H5T_VLEN /* || H5Tget_super(h5_typedefTrajSample) != H5T_FLOAT*/) success=false;
            // obtain the dimensions of the dataset and ensure that it is a 1d array
            hid_t h5_dataspaceTrajSample = H5Dget_space(h5_datasetTrajSample);
            if(H5Sget_simple_extent_ndims(h5_dataspaceTrajSample) != 1) success=false;
            H5Sclose(h5_dataspaceTrajSample);
            H5Tclose(h5_typedefTripleFloat);
            H5Tclose(h5_typedefPosVelPoint);            
            H5Tclose(h5_typedefTrajSample);
        }
        if(!success) {
            // no failure -- continue as if no traj sample file is present
            if(h5_datasetTrajSample) H5Dclose(h5_datasetTrajSample);
            if(h5_fileTrajSample) H5Fclose(h5_fileTrajSample);
            h5_fileTrajSample=0;
            h5_datasetTrajSample=0;
        }
#else
        strmSchwData.open((baseFileName+".schw").c_str(), std::ios::in | std::ios::binary);
        if(!strmSchwData) status=false;
        strmTrajSample.open((baseFileName+".smpl").c_str(), std::ios::in | std::ios::binary);
        // no error if smpl doesn't exist
#endif
    }
    while(strmText && status && !strmText.eof())
    {
        COrbitDesc* od=readOrbitDesc(configOrbit, potential, withModelData);
        if(od!=NULL)
        {
            if(orbitlib==NULL) orbitlib=new COrbitLibrary(configOrbit, potential, NULL);
            orbitlib->addOrbitDesc(od);
        }
    }
    return orbitlib;
}

/* ----- Output to text file (and possibly additional binary files with bulky orbit data) ---- */

void CFileOutputText::open(const std::string &_baseFileName)
{
    baseFileName=_baseFileName;
    if(isopen) 
    {
        isopen=false;
        if(strmText.is_open()) strmText.close();
#ifndef HAVE_HDF5
        if(strmSchwData.is_open()) strmSchwData.close();
        if(strmTrajSample.is_open()) strmTrajSample.close();
#else
        if(h5_fileSchwData) H5Fclose(h5_fileSchwData);
        if(h5_datasetSchwData) H5Dclose(h5_datasetSchwData);
        if(h5_fileTrajSample) H5Fclose(h5_fileTrajSample);
        if(h5_datasetTrajSample) H5Dclose(h5_datasetTrajSample);
#endif
    }
    started=false;
    status=true;
#ifdef HAVE_HDF5
    h5_fileTrajSample=0;
    h5_datasetTrajSample=0;
    h5_fileSchwData=0;
    h5_datasetSchwData=0;
#endif
    strmText.open(baseFileName.c_str(), std::ios::out);
    if(!strmText) { status=false; return; }
    isopen=status;
}

CFileOutputText::~CFileOutputText()
{
    if(strmText.is_open()) strmText.close();
#ifndef HAVE_HDF5
    if(strmSchwData.is_open()) strmSchwData.close();
    if(strmTrajSample.is_open()) strmTrajSample.close();
#else
    if(h5_fileTrajSample) H5Fclose(h5_fileTrajSample);
    if(h5_datasetTrajSample) H5Dclose(h5_datasetTrajSample);
    if(h5_fileSchwData) H5Fclose(h5_fileSchwData);
    if(h5_datasetSchwData) H5Dclose(h5_datasetSchwData);
#endif
}

bool CFileOutputText::writeOrbit(const COrbit* orbit)
{
    if(!status || !isopen) return false;
    const COrbitRuntimeTrajectory* trj = NULL;
    const COrbitInitData<double> InitData=orbit->getInitData();
    for(size_t rf=0; rf<orbit->getInfoNum(); rf++)
        if(orbit->getRuntimeFnc(rf)->FncType() == CBasicOrbitRuntimeFnc::FT_TRAJ_ANALYSIS)
            trj = static_cast<const COrbitRuntimeTrajectory*>(orbit->getRuntimeFnc(rf));
    if(trj==NULL || trj->getTrajSize()==0) 
        return false;
    if(!started)
    {
        strmText << "t\tx\ty\tz\tvx\tvy\tvz" << std::endl;
        started=true;
    }
    for(unsigned int indx=0; indx<trj->getTrajSize(); indx++)
    {
        CPosVelPoint<double> pt(trj->getTraj(indx));
        strmText << (indx*InitData.timeStep) << "\t" << 
            pt.Pos[0] << "\t" << pt.Pos[1] << "\t" << pt.Pos[2] << "\t" << 
            pt.Vel[0] << "\t" << pt.Vel[1] << "\t" << pt.Vel[2] << std::endl;
        if(!strmText) status=false;
    }
    return status;
}

bool CFileOutputText::writeOrbitDesc(const COrbitDesc* value, bool withModelData)
{
    if(!status || !isopen) return false;
    const COrbitInformation<float>* infoOrbit=NULL;
    const CPericenterInformation<float>* infoPeri=NULL;
    const CTrajSampleInformation<float>* infoTrajSample=NULL;
    std::vector<const CSchwInformation*> infoSchw;   // list of all existing Schwarzschild data objects
    for(size_t inf=0; inf<value->getInfoNum(); inf++)
    {
        const CBasicInformation* info=value->getInfo(inf);
        if(info->infoType()==CBasicInformation::IT_TRAJ_ANALYSIS) 
            infoOrbit=static_cast<const COrbitInformation<float>* >(info);
        if(info->infoType()==CBasicInformation::IT_TRAJ_SAMPLE) 
            infoTrajSample=static_cast<const CTrajSampleInformation<float>* >(info);
        if(info->infoType()==CBasicInformation::IT_SCHW) 
            infoSchw.push_back(static_cast<const CSchwInformation* >(info));
        if(info->infoType()==CBasicInformation::IT_PERICENTER) 
            infoPeri=static_cast<const CPericenterInformation<float>* >(info);
    }
    if(!started)  // do some preparatory work before writing out the first orbit
    {
        // write header line
        strmText << "x\ty\tz\tvx\tvy\tvz\tweight\ttimeunit\ttimestep\tinttime\tmaxtime";
        if(infoOrbit!=NULL) strmText << "\tEtot\tEdif\tlfx\tlfy\tlfz\tlfdiff\tlambda\tdescription\tinertx\tinerty\tinertz\tLXavg\tLXvar\tLYavg\tLYvar\tLZavg\tLZvar";
        if(infoPeri!=NULL) strmText << "\tL2min\tL2slope\tscatter\tsignificance\tL2circ";
        strmText << std::endl;
        started=true;
        orbitIndex=0;
        if(withModelData)   // try to open additional binary streams
        {
            if(infoSchw.size()>0) {  // open only if 1st OrbitDesc contains schwarzschild data, otherwise don't write it for any subsequent orbits
#ifndef HAVE_HDF5
                strmSchwData.open((baseFileName+".schw").c_str(), std::ios::out | std::ios::binary);
                if(!strmSchwData) status=false;
#else
                // create a new HDF5 file; if it exists it will be overwritten
                h5_fileSchwData = H5Fcreate((baseFileName+".schw.h5").c_str(), H5F_ACC_TRUNC, H5P_DEFAULT, H5P_DEFAULT);
                // create the datatype to be stored in the file, according to the Schw.data of the 1st orbit 
                // (assuming all orbits have the same set of data). 
                // First determine the total size of all Schw.data for a single orbit
                size_t totalsize=0, offset=0;
                for(size_t i=0; i<infoSchw.size(); i++)
                    totalsize+=infoSchw[i]->getSchwData().size();
                hid_t h5_typedefSchwData = H5Tcreate(H5T_COMPOUND, totalsize*sizeof(sdNumType));
                // atomic type to be used in the data storage (float or double)
                hid_t h5_typedefNumType = sizeof(sdNumType)==sizeof(float) ? H5T_NATIVE_FLOAT : sizeof(sdNumType)==sizeof(double) ? H5T_NATIVE_DOUBLE : H5I_INVALID_HID;
                // next initialize the compound type with sub-arrays
                for(size_t i=0; i<infoSchw.size(); i++) {
                    // create a temporary array type with the appropriate length
                    hsize_t datadims[1] = { infoSchw[i]->getSchwData().size() };
                    hid_t h5_typedefSchwDataElement = H5Tarray_create(h5_typedefNumType, 1, datadims);
                    // add a named field with the name of the current Schw.data type
                    H5Tinsert(h5_typedefSchwData, getSchwDataNameByType(infoSchw[i]->schwDataType()).c_str(), 
                        offset*sizeof(sdNumType), h5_typedefSchwDataElement);
                    H5Tclose(h5_typedefSchwDataElement);
                    offset+=static_cast<size_t>(datadims[0]);
                }
                // specify dimensions of the dataspace: unlimited 1d array, initially empty, and define chunked storage size
                hsize_t dimensions[1] = { 0 };
                hsize_t maxdimensions[1] = { H5S_UNLIMITED };
                hsize_t chunkdimensions[1] = { 256 };
                // create the dataspace (it will only be used for the subsequent creation of a dataset)
                hid_t h5_dataspaceSchwData = H5Screate_simple(1, dimensions, maxdimensions);                
                // modify dataset creation properties, i.e. enable chunking
                hid_t h5_propertiesSchwData = H5Pcreate(H5P_DATASET_CREATE);
                H5Pset_chunk(h5_propertiesSchwData, 1, chunkdimensions);
                // create a new dataset based on the dataspace and element type
                h5_datasetSchwData = H5Dcreate2(h5_fileSchwData, H5DATASETNAME_SCHW, h5_typedefSchwData, 
                    h5_dataspaceSchwData, H5P_DEFAULT, h5_propertiesSchwData, H5P_DEFAULT);
                // close temporary dataspace,typedefs and properties
                H5Sclose(h5_dataspaceSchwData);
                H5Pclose(h5_propertiesSchwData);
                H5Tclose(h5_typedefSchwData);
                if(h5_fileSchwData<0 || h5_datasetSchwData<0) status=false;
#endif
            }
            if(infoTrajSample!=NULL) {  // open only if 1st OrbitDesc contains trajectory sample data, otherwise don't write it for any subsequent orbits
#ifndef HAVE_HDF5
                strmTrajSample.open((baseFileName+".smpl").c_str(), std::ios::out | std::ios::binary);
                if(!strmTrajSample) status=false;
#else
                // create a new HDF5 file; if it exists it will be overwritten
                h5_fileTrajSample = H5Fcreate((baseFileName+".smpl.h5").c_str(), H5F_ACC_TRUNC, H5P_DEFAULT, H5P_DEFAULT);
                // create the data type to be stored in the file: a variable-length array of 3+3 floats
                hsize_t dims3[1] = { 3 };
                hid_t h5_typedefTripleFloat = H5Tarray_create(H5T_NATIVE_FLOAT, 1, dims3);
                hid_t h5_typedefPosVelPoint = H5Tcreate(H5T_COMPOUND, sizeof(CPosVelPoint<float>));
                H5Tinsert(h5_typedefPosVelPoint, "Pos", /*HOFFSET(CPosVelPoint<float>, Pos)*/ 0*sizeof(float), h5_typedefTripleFloat);
                H5Tinsert(h5_typedefPosVelPoint, "Vel", /*HOFFSET(CPosVelPoint<float>, Vel)*/ 3*sizeof(float), h5_typedefTripleFloat);
                hid_t h5_typedefTrajSample = H5Tvlen_create(h5_typedefPosVelPoint);
                // specify dimensions of the dataspace: unlimited 1d array, initially empty, and define chunked storage size
                hsize_t dimensions[1] = { 0 };
                hsize_t maxdimensions[1] = { H5S_UNLIMITED };
                hsize_t chunkdimensions[1] = { 256 };
                // create the dataspace (it will only be used for the subsequent creation of a dataset)
                hid_t h5_dataspaceTrajSample = H5Screate_simple(1, dimensions, maxdimensions);                
                // modify dataset creation properties, i.e. enable chunking
                hid_t h5_propertiesTrajSample = H5Pcreate(H5P_DATASET_CREATE);
                H5Pset_chunk(h5_propertiesTrajSample, 1, chunkdimensions);
                // create a new dataset based on the dataspace and element type
                h5_datasetTrajSample = H5Dcreate2(h5_fileTrajSample, H5DATASETNAME_TRAJ, h5_typedefTrajSample, 
                    h5_dataspaceTrajSample, H5P_DEFAULT, h5_propertiesTrajSample, H5P_DEFAULT);
                // close temporary dataspace,typedefs and properties
                H5Sclose(h5_dataspaceTrajSample);
                H5Pclose(h5_propertiesTrajSample);
                H5Tclose(h5_typedefTripleFloat);
                H5Tclose(h5_typedefPosVelPoint);
                H5Tclose(h5_typedefTrajSample);
                if(h5_fileTrajSample<0 || h5_datasetTrajSample<0) status=false;
#endif
            }
        }
    }   // end preparatory work
    COrbitInitData<float> InitData=value->getInitData();
    float intTime=value->getIntTime();
    float maxTime=value->getMaxTime();
    if(maxTime==intTime) maxTime=0;  // default
    strmText <<  std::setprecision(8) << 
        InitData.initCond.Pos[0] << "\t" << InitData.initCond.Pos[1] << "\t" << InitData.initCond.Pos[2] << "\t" << 
        InitData.initCond.Vel[0] << "\t" << InitData.initCond.Vel[1] << "\t" << InitData.initCond.Vel[2] << "\t" << 
        std::setprecision(6) << value->getWeight() << "\t" <<
        InitData.timeUnit << "\t" << InitData.timeStep << "\t" << intTime << "\t" << maxTime;
    if(infoOrbit!=NULL)
    {
        std::string Description=infoOrbit->getDescription();
        std::string::size_type ind;
        while((ind = Description.find(" ")) != std::string::npos)
            Description=Description.replace(ind, 1, "_");
        strmText << "\t" << infoOrbit->getEinit() << "\t" << infoOrbit->getEdiff() << "\t" << 
            infoOrbit->getlf(0) << "\t" << infoOrbit->getlf(1) << "\t" << infoOrbit->getlf(2) << "\t" << 
            infoOrbit->getlfccdiff() << "\t" << (InitData.calcLyapunov ? infoOrbit->getlambda() : -1) << "\t" << Description << "\t" << 
            infoOrbit->getinertia(0) << "\t" << infoOrbit->getinertia(1) << "\t" << infoOrbit->getinertia(2) << "\t" <<
            infoOrbit->getLavg(0) << "\t" << infoOrbit->getLvar(0) << "\t" <<
            infoOrbit->getLavg(1) << "\t" << infoOrbit->getLvar(1) << "\t" <<
            infoOrbit->getLavg(2) << "\t" << infoOrbit->getLvar(2);
    }
    if(infoPeri!=NULL)
    {
        float fitIntercept=0, fitSlope=0, fitScatter=0, fitSignificance=0, Lcirc2=0;
        infoPeri->getParams(&fitIntercept, &fitSlope, &fitScatter, &fitSignificance, &Lcirc2);
        strmText << "\t" << fitIntercept << "\t" << fitSlope << "\t" << fitScatter << "\t" << fitSignificance << "\t" << Lcirc2;
    }
    strmText << std::endl;
    if(!strmText) status=false;
    if(withModelData)
    {
#ifndef HAVE_HDF5
        if(strmSchwData.is_open()) {
            unsigned int numdata=static_cast<unsigned int>(infoSchw.size());
            strmSchwData.write( reinterpret_cast<const char*>(&numdata), sizeof(numdata) );
            for(size_t infoind=0; infoind<numdata; infoind++)
            {
                // for each item of Schwarzschild data of the given orbit, store the datatype, data length and bulk data array
                const std::vector<sdNumType>& schwdata=infoSchw[infoind]->getSchwData();
                CBasicSchwData::SCHWDATATYPE  datatype=infoSchw[infoind]->schwDataType(); 
                strmSchwData.write( reinterpret_cast<const char*>(&datatype), sizeof(datatype) );
                unsigned int datasize = static_cast<unsigned int>(schwdata.size());
                strmSchwData.write( reinterpret_cast<const char*>(&datasize), sizeof(datasize) );
                if(datasize>0) strmSchwData.write( reinterpret_cast<const char*>(&(schwdata.front()) ), datasize*sizeof(sdNumType));
            }
            if(!strmSchwData) status=false;
        }
#else
        // extend the size of dataset by one
        hsize_t dimensions[1] = { orbitIndex+1 };
        H5Dset_extent(h5_datasetSchwData, dimensions);
        // obtain the file dataspace 
        hid_t h5_dataspaceSchwData = H5Dget_space(h5_datasetSchwData);
        // create a memory dataspace of one element
        hid_t h5_memspaceSchwData = H5Screate(H5S_SCALAR);
        // obtain the compound datatype and create an instance of this datatype, serialized into a 1d array
        hid_t h5_typedefSchwData = H5Dget_type(h5_datasetSchwData);
        std::vector<sdNumType> schwDataArray;
        // check that all orbits share the same kind of data
        bool correct = h5_fileSchwData>0;
        if(H5Tget_nmembers(h5_typedefSchwData)!=static_cast<int>(infoSchw.size())) correct=false;
        if(correct) {
            // loop through all attached Schw.data and fill the serialized array
            for(unsigned int infoind=0; infoind<infoSchw.size(); infoind++)
            {
                const std::vector<sdNumType>& schwdata=infoSchw[infoind]->getSchwData();
                unsigned int datasize = static_cast<unsigned int>(schwdata.size());
                // ensure that the data type and size are the same as defined in the HDF5 compound type
                hid_t h5_typedefSchwDataElement = H5Tget_member_type(h5_typedefSchwData, infoind);
                //if(std::string(H5Tget_member_name(h5_typedefSchwData, infoind)) 
                //      != getSchwDataNameByType(infoSchw[i].schwDataType())) correct=false;  // skip this check
                hsize_t dims[1];
                H5Tget_array_dims2(h5_typedefSchwDataElement, dims);
                if(dims[0] != datasize) correct=false;
                if(correct)  // copy data into serialized array
                    schwDataArray.insert(schwDataArray.end(), schwdata.begin(), schwdata.end());
            }
        }
        if(correct) {
            // select the last element in the dataset to write the data into it
            hsize_t dsetstart[1] = { orbitIndex };
            hsize_t dsetcount[1] = { 1 };
            H5Sselect_hyperslab(h5_dataspaceSchwData, H5S_SELECT_SET, dsetstart, NULL, dsetcount, NULL);
            // write the content of the serialized array into the last element of dataset in the file
            herr_t h5_status = H5Dwrite(h5_datasetSchwData, h5_typedefSchwData, h5_memspaceSchwData, 
                h5_dataspaceSchwData, H5P_DEFAULT, &(schwDataArray.front()));
            if(h5_status<0) status=false;
        } else {
            my_error("Error, amount of Schwarzschild data is not the same for all orbits!");
            status=false;
        }
        // close the dataspaces and typedef
        H5Sclose(h5_dataspaceSchwData);
        H5Sclose(h5_memspaceSchwData);
        H5Tclose(h5_typedefSchwData);
#endif
    }
    if(withModelData)
    {
#ifndef HAVE_HDF5
        if(strmTrajSample.is_open()) {
            // for each orbit, store the trajectory sample length (even if it is zero) and the bulk data array
            unsigned int nump=0;
            if(infoTrajSample!=NULL && infoTrajSample->getTraj().size()>0)
            {
                nump=static_cast<unsigned int>(infoTrajSample->getTraj().size());
                const CPosVelDataFloat& data=infoTrajSample->getTraj();
                strmTrajSample.write( reinterpret_cast<const char*>(&nump), sizeof(nump) );
                strmTrajSample.write( reinterpret_cast<const char*>( &(data.front()) ), nump * sizeof(CPosVelPoint<float>));
            }
            else
                strmTrajSample.write( reinterpret_cast<const char*>(&nump), sizeof(nump) );
            if(!strmTrajSample) status=false;
        }
#else
        // extend the size of dataset by one
        hsize_t dimensions[1] = { orbitIndex+1 };
        H5Dset_extent(h5_datasetTrajSample, dimensions);
        // obtain the file dataspace 
        hid_t h5_dataspaceTrajSample = H5Dget_space(h5_datasetTrajSample);
        // create a memory dataspace of one element
        hid_t h5_memspaceTrajSample = H5Screate(H5S_SCALAR);
        // obtain the datatype
        hid_t h5_typedefTrajSample = H5Dget_type(h5_datasetTrajSample);
        // create hdf5 variable-length data element
        hvl_t h5_varlengthdata;
        size_t nump=infoTrajSample->getTraj().size();
        if(infoTrajSample!=NULL && nump>0)
            h5_varlengthdata.p = const_cast<void*>(static_cast<const void*>(&(infoTrajSample->getTraj().front())));
        else 
            h5_varlengthdata.p = dimensions;   // dummy address
        h5_varlengthdata.len = nump;
        // select the last element in the dataset to write the data into it
        hsize_t dsetstart[1] = { orbitIndex };
        hsize_t dsetcount[1] = { 1 };
        H5Sselect_hyperslab(h5_dataspaceTrajSample, H5S_SELECT_SET, dsetstart, NULL, dsetcount, NULL);
        // write the content of the serialized array into the last element of dataset in the file
        herr_t h5_status = H5Dwrite(h5_datasetTrajSample, h5_typedefTrajSample, h5_memspaceTrajSample, 
            h5_dataspaceTrajSample, H5P_DEFAULT, &h5_varlengthdata);
        if(h5_status<0) status=false;
        // close the dataspaces and typedef
        H5Sclose(h5_dataspaceTrajSample);
        H5Sclose(h5_memspaceTrajSample);
        H5Tclose(h5_typedefTrajSample);
#endif
    }
    orbitIndex++;
    return status;
}

bool CFileOutputText::writeOrbitLibrary(const COrbitLibrary* orbitLib, bool withModelData)
{
    if(!status || !isopen) return false;
    for(size_t o=0; status && o<orbitLib->size(); o++)
        writeOrbitDesc(orbitLib->getOrbitDesc(o), withModelData);
    return status;
}

}   // namespace