/** \file    iodata.h
    \brief   Input/output of various kind of data to/from text or binary files
    \author  EV
    \date    2009-2013

    This file defines two classes, for input and output using (mostly) text files.
    Each one can handle different sorts of data (single orbit trajectories and 
    orbit library), but they are expected not to be mixed in a single file. 
    Loading and storing the orbit library also makes use of additional binary 
    files containing bulky data.
*/

#pragma once
#include "common.h"
#include "schwarzschild.h"
#include <fstream>
#ifdef HAVE_HDF5
#include <hdf5.h>
/// names of data arrays in HDF5 files
#define H5DATASETNAME_SCHW "/Schw"
#define H5DATASETNAME_TRAJ "/Traj"
#endif

namespace smile{

struct CConfigOrbit;  // forward declaration

/// Class for reading data from a text file, possibly accompanied by binary files for the orbit library.
class CFileInputText
{
public:
    /// create object but don't open file
    CFileInputText() { status=isopen=false; };
    /// create object and open the file
    explicit CFileInputText(const std::string &_baseFileName)
    { status=isopen=false; open(_baseFileName); };
    /*virtual*/ ~CFileInputText();                ///< close the file(s) if they were opened
    void open(const std::string &_baseFileName);  ///< open a new file for reading; if another file was opened then close it beforehand
    bool ok() const { return status; }            ///< status of last file operation
    /** load a single trajectory from file, create and return a COrbit object. 
        File must be open (in the constructor or using open());
        \param[in] configOrbit,potential are passed to the constructor of orbit;
        \return a new COrbit object on success or NULL on failure.
    */
    COrbit* readOrbit(const CConfigOrbit* configOrbit, const CPotential* potential);
    /** load an orbit library from the text file (possibly with binary data files).
        \param[in] configOrbit,potential are passed to the constructor of orbit library;
        \param[in] withModelData specifies whether an attempt to open and read additional 
                   binary files is made;  these are filename.schw and filename.smpl files 
                   containing CSchwInformation and CTrajSampleInformation for each orbit, 
                   correspondingly.
        \return    a new COrbitLibrary object on success or NULL on failure.
    */
    COrbitLibrary* readOrbitLibrary(const CConfigOrbit* configOrbit, const CPotential* potential, bool withModelData);
private:
    std::string baseFileName;     ///< name of the text file; if necessary to load binary files, extensions are added to this name
    bool status;                  ///< status of last i/o operation
    bool isopen;                  ///< indicates whether the file is open
    size_t orbitIndex;            ///< index of current orbit in the orbit library (for keeping track of the position in binary files)
    std::ifstream strmText;       ///< stream object for the main text file
#ifndef HAVE_HDF5
    std::ifstream strmSchwData;   ///< stream object for the binary file with CSchwInformation
    std::ifstream strmTrajSample; ///< stream object for the binary file with CTrajSampleInformation
#else
    std::vector<CBasicSchwData::SCHWDATATYPE> schwtypes; ///< list of Schw.data objects for each orbit in the library
    std::vector<size_t> schwsizes;///< list of lengths of Schw.data objects for each orbit in the library
    hid_t h5_fileSchwData;        ///< identificator of HDF5 file with Schwarzschild model data arrays (CSchwInformation)
    hid_t h5_datasetSchwData;     ///< identificator of HDF5 dataset for Schwarzschild data
    hid_t h5_fileTrajSample;      ///< identificator of HDF5 file with trajectory sample (CTrajSampleInformation)
    hid_t h5_datasetTrajSample;   ///< identificator of HDF5 dataset with trajectory sample
#endif
    /** load orbit description from a single line in the orbit library text file, 
        and possibly also reads binary data for trajectory sample and Schwarzschild model. */
    COrbitDesc* readOrbitDesc(const CConfigOrbit* configOrbit, const CPotential* potential, bool withModelData);
};

/// Class for writing data to a text file, possibly accompanied by binary files for the orbit library.
class CFileOutputText
{
public:
    /// create object but don't open file
    CFileOutputText() { status=isopen=started=false; };
    /// create object and open the file
    explicit CFileOutputText(const std::string &_baseFileName)
    { status=isopen=started=false; open(_baseFileName); };
    void open(const std::string &_baseFileName);   ///< open a new file for writing; if another file was opened then close it beforehand
    ~CFileOutputText();                            ///< close file(s) if they were opened
    bool ok() const { return status; }             ///< status of last file operation
    bool writeOrbit(const COrbit* value);          ///< write a trajectory to the text file
    /// write orbit library to the text file, and possibly the data for Schwarzschild modelling to binary files.
    bool writeOrbitLibrary(const COrbitLibrary* value, bool withModelData);
private:
    std::string baseFileName;     ///< name of the text file; if necessary to write binary files, extensions are added to this name
    bool status;                  ///< status of last i/o operation
    bool isopen;                  ///< indicates whether the file is open
    bool started;                 ///< whether the output to the text file has begun and the header line has been written
    size_t orbitIndex;            ///< index of current orbit in the orbit library (for keeping track of the position in binary files)
    std::ofstream strmText;       ///< stream object for the main text file
#ifndef HAVE_HDF5
    std::ofstream strmSchwData;   ///< stream object for the binary file with CSchwInformation
    std::ofstream strmTrajSample; ///< stream object for the binary file with CTrajSampleInformation
#else
    hid_t h5_fileSchwData;        ///< identificator of HDF5 file with Schwarzschild model data arrays (CSchwInformation)
    hid_t h5_datasetSchwData;     ///< identificator of HDF5 dataset for Schwarzschild data
    hid_t h5_fileTrajSample;      ///< identificator of HDF5 file with trajectory sample (CTrajSampleInformation)
    hid_t h5_datasetTrajSample;   ///< identificator of HDF5 dataset with trajectory sample
#endif
    /// write one line to the orbit library text file, and possibly the information objects to binary files
    bool writeOrbitDesc(const COrbitDesc* value, bool withModelData);
};

}  // namespace