/** \file    optimization.h
    \brief   Interface to linear/quadratic optimization solvers for Schwarzshild modelling
    \author  EV
    \date    2009-2013

    This file defines several examples of linear/quadratic optimization solvers derived 
    from smile::CBasicQuadraticOptimizationSolver defined and explained in common.h

    Not all of them may be supported at the compile time; this is regulated 
    by HAVE_CVXOPT and HAVE_GLPK macros.
*/

#pragma once
#include "common.h"
#include <QProcess>

namespace smile{

/// Interface to the standalone program BPMPD.
template<typename NumT> class COptimizationSolverBPMPD: public CBasicQuadraticOptimizationSolver<NumT>
{
private:
    std::string execName, tempDir;
public:
    /** Construct the instance of the solver.
        \param[in] _execName is the name of program (if it doesn't exist, an error is raised in callSolver);
        \param[in] _tempDir  is the directory where the data file will be written and the result file will be read.
    */
    COptimizationSolverBPMPD(const std::string &_execName, const std::string &_tempDir): 
        execName(_execName), tempDir(_tempDir) {};
    virtual int callSolver(
        const CMatrix<NumT> *linearMatrix,      // matrix M of size N_v * N_c. Passed as an interface to an abstract class, must not be NULL
        const std::vector<NumT> &rhs,           // N_c constraints to be satisfied
        const std::vector<NumT> &maxSolValue,   // N_v upper limits for variables (zero means no upper limit). Lower limits are zeroes
        const std::vector<NumT> &objFuncLinear, // N_v linear penalty factors in the objective function
        const CMatrix<NumT> *objFuncQuadratic,  // N_v * N_v quadratic penalty factors, may be NULL
        std::vector<NumT> *sol) const;          // returns solution in this vector of size N_v (must not be NULL but may be empty on call)
    virtual std::string errorDescription(int result) const;
};

/// Helper class to capture diagnostic and progress messages from running BPMPD process,
/// and redirect them to the main SMILE program via my_message()
class COutputHookQProcess: public QObject {
Q_OBJECT
public:
    COutputHookQProcess(QProcess* _proc): proc(_proc) {};
public slots:
    void readLineOutput();
private:
    QProcess* proc;
};

#ifdef HAVE_GLPK
/// Interface to the linear optimization solver GLPK
template<typename NumT> class COptimizationSolverGLPK: public CBasicQuadraticOptimizationSolver<NumT>
{
public:
    COptimizationSolverGLPK() {};
    virtual int callSolver(
        const CMatrix<NumT> *linearMatrix,
        const std::vector<NumT> &rhs,
        const std::vector<NumT> &maxSolValue,
        const std::vector<NumT> &objFuncLinear,
        const CMatrix<NumT> *objFuncQuadratic,
        std::vector<NumT> *sol) const;
    virtual std::string errorDescription(int result) const;
};
#endif

#ifdef HAVE_CVXOPT
/// Interface to the LP/QP solver CVXOPT written in Python
template<typename NumT> class COptimizationSolverCVXOPT: public CBasicQuadraticOptimizationSolver<NumT>
{
public:
    COptimizationSolverCVXOPT() {};
    virtual int callSolver(
        const CMatrix<NumT> *linearMatrix,
        const std::vector<NumT> &rhs,
        const std::vector<NumT> &maxSolValue,
        const std::vector<NumT> &objFuncLinear,
        const CMatrix<NumT> *objFuncQuadratic,
        std::vector<NumT> *sol) const;
    virtual std::string errorDescription(int result) const;
};
#endif

}