#include "orbitan.h"
#include <cassert>
#include <algorithm>
#include <cmath>
#include <set>
#include <gsl/gsl_fft_real.h>
#include <gsl/gsl_fft_halfcomplex.h>
#include <gsl/gsl_integration.h>
#include <gsl/gsl_math.h>
#include <gsl/gsl_fit.h>
#include <gsl/gsl_cdf.h>
#include <gsl/gsl_rng.h>
#include <gsl/gsl_errno.h>
#include "potential.h"
#include "orbitint.h"
#include "utils.h"

namespace smile{
 
/*-----------------   Below follows COrbit class  -------------*/

/* constructor for an empty orbit */
template<typename NumT> COrbit::COrbit(const CConfigOrbit* _configOrbit, 
    const COrbitInitData<NumT> &_InitData, const vectorRuntimeFncCreators* creators) :
  configOrbit(_configOrbit),
  InitData(completeInitData(_InitData))
{
    intTime=0;
    state=OS_INITIALIZED;
    orbInt=NULL;
    if(creators!=NULL)  // construct timestep function objects using creator classes
        for(vectorRuntimeFncCreators::const_iterator iter=creators->begin(); iter!=creators->end(); iter++)
        {
            CBasicOrbitRuntimeFnc* fnc=(*iter)->createRuntimeFnc(this);
            if(fnc!=NULL) runtimeFncs.push_back(fnc);
        }
}
// explicit instantiations of constructor for two variants of template argument
template COrbit::COrbit(const CConfigOrbit* _configOrbit, const COrbitInitData<float > &_InitData, const vectorRuntimeFncCreators* creators);
template COrbit::COrbit(const CConfigOrbit* _configOrbit, const COrbitInitData<double> &_InitData, const vectorRuntimeFncCreators* creators);

/* constructor for an existing orbit */
template<typename NumT> COrbit::COrbit(const CConfigOrbit* _configOrbit, 
    const CPotential* _potential, const double _timeStep, const double _timeUnit, const std::vector< CPosVelPoint<NumT> > &_traj) :
  configOrbit(_configOrbit), 
  InitData(completeInitData(COrbitInitData<double>(_potential, _timeStep, _timeUnit, /*initCond*/ _traj[0], /*calcLyapunov*/ false)))
{
    state=OS_DONE;
    orbInt=NULL;
    intTime=InitData.timeStep*_traj.size();
    runtimeFncs.push_back( new COrbitRuntimeTrajectory(this, _traj));
    runtimeFncs.push_back( new COrbitRuntimePericenter(this, _traj));
    for(vectorRuntimeFncs::iterator iter=runtimeFncs.begin(); iter!=runtimeFncs.end(); iter++)
        (*iter)->Finish();
};
// explicit instantiations of constructor for two variants of template argument
template COrbit::COrbit(const CConfigOrbit* _configOrbit, const CPotential* _potential, const double _timeStep, const double _timeUnit, const std::vector< CPosVelPoint<float > > &_traj);
template COrbit::COrbit(const CConfigOrbit* _configOrbit, const CPotential* _potential, const double _timeStep, const double _timeUnit, const std::vector< CPosVelPoint<double> > &_traj);

COrbit::~COrbit()
{
    // if there were any timestep functions attached, destroy 'em all!!
    for(vectorRuntimeFncs::iterator iter=runtimeFncs.begin(); iter!=runtimeFncs.end(); iter++)
        delete (*iter);
}

void COrbit::halt()
{ 
    if(orbInt!=NULL) orbInt->halt();
}

double COrbit::getInterpolatedTrajectory(unsigned int c, double t) const 
{
    if(orbInt!=NULL) 
        return orbInt->getInterpolatedTrajectory(c, t); 
    else return 0; 
}

double COrbit::getIntTime() const
{ 
    return orbInt!=NULL ? orbInt->getTime() : intTime; 
}

std::string COrbit::toString(bool runtimeFncInfo) const 
{
    std::string strInfo;
    if(runtimeFncInfo)  
        for(size_t rf=0; rf<getInfoNum(); rf++)
        {
            const CBasicInformation* info=getNewInfo(rf);
            strInfo += info->toString();
            delete info;
        }
    //strInfo+="fnc.eval="+convertToString(nfcn)+", naccp="+convertToString(naccpt)+", nrej="+convertToString(nrejct)+"\n";
    return strInfo;
}

void COrbit::integrateToTime(double time, double maxTime)
{
    if(orbInt!=NULL) 
        return;   // can't run orbit integrator twice
    if(InitData.potential->PotentialType() == CPotential::PT_NB) 
        orbInt = new COrbitIntegratorLeapfrog(InitData.potential, &runtimeFncs, 
            configOrbit->accTreecode, configOrbit->treecodeSymmetrizeTimestep);
    else
        orbInt = new COrbitIntegratorDOP853(InitData.potential, &runtimeFncs, 
            InitData.calcLyapunov? configOrbit->calcLyapunovMethod : 0, 
            configOrbit->accuracyAbs, configOrbit->accuracyRel);
    state=OS_RUNNING;
    orbInt->integrateToTime(InitData.initCond, time, maxTime);
    intTime=orbInt->getTime();
    for(vectorRuntimeFncs::iterator iter=runtimeFncs.begin(); iter!=runtimeFncs.end(); iter++)
        (*iter)->Finish();
    state=OS_DONE;
    delete orbInt;
    orbInt=NULL;
}

//----------------------------------------------------------------------------//

template<typename NumT> std::string CPericenterInformation<NumT>::toString() const
{
    return "Lmin\xB2="+convertToString(fitIntercept,4)+" + "+convertToString(fitSlope,3)+"*w [sct="+convertToString(fitScatter,3)+",sig="+convertToString(fitSignificance,5)+"]\nLcirc\xB2="+convertToString(Lcirc2,3)+"\n";
}
template std::string CPericenterInformation<float >::toString() const;
template std::string CPericenterInformation<double>::toString() const;

template<typename NumT> std::string COrbitInformation<NumT>::toString() const
{
    return Description 
        + "\nlfx=" + convertToString(lf[0],8)
        + "\nlfy=" + convertToString(lf[1],8)
        + "\nlfz=" + convertToString(lf[2],8)
        + "\nFreqDiff=" + convertToString(lfccdiff,3) 
        + (lambda>=0 ? "\nLyapunovExp=" + convertToString(lambda,3) : "")
        + "\nInertia=" + convertToString(inertia[0],3) + "," +  convertToString(inertia[1],3) + "," +  convertToString(inertia[2],3)
        + "\nLz=" + convertToString(Lavg[2],3)+"\xB1"+convertToString(Lvar[2],3) 
        + (Lavg[0]!=0 ? ", Lx=" + convertToString(Lavg[0],3)+"\xB1"+convertToString(Lvar[0],3) : "") 
        + "\nE=" + convertToString(Einit) + ", deltaE=" + convertToString(Ediff,3) + "\n";
}
template std::string COrbitInformation<float >::toString() const;
template std::string COrbitInformation<double>::toString() const;

template<typename NumT> std::string CTrajSampleInformation<NumT>::toString() const
{
    if(trajSample.empty()) return std::string(); 
    else return convertToString(trajSample.size())+" sampling points\n";
}
template std::string CTrajSampleInformation<float >::toString() const;
template std::string CTrajSampleInformation<double>::toString() const;

// Runtime function for recording and analyzing trajectory by means of frequency analysis and Lyapunov exponent 

COrbitRuntimeTrajectory::COrbitRuntimeTrajectory(const smile::COrbit *_orbit) :
  CBasicOrbitRuntimeFnc(_orbit), 
  N_dim(orbit->getPotential()->N_dim), 
  timeStep(orbit->getInitData().timeStep),
  timeUnit(orbit->getInitData().timeUnit),
  calcLyapunov(orbit->getInitData().calcLyapunov)
{
    Einit=orbit->getPotential()->totalEnergy(orbit->getInitData().initCond);
    Ediff=0;
}

template<typename NumT> COrbitRuntimeTrajectory::COrbitRuntimeTrajectory( 
    const smile::COrbit *_orbit, const std::vector< CPosVelPoint<NumT> > &_traj) :
  CBasicOrbitRuntimeFnc(_orbit), 
  N_dim(orbit->getPotential()->N_dim), 
  timeStep(orbit->getInitData().timeStep),
  timeUnit(orbit->getInitData().timeUnit),
  calcLyapunov(orbit->getInitData().calcLyapunov),
  traj(_traj.begin(), _traj.end())
{
    Einit=orbit->getPotential()->totalEnergy(orbit->getInitData().initCond);
    Ediff=0;  // not quite true, but we won't bother calculating it..
}

bool COrbitRuntimeTrajectory::Timestep(const double told, const double t, const double y[])
{
    int i1=static_cast<int>(told/timeStep); 
    int i2=static_cast<int>(t/timeStep);

    // check energy conservation
    if(i2>i1 && i2%10==0)  // do not check too frequently (potential evaluation may be costly)
    {
        double Z=(N_dim==3)?y[2]:0;
        double vZ=(N_dim==3)?y[N_dim+2]:0;
        double Ecur = orbit->getPotential()->Phi(y[0], y[1], Z) + (pow_2(y[N_dim]) + pow_2(y[N_dim+1]) + pow_2(vZ))/2;
        double Edifcur = fabs(Einit - Ecur);
        Ediff=std::max<double>(Edifcur, Ediff);
    }
    // store trajectory at regular intervals of time    
    for (int i=i1; i<=i2; i++)
    {
        double tt=timeStep*i;
        if ((tt>=told) && (tt<t))
        {
            CPosVelPoint<double> pt;
            for(unsigned int c=0; c<N_dim; c++)
            {
                pt.Pos[c]=orbit->getInterpolatedTrajectory(c, tt);
                pt.Vel[c]=orbit->getInterpolatedTrajectory(c+N_dim, tt);
            }
            traj.push_back(pt);
            if(orbit->getInitData().calcLyapunov)
                lnw.push_back(orbit->getInterpolatedTrajectory(2*N_dim, tt));
        }
    }
    return true;
}

template<typename NumT> COrbitInformation<NumT>* COrbitRuntimeTrajectory::analyzeOrbit(size_t start, size_t count, CSpectrum *spectrum, vectord *fa) const
{
    if(start>=traj.size())  // invalid call
        return new COrbitInformation<NumT>("Invalid call: start>=size");
    if(count<=0 || count>traj.size()-start) count = traj.size() - start;   // passing zero for count means analysing entire orbit
    if(count<=1)  // empty orbit
        return new COrbitInformation<NumT>("?");
    CSpectrum tempSpectrum[N_DIM];
    CSpectrum* sl = spectrum!=NULL ? spectrum : tempSpectrum;   // if outputting spectrum is not necessary, use temp. array for storing the information
    double Lavg_d[N_DIM]={0,0,0}, Lvar_d[N_DIM]={0,0,0};
    std::string Description = performOrbitAnalysis(start, count, sl, fa, Lavg_d, Lvar_d);
    double lfccdiff = calcLFCCdiffusion(start, count);
    double lambda = calcLyapunov ? calcLambda() : -1;
    double inertia_d[3];
    calcInertiaOverInterval(start, count, inertia_d);
    NumT leadfreq[N_DIM]={static_cast<NumT>(sl[0][0]), static_cast<NumT>(sl[1][0]), static_cast<NumT>((N_dim==3)?sl[2][0].freq:0)};
    NumT inertia[N_DIM] ={static_cast<NumT>(inertia_d[0]), static_cast<NumT>(inertia_d[1]), static_cast<NumT>(inertia_d[2])};
    NumT Lavg[N_DIM] ={static_cast<NumT>(Lavg_d[0]), static_cast<NumT>(Lavg_d[1]), static_cast<NumT>(Lavg_d[2])};
    NumT Lvar[N_DIM] ={static_cast<NumT>(Lvar_d[0]), static_cast<NumT>(Lvar_d[1]), static_cast<NumT>(Lvar_d[2])};
    // note that Ediff is returned for entire orbit integration time, not for the given interval
    return new COrbitInformation<NumT>(Description, static_cast<NumT>(Einit), static_cast<NumT>(Ediff), leadfreq, static_cast<NumT>(lfccdiff), static_cast<NumT>(lambda), inertia, Lavg, Lvar);
}
// explicit instantiations of a template function
template COrbitInformation<double>* COrbitRuntimeTrajectory::analyzeOrbit<double>(size_t start, size_t count, CSpectrum *spectrum, vectord *fa) const;

// perform FFT of orbit (output amplitudes to 'fa' array) and find 'maxnlines' lines, which will be placed into 'sl' array. 
// If fa=NULL, do not output frequency amplitudes
void COrbitRuntimeTrajectory::findLines(size_t start, size_t npoints, CSpectrum *sl, vectord *fa) const
{
    vectord trajSlice(npoints);
    const CConfigOrbit* configOrbit=orbit->getConfig();
    for(unsigned int c=0; c<N_dim; c++)
    {
        for(size_t i=0; i<npoints; i++)
            trajSlice[i]=traj[start+i].Pos[c];
        CSpectrumAnalyzerHunter spect(trajSlice, timeStep/timeUnit);
        spect.findSpectralLines(&(sl[c]), fa!=NULL? &(fa[c]) : NULL);
        int sw=0; CSpectralLine ftmp;
        switch(c)   // ensure that the first line in each coordinate has frequency >1 (for X), >X (for Y), and >Y (for Z); tolerate small deviations given in MIN_FREQ_*
        {
        case 0:
            if(configOrbit->freqStrictOrder) 
                while((sw<(int)sl[0].size()-1) && (sl[0][0].freq<configOrbit->minFreqX)) 
                { 
                    ftmp=sl[0][sw+1]; 
                    sl[0][sw+1]=sl[0][0]; 
                    sl[0][0]=ftmp; sw++; 
                }
            break;
        case 1:
            if(configOrbit->freqStrictOrder) 
                while((sw<(int)sl[1].size()-1) && (sl[1][0].freq<configOrbit->minFreqYZ*sl[0][0].freq)) 
                { 
                    ftmp=sl[1][sw+1]; 
                    sl[1][sw+1]=sl[1][0]; 
                    sl[1][0]=ftmp; sw++; 
                }
            break;
        case 2:
            if(configOrbit->freqStrictOrder && N_dim==3) 
                while((sw<(int)sl[2].size()-1) && (sl[2][0].freq<configOrbit->minFreqYZ*sl[1][0].freq)) 
                { 
                    ftmp=sl[2][sw+1]; 
                    sl[2][sw+1]=sl[2][0]; 
                    sl[2][0]=ftmp; sw++; 
                }
            break;
        }
    }
}

/// \cond INTERNAL_DOCS
void simplify(int &n1, int &n2)    // remove common factors
{
    int nmin=std::min<int>(n1, n2);
    int f=1;
    for(int k=1; k<nmin/2; k++)
        if((n1%k==0) && (n2%k==0)) f=k;
    n1/=f;
    n2/=f;
}

bool commensurable(double a, double b, double acc, int &na, int &nb)   // check that a:b ~=~ na:nb
{
    if(a*b==0) return(false);
    na=1; nb=0;
    while(na<=FREQ_RES_2)
    {
        int nb1=(int)floor(b/a*na+.5);
        if(nb1 && (fabs(nb1*a-na*b) < acc))
        {
            nb=nb1;
            return true;
        }
        else na++;
    }
    return false;
}

bool lineardepend(double f, double a, double b, double acc, int &ma, int &mb)   // check linear dependency: f ~=~ ma*a + mb*b
{
    ma=mb=0; 
    while(abs(ma)<=FREQ_RES_2)
    {
        mb=(int)floor((f-ma*a)/b+0.5);
        if(fabs(ma*a+mb*b-f) < acc) 
            return true;
        else 
            if(ma>0) ma=-ma; else ma=-ma+1;
    }
    return false;
}

bool lineardepend3(double f, double a, double b, double c, double acc, int &ma, int &mb, int &mc)   // check 3d linear dependency: f ~=~ ma*a + mb*b + mc*c
{
    ma=mb=mc=0; 
    while(abs(ma)<=FREQ_RES_2)
    {
      mb=0;
      while(abs(mb)<=FREQ_RES_2)
      {
        mc=(int)floor((f-ma*a-mb*b)/c+0.5);
        if(fabs(ma*a+mb*b+mc*c-f) < acc) 
            return true;
        else 
            if(mb>0) mb=-mb; else mb=-mb+1;
      }
      if(ma>0) ma=-ma; else ma=-ma+1;
    }
    return false;
}

bool commensurable3(double a, double b, double c, double acc, int &ma, int &mb, int &mc)
{
    for(int na=1; na<=FREQ_RES_3; na++)
        for(int nb=-FREQ_RES_3; nb<=FREQ_RES_3; nb++)
            for(int nc=-FREQ_RES_3; nc<=FREQ_RES_3; nc++)
                if(fabs(a*na+b*nb+c*nc) < acc)
                {
                    ma=na; mb=nb; mc=nc;
                    return true;
                }
    return false;
}
/// \endcond

std::string COrbitRuntimeTrajectory::performOrbitAnalysis(size_t start, size_t count, 
    CSpectrum out_lines[N_DIM], vectord *out_ampl, double Lavg[N_DIM], double Lvar[N_DIM]) const
{
    findLines(start, count, out_lines, out_ampl);
    std::string Description;
    double Labs[N_DIM];  // abs.value of ang.mom.components
    double Cavg[N_DIM];  // avg.value of C-th coordinate
    double Cmax[N_DIM];  // max.value of this coordinate
    for(unsigned int d=0; d<N_DIM; d++)
    {
        Lavg[d]=0;  // mean value of d-axis component of angular momentum
        Lvar[d]=0;  // variation in this component (r.m.s.)
        Labs[d]=0;  // mean absolute value of d-axis component of ang.mom.
        Cavg[d]=0;
        Cmax[d]=0;
    }
    for(size_t i=start; i<start+count; i++)
    {
        for(unsigned int d=0; d<N_DIM; d++)
        {
            double L = traj[i].Pos[(d+1)%N_DIM]*traj[i].Vel[(d+2)%N_DIM] - traj[i].Pos[(d+2)%N_DIM]*traj[i].Vel[(d+1)%N_DIM];
            Lavg[d] += L;
            Labs[d] += fabs(L);
            Lvar[d] += pow_2(L);
            Cavg[d] += traj[i].Pos[d];
            Cmax[d] = std::max<double>(Cmax[d], traj[i].Pos[d]);
        }
    }
    for(unsigned int d=0; d<N_DIM; d++)
    {
        Lavg[d]/=count;
        Labs[d]/=count;
        Cavg[d]/=count;
        Lvar[d]=sqrt(std::max<double>(0, Lvar[d]/count-pow_2(Lavg[d])));
    }
    // determine orbit class
    bool tubez=(fabs(Lavg[2])>Labs[2]*OC_TUBE_AVGL);  // has definite sense of rotation about z axis (more robust indicator than 1:1 correspondence of leading freqencies)
    bool tubex=(N_dim==3) && (fabs(Lavg[0])>Labs[0]*OC_TUBE_AVGL);  // same for x axis
    bool tubey=(N_dim==3) && (fabs(Lavg[1])>Labs[1]*OC_TUBE_AVGL);  // same for y axis (although we don't expect to find y-axis tubes if axes are sorted properly)
    double lfX=out_lines[0][0];
    double lfY=out_lines[1][0];
    double acc = timeUnit/(timeStep*count) * FREQ_ACCURACY;  // accuracy of comparison
    if(acc<FREQ_DIFF_REG_MIN) acc=FREQ_DIFF_REG_MIN;  // don't be too demanding.. (actually the accuracy limit is determined by energy conservation error, for simplicity ignored here)
    if(N_dim==2)
    {
        // try to find nx:ny correspondence of leading frequencies
        int nx, ny;
        if(commensurable(lfX, lfY, acc, nx, ny))    // boxlet family or tube (in case lfX==lfY)
        {
            double lf = (lfX/nx + lfY/ny)/2;  // mean value of common denominator 
            // find additional (libration) frequency
            double lfa = 0;
            bool regular=true;
            int rank=1;
            int ma, mb;
            for(unsigned int c=0; c<N_dim; c++)
            {
                for(size_t i=1; regular&&(i<out_lines[c].size()); i++)
                {
                    if(commensurable(out_lines[c][i], lf, acc, ma, mb)) 
                    {
                        if(((ma==1) || (abs(ma-mb)==1)) && (mb>1))
                            rank=mb;
                    }
                    else
                    {
                        double lfa1=std::min<double>(out_lines[c][i].freq, fabs(out_lines[c][0].freq-out_lines[c][i].freq));
                        if(lfa)
                        {
                            if(!lineardepend(out_lines[c][i], lf, lfa, acc, ma, mb)) 
                            {
                                regular=false;
                            }
                            else 
                                if(lfa1<lfa) 
                                {
                                    lfa=lfa1;
                                }
                        }
                        else
                        {
                            lfa=lfa1;
                        }
                    }
                }
            }
            if(tubez)
                Description = "Tube";
            else if((nx==1)&&(ny==2))
                Description = "banana";
            else if((nx==2)&&(ny==3))
                Description = "fish";
            else if((nx==3)&&(ny==4))
                Description = "pretzel";
            else 
                Description = convertToString(nx)+":"+convertToString(ny)+" resonant boxlet";
            Description += " rank "+convertToString(rank);
            if(!regular)
                Description = "chaotic "+Description;
        }
        else 
        {
            // for regular box orbit the remaining lines must be linear-dependent on two leading frequencies
            bool regular=true;
            int ma, mb;
            for(int c=0; c<2; c++)
                for(size_t i=1; regular&&(i<out_lines[c].size()); i++)
                    if(!lineardepend(out_lines[c][i], lfX, lfY, acc, ma, mb)) regular=false;
            if(regular)
                Description = "box";
            else
                Description = "chaotic";
        }
    }
    else    // N_dim==3
    {
        // determine if there are resonances between each pair of coordinates
        bool res[3];
        int nres=0;          // number of resonances
        int na[3], nb[3];    // resonance quotients
        double lf[3]={0,0,0};// leading frequencies (1 for nres=3, 2 for nres=1, 3 for nres=0)
        int nlf=0;           // number of leading frequencies
        for(int d=0; d<3; d++)
        {
            res[d] = commensurable(out_lines[d][0], out_lines[(d+1)%3][0], acc, na[d], nb[d]);
            if(res[d]) nres++;
        }
        
        if(nres==2)  // check for consistency (there cannot be two resonances)
        {   // reconstruct missing resonance number dn
            int dn=!res[0] ? 0 : (!res[1] ? 1 : 2);
            na[dn]=nb[(dn+2)%3]*nb[(dn+1)%3];
            nb[dn]=na[(dn+1)%3]*na[(dn+2)%3];
            simplify(na[dn], nb[dn]);
            nres=3;
        }
        if(nres==0)
        {
            nlf=3;
            lf[0]=out_lines[0][0];
            lf[1]=out_lines[1][0];
            lf[2]=out_lines[2][0];
            int n1, n2, n3;
            if(commensurable3(lf[0],lf[1],lf[2],acc,n1,n2,n3))
            {
                Description = "("+convertToString(n1)+","+convertToString(n2)+","+convertToString(n3)+") thin orbit";
                nlf = 2;
            }
            else
            {
                // either box or pyramid (near black hole)
                if(fabs(Cavg[2])>fabs(Cmax[2]*OC_PYRAMID_AVGZ))
                    Description = "pyramid";
                else if(fabs(Cavg[0])>fabs(Cmax[0]*OC_PYRAMID_AVGZ))
                    Description = "x-pyramid";   // shouldn't exist normally
                else if(fabs(Cavg[1])>fabs(Cmax[1]*OC_PYRAMID_AVGZ))
                    Description = "y-pyramid";   // shouldn't exist normally
                else
                    Description = "box";
            }
        }
        if(nres==1)
        {   
            int de=res[0] ? 0 : (res[1] ? 1 : 2);
            nlf=2;
            lf[0] = (out_lines[de][0]/na[de] + out_lines[(de+1)%3][0]/nb[de])/2;
            lf[1] = out_lines[(de+2)%3][0];
            if(lf[1]==0)
                nlf=1;  // avoid degenerate cases
            switch(de)
            {
            case 0: Description = convertToString(na[0])+":"+convertToString(nb[0])+":*"; break;
            case 1: Description = "*:"+convertToString(na[1])+":"+convertToString(nb[1]); break;
            case 2: Description = convertToString(nb[2])+":*:"+convertToString(na[2]); break;
            }
            Description +=" resonance";
        }
        if(nres==3)
        {
            nlf=1;
            lf[0]=std::min<double>((out_lines[0][0].freq/na[0] + out_lines[1][0].freq/nb[0])/2, 
                std::min<double>  ((out_lines[1][0].freq/na[1] + out_lines[2][0].freq/nb[1])/2, 
                                   (out_lines[2][0].freq/na[2] + out_lines[0][0].freq/nb[2])/2));
            int nx=(int)floor(out_lines[0][0]/lf[0]+0.5); 
            int ny=(int)floor(out_lines[1][0]/lf[0]+0.5); 
            int nz=(int)floor(out_lines[2][0]/lf[0]+0.5); 
            Description = convertToString(nx)+":"+convertToString(ny)+":"+convertToString(nz)+" resonance";
        }
        if(tubex && !tubez && !tubey) 
        {
            Description = "LAT "+Description;
            // try to determine whether it's inner or outer LAT
            // ..remains to be implemented
        }
        if(tubez && !tubey && !tubex) 
        {
            if(fabs(Cavg[2])>fabs(Cmax[2]*OC_PYRAMID_AVGZ))
                Description = "SAU "+Description;  // saucer orbit (a variant of short-axis tube)
            else
                Description = "SAT "+Description;
        }
        if(tubey && !tubex && !tubez)
        {
            Description = "IAT "+Description;  // intermediate-axis tube, they simply don't exist.
        }
        // check that every other freq is commensurable with base
        int ma, mb, mc;
        bool regular=true;
        for(int c=0; c<3; c++)
            for(size_t i=1; regular&&(i<out_lines[c].size()); i++)
            {
                switch(nlf)
                {
                case 1:
                    if(!commensurable(out_lines[c][i], lf[0], acc, ma, mb)) 
                    {
                        nlf++;
                        lf[1] = std::min<double>(out_lines[c][i].freq, fabs(lf[0]-out_lines[c][i].freq));
                    }
                    break;
                case 2:
                    if(!lineardepend(out_lines[c][i], lf[0], lf[1], acc, ma, mb))
                    {
                        nlf++;
                        lf[2] = std::min<double>(out_lines[c][i].freq, 
                            std::min<double>(fabs(lf[0]-out_lines[c][i].freq), fabs(lf[1]-out_lines[c][i].freq)));
                    }
                    break;
                case 3:
                    if(!lineardepend3(out_lines[c][i], lf[0], lf[1], lf[2], acc, ma, mb, mc))
                        regular=false;
                }
            }
        if(!regular)
            Description = "chaotic "+Description;
    }
    return Description;
}

void COrbitRuntimeTrajectory::calcLFCCoverInterval(size_t start, size_t count, double leadfreq[]) const
{
    CSpectrum tempSpectrum[N_DIM];
    findLines(start, count, tempSpectrum, NULL);
    for(unsigned int c=0; c<N_dim; c++)
        leadfreq[c]=(tempSpectrum[c].size()>0)? tempSpectrum[c][0].freq : -1;
}

// calculate rate of diffusion of leading frequencies in cartesian coordinates (LFCCs), which are related to the degree of chaos in the orbit
// determines the LFCCs for first and second halves of orbit, computes difference, returns delta f/f averaged over all coordinates
double COrbitRuntimeTrajectory::calcLFCCdiffusion(size_t start, size_t count) const
{
    if(count<=0) count=traj.size()-start;  // whole orbit by default
    size_t count2=count/2;
    double lf1[N_DIM], lf2[N_DIM];  // leading frequencies of 1st and 2nd half for each coordinate
    calcLFCCoverInterval(start, count2, lf1);
    calcLFCCoverInterval(start+count2, count2, lf2);
    double FreqDiff = 0;
#if FREQ_DIFF_METHOD == 0
    for(int c=N_dim-1; c>=0 && FreqDiff==0; c++)
      if(lf1[c] + lf2[c] > 0 )
        FreqDiff = fabs(lf1[c] - lf2[c]) * 2 / (lf1[c] + lf2[c]);
#else
    int nf=0;
    for(unsigned int c=0; c<N_dim; c++)
        if(lf1[c] + lf2[c] > 0 )
#if FREQ_DIFF_METHOD == 2
          if(fabs(lf1[c] - lf2[c]) * 2 / (lf1[c] + lf2[c]) < 1)
#endif
        {
            FreqDiff += fabs(lf1[c] - lf2[c]) * 2 / (lf1[c] + lf2[c]);
            nf++;
        }
    if(nf>0) FreqDiff /= nf;
#endif
#if 0
    double maxFreqDiff = pow_2(timeUnit/intTime);  // rough estimate of regular/chaotic divide
    if(maxFreqDiff<FREQ_DIFF_REG_MIN) maxFreqDiff=FREQ_DIFF_REG_MIN;

    if(FreqDiff<1*maxFreqDiff)
        DescriptionLFCCdiff = "regular";
    else 
    if(FreqDiff<2*maxFreqDiff)
        DescriptionLFCCdiff = "probably regular";
    else 
    if(FreqDiff<4*maxFreqDiff)
        DescriptionLFCCdiff = "probably chaotic";
    else 
        DescriptionLFCCdiff = "chaotic";
#endif
    return FreqDiff;
}

void COrbitRuntimeTrajectory::calcInertiaOverInterval(size_t start, size_t count, double inertia[]) const
{
    for(unsigned int c=0; c<N_DIM; c++)
        inertia[c] = 0;
    for(size_t i=0; i<count; i++)
    {
        for(unsigned int c=0; c<N_dim; c++)
            inertia[c] += pow_2(traj[i+start].Pos[c]);
    }
    for(unsigned int c=0; c<N_dim; c++)
        inertia[c] = sqrt(inertia[c]/count);
}

// calculates maximal Lyapunov characteristic exponent (if the deviation equation was integrated along with orbit)
double COrbitRuntimeTrajectory::calcLambda() const
{
    if(!calcLyapunov)
        return 0;
    double sumavg=0;     // to calculate average value of ln(|w|/t) over the linear-growth interval
    int cntavg=0;
    double sumlambda=0;  // to calculate average value of ln(|w|)/t over the exponential growth interval
    int cntlambda=0;
    int timeblock=static_cast<int>(timeUnit/timeStep*2);    // averaging interval (roughly orbital period * 2)
    std::vector<double> interval(timeblock);
    bool chaotic=false;
    for(int t=timeblock; t<static_cast<int>(traj.size()); t+=timeblock)
    {
        int cnt = std::min<int>(timeblock, static_cast<int>(traj.size())-t);
        if(!chaotic)
        {
            // compute median value over 'timeblock' points  to avoid outliers
            interval.resize(cnt);
            for(int i=0; i<cnt; i++)
                interval[i] = (lnw[i+t]-log((i+t)*timeStep));
            std::sort(interval.begin(), interval.end());
            double mid = interval[cnt/2];
            sumavg+=mid;
            cntavg++;
            if((cntavg>4) && (mid > sumavg/cntavg + 2))
                chaotic=true;     // from here on, assume that we got into t->infinity regime and start collecting data for estimating lambda
        }
        if(chaotic)
        {
            for(int i=0; i<cnt; i++)
            {
                sumlambda += lnw[i+t]/(i+t);
                cntlambda++;
            }
        }
    }
    if(chaotic)
        return (sumlambda / cntlambda * timeUnit / timeStep);
    else
        return 0;
}

// return maximal distance between adjacent points
double COrbitRuntimeTrajectory::getMaxDist(size_t start, size_t count) const
{
    if(start>=traj.size())  return -1; // invalid call
    if(count<=0 || count>traj.size()-start) count = traj.size() - start;   // passing zero for count means analysing entire orbit
    if(count<=1) return -1; // empty orbit
    double maxd=0;
    CPosPoint<double> prev=traj[start];
    for(size_t i=start+1; i<start+count; i++)
    {
        CPosPoint<double> cur=traj[i];
        maxd = std::max<double>(maxd, sqrt(pow_2(cur.Pos[0]-prev.Pos[0])+pow_2(cur.Pos[1]-prev.Pos[1])+pow_2(cur.Pos[2]-prev.Pos[2])));
        prev=cur;
    }
    return maxd;
}

//----------------------------------------------------------------------------//
// Timestep function for storing a number of sampling points from trajectory
COrbitRuntimeTrajSample::COrbitRuntimeTrajSample(const COrbit* _orbit, size_t _numSamplingPoints, double _timeStep) :
  CBasicOrbitRuntimeFnc(_orbit), 
  numSamplingPoints(_numSamplingPoints), 
  timeStep(_timeStep),
  N_dim(orbit->getPotential()->N_dim)
{}; 

bool COrbitRuntimeTrajSample::Timestep(const double told, const double t, const double /*y*/[])
{
    int i1=static_cast<int>(told/timeStep); 
    int i2=static_cast<int>(t/timeStep);

    // store trajectory at regular intervals of time    
    for (int i=i1; i<=i2; i++)
    {
        double tt=timeStep*i;
        if ((tt>=told) && (tt<t))
        {
            CPosVelPoint<float> pt;
            for(unsigned int c=0; c<N_dim; c++)
            {
                pt.Pos[c]=static_cast<float>(orbit->getInterpolatedTrajectory(c, tt));
                pt.Vel[c]=static_cast<float>(orbit->getInterpolatedTrajectory(c+N_dim, tt));
            }
            trajEntire.push_back(pt);
        }
    }
    return true;
}

void COrbitRuntimeTrajSample::Finish()
{
    size_t numActual=trajEntire.size();
    if(numActual<1) return;
    numSamplingPoints = std::min<size_t>(numSamplingPoints, numActual-1);
    std::set<size_t> samplingIndexes;
    gsl_rng* randn=gsl_rng_alloc(gsl_rng_default);
    for(size_t i=0; i<numSamplingPoints; i++)
    {
        size_t np;
        do{
            np=static_cast<size_t>(floor(gsl_rng_uniform(randn)*(numActual-1)));
        }
        while(samplingIndexes.find(np)!=samplingIndexes.end());
        samplingIndexes.insert(np);   // try to avoid duplication
        trajSample.push_back(trajEntire[np]);
    }
    gsl_rng_free(randn);
}

COrbitRuntimeTrajSampleCreator::COrbitRuntimeTrajSampleCreator(unsigned int _numSamplingPointsDefault, double _intTimeMin, const CPointCountSetFloat* _numSamplingPointsSpecial) : 
  numSamplingPointsDefault(_numSamplingPointsDefault), intTimeMin(_intTimeMin)
{ 
    if(numSamplingPointsDefault==0) {
        my_error("Trajectory sampler: numSamplingPoints=0"); 
        return;
    }
    if(_numSamplingPointsSpecial!=NULL) 
        numSamplingPointsSpecial.assign(_numSamplingPointsSpecial->begin(), _numSamplingPointsSpecial->end());
};

CBasicOrbitRuntimeFnc* COrbitRuntimeTrajSampleCreator::createRuntimeFnc(const COrbit* orbit) const
{ 
    size_t numSamplingPoints=numSamplingPointsDefault;
    // find if this orbit is listed in the 'special' list
    if(!numSamplingPointsSpecial.empty())
    {
        CPosVelPoint<double> initCond(orbit->getInitData().initCond);
        for(size_t i=0; i<numSamplingPointsSpecial.size(); i++)
        {
            CPosVelPoint<double> specialCond(numSamplingPointsSpecial[i].first);
            const double eps=1e-6;  // accuracy of comparison (float rather than double)
            if( gsl_fcmp(initCond.Pos[0], specialCond.Pos[0], eps)==0 &&
                gsl_fcmp(initCond.Pos[1], specialCond.Pos[1], eps)==0 &&
                gsl_fcmp(initCond.Pos[2], specialCond.Pos[2], eps)==0 &&
                gsl_fcmp(initCond.Vel[0], specialCond.Vel[0], eps)==0 &&
                gsl_fcmp(initCond.Vel[1], specialCond.Vel[1], eps)==0 &&
                gsl_fcmp(initCond.Vel[2], specialCond.Vel[2], eps)==0)
                numSamplingPoints = numSamplingPointsSpecial[i].second;
        }
    }
    // compute timestep and create runtime function
    return new COrbitRuntimeTrajSample(orbit, numSamplingPoints, 
        orbit->getInitData().timeUnit*intTimeMin/(numSamplingPoints*10+1)); 
}

//----------------------------------------------------------------------------//
/// \cond INTERNAL_DOCS
// simple trajectory timestep function recording points in Poincare section
struct CPoincareParam {
    const COrbit* orbit;
    unsigned int psindex;
};

double findPS(double t, void* params)
{
    return static_cast<CPoincareParam*>(params)->orbit->getInterpolatedTrajectory(static_cast<CPoincareParam*>(params)->psindex,t);
}
/// \endcond 

bool COrbitRuntimePoincare::Timestep(const double told, const double t, const double y[])
{
    // Poincare section
    unsigned int N_dim=orbit->getPotential()->N_dim;
    double yprev=(told>0)?orbit->getInterpolatedTrajectory(PS2, told):y[PS2];  // previous y-value
    if((ps.size()<MAX_POINCARE_POINTS) && (yprev<=0) && (y[PS2]>=0))  // put a point
    {
        if(y[PS2]>yprev)
        {
            // interpolate coordinates at Poincare section crossing
            gsl_function F;
            CPoincareParam params;
            F.function=&findPS;
            F.params=&params;
            params.orbit=orbit;
            params.psindex=PS2;
            double tcross = findRoot(&F, told, t);
            ps.push_back(paird(orbit->getInterpolatedTrajectory(PS1,tcross), orbit->getInterpolatedTrajectory(PS1+N_dim, tcross)));
        }
        else
        {
            ps.push_back(paird(y[PS1], y[PS1+N_dim]));
        }
    }
    return true;
}

//----------------------------------------------------------------------------//
COrbitRuntimePericenter::COrbitRuntimePericenter(const COrbit* _orbit):
CBasicOrbitRuntimeFnc(_orbit)
{
    fitSlope=fitIntercept=fitScatter=fitSignificance=0;
    Lavg[0]=Lavg[1]=Lavg[2]=L2avg[0]=L2avg[1]=L2avg[2]=0;
    pericenterrmin=gsl_posinf();
    // init sign
    COrbitInitData<double> data = orbit->getInitData();
    drdtsign= data.initCond.Pos[0]*data.initCond.Vel[0]+data.initCond.Pos[1]*data.initCond.Vel[1]+data.initCond.Pos[2]*data.initCond.Vel[2] >0 ? 1 : -1;
    // find approximate L2 for a circular orbit
    Lcirc2=findLcirc2(orbit->getPotential(), orbit->getPotential()->totalEnergy(data.initCond));
}

template<typename NumT> COrbitRuntimePericenter::COrbitRuntimePericenter(const COrbit* _orbit, const std::vector< CPosVelPoint<NumT> > &traj):
CBasicOrbitRuntimeFnc(_orbit)
{
    fitSlope=fitIntercept=fitScatter=fitSignificance=0;
    pericenterrmin=gsl_posinf();
    Lavg[0]=Lavg[1]=Lavg[2]=L2avg[0]=L2avg[1]=L2avg[2]=0;
    drdtsign= traj[0].Pos[0]*traj[0].Vel[0]+traj[0].Pos[1]*traj[0].Vel[1]+traj[0].Pos[2]*traj[0].Vel[2] >0 ? 1 : -1;
    int N_dim=orbit->getPotential()->N_dim;  // shortcut
    double timeStep=orbit->getInitData().timeStep;
    for(size_t i=1; i<traj.size(); i++)
    {
        int prevsign=drdtsign;
        drdtsign= traj[i].Pos[0]*traj[i].Vel[0]+traj[i].Pos[1]*traj[i].Vel[1]+traj[i].Pos[2]*traj[i].Vel[2] >0 ? 1 : -1;
        if(drdtsign*prevsign<0 && drdtsign>0)
        {   // direction reversed - passed pericenter; do not try to refine exact time, just record current squared ang.mom.
            double Lx=(N_dim==3)? traj[i].Pos[1]*traj[i].Vel[2]-traj[i].Pos[2]*traj[i].Vel[1] : 0;
            double Ly=(N_dim==3)? traj[i].Pos[2]*traj[i].Vel[0]-traj[i].Pos[0]*traj[i].Vel[2] : 0;
            double Lz=traj[i].Pos[0]*traj[i].Vel[1]-traj[i].Pos[1]*traj[i].Vel[0];
            double L2=pow_2(Lx)+pow_2(Ly)+pow_2(Lz);
            double rperi=sqrt(pow_2(traj[i].Pos[0])+pow_2(traj[i].Pos[1])+pow_2(traj[i].Pos[2]));
            pericentert.push_back(i*timeStep);
            pericenterr.push_back(rperi);
            pericenterl2.push_back(L2);
            Lavg[0]+=Lx; L2avg[0]+=pow_2(Lx);
            Lavg[1]+=Ly; L2avg[1]+=pow_2(Ly);
            Lavg[2]+=Lz; L2avg[2]+=pow_2(Lz);
            if(rperi<pericenterrmin) pericenterrmin=rperi;
        }
    }
    // find approximate L2 for a circular orbit
    Lcirc2=findLcirc2(orbit->getPotential(), orbit->getPotential()->totalEnergy(traj[0]));
}

/// \cond INTERNAL_DOCS
struct CPericenterParam {
    const COrbit* orbit;
};

double finddrdt(double t, void* params)
{
    const COrbit* o=static_cast<CPericenterParam*>(params)->orbit;
    unsigned int N_dim=o->getPotential()->N_dim;
    return o->getInterpolatedTrajectory(0,t)*o->getInterpolatedTrajectory(N_dim,t) + o->getInterpolatedTrajectory(1,t)*o->getInterpolatedTrajectory(N_dim+1,t) + (N_dim==3?o->getInterpolatedTrajectory(2,t)*o->getInterpolatedTrajectory(5,t):0);
}
/// \endcond 

bool COrbitRuntimePericenter::Timestep(const double told, const double t, const double y[])
{
    unsigned int N_dim=orbit->getPotential()->N_dim;  // shortcut
    int prevsign=drdtsign;
    drdtsign= y[0]*y[N_dim]+y[1]*y[N_dim+1]+(N_dim==3?y[2]*y[5]:0) >0 ? 1 : -1;
    if(t>told && (drdtsign*prevsign<0) && drdtsign>0)
    {
        // direction reversed -- passed pericenter; find exact passagee time
        gsl_function F;
        CPericenterParam params;
        params.orbit=orbit;
        F.function=&finddrdt;
        F.params=&params;
        double tperi=findRoot(&F, told, t);
        pericentert.push_back(tperi);
        double yperi[N_DIM*2];
        for(unsigned int i=0; i<N_dim*2; i++)
            yperi[i]=orbit->getInterpolatedTrajectory(i, tperi);
        double rperi=sqrt(pow_2(yperi[0])+pow_2(yperi[1])+pow_2(yperi[2]));
        pericenterr.push_back(rperi);
        if(rperi<pericenterrmin) pericenterrmin=rperi;
        double Lx=(N_dim==3)? yperi[1]*yperi[5]-yperi[2]*yperi[4] : 0;
        double Ly=(N_dim==3)? yperi[2]*yperi[3]-yperi[0]*yperi[5] : 0;
        double Lz=(N_dim==3)? yperi[0]*yperi[4]-yperi[1]*yperi[3] : yperi[0]*yperi[3]-yperi[1]*yperi[2];
        double L2=pow_2(Lx)+pow_2(Ly)+pow_2(Lz);
        pericenterl2.push_back(L2);
        Lavg[0]+=Lx; L2avg[0]+=pow_2(Lx);
        Lavg[1]+=Ly; L2avg[1]+=pow_2(Ly);
        Lavg[2]+=Lz; L2avg[2]+=pow_2(Lz);
    }
    return true;  // do not require anything special to be able to terminate integration
}

void COrbitRuntimePericenter::Finish()
{
    vectord lp(pericenterl2);  // !! here one may substitute either pericenter radius or pericenter ang.mom.sq., but the latter is more general
    std::sort(lp.begin(), lp.end());
    int maxn=static_cast<int>(lp.size());
    int nump=static_cast<int>(maxn*PERICENTER_FIT_FRACTION);
    if(nump<PERICENTER_FIT_MIN_POINTS) nump=PERICENTER_FIT_MIN_POINTS;
    if(nump>PERICENTER_FIT_MAX_POINTS) nump=PERICENTER_FIT_MAX_POINTS;
    if(nump>=maxn) nump=maxn-1;
    if(nump<3) { fitScatter=fitSignificance=fitSlope=fitIntercept=0; return; }
    // find out whether there are any conserved components of ang.mom.
    double L2min=0;  // minimum possible L^2 due to conserved quantities
    for(unsigned int d=0; d<N_DIM; d++)
    {
        Lavg[d]/=lp.size();
        L2avg[d]/=lp.size();
        const double eps=1e-5;  // relative accuracy of comparison (std.dev. in L[d] < eps*|L[d]| )
        if(gsl_fcmp(pow_2(Lavg[d]), L2avg[d], eps*eps)==0)
            L2min+=L2avg[d];
    }
    if(L2min>0)  // subtract it from the values for fitting
    {
        if(L2min>lp[0]) L2min=lp[0];  // to be sure that first point is not below that...
        for(int i=0; i<nump; i++)
            lp[i]-=L2min;
    }
    vectord np(nump)/*, weight(nump)*/;
    for(int i=0; i<nump; i++)
    {
        np[i]=(i+1.0)/maxn;
        /*weight[i] = (i<PERICENTER_FIT_MIN_POINTS ? 1.0 : (PERICENTER_FIT_MIN_POINTS+0.0)/i);*/  // not used -- may be improved in the future to give more weight to innermost points
    }
    double cov00, cov01, cov11, sumsq;
    // fit linear regression with and without constant term
    gsl_fit_linear(&(np.front()), 1, &(lp.front()), 1, nump, &fitIntercept, &fitSlope, &cov00, &cov01, &cov11, &sumsq);
    // compute expected dispersion assuming that chi^2/(N-2)=1
    double sigma_expected = fitSlope/maxn * sqrt(nump*1.0);
    double sigma_actual = sqrt(sumsq/(nump-2));
#if 0
    double sigma2_used = pow_2(std::max(sigma_expected, sigma_actual));
#else
    double sigma2_used = pow_2(sigma_actual);
#endif
    double chi2 = sumsq/sigma2_used;    // equals to nump-2 or less

    double fitSlopeM, cov11M, sumsqM;
    gsl_fit_mul(&(np.front()), 1, &(lp.front()), 1, nump, &fitSlopeM, &cov11M, &sumsqM);
    double deltachi2 = //(sumsqM/sumsq - 1)*(nump-2);
        sumsqM/sigma2_used - chi2;
    if(deltachi2 < 11.8 || fitIntercept<0)  // <3sigma
    {
        fitSlope=fitSlopeM;
        fitIntercept=0;
        sumsq=sumsqM;
    }
    // compute ratio of actual over expected dispersion
    fitScatter = sigma_actual/sigma_expected;
    // compute how significant (how many sigmas) is intercept!=0 with respect to zero intercept
    double stddevcount=0;
    if(deltachi2<50)  // less than ~7sigma - directly compute equivalent number of 'gaussian' sigmas from chi^2 tail
    {
        stddevcount = gsl_cdf_gaussian_Qinv(gsl_cdf_chisq_Q(deltachi2, 2)/2, 1);
    }
    else
    {
        double intrcept, stddev;
        gsl_fit_linear_est(0, fitIntercept, fitSlope, cov00, cov01, cov11, &intrcept, &stddev);
        stddevcount = intrcept/stddev;  // distance in std.dev.
    }
    fitSignificance=stddevcount;
    fitIntercept += L2min;  // add back conserved value
}

/* ------- Frequency analysis routines -------- */
/** utility function to ensure that n does not have large prime divisors.
    if needed, decrease it (in the FFT primes are great evil). */
void tuneNonPrime(size_t *n)
{
    assert(n);
    size_t m=*n, f=2;
    while(f*f<=m)
        if(m%f==0) m/=f; else f++;
    if(m>=MAX_FFT_PRIME && *n>=MAX_FFT_PRIME)
    {
        (*n)--;
        tuneNonPrime(n);
    }
}

// perform FFT of orbit (output amplitudes to 'fa' array) and find 'maxnlines' lines, which will be placed into 'sl' array. 
// If fa=NULL, do not output frequency amplitudes
void CSpectrumAnalyzerHunter::findSpectralLines(smile::CSpectrum *spectralLines, smile::vectord *spectrumAmplitudes)
{
    size_t npoints=timeSeries.size();
    if(npoints==0 || spectralLines==NULL) return;
    tuneNonPrime(&npoints);
    gsl_fft_real_wavetable * wavetable = gsl_fft_real_wavetable_alloc (npoints);
    gsl_fft_real_workspace * workspace = gsl_fft_real_workspace_alloc (npoints);
    vectord in(timeSeries);   // copy of input array; FFT is performed in-place
    gsl_fft_real_transform (&(in.front()), 1, npoints, wavetable, workspace);
    gsl_fft_real_wavetable_free (wavetable);
    gsl_fft_real_workspace_free (workspace);
    std::vector<complexd> fc(npoints/2+1);  // working array of complex amplitudes
    fc[0]=complexd(in[0],0)  / (0.5*npoints);
    for(size_t i = 1; i < npoints/2; i++)
        fc[i]=complexd(in[2*i-1], in[2*i])/(0.5*npoints);
    if(npoints%2==0)
        fc[npoints/2]=complexd(in[npoints-1], 0)/(0.5*npoints);
    if(spectrumAmplitudes!=NULL)
    {  // output amplitudes only (for illustrative purposes) if necessary
        spectrumAmplitudes->resize(npoints/2+1, 0);
        for(size_t i=0; i<npoints/2+1; i++)
            spectrumAmplitudes->at(i)=abs(fc[i]);    // store amplitude of spectrum if needed - and then forget it; only the true complex spectrum is used hereafter
    }
    // perform frequency analysis
    spectralLines->clear();
    bool finish=false;
    double maxamp=0; double ma=0;
    while(!finish)
    {   // find next greatest line
        int m=0; 
        ma=0;
        for(int i=0; i<static_cast<int>(npoints/2)-2; i++) {
            if(abs(fc[i])>ma)
            {
                ma=abs(fc[i]);
                m=i;
            }
        }
        if(maxamp==0 && m>0) maxamp=ma;  // record max.amplitude if the frequency is non-zero
        if(ma>maxamp*MIN_SPECTRAL_LINE_AMPLITUDE)
        {   // add line
            // peak is somewhere near m.  
            if(m>0)
            {
                int m1=m;
                if(m>1 && abs(fc[m-1])>abs(fc[m+1]))  m1--;
                double s;   // fractional index of peak frequency 
                complexd Ac;// amplitude
                complexd f1 = (fc[m1] - (fc[m1-1]+fc[m1+1])/2.0);
                complexd f2 = (fc[m1+1] - (fc[m1]+fc[m1+2])/2.0);
                // C=f2/f1; check out that C<>0 and C<>infinity
                if(abs(f1) < abs(f2)*1e-4)
                {   
                    m1++;
                    f1=f2;
                    f2 = (fc[m1+1] - (fc[m1]+fc[m1+2])/2.0);
                }
                if((abs(f2) < abs(f1)*1e-4) && (m1>1))  // !!!what happens here???
                {   
                    m1--;
                    f2=f1;
                    f1 = (fc[m1] - (fc[m1-1]+fc[m1+1])/2.0);
                }
                complexd C = f2 / f1;
                double Ca=C.real();  // in ideal case, Im C=0
                double omega=2*M_PI/npoints;
                int method=1;  // Hunter's method
                if((Ca<-1.0/FREQ_PRECISE_THRESHOLD) && (Ca>-FREQ_PRECISE_THRESHOLD)/*abs(C.imag())*0 < abs(Ca)*/)
                {   // try Hunter's precise method
                    double sgn = (Ca<1)?+1:-1;
                    double t = (Ca*cos(omega)-1 + sgn * sqrt( pow_2(Ca*cos(omega)-1) - (1+2*Ca)*pow_2(sin(omega))))/sin(omega);
                    s = m1 + npoints/M_PI * atan(t);
                    complexd z = std::polar(1.0, 2*M_PI*(s-m1)/npoints); //complexd((1-pow_2(t))/(1+pow_2(t)), 2*t/(1+pow_2(t)));
                    complexd beta = std::polar(1.0, omega);
                    complexd P = (std::polar(1.0, 2*M_PI*s)-1.0) / (std::polar(1.0, 2*M_PI*(s-m1)/npoints)-1.0) / (1.0*npoints);
                    complexd g1 = P * (1.0 - 0.5*(z-1.0)*(1.0/(z*beta-1.0) + 1.0/(z*std::conj(beta)-1.0)));
                    Ac = f1/g1;
                }
                else
                {   // Follow the method by Carpintero&Aguilar(1998)
                    method=0;
                    double alpha = atan(sin(omega/2)/(cos(omega/2)+abs(fc[m1]/fc[m1+1])));
                    s = m1 + npoints*alpha/M_PI;
                    complexd P=(std::polar(1.0, 2*alpha*npoints)-1.0) / (std::polar(1.0, 2*alpha)-1.0) / (1.0*npoints);
                    complexd Q=(std::polar(1.0, -2*M_PI*(s+m1))-1.0) / (std::polar(1.0, -2*M_PI*(s+m1)/npoints)-1.0) / (1.0*npoints);
                    Ac=fc[m1] / P;
                }
                spectralLines->push_back(CSpectralLine(s/(timeStep*npoints), abs(Ac), std::arg(Ac), method));
                if(spectralLines->size()>=MAX_SPECTRAL_LINES) finish=true;    // not to iterate forever
                // subtract line from spectrum
                if(!finish)
                {
                    if(s-floor(s) < 1e-5)    // frequency was very close to integer value, so subtract only this point
                    {
                        fc[(int)floor(s)]-=Ac;
                    } else
                        if(s-floor(s) > 1-1e-5)
                        {
                            fc[(int)floor(s)+1]-=Ac;
                        }
                        else   // subtract this line with its wings from the whole spectrum
                        {
                            complexd Amul = Ac / (1.0*npoints) * (std::polar(1.0, 2*M_PI*(s))-1.0);
                            complexd Acmul = std::conj(Ac) / (1.0*npoints) * (std::polar(1.0,-2*M_PI*(s))-1.0);
                            int jmin=std::max<int>(0, m-1000);
                            int jmax=std::min<int>(static_cast<int>(npoints/2)+1, m+1000);
                            for(int j=jmin; j<jmax; j++)
                            {
                                complexd linea = Amul / (std::polar(1.0, 2*M_PI*(s-j)/npoints)-1.0) +
                                    Acmul / (std::polar(1.0,-2*M_PI*(s+j)/npoints)-1.0);
                                fc[j] -= linea;
                            }
                        }
                }
            }
            else
            {   // zero frequency -- just nullify one array element (do not record zero frequencies)
                fc[m]=0;
            }
        }
        else finish=true;
    }
    std::sort(spectralLines->begin(), spectralLines->end());
    if(spectralLines->size()==0)
        spectralLines->push_back(CSpectralLine());  // add zero to ensure that array is non-empty
}

// Utility function to find a 1:1 periodic orbit in a plane specified by coord1:coord2 
bool findPeriodicOrbit(const CPotential* potential, double E, const CConfigOrbit* configOrbit, int coord1, int coord2, double& cross1)
{
    if(coord1>coord2) { int tmp=coord1; coord1=coord2; coord1=tmp; }
    double Xlo=0, Xhi=potential->longaxisradius(E);
    double tol=Xhi*1e-8;
    double Torb=potential->longaxisperiod(E, Xhi);
    COrbitInitData<double> IC(potential, Torb/10, Torb, CPosVelPoint<double>(), false);
    cross1=0;
    int niter=0;
    vectorRuntimeFncCreators crts;
    crts.push_back(new COrbitRuntimePoincareCreator(coord1, coord2));
    do{
        double v2;
        do{
            IC.initCond.Pos[coord1]=(Xlo+Xhi)/2; 
            v2=2*(E-potential->Phi(IC.initCond.Pos[0],IC.initCond.Pos[1],IC.initCond.Pos[2]));
            if(v2<0) Xhi=IC.initCond.Pos[coord1];
            else IC.initCond.Vel[coord2]=sqrt(v2);
        } 
        while(v2<0);
        COrbit* orb=new COrbit(configOrbit, IC, &crts);
        orb->integrateToTime(Torb*2);   // should perform more than one revolution
        const COrbitRuntimePoincare* pi = static_cast<const COrbitRuntimePoincare*>(orb->getRuntimeFnc(0));
        if(pi->size()<2)
        {
            delete crts[0];
            return false;
        }
        if(pi->at(1).first > IC.initCond.Pos[coord1])
        {   Xlo=IC.initCond.Pos[coord1]; }
        else
        {   Xhi=IC.initCond.Pos[coord1]; }
        niter++;
        delete orb;
    }
    while(niter<25 && Xhi-Xlo>tol && Xhi>tol);
    cross1=(Xhi>tol) ? IC.initCond.Pos[coord1] : 0;
    delete crts[0];
    return (Xhi>tol);
}

/// \cond INTERNAL_DOCS
struct CRcircParam{
    const CPotential* P;
    double E;
    unsigned int direction;
};

double findRcirc(double r, void* params)
{
    double X=0, Y=0, Z=0;
    switch(static_cast<CRcircParam*>(params)->direction)
    {
    case 0: X=r; break;
    case 1: Y=r; break;
    case 2: Z=r; break;
    }
    const CPotential* P=static_cast<CRcircParam*>(params)->P;
    double V2 = 2*( static_cast<CRcircParam*>(params)->E - P->Phi(X, Y, Z) );
    double cv[N_DIM*2]={X,Y,Z,0,0,0};
    double force[N_DIM*2]={0,0,0,0,0,0};
    P->DiffEq(2*P->N_dim, 0, cv, force);
    return V2/r + force[static_cast<CRcircParam*>(params)->direction + P->N_dim];
}

// Utility function to find approximate L^2 of a circular orbit with given energy
double findLcirc2(const CPotential* potential, double E)
{
    gsl_function F;
    CRcircParam params;
    params.E=E;
    params.P=potential;
    F.function=&findRcirc;
    F.params=&params;
    // average over 2 or 3 coordinates
    double Lcirc2=0;
    unsigned int numAvg=0;
    for(unsigned int d=0; d<potential->N_dim; d++)
    {
        // find r such that v^2/r = F(r) and v^2/2+Phi(r)=E
        params.direction=d;
        double rupp=potential->findintersection( E, d==0?1.:0, d==1?1.:0, d==2?1.:0);
        double rlow=rupp/4;
        double rcirc=findRoot(&F, rlow, rupp, -1);
        if(rcirc>0) {
            Lcirc2 += 2*(E - potential->Phi( d==0?rcirc:0, d==1?rcirc:0, d==2?rcirc:0))*pow_2(rcirc);
            numAvg++;
        }
    }
    if(numAvg) Lcirc2/=numAvg;
    return Lcirc2;
}
/// \endcond

// helper function for initializing missing timestep/timeunit in incomplete initial conditions data 
template<typename NumT> COrbitInitData<NumT> completeInitData(const COrbitInitData<NumT>& InitData)
{
    NumT timeStep=InitData.timeStep;
    NumT timeUnit=InitData.timeUnit;
    bool changed=false;
    if(timeUnit==0 && InitData.potential!=NULL)
    {
        changed=true;
        timeUnit = static_cast<NumT>(InitData.potential->longaxisperiod( InitData.potential->totalEnergy(InitData.initCond)));
    }
    if(timeStep==0)
    {
        changed=true;
        timeStep=timeUnit/10;  //< somewhat arbitrary number, but should work reasonably
    }
    if(changed)
        return COrbitInitData<NumT>(InitData.potential, timeStep, timeUnit, InitData.initCond, InitData.calcLyapunov);
    else
        return InitData;
}
template COrbitInitData<double> completeInitData(const COrbitInitData<double>& InitData);
template COrbitInitData<float > completeInitData(const COrbitInitData<float >& InitData);

}