/** \file    orbitan.h
    \brief   Orbit analysis classes
    \author  EV
    \date    2009-2013

    This file defines smile::COrbit class which serves as an interface between orbit integrators
    (implemented in orbitint.h) and orbit analysis methods, implemented in various classes 
    derived from smile::CBasicOrbitRuntimeFnc, for instance, routines for spectrum analysis, 
    Poincare surface of section and other stuff.
    Each runtime function class has its own information object class derived from 
    smile::CBasicInformation, and a creator class derived from smile::CBasicRuntimeFncCreator.
*/

#pragma once
#include "common.h"
#include <complex>

namespace smile{

/** general parameters for orbit integration and classification */
struct CConfigOrbit
{
    // integration accuracy and parameters
    double accuracyRel;              ///< relative accuracy parameter to 8th order Runge-Kutta integrator
    double accuracyAbs;              ///< absolute accuracy parameter 
    double accTreecode;              ///< factor in leap-frog timestep selection for treecode potential
    bool treecodeSymmetrizeTimestep; ///< whether to use time-symmetrized leapfrog (Hut, Makino & McMillan 1995) - doubles the number of evaluations but better conserves energy
    int calcLyapunovMethod;          ///< 1 -- nearby orbit, 2 -- variational equation
    bool freqStrictOrder;            ///< whether to require leading frequencies to be sorted so that w_x <= w_y <= w_z  (with small tolerance described below)
    double minFreqX;                 ///< search for leading frequency in X coordinate which is not less than MIN_FREQ_X
    double minFreqYZ;                ///< search for leading freq. in Y and Z which are not less than MIN_FREQ_YZ * (leading frequency in X and Y correspondingly)
};

class CBasicOrbitIntegrator;  // forward declaration, defined in orbitint.h

/** Class for performing a single orbit integration.
    COrbit is an interface class that stores the orbit initial data, calls one of the available 
    orbit integrators (derived from CBasicOrbitIntegrator), and stores and provides access 
    to orbit runtime functions and their information objects 
    (e.g. trajectory and spectrum analysis, Poincare section, etc.)
*/
class COrbit
{
public:
    /** three stages of orbit integration */
enum ORBITSTATE{
    OS_INITIALIZED,
    OS_RUNNING,
    OS_DONE
};
    /** construct from given initial conditions and optionally list of runtime function creators */
    template<typename NumT> COrbit(const CConfigOrbit* _configOrbit, 
        const COrbitInitData<NumT> &_InitData, const vectorRuntimeFncCreators* creators=NULL);

    /** construct from an existing trajectory in a given potential (for instance, loaded from a file */
    template<typename NumT> COrbit(const CConfigOrbit* _configOrbit, const CPotential* _potential, 
        const double _timeStep, const double _timeUnit, const std::vector< CPosVelPoint<NumT> > &_traj);

    /** destructor deletes all associated runtime functions */
    ~COrbit();
    
    /** integrate orbit up to the given time; if any of runtime functions returns false 
        then continue until all of them return true or maxTime is reached; maxTime=0 means maxTime=time */
    void integrateToTime(double time, double maxTime=0);
    
    /** instantly terminate orbit integration */
    void halt();

    /** returns c-th coordinate interpolated at time t within currect timestep interval */
    double getInterpolatedTrajectory(unsigned int c, double t) const;

    // get functions
    /** return current state (unstarted/running/done) */
    ORBITSTATE getState() const { return state; };

    /** if the orbit is running, return integration time to which it has already progressed, 
        otherwise return the total integration time if it is finished, or 0 if it hasn't started yet */
    double getIntTime() const;

    /** return initial conditions */
    const COrbitInitData<double>& getInitData() const { return InitData; };

    /** convenience function: return pointer to potential used in integration */
    const CPotential* getPotential() const { return InitData.potential; };

    /** return pointer to configuration parameters (may be needed by runtime functions) */
    const CConfigOrbit* getConfig() const { return configOrbit; };

    /** return pointer to the runtime function at the given index */
    const CBasicOrbitRuntimeFnc* getRuntimeFnc(size_t index) const { return runtimeFncs.at(index); };

    /** convenience function: create and return result information object for the given runtime function index */
    const CBasicInformation * getNewInfo(size_t index) const { return runtimeFncs.at(index)->getData(); };

    /** return the total number of runtime functions */
    size_t getInfoNum() const { return runtimeFncs.size(); };

    /** return some textual information about the orbit, optionally with contributions 
        from all runtime function information objects */
    std::string toString(bool withRuntimeFncInfo=true) const;

private:
    const CConfigOrbit* configOrbit; //< pointer to the configuration structure keeping various integration and analysis parameters
    const COrbitInitData<double> InitData;  //< initial conditions and parameters like orbit period
    volatile ORBITSTATE state;       //< current state of affairs
    CBasicOrbitIntegrator* orbInt;   //< pointer to orbit integrator if it is running, otherwise NULL
    double intTime;                  //< integration time - 0 if it hasn't started yet, or the actual duration when finished
    vectorRuntimeFncs runtimeFncs;   //< holds all attached 'timestep functions' (if any). They are destroyed along with COrbit!
};

/* -------------- Spectrum analysis data and routines --------------- */

/** Definition of one spectral line in one-dimensional time series (nothing to do with spectroscopy!:) */
struct CSpectralLine
{
    double freq;  ///< frequency
    double ampl;  ///< amplitude
    double phase; ///< phase (complex amplitude = ampl * exp(I*phase)
    int method;   ///< 0: Carpintero&Aguilar, 1: Hunter (more precise); other variants are not implemented yet
    CSpectralLine(double _freq=0, double _ampl=0, double _phase=0, int _method=0) : 
      freq(_freq), ampl(_ampl), phase(_phase), method(_method) {}
    inline bool operator<(CSpectralLine S) const
    {   return (ampl>S.ampl); }    ///< sort in reverse order of amplitude
    inline operator double() const { return freq; }
};
/** An array of spectral lines in one-dimensional time series */
typedef std::vector<CSpectralLine> CSpectrum;
typedef std::complex<double> complexd;   ///< convenience shorthand for a complex number

/** Basic class for various methods of spectral analysis */
class CBasicSpectrumAnalyzer {
public:
    virtual ~CBasicSpectrumAnalyzer() {};
    /** Frequency analysis routine: perform Fourier transform and find up to Nmax lines in the spectrum.
        input is provided by the constructor in derived classes 
        (it may consist of one-dimensional real array of x(t) for the Hunter's method, 
        or a complex array of { x(t),v(t) } for the Laskar's method).
        \param[out]  spectralLines array (must be a pointer to an existing vector); 
        \param[out]  spectrumAmplitudes (absolute values of complex spectrum) if the pointer to this vector is not NULL
    */
    virtual void findSpectralLines(CSpectrum *spectralLines, vectord *spectrumAmplitudes=NULL) = 0;
};

/** Spectrum analysis based on Hunter's DFT method.
    It determines the location of spectral lines with a high accuracy, using Hanning filtering;
    the backup method of Carpintero&Aguilar without filtering is used in case of a poorly resolved line
*/
class CSpectrumAnalyzerHunter: public CBasicSpectrumAnalyzer {
public:
    /** Constructor takes the one-dimensional array of equally spaced x(t), and the timestep */
    CSpectrumAnalyzerHunter(const vectord &_timeSeries, const double _timeStep) : 
      timeSeries(_timeSeries), timeStep(_timeStep) {};
    virtual void findSpectralLines(CSpectrum *spectralLines, vectord *spectrumAmplitudes=NULL);
private:
    const vectord &timeSeries;   ///< reference to the source array of x(t)
    const double timeStep;       ///< spacing between elements of the array (used to scale the frequencies)
};

/* ------------- Definition of various orbit information classes ---------- */

/** information object containing orbit analysis results.
    It represents some general data about the orbit, like the textual description of 
    orbit class, frequencies and chaotic properties, etc. 
*/
template<typename NumT> class COrbitInformation: public CBasicInformation
{
public:
    explicit COrbitInformation(const std::string &_Description) :
      Description(_Description), Einit(0), Ediff(0), lfccdiff(0), lambda(0) 
      { for(unsigned int d=0; d<N_DIM; d++) { lf[d]=0; inertia[d]=0; Lavg[d]=0; Lvar[d]=0; } };
    COrbitInformation(const std::string &_Description, const NumT _Einit, const NumT _Ediff, const NumT _lf[N_DIM], const NumT _lfccdiff, const NumT _lambda, const NumT _inertia[N_DIM], const NumT _Lavg[N_DIM], const NumT _Lvar[N_DIM]) :
      Description(_Description), Einit(_Einit), Ediff(_Ediff), lfccdiff(_lfccdiff), lambda(_lambda) 
      { for(unsigned int d=0; d<N_DIM; d++) { lf[d]=_lf[d]; inertia[d]=_inertia[d]; Lavg[d]=_Lavg[d]; Lvar[d]=_Lvar[d]; } };
    COrbitInformation(const COrbitInformation<float> &src) :
      Description(src.Description), Einit(static_cast<NumT>(src.Einit)), Ediff(static_cast<NumT>(src.Ediff)), lfccdiff(static_cast<NumT>(src.lfccdiff)), lambda(static_cast<NumT>(src.lambda)) 
      { for(unsigned int d=0; d<N_DIM; d++) { lf[d]=static_cast<NumT>(src.lf[d]); inertia[d]=static_cast<NumT>(src.inertia[d]); Lavg[d]=static_cast<NumT>(src.Lavg[d]); Lvar[d]=static_cast<NumT>(src.Lvar[d]); } };
    COrbitInformation(const COrbitInformation<double> &src) :
      Description(src.Description), Einit(static_cast<NumT>(src.Einit)), Ediff(static_cast<NumT>(src.Ediff)), lfccdiff(static_cast<NumT>(src.lfccdiff)), lambda(static_cast<NumT>(src.lambda)) 
      { for(unsigned int d=0; d<N_DIM; d++) { lf[d]=static_cast<NumT>(src.lf[d]); inertia[d]=static_cast<NumT>(src.inertia[d]); Lavg[d]=static_cast<NumT>(src.Lavg[d]); Lvar[d]=static_cast<NumT>(src.Lvar[d]); } };
    virtual INFOTYPE infoType() const { return sizeof(NumT)==sizeof(float) ? IT_TRAJ_ANALYSIS : IT_UNKNOWN; };  // streams support only float instantiations of this class
    virtual std::string toString() const;
    // get functions
    std::string getDescription() const { return Description; }
    NumT getEinit() const { return Einit; }
    NumT getEdiff() const { return Ediff; }
    NumT getlf(unsigned int index) const { if(index<N_DIM) return lf[index]; else return -1; }
    NumT getlfccdiff() const { return lfccdiff; }
    NumT getlambda() const { return lambda; }
    NumT getinertia(unsigned int index) const { if(index<N_DIM) return inertia[index]; else return -1; }
    NumT getLavg(unsigned int index) const { if(index<N_DIM) return Lavg[index]; else return -1; }
    NumT getLvar(unsigned int index) const { if(index<N_DIM) return Lvar[index]; else return -1; }
private:
    std::string Description;     ///< description of orbit type
    NumT Einit, Ediff;           ///< initial energy and energy conservation error after integration
    NumT lf[N_DIM];              ///< leading frequencies
    NumT lfccdiff;               ///< rate of LFCC diffusion
    NumT lambda;                 ///< Lyapunov exponent
    NumT inertia[N_DIM];         ///< inertia tensor components (only diagonal in Cartesian coordinates)
    NumT Lavg[N_DIM];            ///< average components of angular momentum
    NumT Lvar[N_DIM];            ///< variation in angular momentum components
};

/// information object for pericenter data (distribution of angular momentum at pericenter passages)
template<typename NumT> class CPericenterInformation: public CBasicInformation
{
public:
    CPericenterInformation(const NumT _fitIntercept, const NumT _fitSlope, const NumT _fitScatter, const NumT _fitSignificance, const NumT _Lcirc2) : 
      fitIntercept(_fitIntercept), fitSlope(_fitSlope), fitScatter(_fitScatter), fitSignificance(_fitSignificance), Lcirc2(_Lcirc2) {};
    virtual INFOTYPE infoType() const { return sizeof(NumT)==sizeof(float) ? IT_PERICENTER : IT_UNKNOWN; };
    virtual std::string toString() const;
    /// get data (all output parameters must point to existing variables which are filled with the values of parameters)
    void getParams(NumT* _fitIntercept, NumT* _fitSlope, NumT* _fitScatter, NumT* _fitSignificance, NumT* _Lcirc2) const 
      { *_fitIntercept=fitIntercept; *_fitSlope=fitSlope; *_fitScatter=fitScatter; *_fitSignificance=fitSignificance; *_Lcirc2=Lcirc2; };
private:
    /// \name fitting coefficients for distribution of squared angular momentum at pericenter
    NumT fitIntercept;    ///< L_{min}^2 - estimated lower bound on angular momentum at pericenter passage (0 means a centrophilic orbit)
    NumT fitSlope;        ///< slope of the regression  L^2_{k} = L^2_{min} + slope*k/N_{peri}  for k=1..N_{peri}
    NumT fitScatter;      ///< scatter of points around the best-fit regression, normalized to the `typical' value (should generally be <0.5 for the regressionto be reliable)
    NumT fitSignificance; ///< difference of L_{min}^2 from zero, measured in `sigma's (if less than a few, probably the orbit is centrophilic)
    NumT Lcirc2;          ///< approximate squared angular momentum of a circular orbit
};

/// information object keeping a trajectory sample (to generate an N-body model from a Schwarzschild model)
template<typename NumT> class CTrajSampleInformation: public CBasicInformation
{
public:
    explicit CTrajSampleInformation(const std::vector< CPosVelPoint<NumT> > &src) : trajSample(src) {};
    virtual INFOTYPE infoType() const { return sizeof(NumT)==sizeof(float) ? IT_TRAJ_SAMPLE : IT_UNKNOWN; };
    virtual std::string toString() const;
    /// get function: return the const reference to the entire trajectory
    const CPosVelDataFloat& getTraj() const { return trajSample; }
private:
    CPosVelDataFloat trajSample;   //< array of unsorted sampling points from the trajectory
};

/* --------- several orbit runtime functions and their associated creator classes ------------- */

/** Class for recording orbit trajectory and doing various analysis (frequency, Lyapunov exponent, etc).
    It records the trajectory at equal intervals of time (specified by timeStep), 
    and upon completion performs frequency analysis and extracts other information, 
    like the orbit class, chaotic properties, etc. The trajectory itself 
    is used in computing these properties, but not returned in the COrbitInformation class.
*/
class COrbitRuntimeTrajectory: public CBasicOrbitRuntimeFnc
{
public:
    /** constructor for an empty orbit */
    COrbitRuntimeTrajectory(const COrbit* _orbit);
    /** constructor for an existing trajectory */
    template<typename NumT> COrbitRuntimeTrajectory(const COrbit* _orbit, const std::vector< CPosVelPoint<NumT> > &_traj);
    virtual FNCTYPE FncType() const {return FT_TRAJ_ANALYSIS; };
    virtual bool Timestep(const double told, const double t, const double y[]);
    virtual CBasicInformation* getData() const  { return analyzeOrbit<float>(0, 0); }

    /** extract frequencies and perform orbit classification.
        by default operate on the entire whole orbit, but may use part of orbit as well. */
    template<typename NumT> COrbitInformation<NumT>* analyzeOrbit(
        size_t start, size_t count,  ///< [in] specifies the part of orbit to be analyzed; 0 means entire orbit
        CSpectrum *spec_lines=NULL,  ///< [out] list of spectral lines found for each coordinate; if NULL then no output is given
        vectord *spec_ampl=NULL      ///< [out] entire spectrum (amplitudes only) for each coordinate; if NULL then no output is given
    ) const;                         ///< \return a COrbitInformation object with all computed parameters
    // get functions
    /// returns the number of points in recorded trajectory
    size_t getTrajSize() const { return traj.size(); };
    /// retrieve a point from trajectory
    const CPosVelPoint<double>& getTraj(size_t index) const { return traj[index]; }
    /// return the logarithm of magnitude of the deviation vector
    double getLnDevVec(size_t index) const { return lnw.at(index); }
    /// return max distance between adjacent points in the given interval
    double getMaxDist(size_t start=0, size_t count=0) const;
private:
    const unsigned int N_dim;        ///< copied from orbit->potential->N_dim, for convenience
    const double timeStep, timeUnit; ///< copied from orbit->initData.timeStep,timeUnit
    const bool calcLyapunov;         ///< copied from orbit->initData.calcLyapunov

    CPosVelDataDouble traj;          ///< stores entire trajectory
    vectord lnw;                     ///< stores ln(|w|) - norm of deviation vector, lambda - finite-time Lyapunov exponent
    double Einit, Ediff;             ///< initial energy and energy conservation error by the end of integration

    /// estimate Lyapunov exponent
    double calcLambda() const;

    /// perform orbit classification based on its spectrum and angular momentum
    std::string performOrbitAnalysis(
        size_t start, size_t count,  ///< [in] specifies the part of orbit to be analyzed
        CSpectrum spec_lines[N_DIM], ///< [out] list of spectral lines found for each coordinate
        vectord spec_ampl[N_DIM],    ///< [out] entire spectrum (amplitudes only) for each coordinate; if NULL then no output is given
        double Lavg[N_DIM],          ///< [out] average values of angular momentum for each coordinate
        double Lvar[N_DIM]           ///< [out] std.dev. of angular momentum in each coord
    ) const;                         ///< \return a string with the description of orbit class

    /** finds spectral lines in each coordinate.
        uses CBasicSpectrumAnalyzer to find all lines, then sort them in appropriate order 
        taking into account axis ordering and base frequency of a long-axis orbit. 
        \param[out] spec_lines is the array of sorted spectral lines; 
        \param[out] spec_ampl contains the amplitudes of entire spectrum if not NULL. */
    void findLines(size_t start, size_t count, CSpectrum sl[N_DIM], vectord fa[N_DIM]) const;

    /** calculate LFCCs over a part of orbit or the entire orbit.
        \param[out] leadfreq is the array of leading frequencies in Cartesian coordinates. */
    void calcLFCCoverInterval(size_t start, size_t count, double leadfreq[N_DIM]) const;

    /** evaluate the rate of diffusion of leading frequencies over a given interval (whole orbit by default) */
    double calcLFCCdiffusion(size_t start, size_t count) const;

    /** compute the inertia tensor of the orbit or part of it.
        \param[out] inertia is the array of xx,yy and zz components of the inertia tensor, 
        roughly corresponds to the average extent of the orbit in a given direction. */
    void calcInertiaOverInterval(size_t start, size_t count, double inertia[N_DIM]) const;
};

/// corresponding creator class for Trajectory analysis timestep fnc
class COrbitRuntimeTrajectoryCreator: public CBasicOrbitRuntimeFncCreator
{
public:
    virtual CBasicOrbitRuntimeFnc* createRuntimeFnc(const COrbit* orbit) const 
    { return new COrbitRuntimeTrajectory(orbit); };
};

/** Trajectory sampler (record a given number of unsorted points from the trajectory, 
    to be used in creating an Nbody model from the Schwarzschild model).
*/
class COrbitRuntimeTrajSample: public CBasicOrbitRuntimeFnc
{
public:
    COrbitRuntimeTrajSample(const COrbit* _orbit, size_t _numSamplingPoints, double _timeStep);
    virtual FNCTYPE FncType() const {return FT_TRAJ_SAMPLE; };
    virtual bool Timestep(const double told, const double t, const double y[]);  ///< record trajectory at regular intervals
    virtual void Finish();   ///< draw a necessary number of sampling points from the entire trajectory
    virtual CBasicInformation* getData() const { return new CTrajSampleInformation<float>(trajSample); };
private:
    size_t numSamplingPoints;      ///< required number of sampling points at the end of integration
    const double timeStep;         ///< time interval for recording the trajectory (must be less than duration of integration divided by numSamplingPoints)
    const unsigned int N_dim;      ///< dimensions of potential (stored for convenience)
    CPosVelDataFloat trajEntire;   ///< recorded trajectory
    CPosVelDataFloat trajSample;   ///< points sampled from the trajectory (unordered)
};

/** corresponding creator class for trajectory sampler.
    In the Schwarzschild model, orbits that have a high weight may need more sampling points 
    to represent them in the Nbody model. Consequently, if the number of sampling points 
    for an orbit was insufficient, it will be re-integrated before exporting to Nbody model, 
    with an increased number of sampling points. 

    This creator class takes care of these special orbits:
    it requests a default number of sampling points for all orbits except those ones 
    which initial conditions are listed in the numSamplingPointsSpecial array;
    these special points are requested to provide more samples.
    intTimeMin is the (minimal) length of integration time (in periods), 
    it is necessary to compute the interval between sampling points used by COrbitRuntimeTrajSample,
    since the runtime function does not know the total duration of the integration in advance.
*/
class COrbitRuntimeTrajSampleCreator: public CBasicOrbitRuntimeFncCreator
{
public:
    COrbitRuntimeTrajSampleCreator(unsigned int _numSamplingPointsDefault, double _intTimeMin, const CPointCountSetFloat* _numSamplingPointsSpecial);
    virtual CBasicOrbitRuntimeFnc* createRuntimeFnc(const COrbit* orbit) const;
private:
    const unsigned int numSamplingPointsDefault;  ///< standard number of sampling points for all orbits except for the special list
    const double intTimeMin;                      ///< (minimum) integration time in periods, needed to compute timestep
    CPointCountSetFloat numSamplingPointsSpecial; ///< list of special orbits, for which the number of sampling points needs to be larger
};

/** compute the Poincare surface of section for an orbit.
    When coord2=0 and coord2'>0, record a pair (coord1, coord1').
    The Poincare surface of section is used only for visualization in the GUI;
    no information class is stored along with the orbit.
*/
class COrbitRuntimePoincare: public CBasicOrbitRuntimeFnc
{
public:
    COrbitRuntimePoincare(const COrbit* _orbit, unsigned int _PS1, unsigned int _PS2): CBasicOrbitRuntimeFnc(_orbit), PS1(_PS1), PS2(_PS2) {};
    virtual FNCTYPE FncType() const {return FT_POINCARE; };
    virtual bool Timestep(const double told, const double t, const double y[]);
    virtual CBasicInformation* getData() const { return NULL; };
    unsigned int size() const { return static_cast<unsigned int>(ps.size()); }
    const std::pair<double, double>& at(unsigned int index) const { return ps.at(index); }
private:
    const unsigned int PS1, PS2;                  ///< index of coordinates to use in Poincare section
    std::vector< std::pair<double, double> > ps;  ///< Poincare section  (x - xdot plane)
};

/// corresponding creator class for Poincare timestep fnc
class COrbitRuntimePoincareCreator: public CBasicOrbitRuntimeFncCreator
{
public:
    COrbitRuntimePoincareCreator(unsigned int _PS1, unsigned int _PS2) : PS1(_PS1), PS2(_PS2) {};
    virtual CBasicOrbitRuntimeFnc* createRuntimeFnc(const COrbit* orbit) const { return new COrbitRuntimePoincare(orbit, PS1, PS2); };
private:
    const unsigned int PS1, PS2;
};
        
/// class for recording pericenter passages and analyzing their distribution 
/// to figure out if the orbit is centrophilic or not
class COrbitRuntimePericenter: public CBasicOrbitRuntimeFnc
{
public:
    /// create the runtime function for an orbit to be integrated later
    COrbitRuntimePericenter(const COrbit* _orbit);
    /// create the runtime function for an orbit with already existing trajectory
    template<typename NumT> COrbitRuntimePericenter(const COrbit* _orbit, const std::vector< CPosVelPoint<NumT> > &_traj);
    virtual FNCTYPE FncType() const {return FT_PERICENTER; };
    virtual bool Timestep(const double told, const double t, const double y[]);  ///< records pericenter passages
    virtual void Finish();   ///< fit pericenter distribution and compute fitting coefficients
    virtual CBasicInformation* getData() const { return new CPericenterInformation<float>((float)fitIntercept, (float)fitSlope, (float)fitScatter, (float)fitSignificance, (float)Lcirc2); };
    size_t size() const { return pericentert.size(); };
    double gett(size_t index) const { return pericentert[index]; };
    double getr(size_t index) const { return pericenterr[index]; };
    double getl2(size_t index)const { return pericenterl2[index]; };
private:
    int drdtsign;                      ///< whether approaching or receding from center on the previous timestep
    vectord pericentert, pericenterr;  ///< times and radii of consequent pericenter passages
    vectord pericenterl2;              ///< L^2 - squared magnitude of angular momentum
    double pericenterrmin;             ///< minimum pericenter radius reached so far
    double fitIntercept, fitSlope;     ///< fitting coefficients for L2 distribution
    double fitScatter;                 ///< scatter in linear fit (less=better, >0.5-poor fit) 
    double fitSignificance;            ///< significance of non-zero value of intercept (in sigmas)
    double Lcirc2;                     ///< approximate squared angular momentum of a circular orbit with this energy
    double Lavg[N_DIM], L2avg[N_DIM];  ///< accumulator for average value of i-th component of angular momentum and its square (to determine if there is a conserved component)
};

/// corresponding creator class for the pericenter analysis runtime function
class COrbitRuntimePericenterCreator: public CBasicOrbitRuntimeFncCreator
{
public:
    COrbitRuntimePericenterCreator() {};
    virtual CBasicOrbitRuntimeFnc* createRuntimeFnc(const COrbit* orbit) const { return new COrbitRuntimePericenter(orbit); };
};

/* ---------- several non-class utility routines ---------- */

/** Utility function to find a 1:1 periodic orbit in a plane specified by coord1:coord2 */
bool findPeriodicOrbit(const CPotential* potential, double E, const CConfigOrbit* configOrbit, int coord1, int coord2, double& cross1);    // find 1:1 periodic orbit at given energy in coord1-coord2 plane, return starting position in 1st coordinate

/** Utility function to find approximate L^2 of a circular orbit with given energy */
double findLcirc2(const CPotential* potential, double E);

/** helper function for initializing missing timestep/timeunit in incomplete initial conditions data */
template<typename NumT> COrbitInitData<NumT> completeInitData(const COrbitInitData<NumT>& InitData);

}  // namespace