/** \file    orbitint.h
    \brief   Orbit integration classes
    \author  EV
    \date    2008-2013

    Classes for orbit integration, derived from smile::CBasicOrbitIntegrator, 
    are declared in this module.
    The integrators call smile::CPotential::DiffEq for obtaining accelerations, 
    adaptively adjust timestep to maintain desired accuracy, and call 
    runtime functions derived from smile::CBasicOrbitRuntimeFnc after each timestep.

    The COrbitIntegratorDOP853 contains code from Hairer et al.
*/

#pragma once
#include "common.h"

namespace smile{

/** base class for orbit integrator */
class CBasicOrbitIntegrator
{
public:
    CBasicOrbitIntegrator(const CPotential* _potential, vectorRuntimeFncs *_runtimeFncs):
      potential(_potential), runtimeFncs(_runtimeFncs), needToTerminate(false), xout(0) {};

    virtual ~CBasicOrbitIntegrator() {};

    /** perform integration */
    virtual void integrateToTime(const CPosVelPoint<double> &initCond, double time, double maxTime=0)=0;

    /** called if it is necessary to stop integration immediately; raises internal termination flag */
    void halt() { needToTerminate=true; }

    /** return the time to which the integration has proceeded so far */
    double getTime() const { return xout; }

    /** return interpolated value of c-th variable at time t, where t must lie within current timestep interval;
        called from various runtime functions that need to obtain x(t) at arbitrary times */
    virtual double getInterpolatedTrajectory(unsigned int c, double t) const=0;

protected:
    const CPotential* potential;    //< pointer to the potential used in the integration
    vectorRuntimeFncs* runtimeFncs; //< array of runtime functions that are called on each timestep
    volatile bool needToTerminate;  //< a flag raised internally or externally if the integration should be finished immediately
    double xold, xout;              //< previous and current value of time to which the integration has advanced

    /** call each of the runtime functions at each timestep; 
        return true if all runtime functions can stop right now */
    bool solout(double told, double t, double y[]) {
        bool canFinish=true;
        for(vectorRuntimeFncs::iterator iter=runtimeFncs->begin(); iter!=runtimeFncs->end(); iter++)
            canFinish &= (*iter)->Timestep(told, t, y);
        return canFinish;
    }
};

/** 8th order Runge-Kutta integrator from Hairer,Norsett&Wanner.
    It is suitable for integrating orbits in a smooth potential (everything except the treecode potential).
    It can also compute Lyapunov exponent if necessary. */
class COrbitIntegratorDOP853: public CBasicOrbitIntegrator
{
public:
    COrbitIntegratorDOP853(const CPotential* _potential, vectorRuntimeFncs *_runtimeFncs, 
        int _calcLyapunovMethod, double _accAbs, double _accRel):
      CBasicOrbitIntegrator(_potential, _runtimeFncs), 
      calcLyapunovMethod(_calcLyapunovMethod), accAbs(_accAbs), accRel(_accRel) {};

    virtual void integrateToTime(const CPosVelPoint<double> &initCond, double time, double maxTime=0);

    /** return interpolated value of coordinate or velocity within the current timestep;
        if c==2*N_dim and calcLyapunovMethod>0, then return the logarithm of deviation vector ln(|w|)
        (to compute Lyapunov exponent). */
    virtual double getInterpolatedTrajectory(unsigned int c, double t) const;
private:
    /** whether and how to compute the evolution of the deviation vector used in estimating the Lyapunov exponent.
        0 - no dev.vec., 1 - integrate a nearby orbit, 2 - use variational equation */
    const int calcLyapunovMethod;
    unsigned int N_dim;       ///< copied from potential, for convenience
    double accAbs, accRel;    ///< relative and absolute accuracy for the integrator
    unsigned int nfcn,        ///< statistics: number of force evaluations
        nstep, naccpt, nrejct;///< statistics: number of timesteps (total, accepted, rejected)
    double lnwadd;            ///< addition term - required for renormalization so that |w| remains not too large
    /// \name dop853.c encapsulated variables and functions
    ///@{
    double   hout;            ///< current timestep
    unsigned int nrds;
    unsigned int* indir;
    double   *rcont1, *rcont2, *rcont3, *rcont4;
    double   *rcont5, *rcont6, *rcont7, *rcont8;
    ///@}

    /// rhs of the differential equation 
    /// (calls CPotential::DiffEq depending on the method of computing Lyapunov exponent)
    void fcn(unsigned int n, double x, double *y, double *f);

    /// driver routine of the DOP853 integrator
    int dop853(
      unsigned int n,      /**< dimension of the system <= UINT_MAX-1*/
      double x,            /**< initial x-value */
      double* y,           /**< initial values for y */
      double xend,         /**< final x-value (xend-x may be positive or negative), if not using adaptive choice of integration time */
      double xmaxTime,     /**< hard upper limit on integration time for the adaptive choice of it */
      double* rtoler,      /**< relative error tolerance */
      double* atoler,      /**< absolute error tolerance */
      int itoler,          /**< switch for rtoler and atoler */
      int iout,            /**< switch for calling solout */
      double uround,       /**< rounding unit */
      double safe,         /**< safety factor */
      double fac1,         /**< parameters for step size selection */
      double fac2,        
      double beta,         /**< for stabilized step size control */
      double hmax,         /**< maximal step size */
      double h,            /**< initial step size */
      long nmax,           /**< maximal number of allowed steps */
      int meth,            /**< switch for the choice of the coefficients */
      long nstiff,         /**< test for stiffness */
      unsigned int nrdens, /**< number of components for which dense outpout is required */
      unsigned int* icont, /**< indexes of components for which dense output is required, >= nrdens */
      unsigned int licont  /**< declared length of icon */
    );

    /// internal integration initialization routine
    double hinit (unsigned int n, double x, double* y,
      double posneg, double* f0, double* f1, double* yy1, int iord,
      double hmax, double* atoler, double* rtoler, int itoler);
    
    /// internal integration routine (performs actual computation)
    int dopcor (unsigned int n, double x, double* y, double xend, double xmaxTime, 
      double hmax, double h, double* rtoler, double* atoler, int itoler, int iout,
      unsigned int nmax, double uround, int meth, long nstiff, double safe,
      double beta, double fac1, double fac2, unsigned* icont);

    /// computes interpolated coordinate at time t within current timestep
    double contd8(
      unsigned int ii, /**< index of desired variable */
      double t         /**< time to compute the variable */
    ) const;
};

/** low-order leapfrog method provided for treecode, which has discontinuous potential/forces 
    and is poorly handled by Runge-Kutta integrator. It does not compute Lyapunov exponent either. */
class COrbitIntegratorLeapfrog: public CBasicOrbitIntegrator
{
public:
    COrbitIntegratorLeapfrog(const CPotential* _potential, vectorRuntimeFncs *_runtimeFncs, 
        double _accTreecode, bool _symmetrizeTimestep):
      CBasicOrbitIntegrator(_potential, _runtimeFncs), 
      accTreecode(_accTreecode), symmetrizeTimestep(_symmetrizeTimestep) {};
    virtual void integrateToTime(const CPosVelPoint<double> &initCond, double time, double maxTime=0);
    virtual double getInterpolatedTrajectory(unsigned int c, double t) const;
private:
    double accTreecode;        ///< accuracy parameter (timestep is accTreecode*timeToCollision)
    bool symmetrizeTimestep;   ///< whether to use Hut et al.(1995) timestep symmetrization algorithm
    double yprev[4*N_DIM], aprev[2*N_DIM], anew[N_DIM];
};

};  // namespace