/** \file    orbitlib.h
    \brief   Classes for handling the orbit library and creating an Nbody model
    \author  EV
    \date    2009-2013

    This file defines classes for managing the orbit library, i.e. a collection of 
    orbit initial conditions and results of integration and analysis, used for 
    creating a frequency map or a Schwarzschild model. 

    The main class, smile::COrbitLibrary, implements methods for generating initial 
    conditions, and for creating an Nbody model out of the orbit library.
    An interface class smile::COrbitDesc is a container for the initial data and 
    results of integration of a single orbit.
    A multimass scheme for Nbody models is implemented by a couple of classes 
    for mass refinement.

*/

#pragma once
#include "common.h"

namespace smile{

struct CConfigOrbit;  // forward declaration

/** Container for orbit initial conditions and information objects.
    This class serves as an interface between a single COrbit and the entire COrbitLibrary.
    It keeps the initial data for an orbit, and internally creates an instance 
    of COrbit for the integration and analysis; then it records the results in the form 
    of an array of CBasicInformation and deletes the COrbit to free up memory.

    A collection of this objects constitutes the COrbitLibrary.
*/
class COrbitDesc
{
public:
    /// lists all possible states of orbit integration
enum ORBITSTATE{
    OS_INITIALIZED,      ///< the object is created but only contains initial data
    OS_PREPARING,        ///< the run() method is called
    OS_RUNNING,          ///< the internal COrbit::run() is in progress
    OS_NEEDTOTERMINATE,  ///< the user requested to stop orbit integration
    OS_DONE              ///< integration completed, all information objects are stored (also set when all data is loaded from a file).
};
    /// Create the object for the given initial conditions, without the results of orbit integration (OS_INITIALIZED)
    template<typename NumT> COrbitDesc(const CConfigOrbit* _configOrbit, 
        const COrbitInitData<NumT> &_InitData, const NumT _intTime, const NumT _maxTime, 
        const vectorRuntimeFncCreators* _creators);

    /// Create the object containing both initial data and the results of integration (OS_DONE)
    COrbitDesc(const CConfigOrbit* _configOrbit, 
        const COrbitInitData<float> &_InitData, const float _intTime, const float _maxTime, 
        const float _weight, const vectorInformation& _info);
    ~COrbitDesc();       ///< destroy all information containers
    void run();          ///< create an internal COrbit instance and run orbit integration and analysis
    void halt();         ///< if running, immediately stop computation (set the state to OS_NEEDTOTERMINATE)
    // get/set functions
    /// return the initial data for the orbit
    const COrbitInitData<float>& getInitData() const { return InitData; };
    /// return the integration time, whether the actual one if orbit is done, or the intended one if only initialized
    float getIntTime() const { return intTime; };
    /// return the maximum time for adaptive integration
    float getMaxTime() const { return maxTime; };
    /// return the number of CBasicInformation objects
    size_t getInfoNum() const { return info.size(); };
    /// return a pointer to the given information object (a private data member!)
    const CBasicInformation* getInfo(size_t index) const { return info.at(index); };
    /// convenience function to get a pointer to information of a given type, or NULL if it doesn't exist
    const CBasicInformation* getInfoByType(CBasicInformation::INFOTYPE InfoType) const {
        for(size_t i=0; i<info.size(); i++)
            if(info[i]->infoType()==InfoType) return info[i];
        return NULL; };
    /// returns string representation of all information objects associated with the orbit, 
    /// plus the weight and integration time
    std::string toString() const;
    /// return orbit weight in the Schwarzschild model
    double getWeight() const { return orbitWeight; };
    /// assign the orbit weight
    void setWeight(double _weight) { orbitWeight=_weight; }
    /// return the state of orbit integration (init/running/done)
    ORBITSTATE getState() const { return state; };
    /// set the state to OS_PREPARING if it is going to be started by the orbit library class,
    /// or reset it to OS_INITIALIZED if it was completed.
    void setState(ORBITSTATE _state) 
    { if((state==OS_INITIALIZED && _state==OS_PREPARING)||(state==OS_DONE && _state==OS_INITIALIZED)) state=_state; };
private:
    const CConfigOrbit* configOrbit;          ///< pointer to orbit integration/analysis parameters
    const COrbitInitData<float> InitData;     ///< orbit initial conditions and other data
    COrbit* orbit;                            ///< pointer to the internal object created in run()
    float intTime, maxTime;                   ///< integration and max integration time
    const vectorRuntimeFncCreators* creators; ///< pointer to CBasicOrbitRuntimeFncCreator classes collection
    vectorInformation info;                   ///< array of CBasicOrbitInformation objects
    double orbitWeight;                       ///< weight of the orbit in the Schwarzschild model
    volatile ORBITSTATE state;                ///< current state of affairs
};

/** The collection of orbits.
    Contains the array of COrbitDesc objects, methods to generate initial conditions 
    for the orbit library, to get the statistical information about the orbit ensemble, 
    and to create an Nbody model from the orbit library.
*/
class COrbitLibrary
{
public:
    /// Construct an empty orbit library with the given potential and array of runtime function creators.
    COrbitLibrary(const CConfigOrbit* _configOrbit, const CPotential* _potential, const vectorRuntimeFncCreators* _creators) : 
      configOrbit(_configOrbit), potential(_potential), creators(_creators) {};
    /// Destroy the orbit library and delete all orbit data
    ~COrbitLibrary();
    /// replace the library by an array of initial conditions for the given energy (for the frequency map)
    void prepareInitialConditionsE(
        size_t NStationary,           ///< number of orbits in the stationary start-space (approximate)
        size_t NPrincipalPlane,       ///< number of orbits in the principal-plane start-space (approx)
        size_t NYalpha,               ///< number of orbits in the Y-alpha start space (approx)
        size_t NRandom,               ///< number of orbits with random location
        double E,                     ///< the energy of each orbit
        double intTimeStepsPerPeriod, ///< number of output timesteps per one dynamical time
        double intTimeInPeriods,      ///< integration time in units of dynamical time
        double intTimeMaxAdaptive,    ///< maximal integration time (if 0, equal to the previous)
        bool calcLyapunov             ///< whether to compute Lyapunov exponent for each orbit
    );
    /// replace the library by initial conditions for the entire range of energies (for Schwarzschild modelling)
    void prepareInitialConditions(
        size_t NRandom,               ///< number of orbits in the model
        double maxRadius,             ///< max energy of the orbit is given by Phi(maxRadius), if it is 0 then maxEnergy=0
        double intTimeStepsPerPeriod, ///< number of output timesteps per one dynamical time
        double intTimeInPeriods,      ///< integration time in units of dynamical time
        double intTimeMaxAdaptive,    ///< maximal integration time (if 0, equal to the previous)
        bool calcLyapunov             ///< whether to compute Lyapunov exponent for each orbit
    );
    /// modify the potential and integration time/timestep for each orbit in an existing orbit library
    void modifyInitialConditions(
        const CPotential* _potential, ///< pointer to the new potential
        const vectorRuntimeFncCreators* _creators, ///< pointer to the new runtime function creators array
        double intTimeStepsPerPeriod, ///< new number of output timesteps per one dynamical time
        double intTimeInPeriods,      ///< new integration time in units of dynamical time
        double intTimeMaxAdaptive,    ///< new maximal integration time (if 0, equal to the previous)
        bool calcLyapunov,            ///< whether to compute Lyapunov exponent for each orbit
        bool modifyAll                ///< if true, replace all of the above data; if false, replace only potential and creators
    );
    /// if integration of the whole library was aborted by user, delete unfinished orbits
    void removeUnused();
    /// return statistics about the orbit library.
    std::string getOrbitPopulation(
        const CBasicOrbitFilteringFnc* filter, ///< if not NULL, select only orbits for which this function returns nonzero
        const CBasicOrbitFilteringFnc* chaos,  ///< if not NULL, provide information separately for regular and chaotic orbits determined by this function
        bool useWeight                         ///< whether to use orbit weights in computing the fraction of orbits, or use equal weighting
    ) const;  ///< \return a string containing orbit population (5 most abundant orbit families with their combined number, possibly weighted).
    double totalMass() const;                  ///< returns sum of all orbit weights (if they were assigned)
    /** try to create Nbody model from orbit library (with trajectory sample data).
        If number of sampling points was not enough for some orbits, put duplicate points 
        in the output for orbits with insufficient sampling, mark those orbits as unstarted 
        and return list of orbits for reintegration.
        \param[in] numPoints  is number of points to be created in the Nbody snapshot;
        \param[in] massRefineFactor  may be used to generate a multimass model: 
                   the mass of points sampled from a given orbit is proportional to 
                   the number returned by this function for that orbit), 
                   if this parameter is NULL then just assign equal mass to all points.
        \param[in,out] result  is a pointer to the point mass set that is created in this routine;
                   if this parameter points to a NULL pointer then a new CPointMassSet* is allocated, 
                   otherwise the existing snapshot is replaced with the new points;
        \param[in,out] numSamplingPointsMin  is a pointer to an associative array which will 
                   contain the initial conditions for those orbits that need to be reintegrated, 
                   and their minimum required number of sampling points;
                   if this parameter points to an existing pointer then it is replaced by the new 
                   list of orbits for reintegration, otherwise a new array is allocated if needed.
        \return some statistics about number of orbits and points.
    **/
    std::string exportNbody(
        size_t numPoints,
        const CBasicOrbitFilteringFnc* massRefineFactor,
        CPointMassSetFloat* *result,
        CPointCountSetFloat* *numSamplingPointsMin
        ) const;  
    // get functions
    /// return the pointer to the potential used in orbit integration
    const CPotential* getPotential() const { return potential; };
    /// return total number of orbits in the library (both completed and unstarted)
    size_t size() const { return OrbitList.size(); };
    /// return a pointer to a particular orbit descriptor
    const COrbitDesc* getOrbitDesc(size_t index) const { return OrbitList.at(index); };
    /// find how many orbits are already finished
    size_t numComplete() const;
    /** find an orbit that hasn't been started yet.
        search for the first orbit with state=OS_INITIALIZED and set it to OS_PREPARING.
        Should be called with external mutex/lock in a multi-threaded environment.
        \return the index of this orbit or -1 if none was found.
    */
    int findUnstartedOrbit() const;
    // set functions
    /// append an orbit descriptor to the library (used in loading from a file)
    void addOrbitDesc(COrbitDesc* const &od) { OrbitList.push_back(od); };
    /// assign the orbit weight for the given orbit index
    void setOrbitWeight(size_t index, double weight) { OrbitList.at(index)->setWeight(weight); };
    /// integrate the orbit with the given index
    void runOrbit(size_t index) { OrbitList.at(index)->run(); };
    /// stop all orbit integrations (call halt for all orbit descriptors)
    void halt() {
        for(std::vector<COrbitDesc*>::iterator iter=OrbitList.begin(); iter!=OrbitList.end(); ++iter)
            (*iter)->halt();
    }
private:
    const CConfigOrbit* configOrbit;          ///< points to the common orbit integration parameter block (accuracy, etc.)
    const CPotential* potential;              ///< potential assigned for all orbits
    const vectorRuntimeFncCreators* creators; ///< array of runtime function creators
    std::vector<COrbitDesc*> OrbitList;       ///< array of orbit descriptors
};

/// A helper object for computing the virial ratio for a set of point masses
template<typename NumT> class CPointMassSetHandler
{
public:
    CPointMassSetHandler(const CPointMassSet<NumT>* _points) : points(_points), state(PS_IDLE) {};
    /// compute the kinetic and potential energy and return them in these two pointers
    void computeVirialRatio(double* T, double* W);
    /// stop the computation
    void halt() { if(state==PS_RUNNING) state=PS_NEEDTOTERMINATE; };
private:
    /// internal state of the object (created/running/done)
enum PSTATE{
    PS_IDLE,
    PS_RUNNING,
    PS_NEEDTOTERMINATE,
    PS_DONE
};
    const CPointMassSet<NumT>* points;  ///< pointer to the array of bodies
    volatile PSTATE state;              ///< computation state
};

/** A mass refinement function is used to create an Nbody snapshot with unequal mass particles. 
    It ranks orbits into several bins, based on another evaluation function passed as a parameter; 
    mass factors of adjacent bins differ by a factor of two, and each bin contains the number of 
    orbits that also differ by a factor of two.
    Used in creating multimass Nbody models: one may wish to have a better mass resolution for 
    the inner parts of the model, based on orbit energy, or pericenter distance, or anything else.
*/
class CMassRefinementFnc: public CBasicOrbitFilteringFnc
{
public:
    /// initialize bin boundaries by sorting the output of evalFnc for all orbits in the library
    CMassRefinementFnc(const COrbitLibrary* orbitlib, size_t numBins, const CBasicOrbitFilteringFnc* _evalFnc);
    /// return the bin into which the orbit falls, using the pre-initialized array of bins
    virtual double eval(const COrbitDesc* orbit) const;
private:
    /// the evaluation function; orbits are sorted according to the value returned by this function
    const CBasicOrbitFilteringFnc* rankFnc;
    /// divide the entire orbit library into numBins; the bin boundaries 
    /// in terms of evaluation function values are stored in this array
    vectord binBoundary;
};

/// an example of evaluation function for multimass Nbody models, which sorts orbits in total energy.
class CEnergyOrbitFilteringFnc: public CBasicOrbitFilteringFnc
{
public:
    virtual double eval(const COrbitDesc* orbit) const;   ///< return total energy of an orbit
};

}