#include "potential.h"
#include <cmath>
#include <algorithm>
#include <cassert>
#include <gsl/gsl_integration.h>
#include <gsl/gsl_math.h>
#include <gsl/gsl_errno.h>
#include <gsl/gsl_sf_legendre.h>
#include <gsl/gsl_sf_gegenbauer.h>
#include <gsl/gsl_sf_gamma.h>
#include <gsl/gsl_multifit.h>
#include "utils.h"
#include "massmodel.h"

#ifdef DEBUGPRINT
#include <iostream>
#endif

namespace smile{

//----------------------------------------------------------------------------//

/// \cond INTERNAL_DOCS
// helper functions for integrating density over volume
struct CDensityParam{
    const CDensity* P;
    double r, rcostheta, rsintheta;
};
double getDensityPhi(double phi, void* params)
{
    return ((CDensityParam*)params)->P->Rho(((CDensityParam*)params)->rsintheta*cos(phi), ((CDensityParam*)params)->rsintheta*sin(phi), ((CDensityParam*)params)->rcostheta);
}
double getDensityTheta(double theta, void* params)
{
    gsl_function F;
    F.function=&getDensityPhi;
    F.params=params;
    ((CDensityParam*)params)->rcostheta=((CDensityParam*)params)->r*cos(theta);
    ((CDensityParam*)params)->rsintheta=((CDensityParam*)params)->r*sin(theta);
    double result, error;
    size_t neval;
    gsl_integration_qng(&F, 0, M_PI/2, 0, EPSREL_DENSITY_INT, &result, &error, &neval);
    return result*sin(theta);
}
double getDensityR(double r, void* params)
{
    gsl_function F;
    F.function=&getDensityTheta;
    F.params=params;
    ((CDensityParam*)params)->r=r;
    double result, error;
    size_t neval;
    gsl_integration_qng(&F, 0, M_PI/2, 0, EPSREL_DENSITY_INT, &result, &error, &neval);
    return result*8*r*r;
}
/// \endcond

// base class for density-only model, without corresponding potential: CDensity

double CDensity::Mass(const double r) const
{
    if(r<=0) return 0;
    // by default, integrate over density inside given radius; may be replaced by cheaper and more approximate evaluation for child classes
    gsl_function F;
    CDensityParam params;
    F.function=&getDensityR;
    F.params=&params;
    params.P=this;
    double result, error;
    gsl_integration_workspace * w = gsl_integration_workspace_alloc(1000);
    gsl_integration_qag(&F, 0, r, 0, EPSREL_DENSITY_INT, 1000, GSL_INTEG_GAUSS51, w, &result, &error);
    gsl_integration_workspace_free(w);
    return result;
}

double CDensity::totalMass() const
{
    double rad=32;
    double mass1, mass2=Mass(rad), mass3=Mass(rad*2), massest=0, massestprev;
    int num_neg=0;
    const int max_num_neg=4;
    do{
        rad*=2;
        mass1=mass2; mass2=mass3; mass3=Mass(rad*2);
        if(mass2==mass3) return mass3;  // mass doesn't seem to grow with raduis anymore
        massestprev=massest>0 ? massest : mass3;
        massest=(mass2*mass2-mass1*mass3)/(2*mass2-mass1-mass3);
        if(gsl_isnan(massest) || massest<0) num_neg++;  // increase counter of 'bad' attempts (negative means that mass is growing at least logarithmically with radius)
    } while(rad<MAX_RADIUS_CUTOFF && mass2!=mass3 && (massest<0 || fabs(massestprev-massest)/massest>EPSREL_DENSITY_INT) && num_neg<max_num_neg);
    if(fabs((massestprev-massest)/massest)>EPSREL_DENSITY_INT) return -1;   // total mass seems to be infinite
    else return massest;
}

/// \cond INTERNAL_DOCS
double findRadiusByMass(double r, void* params)
{
    return ((CDensityParam*)params)->P->Mass(r) - ((CDensityParam*)params)->r;  // params->r holds mass fraction
}
/// \endcond
double CDensity::getRadiusByMass(const double m) const
{
    double totMass=totalMass();
    if(totMass<0 || m>=totMass) return -1;
    double Xlow=0, Xupp=32;
    while(Mass(Xupp)<m && Xupp<MAX_RADIUS_CUTOFF) Xupp*=2;
    if(Xupp>=MAX_RADIUS_CUTOFF) return MAX_RADIUS_CUTOFF; // error, but not a fatal one...
    gsl_function F;
    CDensityParam PP;
    PP.r=m;
    PP.P=this;
    F.function=&findRadiusByMass;
    F.params=&PP;
    return findRoot(&F, Xlow, Xupp, -1);
}

bool CDensity::checkMassMonotonic() const
{
    double massprev=Mass(0);
    if(gsl_isnan(massprev) || massprev!=0) return false;  // mass should be zero at origin (not including Mbh)
    for(double rad=0.01; rad<std::min(MAX_RADIUS_CUTOFF,10000.0); rad*=1.25)
    {
        double mass=Mass(rad);
        if(mass<massprev*(1-1e-10)) return false;
        massprev=mass;
    }
    return true;  // no apparent evidence for weird behaviour..
}

bool CDensity::checkDensityNonzero() const
{
    for(double rad=1e-4; rad<std::min(MAX_RADIUS_CUTOFF,10000.0); rad*=1.5)
    {
        if(Rho(rad, 0, 0) <=0 || Rho(0, rad, 0) <=0 || Rho(0, 0, rad) <=0)
            return false;
    }
    return true;  // no apparent evidence for weird behaviour..
}

double CDensity::getGamma() const
{
    double mass1, mass2, mass3;
    double rad=1./1024;
    do{
        mass2=Mass(rad);
        if(mass2<=0) rad*=2;
    }while(rad<1 && mass2==0);
    mass3=Mass(rad*2);
    if(mass2<=0 || mass3<=0) return -1; // apparent error
    double alpha1, alpha2=log(mass3/mass2)/log(2.), gamma1=-1, gamma2=3-alpha2;
    do{
        rad/=2;
        mass1=Mass(rad);
        if(mass1<=0) return gamma2;
        alpha1=log(mass2/mass1)/log(2.);
        gamma2=gamma1<0 ? 3-alpha1 : gamma1;  // rough estimate
        gamma1=3- (2*alpha1-alpha2); // extrapolated estimate
        alpha2=alpha1;
        mass3=mass2; mass2=mass1;
    } while(fabs(gamma1-gamma2)>0.001);
    if(fabs(gamma1)<1e-3) gamma1=0;
    return gamma1;
}
//----------------------------------------------------------------------------//
// base class for potential/density model: CPotential

/// \cond INTERNAL_DOCS
// helper functions for long axis radius/period evaluation
struct CPotentialParam1{
    const CPotential* P;
    double E;
};
struct CPotentialParam4{
    const CPotential* P;
    double E, X, Y, Z;
};

// general functions for any potential
double twoV2(double x, void* params)
{
    double E=((CPotentialParam1*)params)->E;
    return E-((CPotentialParam1*)params)->P->Phi(x,0,0) -0.5*pow_2(((CPotentialParam1*)params)->P->Omega*x);
}
double Torb(double x, void* params)
{
    if(x==0) return 0;
    double E=((CPotentialParam1*)params)->E;
    double V2=2*(E-((CPotentialParam1*)params)->P->Phi(x,0,0) -0.5*pow_2(((CPotentialParam1*)params)->P->Omega*x));
    if(V2<=0) return 0; else return 1/sqrt(V2);
}
double k_intersect(double k, void* params)
{
    double E= ((CPotentialParam4*)params)->E;
    double X= ((CPotentialParam4*)params)->X;
    double Y= ((CPotentialParam4*)params)->Y;
    double Z= ((CPotentialParam4*)params)->Z;
    return E - ((CPotentialParam4*)params)->P->Phi(X*k,Y*k,Z*k) -0.5*pow_2(((CPotentialParam4*)params)->P->Omega*k)*(X*X+Y*Y);
}
double minusPhi(double x, void* params)
{
    if(((CPotentialParam4*)params)->E)
        return -((CPotentialParam1*)params)->P->Phi(0,x,0);
    else
        return -((CPotentialParam1*)params)->P->Phi(x,0,0);
}
/// \endcond

double CPotential::longaxisradius(double E) const
{
    // find maximal x-axis elongation for given energy
    double Xlow=1, Xupp=1;
    while ((Phi(Xlow,0,0)+0.5*pow_2(Omega*Xlow)>E) && (Xlow>MIN_RADIUS_CUTOFF)) Xlow/=2;
    if(Xlow<MIN_RADIUS_CUTOFF) 
    {
#if 0  // reserved for rotating potentials
        if(Omega>0) // it may happen that we're outside corotation radius and the effective potential is declining with radius
        {
            Xlow=1;
            while((Phi(Xupp,0,0)>E) && (Xupp<1e4)) Xupp*=2;
            if(Xupp>1e4)
                return -1;  // still something erroneous happened
        }
        else
#endif
        return -1;  // signal of error (E lower than minimum of the potential)
    }
    else
    {
        while ((Phi(Xupp,0,0)+0.5*pow_2(Omega*Xupp)<E) && (Xupp<MAX_RADIUS_CUTOFF)) Xupp*=2;
        if(Xupp>MAX_RADIUS_CUTOFF) 
        {
#if 0   // reserved for rotating potentials
            if(Omega>0) // we might catch potential maximum between Xlow and Xupp
            {
                int niter=0;
                double Eupp, Elow=Phi(Xlow,0,0);
                Xupp=2*Xlow;
                while(niter<100)
                {
                    Eupp=Phi(Xupp,0,0);
                    if(Eupp>E) niter=100;
                    else if(Eupp>Elow) Xupp*=1.25; else Xupp=(Xupp+Xlow)/2;
                    niter++;
                }
                if(Eupp<E) return -1;  // still unable to enclose root
            }
            else
#endif
            return -1;  // signal of error (e.g. E>0 and potential is everywhere negative)
        }
    }
    gsl_function F;
    CPotentialParam1 PP;
    PP.E=E;
    PP.P=this;
    F.function=&twoV2;
    F.params=&PP;
    return findRoot(&F, Xlow, Xupp);
}

double CPotential::longaxisperiod(double E, double Xmax) const
{
    // calculate T_orb (for x-axis orbit)
    if(Xmax<0)
        Xmax=longaxisradius(E);
    if(Xmax<0) return -1;  // error
    double result, error;
    gsl_function F;
    CPotentialParam1 PP;
    PP.E=E;
    PP.P=this;
    F.function=&Torb;
    F.params=&PP;
    gsl_integration_workspace * w = gsl_integration_workspace_alloc(1000);
    gsl_integration_qags (&F, 0, Xmax, 0, EPSREL_POTENTIAL_INT, 1000, w, &result, &error); 
    gsl_integration_workspace_free (w);
    return 4*result;
}

double CPotential::findintersection(double E, double X, double Y, double Z) const
{
    double klow=1, kupp=1;
    while ((Phi(klow*X,klow*Y,klow*Z)>E) && (klow>MIN_RADIUS_CUTOFF)) klow/=2;
    if(klow<MIN_RADIUS_CUTOFF) 
        return -1;  // signal of error (E lower than minimum of the potential)
    while ((Phi(kupp*X,kupp*Y,kupp*Z)<E) && (kupp<MAX_RADIUS_CUTOFF)) kupp*=2;
    if(kupp>MAX_RADIUS_CUTOFF) 
        return -1;  // signal of error (e.g. E>0 and potential is everywhere negative)
    gsl_function F;
    CPotentialParam4 PP;
    PP.E=E;
    PP.P=this;
    PP.X=X;
    PP.Y=Y;
    PP.Z=Z;
    F.function=&k_intersect;
    F.params=&PP;
    return findRoot(&F, klow, kupp);
}

// presently unused [rotating potential]
double CPotential::corotationradius() const
{
    if(Omega==0) return 0;  // actually 1/0

    gsl_function F;
    CPotentialParam1 PP;
    PP.P=this;
    F.function=&minusPhi;
    F.params=&PP;
    // X axis
    double Xlow=1, Xupp=3, Xmax, Ymax;
    while ((Phi(Xlow,0,0)>Phi(Xlow*2,0,0)) && (Xlow>MIN_RADIUS_CUTOFF)) Xlow/=2;
    if(Xlow<MIN_RADIUS_CUTOFF) return -1;  // signal of error
    while ((Phi(Xupp,0,0)>Phi(Xlow*2,0,0)) && (Xupp<MAX_RADIUS_CUTOFF)) Xupp*=2;
    if(Xupp>MAX_RADIUS_CUTOFF) return -1;  
    PP.E=0;
    Xmax = findMin(&F, Xlow, Xupp, Xlow*2);
    // Y axis
    Xlow=1; Xupp=3;
    while ((Phi(0,Xlow,0)>Phi(0,Xlow*2,0)) && (Xlow>MIN_RADIUS_CUTOFF)) Xlow/=2;
    if(Xlow<MIN_RADIUS_CUTOFF) return -1;  // signal of error
    while ((Phi(0,Xupp,0)>Phi(0,Xlow*2,0)) && (Xupp<MAX_RADIUS_CUTOFF)) Xupp*=2;
    if(Xupp>MAX_RADIUS_CUTOFF) return -1;  
    PP.E=1;
    Ymax = findMin(&F, Xlow, Xupp, Xlow*2);
    return (Xmax+Ymax)/2;
}

//----------------------------------------------------------------------------//
// Logarithmic potential

double CPotentialLog::Rho(double X, double Y, double Z, double /*t*/) const
{
    double m2=pow_2(Rc) + pow_2(X) + pow_2(Y)/pow_2(q) + pow_2(Z)/pow_2(p);
    return (0.5*(1+1/pow_2(q)+1/pow_2(p))/m2 - (pow_2(X) + pow_2(Y/pow_2(q)) + pow_2(Z/pow_2(p)))/pow_2(m2))/M_PI;
}

double CPotentialLog::Phi(double X, double Y, double Z, double /*t*/) const
{
    return log(pow_2(Rc) + pow_2(X) + pow_2(Y)/pow_2(q) + pow_2(Z)/pow_2(p)) - ((Mbh>0)?(Mbh/sqrt(sqrt(BH_SMOOTH) + pow_2(X) + pow_2(Y) + pow_2(Z))):0);
}

// r.h.s. of the differential equation: for given time 't' and variables 'v', return their difference 'f'
void CPotentialLog::DiffEq(unsigned n, double /*t*/, const double v[], double *f) const
{
    bool calcLyapunov = n>2*N_DIM;
    double X=v[0];
    double Y=v[1];
    double Z=(N_dim==3)?v[2]:0;    // for universality: if only 2 dimensions, Z=0 will add nothing
    f[0] = v[N_dim];
    f[1] = v[N_dim+1];
    if(N_dim==3)
        f[2] = v[N_dim+2];
    double m2=(pow_2(X) + pow_2(Y)/pow_2(q) + pow_2(Z)/pow_2(p) + pow_2(Rc));
    double r3=pow(pow_2(BH_SMOOTH) + pow_2(X) + pow_2(Y) + pow_2(Z), 1.5);
    f[N_dim]   = -2*X/m2    - ((Mbh>0)?(Mbh*X/r3):0);
    f[N_dim+1] = -2*Y/m2/pow_2(q) - ((Mbh>0)?(Mbh*Y/r3):0);
    if(N_dim==3)
        f[N_dim+2] = -2*Z/m2/pow_2(p) - ((Mbh>0)?(Mbh*Z/r3):0);
    if(calcLyapunov)
    {   // variational equation
        double r5=pow(pow_2(X) + pow_2(Y) + pow_2(Z), 2.5);
        double wx = v[N_dim*2];     // variation vector
        double wy = v[N_dim*2+1];
        double wz = (N_dim==3)? v[N_dim*2+2] : 0;
        f[N_dim*2]   = v[N_dim*3];
        f[N_dim*2+1] = v[N_dim*3+1];
        if(N_dim==3)
            f[N_dim*2+2]=v[N_dim*3+2];
        double d2Vdxdy = -X*Y * (4/pow_2(q*m2) + ((Mbh>0)?(3*Mbh/r5):0));
        double d2Vdxdz = (N_dim==3)?( -X*Z * (4/pow_2(p*m2)   + ((Mbh>0)?(3*Mbh/r5):0)) ):0;
        double d2Vdydz = (N_dim==3)?( -Y*Z * (4/pow_2(q*p*m2) + ((Mbh>0)?(3*Mbh/r5):0)) ):0;
        f[N_dim*3]   = -wx * (2/m2   -4*pow_2(X/m2   ) - 3*Mbh*pow_2(X)/r5 + Mbh/r3) - wy*d2Vdxdy - wz*d2Vdxdz;
        f[N_dim*3+1] = -wy * (2/m2/pow_2(q)-4*pow_2(Y/m2/pow_2(q)) - 3*Mbh*pow_2(Y)/r5 + Mbh/r3) - wx*d2Vdxdy - wz*d2Vdydz;
        if(N_dim==3)
            f[N_dim*3+2] = -wz * (2/m2/pow_2(p)-4*pow_2(Z/m2/pow_2(p)) - 3*Mbh*pow_2(Z)/r5 + Mbh/r3) - wx*d2Vdxdz - wy*d2Vdydz;
    }
}

//----------------------------------------------------------------------------//
// simple harmonic oscillator potential

double CPotentialHarmonic::Rho(double /*X*/, double /*Y*/, double /*Z*/, double /*t*/) const
{
    return 2*(1+1/q/q+1/p/p);
}

double CPotentialHarmonic::Phi(double X, double Y, double Z, double /*t*/) const
{
    return X*X + Y*Y/q/q + Z*Z/p/p - Mbh/sqrt(pow_2(BH_SMOOTH)+ pow_2(X) + pow_2(Y) + pow_2(Z)) - ((Omega>0)?pow_2(Omega)*(pow_2(X)+pow_2(Y))/2:0);
}

void CPotentialHarmonic::DiffEq(unsigned n, double /*t*/, const double v[], double *f) const
{
    bool calcLyapunov=n>2*N_DIM;
    double X=v[0];
    double Y=v[1];
    double Z=(N_dim==3)?v[2]:0;    // for universality: if only 2 dimensions, Z=0 will add nothing
    f[0] = v[N_dim];
    f[1] = v[N_dim+1];
    if(N_dim==3)
        f[2] = v[N_dim+2];
    double gradr= Mbh/pow( pow_2(BH_SMOOTH) + pow_2(X) + pow_2(Y) + pow_2(Z), 1.5);

    f[N_dim]   = -X*(gradr + 2)     + ((Omega>0)?  2*Omega*v[N_dim+1]+pow_2(Omega)*X : 0);
    f[N_dim+1] = -Y*(gradr + 2/q/q) + ((Omega>0)? -2*Omega*v[N_dim]  +pow_2(Omega)*Y : 0);
    if((N_dim==3))
        f[N_dim+2] = -Z*(gradr + 2/p/p);
    if(calcLyapunov)
    {   // variational equation
        double Mr3=Mbh/pow(pow_2(X) + pow_2(Y) + pow_2(Z), 1.5);
        double Mr5=Mbh/pow(pow_2(X) + pow_2(Y) + pow_2(Z), 2.5);
        double wx = v[N_dim*2];     // variation vector
        double wy = v[N_dim*2+1];
        double wz = (N_dim==3)? v[N_dim*2+2] : 0;
        f[N_dim*2]   = v[N_dim*3];
        f[N_dim*2+1] = v[N_dim*3+1];
        if(N_dim==3)
            f[N_dim*2+2]=v[N_dim*3+2];
        f[N_dim*3]   = -wx * (2     - 3*Mr5*pow_2(X) + Mr3) + wy*X*Y*3*Mr5 + wz*X*Z*3*Mr5;
        f[N_dim*3+1] = -wy * (2/q/q - 3*Mr5*pow_2(Y) + Mr3) + wx*Y*X*3*Mr5 + wz*Y*Z*3*Mr5;
        if(N_dim==3)
            f[N_dim*3+2] = -wz * (2/p/p - 3*Mr5*pow_2(Z) + Mr3) - wx*Z*X*3*Mr5 + wy*Z*Y*3*Mr5;
    }
}

//----------------------------------------------------------------------------//
// Dehnen potential
CPotentialDehnen::CPotentialDehnen(unsigned int _N_dim, double _q, double _p, double _Mbh, double _Gamma): 
    CPotential(_N_dim, _Mbh), q(_q), p(_p),
    Gamma(std::max<double>(0, std::min<double>(_Gamma, 2.0))) {}

double CPotentialDehnen::Rho(double X, double Y, double Z, double /*t*/) const
{
    double m = sqrt(pow_2(X) + pow_2(Y/q) + pow_2(Z/p));
    return (3-Gamma)/(4*M_PI*p*q) * pow(m, -Gamma) * pow(1+m, Gamma-4);
}

/// \cond INTERNAL_DOCS
struct CPotentialParamD{
    double X, Y, Z;
    double wx, wy, wz;
    double p, q, Gamma;
};
double intPhiD(double s, void* params)
{
    double X= ((CPotentialParamD*)params)->X;
    double Y= ((CPotentialParamD*)params)->Y;
    double Z= ((CPotentialParamD*)params)->Z;
    double p= ((CPotentialParamD*)params)->p;
    double q= ((CPotentialParamD*)params)->q;
    double Gamma= ((CPotentialParamD*)params)->Gamma;
    double m = s * sqrt( pow_2(X) + pow_2(Y)/(1-(1-pow_2(q))*pow_2(s)) + pow_2(Z)/(1-(1-pow_2(p))*pow_2(s)));
    double numerator = (Gamma==2)? (log((1+m)*s/m) - 1/(1+m) - log(s)) :
        (1 - (3-Gamma)*pow(m/(m+1), 2-Gamma) + (2-Gamma)*pow(m/(m+1), 3-Gamma))/(2-Gamma);
    return numerator / sqrt( (1-(1-pow_2(q))*pow_2(s)) * (1-(1-pow_2(p))*pow_2(s)));
}
/// \endcond

// accurate adaptive integration of potential and forces (muuuuch slower, used only for a high-accuracy test run)
#define DEHNENPOTENTIAL_ACCURATE_QUADRATURE 0

double CPotentialDehnen::Phi(double X, double Y, double Z, double /*t*/) const
{
    double result, error;
    CPotentialParamD PP;
    PP.X=X; PP.Y=Y; PP.Z=Z;
    PP.p=p; PP.q=q; PP.Gamma=Gamma;
    gsl_function F;
    F.function=&intPhiD;
    F.params=&PP;
    size_t neval;
    gsl_integration_qng (&F, 0, 1, 0, EPSREL_POTENTIAL_INT, &result, &error, &neval); 
#if DEHNENPOTENTIAL_ACCURATE_QUADRATURE
    if(fabs(error/result)>1e-3) {
        gsl_integration_workspace * ws = gsl_integration_workspace_alloc (1000);
        gsl_integration_qag (&F, 0, 1, 0, EPSREL_POTENTIAL_INT, 1000, 3,ws, &result, &error); 
        gsl_integration_workspace_free(ws);
    }
#endif
    return -result - ((Mbh>0)?(Mbh/sqrt(pow_2(BH_SMOOTH) + pow_2(X) + pow_2(Y) + pow_2(Z))):0);
}

/// \cond INTERNAL_DOCS
double intForceDx(double s, void* params)
{
    double X= ((CPotentialParamD*)params)->X;
    double Y= ((CPotentialParamD*)params)->Y;
    double Z= ((CPotentialParamD*)params)->Z;
    double p= ((CPotentialParamD*)params)->p;
    double q= ((CPotentialParamD*)params)->q;
    double Gamma= ((CPotentialParamD*)params)->Gamma;
    double m = s * sqrt( pow_2(X) + pow_2(Y)/(1+(pow_2(q)-1)*pow_2(s)) + pow_2(Z)/(1+(pow_2(p)-1)*pow_2(s)));
    return pow_2(s) * pow(m, -Gamma) * pow(1+m, Gamma-4) /
        sqrt( (1+(pow_2(q)-1)*pow_2(s)) * (1+(pow_2(p)-1)*pow_2(s)));
}
double intForceDy(double s, void* params)
{
    double X= ((CPotentialParamD*)params)->X;
    double Y= ((CPotentialParamD*)params)->Y;
    double Z= ((CPotentialParamD*)params)->Z;
    double p= ((CPotentialParamD*)params)->p;
    double q= ((CPotentialParamD*)params)->q;
    double Gamma= ((CPotentialParamD*)params)->Gamma;
    double m = s * sqrt( pow_2(X)/(pow_2(q)+(1-pow_2(q))*pow_2(s)) + pow_2(Y)/pow_2(q) + pow_2(Z)/(pow_2(q)+(pow_2(p)-pow_2(q))*pow_2(s)));
    return pow_2(s) * pow(m, -Gamma) * pow(1+m, Gamma-4) /
        sqrt( (pow_2(q)+(pow_2(p)-pow_2(q))*pow_2(s)) * (pow_2(q)+(1-pow_2(q))*pow_2(s)));
}
double intForceDz(double s, void* params)
{
    double X= ((CPotentialParamD*)params)->X;
    double Y= ((CPotentialParamD*)params)->Y;
    double Z= ((CPotentialParamD*)params)->Z;
    double p= ((CPotentialParamD*)params)->p;
    double q= ((CPotentialParamD*)params)->q;
    double Gamma= ((CPotentialParamD*)params)->Gamma;
    double m = s * sqrt( pow_2(X)/(pow_2(p)+(1-pow_2(p))*pow_2(s)) + pow_2(Y)/(pow_2(p)+(pow_2(q)-pow_2(p))*pow_2(s)) + pow_2(Z)/pow_2(p));
    return pow_2(s) * pow(m, -Gamma) * pow(1+m, Gamma-4) /
        sqrt( (pow_2(p)+(1-pow_2(p))*pow_2(s)) * (pow_2(p)+(pow_2(q)-pow_2(p))*pow_2(s)));
}

double intForceD2x(double s, void* params)
{
    double X= ((CPotentialParamD*)params)->X;
    double Y= ((CPotentialParamD*)params)->Y;
    double Z= ((CPotentialParamD*)params)->Z;
    double wx=((CPotentialParamD*)params)->wx;
    double wy=((CPotentialParamD*)params)->wy;
    double wz=((CPotentialParamD*)params)->wz;
    double p= ((CPotentialParamD*)params)->p;
    double q= ((CPotentialParamD*)params)->q;
    double Gamma= ((CPotentialParamD*)params)->Gamma;
    double m = s * sqrt( pow_2(X) + pow_2(Y)/(1+(pow_2(q)-1)*pow_2(s)) + pow_2(Z)/(1+(pow_2(p)-1)*pow_2(s)));
    return pow_2(pow_2(s)) * pow(m, -Gamma-2) * pow(1+m, Gamma-5) * (Gamma+4*m) * 
      ( X * wx / sqrt( (1+(pow_2(q)-1)*pow_2(s)) * (1+(pow_2(p)-1)*pow_2(s))) +
        Y * wy * pow( 1 + (pow_2(q)-1)*pow_2(s), -1.5) / sqrt(1 + (pow_2(p)-1)*pow_2(s)) +
        Z * wz * pow( 1 + (pow_2(p)-1)*pow_2(s), -1.5) / sqrt(1 + (pow_2(q)-1)*pow_2(s)) );
}
double intForceD2y(double s, void* params)
{
    double X= ((CPotentialParamD*)params)->X;
    double Y= ((CPotentialParamD*)params)->Y;
    double Z= ((CPotentialParamD*)params)->Z;
    double wx=((CPotentialParamD*)params)->wx;
    double wy=((CPotentialParamD*)params)->wy;
    double wz=((CPotentialParamD*)params)->wz;
    double p= ((CPotentialParamD*)params)->p;
    double q= ((CPotentialParamD*)params)->q;
    double Gamma= ((CPotentialParamD*)params)->Gamma;
    double m = s * sqrt( pow_2(X)/(pow_2(q)+(1-pow_2(q))*pow_2(s)) + pow_2(Y)/pow_2(q) + pow_2(Z)/(pow_2(q)+(pow_2(p)-pow_2(q))*pow_2(s)));
    return pow_2(pow_2(s)) * pow(m, -Gamma-2) * pow(1+m, Gamma-5) * (Gamma+4*m) * 
      ( Y * wy / pow_2(q) / sqrt( (pow_2(q)+(pow_2(p)-pow_2(q))*pow_2(s)) * (pow_2(q)+(1-pow_2(q))*pow_2(s))) +
        Z * wz * pow( pow_2(q) + (pow_2(p)-pow_2(q))*pow_2(s), -1.5) / sqrt(pow_2(q) + (1-pow_2(q))*pow_2(s)) +
        X * wx * pow( pow_2(q) + (1-pow_2(q))*pow_2(s), -1.5) / sqrt(pow_2(q) + (pow_2(p)-pow_2(q))*pow_2(s)) );
}
double intForceD2z(double s, void* params)
{
    double X= ((CPotentialParamD*)params)->X;
    double Y= ((CPotentialParamD*)params)->Y;
    double Z= ((CPotentialParamD*)params)->Z;
    double wx=((CPotentialParamD*)params)->wx;
    double wy=((CPotentialParamD*)params)->wy;
    double wz=((CPotentialParamD*)params)->wz;
    double p= ((CPotentialParamD*)params)->p;
    double q= ((CPotentialParamD*)params)->q;
    double Gamma= ((CPotentialParamD*)params)->Gamma;
    double m = s * sqrt( pow_2(X)/(pow_2(p)+(1-pow_2(p))*pow_2(s)) + pow_2(Y)/(pow_2(p)+(pow_2(q)-pow_2(p))*pow_2(s)) + pow_2(Z)/pow_2(p));
    return pow_2(pow_2(s)) * pow(m, -Gamma-2) * pow(1+m, Gamma-5) * (Gamma+4*m) * 
      ( Z * wz / pow_2(p) / sqrt( (pow_2(p)+(1-pow_2(p))*pow_2(s)) * (pow_2(p)+(pow_2(q)-pow_2(p))*pow_2(s))) +
        X * wx * pow( pow_2(p) + (1-pow_2(p))*pow_2(s), -1.5) / sqrt(pow_2(p) + (pow_2(q)-pow_2(p))*pow_2(s)) +
        Y * wy * pow( pow_2(p) + (pow_2(q)-pow_2(p))*pow_2(s), -1.5) / sqrt(pow_2(p) + (1-pow_2(p))*pow_2(s)) );
}
/// \endcond

void CPotentialDehnen::DiffEq(unsigned n, double /*t*/, const double v[], double *f) const
{
    bool calcLyapunov=n>2*N_DIM;
    double X=v[0];
    double Y=v[1];
    double Z=(N_dim==3)?v[2]:0;    // for universality: if only 2 dimensions, Z=0 will add nothing
    f[0] = v[N_dim];
    f[1] = v[N_dim+1];
    if(N_dim==3)
        f[2] = v[N_dim+2];
    double r3=pow(pow_2(BH_SMOOTH) + pow_2(X) + pow_2(Y) + pow_2(Z), 1.5);
    // force calculation requires integration
    double resultx, resulty, resultz, error; 
    size_t neval;
    CPotentialParamD PP;
    PP.X=X; PP.Y=Y; PP.Z=Z;
    PP.p=p; PP.q=q; PP.Gamma=Gamma;
    gsl_function F;
    F.params=&PP;
    F.function=&intForceDx;
    gsl_integration_qng (&F, 0, 1, 0, EPSREL_POTENTIAL_INT, &resultx, &error, &neval); 
#if DEHNENPOTENTIAL_ACCURATE_QUADRATURE
    gsl_integration_workspace * ws = NULL;
    if(fabs(error/resultx)>0.001) {
        if(!ws) ws=gsl_integration_workspace_alloc (1000);
        gsl_integration_qag (&F, 0, 1, 0, EPSREL_POTENTIAL_INT, 1000, 3,ws, &resultx, &error); 
    }
#endif
    f[N_dim]   = -(3-Gamma)*X * resultx   - ((Mbh>0)?(Mbh*X/r3):0);
    F.function=&intForceDy;
    gsl_integration_qng (&F, 0, 1, 0, EPSREL_POTENTIAL_INT, &resulty, &error, &neval); 
#if DEHNENPOTENTIAL_ACCURATE_QUADRATURE
    if(fabs(error/resulty)>0.001) {
        if(!ws) ws=gsl_integration_workspace_alloc (1000);
        gsl_integration_qag (&F, 0, 1, 0, EPSREL_POTENTIAL_INT, 1000, 3,ws, &resulty, &error); 
    }
#endif
    f[N_dim+1] = -(3-Gamma)*Y/q * resulty - ((Mbh>0)?(Mbh*Y/r3):0);
    if(N_dim==3)
    {
        F.function=&intForceDz;
        gsl_integration_qng (&F, 0, 1, 0, EPSREL_POTENTIAL_INT, &resultz, &error, &neval); 
#if DEHNENPOTENTIAL_ACCURATE_QUADRATURE
        if(fabs(error/resultz)>0.001) {
            if(!ws) ws=gsl_integration_workspace_alloc (1000);
            gsl_integration_qag (&F, 0, 1, 0, EPSREL_POTENTIAL_INT, 1000, 3,ws, &resultz, &error); 
        }
#endif
        f[N_dim+2] = -(3-Gamma)*Z/p * resultz - ((Mbh>0)?(Mbh*Z/r3):0);
    }
#if DEHNENPOTENTIAL_ACCURATE_QUADRATURE
    gsl_integration_workspace_free(ws);
#endif
    if(calcLyapunov)
    {   // variational equation
        double r5=pow(pow_2(X) + pow_2(Y) + pow_2(Z), 2.5);
        double wx = v[N_dim*2];     // variation vector
        double wy = v[N_dim*2+1];
        double wz = (N_dim==3)? v[N_dim*2+2] : 0;
        f[N_dim*2]   = v[N_dim*3];
        f[N_dim*2+1] = v[N_dim*3+1];
        if(N_dim==3)
            f[N_dim*2+2]=v[N_dim*3+2];
        PP.wx=wx; PP.wy=wy; PP.wz=wz;
        double result;
        F.function=&intForceD2x;
        gsl_integration_qng (&F, 0, 1, 0, EPSREL_POTENTIAL_INT, &result, &error, &neval); 
        f[N_dim*3]   = -(3-Gamma) * (wx*resultx - X*result) + ((Mbh>0)?Mbh*( (3*pow_2(X)/r5-1/r3)*wx + 3*X*Y/r5*wy + 3*X*Z/r5*wz ):0);
        F.function=&intForceD2y;
        gsl_integration_qng (&F, 0, 1, 0, EPSREL_POTENTIAL_INT, &result, &error, &neval); 
        f[N_dim*3+1] = -(3-Gamma)/q * (wy*resulty - Y*result) + ((Mbh>0)?Mbh*( (3*pow_2(Y)/r5-1/r3)*wy + 3*Y*X/r5*wx + 3*Y*Z/r5*wz ):0);
        if(N_dim==3)
        {
            F.function=&intForceD2z;
            gsl_integration_qng (&F, 0, 1, 0, EPSREL_POTENTIAL_INT, &result, &error, &neval); 
            f[N_dim*3+2] = -(3-Gamma)/p * (wz*resultz - Z*result) + ((Mbh>0)?Mbh*( (3*pow_2(Z)/r5-1/r3)*wz + 3*Z*X/r5*wx + 3*Z*Y/r5*wy ):0);
        }
    }
}

//----------------------------------------------------------------------------//
// CPotentialSH -- parent class for all potentials using angular expansion in spherical harmonics

void CPotentialSH::assignlmrange()
{
    lmax = mysymmetry & ST_SPHSYM ? 0 : static_cast<int>(Ncoefs_angular);  // if spherical model, use only l=0,m=0 term
    lstep= mysymmetry & ST_REFLECTION ? 2 : 1;    // if reflection symmetry, use only even l
    mmax = mysymmetry & ST_ZROTSYM ? 0 : 1;       // if axisymmetric model, use only m=0 terms, otherwise all terms up to l (1 is the multiplying factor)
    mmin = mysymmetry & ST_PLANESYM ? 0 : -1;     // if triaxial symmetry, do not use sine terms which correspond to m<0
    mstep= mysymmetry & ST_PLANESYM ? 2 : 1;      // if triaxial symmetry, use only even m
}

/// \cond INTERNAL_DOCS
// helper functions for integration over angles (common for BSE and Spline potentials), used in computing coefficients from analytic density model
struct CPotentialParamSH{
    const CDensity* P;
    int sh_n, sh_l, sh_m;
    double costheta, sintheta, r, Alpha;
    double theta_max, phi_max;    // Pi/2, Pi or even 2*Pi, depending on potential symmetry
};
double intSH_phi(double phi, void* params)
{
    double costheta= ((CPotentialParamSH*)params)->costheta;
    double sintheta= ((CPotentialParamSH*)params)->sintheta;
    double r= ((CPotentialParamSH*)params)->r;
    const CDensity* P=((CPotentialParamSH*)params)->P;
    int sh_l=((CPotentialParamSH*)params)->sh_l;
    int sh_m=((CPotentialParamSH*)params)->sh_m;
    return P->Rho(r*sintheta*cos(phi), r*sintheta*sin(phi), r*costheta)
        * gsl_sf_legendre_sphPlm(sh_l, abs(sh_m), costheta) * (sh_m>=0 ? cos(sh_m*phi) : sin(-sh_m*phi));
}
double intSH_theta(double theta, void* params)
{
    double result, error;
    size_t neval;
    ((CPotentialParamSH*)params)->costheta=cos(theta);
    ((CPotentialParamSH*)params)->sintheta=sin(theta);
    gsl_function F;
    F.function=&intSH_phi;
    F.params=params;
    gsl_integration_qng(&F, 0, ((CPotentialParamSH*)params)->phi_max, 0, EPSREL_POTENTIAL_INT, &result, &error, &neval);
    return result * sin(theta) / (2*M_SQRTPI);   // factor \sqrt{4\pi} coming from the definition of spherical function Y_l^m
}
/// \endcond

double CPotentialSH::Rho(double X, double Y, double Z, double /*t*/) const
{
    double costheta=(X==0&&Y==0&&Z==0)?0:(Z/sqrt(pow_2(X)+pow_2(Y)+pow_2(Z)));
    double phi=atan2(Y, X);
    double r=sqrt(X*X+Y*Y+Z*Z);
    double result=0;
    // arrays where angular expansion coefficients will be accumulated by calling computeSHcoefs() for derived classes
    double coefsF[MAX_NCOEFS_ANGULAR*MAX_NCOEFS_ANGULAR];      // F(theta,phi)
    double coefsdFdr[MAX_NCOEFS_ANGULAR*MAX_NCOEFS_ANGULAR];   // dF(theta,phi)/dr
    double coefsd2Fdr2[MAX_NCOEFS_ANGULAR*MAX_NCOEFS_ANGULAR]; // d2F(theta,phi)/dr2
    computeSHCoefs(r, coefsF, coefsdFdr, coefsd2Fdr2);         // implemented in derived classes
    double legendre_array[MAX_NCOEFS_ANGULAR-1];
    for(int m=0; m<=mmax*lmax; m+=mstep)
    {
        gsl_sf_legendre_sphPlm_array(lmax, m, costheta, legendre_array);
        int lmin = lstep==2 ? (m+1)/2*2 : m;   // if lstep is even and m is odd, start from next even number greater than m
        double cosmphi = (m==0 ? 1 : cos(m*phi)*M_SQRT2) * 2*M_SQRTPI;   // factor \sqrt{4\pi} from the definition of spherical function Y_l^m absorbed into this term
        for(int l=lmin; l<=lmax; l+=lstep)
        {
            result += (coefsd2Fdr2[l*(l+1)+m] + 2/r*coefsdFdr[l*(l+1)+m] - l*(l+1)/(r*r)*coefsF[l*(l+1)+m]) 
                * legendre_array[l-m] * cosmphi;
        }
        if(mmin<0 && m>0)  // use sin terms as well
        {
            double sinmphi = sin(m*phi)*M_SQRT2 * 2*M_SQRTPI;
            for(int l=lmin; l<=lmax; l+=mstep)
                result += (coefsd2Fdr2[l*(l+1)-m] + 2/r*coefsdFdr[l*(l+1)-m] - l*(l+1)/(r*r)*coefsF[l*(l+1)-m]) 
                    * legendre_array[l-m] * sinmphi;
        }
    }
    result *= 1/(4*M_PI);
    if(result<0 || gsl_isnan(result)) result=0;
    return result;
}

double CPotentialSH::Phi(double X, double Y, double Z, double /*t*/) const
{
    double costheta=(X==0&&Y==0&&Z==0)?0:(Z/sqrt(pow_2(X)+pow_2(Y)+pow_2(Z)));
    double phi=atan2(Y, X);
    double r=sqrt(X*X+Y*Y+Z*Z);
    double res=0;
    // arrays where angular expansion coefficients will be accumulated by calling computeSHcoefs() for derived classes
    double coefsF[MAX_NCOEFS_ANGULAR*MAX_NCOEFS_ANGULAR];      // F(theta,phi)
    computeSHCoefs(r, coefsF, NULL, NULL);  // implemented in derived classes
    double legendre_array[MAX_NCOEFS_ANGULAR-1];
    for(int m=0; m<=mmax*lmax; m+=mstep)
    {
        gsl_sf_legendre_sphPlm_array(lmax, m, costheta, legendre_array);
        double cosmphi = (m==0 ? 1 : cos(m*phi)*M_SQRT2) * 2*M_SQRTPI;   // factor \sqrt{4\pi} from the definition of spherical function Y_l^m absorbed into this term
        int lmin = lstep==2 ? (m+1)/2*2 : m;   // if lstep is even and m is odd, start from next even number greater than m
        for(int l=lmin; l<=lmax; l+=lstep)
            res +=  coefsF[l*(l+1)+m] * legendre_array[l-m] * cosmphi;
        if(mmin<0 && m>0)
        {
            double sinmphi = sin(m*phi)*M_SQRT2 * 2*M_SQRTPI;
            for(int l=lmin; l<=lmax; l+=lstep)
                res +=  coefsF[l*(l+1)-m] * legendre_array[l-m] * sinmphi;
        }
    }
    return res - ((Mbh>0)?(Mbh/sqrt(pow_2(BH_SMOOTH) + pow_2(X) + pow_2(Y) + pow_2(Z))):0) - ((Omega>0)?pow_2(Omega)*(pow_2(X)+pow_2(Y))/2:0);
}

void CPotentialSH::DiffEq(unsigned n, double /*t*/, const double v[], double *f) const
{
    bool calcLyapunov=n>2*N_DIM;
    double X=v[0];
    double Y=v[1];
    double Z=(N_dim==3)?v[2]:0;    // for universality: if only 2 dimensions, Z=0 will add nothing
    f[0] = v[N_dim];
    f[1] = v[N_dim+1];
    if(N_dim==3)
        f[2] = v[N_dim+2];

    double dFdR=0, dFdTheta=0, dFdPhi=0;   
    double d2FdR2=0, d2FdRdTheta=0, d2FdRdPhi=0, d2FdTheta2=0, d2FdThetadPhi=0, d2FdPhi2=0;
    double r2 = pow_2(X) + pow_2(Y) + pow_2(Z);
    double r  = sqrt(r2);
    double rxy2=pow_2(X)+pow_2(Y);
    double rxy =sqrt(rxy2);
    double theta=atan2(rxy, Z); 
    double phi  =atan2(Y, X);
    double costheta=(Z==0)?0:cos(theta);
    double sintheta=sin(theta);
    double cosphi=(X==0)?0:cos(phi);
    double sinphi=(Y==0)?0:sin(phi);
    // arrays where angular expansion coefficients will be accumulated by calling computeSHcoefs() for derived classes
    double coefsF[MAX_NCOEFS_ANGULAR*MAX_NCOEFS_ANGULAR];      // F(theta,phi)
    double coefsdFdr[MAX_NCOEFS_ANGULAR*MAX_NCOEFS_ANGULAR];   // dF(theta,phi)/dr
    double coefsd2Fdr2[MAX_NCOEFS_ANGULAR*MAX_NCOEFS_ANGULAR]; // d2F(theta,phi)/dr2
    computeSHCoefs(r, coefsF, coefsdFdr, calcLyapunov ? coefsd2Fdr2 : NULL);  // implemented in derived classes
    double legendre_array[MAX_NCOEFS_ANGULAR-1];
    double legendre_deriv_array[MAX_NCOEFS_ANGULAR-1];
    for(int m=0; m<=mmax*lmax; m+=mstep)
    {
#ifdef NO_OPTIMIZATION
        gsl_sf_legendre_sphPlm_deriv_array(lmax, m, costheta, legendre_array, legendre_deriv_array);
#else
        // compute unnormalized polynomials and then normalize manually, which is faster than computing normalized ones.
        // This is not suitable for large l,m (when overflow may occur), but in our application we aren't going to have such large values.
        gsl_sf_legendre_Plm_deriv_array(lmax, m, costheta, legendre_array, legendre_deriv_array);
        double prefact=sqrt(1.0/4/M_PI/gsl_sf_fact(2*m));
        for(int l=m; l<=lmax; l++)
        {
            double prefactl=sqrt(2*l+1.0)*prefact;
            legendre_array[l-m] *= prefactl;
            legendre_deriv_array[l-m] *= prefactl;
            prefact *= sqrt((l+1.0-m)/(l+1.0+m));
        }
#endif
        double cosmphi = (m==0 ? 1 : cos(m*phi)*M_SQRT2) * 2*M_SQRTPI;   // factor \sqrt{4\pi} from the definition of spherical function Y_l^m absorbed into this term
        double sinmphi = (sin(m*phi)*M_SQRT2) * 2*M_SQRTPI;
        int lmin = lstep==2 ? (m+1)/2*2 : m;   // if lstep is even and m is odd, start from next even number greater than m
        for(int l=lmin; l<=lmax; l+=lstep)
        {
            int indx=l*(l+1)+m;
            dFdR +=  coefsdFdr[indx] * legendre_array[l-m] * cosmphi;
            dFdTheta += coefsF[indx] * legendre_deriv_array[l-m] * (-sintheta) * cosmphi;
            dFdPhi   += coefsF[indx] * legendre_array[l-m] * (-m)*sinmphi;
            if(calcLyapunov)
            {
                d2FdR2 +=  coefsd2Fdr2[indx] * legendre_array[l-m] * cosmphi;
                d2FdRdTheta+=coefsdFdr[indx] * legendre_deriv_array[l-m] * (-sintheta) * cosmphi;
                d2FdRdPhi += coefsdFdr[indx] * legendre_array[l-m] * (-m)*sinmphi;
                d2FdTheta2   += coefsF[indx] * (costheta * legendre_deriv_array[l-m] - (l*(l+1)-pow_2(m/sintheta)) * legendre_array[l-m]) * cosmphi;
                d2FdThetadPhi+= coefsF[indx] * legendre_deriv_array[l-m] * (-sintheta) * (-m)*sinmphi;
                d2FdPhi2     += coefsF[indx] * legendre_array[l-m] * cosmphi * -m*m;
            }
            if(mmin<0 && m>0)
            {
                indx=l*(l+1)-m;
                dFdR +=  coefsdFdr[indx] * legendre_array[l-m] * sinmphi;
                dFdTheta += coefsF[indx] * legendre_deriv_array[l-m] * (-sintheta) * sinmphi;
                dFdPhi   += coefsF[indx] * legendre_array[l-m] * m*cosmphi;
                if(calcLyapunov)
                {
                    d2FdR2 +=  coefsd2Fdr2[indx] * legendre_array[l-m] * sinmphi;
                    d2FdRdTheta+=coefsdFdr[indx] * legendre_deriv_array[l-m] * (-sintheta) * sinmphi;
                    d2FdRdPhi += coefsdFdr[indx] * legendre_array[l-m] * m*cosmphi;
                    d2FdTheta2   += coefsF[indx] * (costheta * legendre_deriv_array[l-m] - (l*(l+1)-pow_2(m/sintheta)) * legendre_array[l-m]) * sinmphi;
                    d2FdThetadPhi+= coefsF[indx] * legendre_deriv_array[l-m] * (-sintheta) * m*cosmphi;
                    d2FdPhi2     += coefsF[indx] * legendre_array[l-m] * sinmphi * -m*m;
                }
            }
        }
    }
    if((mysymmetry & ST_TRIAXIAL) == ST_TRIAXIAL) {
        if((X==0)||(Y==0)) dFdPhi=0; // to avoid singularity
        if(Z==0) dFdTheta=0;
    }
    if(rxy==0) rxy=BH_SMOOTH; 
    double r3=pow(pow_2(BH_SMOOTH) + pow_2(X) + pow_2(Y) + pow_2(Z), 1.5);
    double J[3][3]   = { {sintheta*cosphi, sintheta*sinphi, costheta}, {costheta*cosphi/r, costheta*sinphi/r, -sintheta/r}, {-sinphi/rxy, cosphi/rxy, 0} };
    f[N_dim]   = -( dFdR * J[0][0] + dFdTheta * J[1][0] + dFdPhi * J[2][0]) - ((Mbh>0)?(Mbh*X/r3):0) + ((Omega>0)?  2*Omega*v[N_dim+1]+pow_2(Omega)*X : 0) ;
    f[N_dim+1] = -( dFdR * J[0][1] + dFdTheta * J[1][1] + dFdPhi * J[2][1]) - ((Mbh>0)?(Mbh*Y/r3):0) + ((Omega>0)? -2*Omega*v[N_dim]  +pow_2(Omega)*Y : 0) ;
    if((N_dim==3))
        f[N_dim+2] = -( dFdR * J[0][2] + dFdTheta * J[1][2]) - ((Mbh>0)?(Mbh*Z/r3):0);
    if(calcLyapunov)
    {
        double X2=X*X, Y2=Y*Y, Z2=Z*Z;
        double d2F[3][3] = { { d2FdR2, d2FdRdTheta, d2FdRdPhi }, {d2FdRdTheta, d2FdTheta2, d2FdThetadPhi}, {d2FdRdPhi, d2FdThetadPhi, d2FdPhi2} };
        double m=1/(r*r2);
        double d2R[3][3] = { {(Y2+Z2)*m, -X*Y*m, -X*Z*m}, {-X*Y*m, (X2+Z2)*m, -Y*Z*m}, {-X*Z*m, -Y*Z*m, (X2+Y2)*m} };
        m=1/(rxy*rxy2*r2*r2);
        double d2Theta[3][3] = { 
            {-Z*(2*X2*X2+Y2*X2-Y2*Y2-Z2*Y2)*m, -X*Y*Z*(3*X2+3*Y2+Z2)*m, X*(X2+Y2-Z2)*rxy2*m},
            {-X*Y*Z*(3*X2+3*Y2+Z2)*m, -Z*(2*Y2*Y2+Y2*X2-X2*X2-Z2*X2)*m, Y*(X2+Y2-Z2)*rxy2*m},
            {X*(X2+Y2-Z2)*rxy2*m, Y*(X2+Y2-Z2)*rxy2*m, 2*Z*rxy2*rxy2*m} };
        m=1/(rxy2*rxy2);
        double d2Phi[3][3] = { {2*X*Y*m, (Y2-X2)*m, 0}, {(Y2-X2)*m, -2*X*Y*m, 0}, {0, 0, 0} };
        // convert to Cartesian coordinates
        double r5=pow(pow_2(BH_SMOOTH) + pow_2(X) + pow_2(Y) + pow_2(Z), 2.5);
        double k0 = (J[0][0]*d2F[0][0]+J[1][0]*d2F[0][1]+J[2][0]*d2F[0][2]);
        double k1 = (J[0][0]*d2F[0][1]+J[1][0]*d2F[1][1]+J[2][0]*d2F[1][2]);
        double k2 = (J[0][0]*d2F[0][2]+J[1][0]*d2F[1][2]+J[2][0]*d2F[2][2]);
        double d2Fdx2 = (k0*J[0][0] + k1*J[1][0] + k2*J[2][0] + dFdR*d2R[0][0] + dFdTheta*d2Theta[0][0] + dFdPhi*d2Phi[0][0]) +Mbh*(3*X2/r5-1/r3);
        double d2Fdxdy= (k0*J[0][1] + k1*J[1][1] + k2*J[2][1] + dFdR*d2R[0][1] + dFdTheta*d2Theta[0][1] + dFdPhi*d2Phi[0][1]) +Mbh*(3*X*Y/r5);
        double d2Fdxdz= (k0*J[0][2] + k1*J[1][2] + k2*J[2][2] + dFdR*d2R[0][2] + dFdTheta*d2Theta[0][2] + dFdPhi*d2Phi[0][2]) +Mbh*(3*X*Z/r5);
        k0 = (J[0][1]*d2F[0][0]+J[1][1]*d2F[0][1]+J[2][1]*d2F[0][2]);
        k1 = (J[0][1]*d2F[0][1]+J[1][1]*d2F[1][1]+J[2][1]*d2F[1][2]);
        k2 = (J[0][1]*d2F[0][2]+J[1][1]*d2F[1][2]+J[2][1]*d2F[2][2]);
        double d2Fdydx= (k0*J[0][0] + k1*J[1][0] + k2*J[2][0] + dFdR*d2R[1][0] + dFdTheta*d2Theta[1][0] + dFdPhi*d2Phi[1][0]) +Mbh*(3*Y*X/r5);
        double d2Fdy2 = (k0*J[0][1] + k1*J[1][1] + k2*J[2][1] + dFdR*d2R[1][1] + dFdTheta*d2Theta[1][1] + dFdPhi*d2Phi[1][1]) +Mbh*(3*Y2/r5-1/r3);
        double d2Fdydz= (k0*J[0][2] + k1*J[1][2] + k2*J[2][2] + dFdR*d2R[1][2] + dFdTheta*d2Theta[1][2] + dFdPhi*d2Phi[1][2]) +Mbh*(3*Y*Z/r5);
        k0 = (J[0][2]*d2F[0][0]+J[1][2]*d2F[0][1]+J[2][2]*d2F[0][2]);
        k1 = (J[0][2]*d2F[0][1]+J[1][2]*d2F[1][1]+J[2][2]*d2F[1][2]);
        k2 = (J[0][2]*d2F[0][2]+J[1][2]*d2F[1][2]+J[2][2]*d2F[2][2]);
        double d2Fdzdx= (k0*J[0][0] + k1*J[1][0] + k2*J[2][0] + dFdR*d2R[2][0] + dFdTheta*d2Theta[2][0] + dFdPhi*d2Phi[2][0]) +Mbh*(3*Z*X/r5);
        double d2Fdzdy= (k0*J[0][1] + k1*J[1][1] + k2*J[2][1] + dFdR*d2R[2][1] + dFdTheta*d2Theta[2][1] + dFdPhi*d2Phi[2][1]) +Mbh*(3*Z*Y/r5);
        double d2Fdz2 = (k0*J[0][2] + k1*J[1][2] + k2*J[2][2] + dFdR*d2R[2][2] + dFdTheta*d2Theta[2][2] + dFdPhi*d2Phi[2][2]) +Mbh*(3*Z2/r5-1/r3);

        double wx = v[N_dim*2];     // variation vector
        double wy = v[N_dim*2+1];
        double wz = (N_dim==3)? v[N_dim*2+2] : 0;
        f[N_dim*2]   = v[N_dim*3];
        f[N_dim*2+1] = v[N_dim*3+1];
        if(N_dim==3)
            f[N_dim*2+2]=v[N_dim*3+2];
        f[N_dim*3]   = -(wx*d2Fdx2  + wy*d2Fdxdy + wz*d2Fdxdz);
        f[N_dim*3+1] = -(wx*d2Fdydx + wy*d2Fdy2  + wz*d2Fdydz);
        if(N_dim==3)
            f[N_dim*3+2] = -(wx*d2Fdzdx + wy*d2Fdzdy + wz*d2Fdz2);
    }
}

//----------------------------------------------------------------------------//
// Basis-set expansion for arbitrary potential (using Zhao(1995) basis set)

template<typename NumT> CPotentialBSE::CPotentialBSE(
    unsigned int _N_dim, double _Mbh, double _Alpha, size_t _Ncoefs_radial, size_t _Ncoefs_angular, 
    const CPointMassSet<NumT> &points, SYMMETRYTYPE _sym, bool checkMonotonic):
  CPotentialSH(_N_dim, _Mbh, _Ncoefs_angular), 
  Ncoefs_radial(std::min<size_t>(MAX_NCOEFS_RADIAL-1, _Ncoefs_radial)),
  Alpha(_Alpha)
{
    mysymmetry=_sym;
    assignlmrange();
    if(points.size()==0) { initDefault(); return; }
    prepareCoefsDiscrete(points);
    if(checkMonotonic) checkSymmetry();  // not called from BSE-Schwarzschild model
}
template CPotentialBSE::CPotentialBSE(unsigned int _N_dim, double _Mbh, double _Alpha, size_t _Ncoefs_radial, size_t _Ncoefs_angular, const smile::CPointMassSet<double>&points, smile::CDensity::SYMMETRYTYPE _sym, bool checkMonotonic);
template CPotentialBSE::CPotentialBSE(unsigned int _N_dim, double _Mbh, double _Alpha, size_t _Ncoefs_radial, size_t _Ncoefs_angular, const smile::CPointMassSet<float> &points, smile::CDensity::SYMMETRYTYPE _sym, bool checkMonotonic);

CPotentialBSE::CPotentialBSE(unsigned int _N_dim, double _Mbh, double _Alpha, const std::vector< vectord > &coefs):
  CPotentialSH(_N_dim, _Mbh, (assert(coefs.size()>0), static_cast<size_t>(sqrt(coefs[0].size()*1.0)-1))), 
  Ncoefs_radial(std::min<size_t>(MAX_NCOEFS_RADIAL-1, static_cast<size_t>(coefs.size()-1))),
  Alpha(std::max<double>(0.5,_Alpha))  // here Alpha!=0 - no autodetect
{
    if(_Alpha<0.5) { my_error("Error, invalid parameter Alpha="+convertToString(_Alpha)); initDefault(); return; }
    assert(coefs[0].size()==pow_2(Ncoefs_angular+1));
    SHcoefs = coefs;  // do not check array size -- assumed to be done beforehand (in load routine)...
    checkSymmetry();
}

CPotentialBSE::CPotentialBSE(unsigned int _N_dim, double _Mbh, double _Alpha, size_t _Ncoefs_radial, size_t _Ncoefs_angular, 
    const CDensity* density):    // init potential from analytic mass model
  CPotentialSH(_N_dim, _Mbh, _Ncoefs_angular), 
  Ncoefs_radial(std::min<size_t>(MAX_NCOEFS_RADIAL-1, _Ncoefs_radial)),
  Alpha(_Alpha)
{
    mysymmetry=density->symmetry();
    assignlmrange();
    prepareCoefsAnalytic(density);
    checkSymmetry();
}
void CPotentialBSE::initDefault()
{
    Ncoefs_radial=0; 
    Ncoefs_angular=0;
    assignlmrange();
    Alpha=1;
    SHcoefs.resize(1);
    SHcoefs[0].assign(1,1);
}

void CPotentialBSE::checkSymmetry()
{   
    SYMMETRYTYPE sym=ST_SPHERICAL;  // too optimistic:))
    const double MINCOEF=1e-8;
    for(size_t n=0; n<=Ncoefs_radial; n++)
    {
        for(int l=0; l<=(int)Ncoefs_angular; l++)
            for(int m=-l; m<=l; m++)
                if(fabs(SHcoefs[n][l*(l+1)+m])>MINCOEF)  
                {   // nonzero coef.: check if that breaks any symmetry
                    if(l%2==1)  sym = (SYMMETRYTYPE)(sym & ~ST_REFLECTION);
                    if(m<0 || m%2==1)  sym = (SYMMETRYTYPE)(sym & ~ST_PLANESYM);
                    if(m!=0) sym = (SYMMETRYTYPE)(sym & ~ST_ZROTSYM);
                    if(l>0) sym = (SYMMETRYTYPE)(sym & ~ST_SPHSYM);
                }
    }
    // now set all coefs excluded by the inferred symmetry  to zero
    for(size_t n=0; n<=Ncoefs_radial; n++)
    {
        for(int l=0; l<=(int)Ncoefs_angular; l++)
            for(int m=-l; m<=l; m++)
                if( (l>0 && (sym & ST_SPHSYM)) ||
                    (m!=0 && (sym & ST_ZROTSYM)) ||
                    ((m<0 || m%2==1) && (sym & ST_PLANESYM)) ||
                    (l%2==1 && (sym & ST_REFLECTION)) )  
                        SHcoefs[n][l*(l+1)+m] = 0;
    }
    mysymmetry = sym; 
    assignlmrange();
    bool densityNonzero = checkDensityNonzero();
    bool massMonotonic  = checkMassMonotonic();
    if(!massMonotonic || !densityNonzero) 
        my_message(std::string("Warning, ") + 
        (!massMonotonic ? "mass does not monotonically increase with radius" : "") +
        (!massMonotonic && !densityNonzero ? " and " : "") + 
        (!densityNonzero ? "density drops to zero at a finite radius" : "") + "!");
}

/// \cond INTERNAL_DOCS
double intBSE_xi(double xi, void* params)   // integrate over scaled radial variable
{
    if(xi>=1.0 || xi<=-1.0) return 0;
    double Alpha=((CPotentialParamSH*)params)->Alpha;
    double ralpha=(1+xi)/(1-xi);
    double r=pow(ralpha, Alpha);
    int n=((CPotentialParamSH*)params)->sh_n;
    int l=((CPotentialParamSH*)params)->sh_l;
    double phi_nl = gsl_sf_gegenpoly_n(n, (2*l+1)*Alpha+0.5, xi) * pow(r, l) * pow(1+ralpha, -(2*l+1)*Alpha);
    ((CPotentialParamSH*)params)->r=r;
    gsl_function F;
    F.function=&intSH_theta;
    F.params=params;
    double result, error;
    size_t neval;
    gsl_integration_qng(&F, 0, ((CPotentialParamSH*)params)->theta_max, 0, EPSREL_POTENTIAL_INT, &result, &error, &neval);
    return result * phi_nl * 8*M_PI*Alpha *r*r*r/(1-xi*xi);
}
/// \endcond

void CPotentialBSE::prepareCoefsAnalytic(const CDensity* density)
{
    if(Alpha<0.5)
    {  // find best-suited value depending on inner density slope
        double Gamma=density->getGamma();
        Alpha = (Gamma<=1? 1.0+Gamma :
        (Gamma<=1.5? 2.0 :
        (Gamma<1.75? 1./(2-Gamma) : 4.0)));
        if(density->PotentialType()==CDensity::PT_PLUMMER) Alpha=0.5;  // the default value for plummer potential only; however even in this case alpha=1 might perform better..
    }
    SHcoefs.resize(Ncoefs_radial+1);
    for(size_t n=0; n<=Ncoefs_radial; n++)
        SHcoefs[n].assign(pow_2(Ncoefs_angular+1), 0);
    CPotentialParamSH PP;
    PP.P=density;
    PP.Alpha=Alpha;
    PP.theta_max= mysymmetry & ST_REFLECTION ? M_PI_2 : M_PI;  // if symmetries exist, no need to integrate over whole space
    PP.phi_max= mysymmetry & ST_PLANESYM ? M_PI_2 : 2*M_PI;
    int multfactor = (mysymmetry & ST_PLANESYM ? 4 : 1) * (mysymmetry & ST_REFLECTION ? 2 : 1);  // compensates integration of only half- or 1/8-space
    gsl_integration_workspace * ws = gsl_integration_workspace_alloc (1000);
    double interval[2]={-1.0, 1.0};
    gsl_function F;
    F.function=&intBSE_xi;
    F.params=&PP;
    for(size_t n=0; n<=Ncoefs_radial; n++)
        for(int l=0; l<=lmax; l+=lstep)
        {
            PP.sh_n=static_cast<int>(n);  PP.sh_l=l;
            double w=(2*l+1)*Alpha+0.5;
            double Knl = (4*pow_2(n+w)-1)/8/pow_2(Alpha);
            double Inl = Knl * 4*M_PI*Alpha * 
                exp( gsl_sf_lngamma(n+2*w) - 2*gsl_sf_lngamma(w) - gsl_sf_lnfact(PP.sh_n) - 4*w*log(2.0)) / (n+w);
            for(int m=l*mmin; m<=l*mmax; m+=mstep)
            {
                PP.sh_m=m;
                double result, error;
                gsl_integration_qagp(&F, interval, 2, 0, EPSREL_POTENTIAL_INT, 1000, ws, &result, &error);
                SHcoefs[n][l*(l+1)+m] = result * multfactor * (m==0 ? 1 : M_SQRT2) / Inl;
            }
        }
    gsl_integration_workspace_free (ws);
}

template<typename NumT> void CPotentialBSE::prepareCoefsDiscrete(const CPointMassSet<NumT> &points)
{
    SHcoefs.resize(Ncoefs_radial+1);
    for(size_t n=0; n<=Ncoefs_radial; n++)
        SHcoefs[n].assign(pow_2(1+Ncoefs_angular), 0);
    size_t npoints=points.size();
    if(Alpha<0.5)  // autodetect.. the heuristic below is not terribly good, so if you know the inner density slope beforehand, it's better to set alpha=1/(2-gamma) manually
    {
        vectorpd radmass(npoints);
        for(size_t i=0; i<npoints; i++) 
            radmass[i] = paird(sqrt(pow_2(points[i].first.Pos[0])+pow_2(points[i].first.Pos[1])+pow_2(points[i].first.Pos[2])), points[i].second);
        std::sort(radmass.begin(), radmass.end());
        size_t n_est=std::min<size_t>(std::max<size_t>(npoints/50,1000),npoints-1); 
        size_t n_min=n_est/100;
        double inm=0, r_n_est=radmass[n_est].first;
        while(n_min<n_est && radmass[n_min].first<r_n_est*0.03) n_min++;  // neglect some first entries
        gsl_matrix* X=gsl_matrix_alloc(n_est-n_min, 3);  // matrix of coefficients
        gsl_vector* Y=gsl_vector_alloc(n_est-n_min);     // vector of r.h.s. values (mass(r))
        gsl_vector* W=gsl_vector_alloc(n_est-n_min);     // weights 
        for(size_t i=0; i<n_est; i++)
        {
            inm+=radmass[i].second;
            if(i>=n_min){
            gsl_matrix_set(X, i-n_min, 0, 1);  // constant term
            gsl_matrix_set(X, i-n_min, 1, log(radmass[i].first+BH_SMOOTH));  // to avoid log zero values
            gsl_matrix_set(X, i-n_min, 2, radmass[i].first);
            gsl_vector_set(Y, i-n_min, log(inm));
            gsl_vector_set(W, i-n_min, 1/*radmass[i].first/r_n_est*(1-radmass[i].first/r_n_est)*/);  // simple weighting function disfavouring lower and higher radii
            }
        }
        if(n_est-n_min>100)  { // meaningful fit
            // approximate m(r) by power law function:  m(r) = C * r^(3-Gamma) * (1 + k*r), three parameters: C, Gamma, k
            gsl_vector* fit=gsl_vector_alloc(3);
            gsl_matrix* cov=gsl_matrix_alloc(3,3);
            double chisq;
            gsl_multifit_linear_workspace* ws=gsl_multifit_linear_alloc(n_est-n_min, 3);
            gsl_multifit_wlinear(X, W, Y, fit, cov, &chisq, ws);
            gsl_multifit_linear_free(ws);
            double Gamma = 3-gsl_vector_get(fit, 1);   // mass ~ r^(3-gamma)
            Alpha = (Gamma<=0.5? 1.0 :
              (Gamma<=1.5? 0.5+Gamma :
              (Gamma<1.75? 1./(2-Gamma) : 4.0)));
            gsl_vector_free(fit);
            gsl_matrix_free(cov);
#ifdef DEBUGPRINT
            std::cerr << "Gamma_est=" << Gamma << ", alpha=" << Alpha << "\n";
#endif
        }  else Alpha=1.;  // default value
        gsl_matrix_free(X);
        gsl_vector_free(Y);
        gsl_vector_free(W);
    }
    double legendre_array[MAX_NCOEFS_ANGULAR][MAX_NCOEFS_ANGULAR-1];
    double gegenpoly_array[MAX_NCOEFS_RADIAL];
    double Inl[MAX_NCOEFS_RADIAL][MAX_NCOEFS_ANGULAR];
    for(int l=0; l<=lmax; l+=lstep)
    {   // pre-compute coefficients
        double w=(2*l+1)*Alpha+0.5;
        for(unsigned int n=0; n<=Ncoefs_radial; n++)
            Inl[n][l] = 4*M_PI*Alpha * 
              exp( gsl_sf_lngamma(n+2*w) - 2*gsl_sf_lngamma(w) - gsl_sf_lnfact(n) - 4*w*log(2.0)) /
              (n+w) * (4*(n+w)*(n+w)-1)/(8*Alpha*Alpha);
    }
    for(size_t i=0; i<npoints; i++)
    {
        double massi=points[i].second;
        double r=sqrt(pow_2(points[i].first.Pos[0])+pow_2(points[i].first.Pos[1])+pow_2(points[i].first.Pos[2]));
        double costheta= (r==0) ? 0 : (points[i].first.Pos[2]/r);
        double phi=atan2(points[i].first.Pos[1], points[i].first.Pos[0]);
        double ralpha=pow(r, 1/Alpha);
        double xi=(ralpha-1)/(ralpha+1);
        for(int m=0; m<=lmax; m+=mstep)
            gsl_sf_legendre_sphPlm_array(lmax, m, costheta, legendre_array[m]);

        for(int l=0; l<=lmax; l+=lstep)
        {
            double w=(2*l+1)*Alpha+0.5;
            double phil=pow(r, l) * pow(1+ralpha, -(2*l+1)*Alpha);
            gsl_sf_gegenpoly_array(static_cast<int>(Ncoefs_radial), w, xi, gegenpoly_array);
            for(size_t n=0; n<=Ncoefs_radial; n++)
            {
                double mult= massi * gegenpoly_array[n] * phil * 2*M_SQRTPI / Inl[n][l];
                for(int m=0; m<=l*mmax; m+=mstep)
                    SHcoefs[n][l*(l+1)+m] += mult * legendre_array[m][l-m] * cos(m*phi) * (m==0 ? 1 : M_SQRT2);
                if(mmin)
                    for(int m=mmin*l; m<0; m+=mstep)
                        SHcoefs[n][l*(l+1)+m] += mult * legendre_array[-m][l+m] * sin(-m*phi) * M_SQRT2;
            }
        }
    }
}

double CPotentialBSE::Mass(const double r) const
{
    if(r<=0) return 0;
    double ralpha=pow(r, 1/Alpha);
    double xi=(ralpha-1)/(ralpha+1);
    double gegenpoly_array[MAX_NCOEFS_RADIAL];
    gsl_sf_gegenpoly_array(static_cast<int>(Ncoefs_radial), Alpha+0.5, xi, gegenpoly_array);
    double multr = pow(1+ralpha, -Alpha);
    double multdr= -ralpha/((ralpha+1)*r);
    if(gsl_isnan(multdr)) multdr=0;  // safety measure to avoid NaN
    double result=0;
    for(int n=0; n<=static_cast<int>(Ncoefs_radial); n++)
    {
        double dGdr=(n>0 ? (-n*xi*gegenpoly_array[n] + (n+2*Alpha)*gegenpoly_array[n-1])/(2*Alpha*r) : 0);
        result += SHcoefs[n][0] * multr * (multdr * gegenpoly_array[n] + dGdr);
    }
    return -result * r*r;   // d Phi(r)/d r = - G M(r) /r^2
}

void CPotentialBSE::computeSHCoefs(const double r, double coefsF[], double coefsdFdr[], double coefsd2Fdr2[]) const
{
    double ralpha=pow(r, 1/Alpha);
    double xi=(ralpha-1)/(ralpha+1);
    double gegenpoly_array[MAX_NCOEFS_RADIAL];
    if(coefsF)      for(size_t k=0; k<pow_2(Ncoefs_angular+1); k++) coefsF     [k] = 0;
    if(coefsdFdr)   for(size_t k=0; k<pow_2(Ncoefs_angular+1); k++) coefsdFdr  [k] = 0;
    if(coefsd2Fdr2) for(size_t k=0; k<pow_2(Ncoefs_angular+1); k++) coefsd2Fdr2[k] = 0;
    for(int l=0; l<=lmax; l+=lstep)
    {
        double w=(2*l+1)*Alpha+0.5;
        gsl_sf_gegenpoly_array(static_cast<int>(Ncoefs_radial), w, xi, gegenpoly_array);
        double multr = -pow(r, l) * pow(1+ralpha, -(2*l+1)*Alpha);
        double multdr= (l-(l+1)*ralpha)/((ralpha+1)*r);
        if(gsl_isnan(multdr)) multdr=0;  // safety measure to avoid NaN
        for(int n=0; n<=(int)Ncoefs_radial; n++)
        {
            double multdFdr=0, multd2Fdr2=0, dGdr=0;
            if(coefsdFdr!=NULL) {
                dGdr=(n>0 ? (-n*xi*gegenpoly_array[n] + (n+2*w-1)*gegenpoly_array[n-1])/(2*Alpha*r) : 0);
                multdFdr= multdr * gegenpoly_array[n] + dGdr;
                if(coefsd2Fdr2!=NULL)
                    multd2Fdr2 = ( (l+1)*(l+2)*pow_2(ralpha) + 
                                   ((1-2*l*(l+1)) - (2*n+1)*(2*l+1)/Alpha - n*(n+1)/pow_2(Alpha))*ralpha + 
                                   l*(l-1) 
                                 )/pow_2((1+ralpha)*r)*gegenpoly_array[n] - dGdr*2/r;
            }
            for(int m=l*mmin; m<=l*mmax; m+=mstep)
            {
                int indx=l*(l+1)+m;
                double coef = SHcoefs[n][indx] * multr;
                if(coefsF)      coefsF     [indx] += coef * gegenpoly_array[n];
                if(coefsdFdr)   coefsdFdr  [indx] += coef * multdFdr;
                if(coefsd2Fdr2) coefsd2Fdr2[indx] += coef * multd2Fdr2;
            }
        }
    }
}

//----------------------------------------------------------------------------//
// Spherical-harmonic expansion of arbitrary potential, radial part is spline interpolated on a grid

CPotentialSpline::CPotentialSpline(unsigned int _N_dim, double _Mbh, size_t _Ncoefs_radial, size_t _Ncoefs_angular, 
    const CPointMassSetDouble &points, SYMMETRYTYPE _sym, double smoothfactor, const vectord *radii):  // init coefs from point mass set
    CPotentialSH(_N_dim, _Mbh, _Ncoefs_angular),
    Ncoefs_radial(std::max<size_t>(5,_Ncoefs_radial))
{
    mysymmetry=_sym;
    assignlmrange();
    if(points.size() <= Ncoefs_radial*10) { 
        if(points.size()>0)
            my_error("Number of points is too small in Spline potential initialization");
        initDefault(); 
        return; 
    }
    prepareCoefsDiscrete(points, smoothfactor, radii);
}

CPotentialSpline::CPotentialSpline(unsigned int _N_dim, double _Mbh, 
    const vectord &_gridradii, const std::vector< vectord > &_coefs):
    CPotentialSH(_N_dim, _Mbh, (assert(_coefs.size()>0), static_cast<size_t>(sqrt(_coefs[0].size()*1.0)-1))), 
    Ncoefs_radial(std::min<size_t>(MAX_NCOEFS_RADIAL-1, _coefs.size()-1))  // init from existing coefs
{
    assert(_coefs[0].size()==pow_2(Ncoefs_angular+1));
    initSpline(_gridradii, _coefs);
}

CPotentialSpline::CPotentialSpline(unsigned int _N_dim, double _Mbh, size_t _Ncoefs_radial, size_t _Ncoefs_angular, 
    const CDensity* density, const vectord *radii):    // init potential from analytic mass model
    CPotentialSH(_N_dim, _Mbh, _Ncoefs_angular), 
    Ncoefs_radial(std::max<size_t>(5,_Ncoefs_radial))
{
    mysymmetry=density->symmetry();
    assignlmrange();
    prepareCoefsAnalytic(density, radii);
}

// not a constructor, but a method to compute coefs at given radii
CPotentialSpline::CPotentialSpline(size_t _Ncoefs_angular, 
        const CPointMassSetDouble &points, const vectord *radii, std::vector< vectord > *coefsArray):
    CPotentialSH(N_DIM, 0, _Ncoefs_angular),
    Ncoefs_radial(radii->size()-1)
{
    computeCoefsFromPoints(points, radii, NULL, coefsArray);
}

void CPotentialSpline::initDefault()
{
    mysymmetry=ST_SPHERICAL;
    assignlmrange();
    Ncoefs_radial=5; Ncoefs_angular=0;
    vectord nodes(Ncoefs_radial+1);
    std::vector< vectord > values(Ncoefs_radial+1);
    nodes[0]=0;  
    nodes[1]=0.01;
    nodes[2]=0.1;
    nodes[3]=1.0;
    nodes[4]=10.;
    nodes[5]=100;
    for(size_t i=0; i<=5; i++) //default initialization for Dehnen gamma=1 model
        values[i].push_back(-1.0/(nodes[i]+1.0));
    initSpline(nodes, values);
}

CPotentialSpline::~CPotentialSpline()
{
    for(size_t i=0; i<splines.size(); i++)
        if(splines[i]!=NULL)  gsl_spline_free(splines[i]);
}

CPotential* CPotentialSpline::clone() const 
{   // need to duplicate spline arrays explicitly; everything else is copied automatically
    CPotentialSpline* newpot = new CPotentialSpline(*this); 
    for(size_t i=0; i<splines.size(); i++)
        if(splines[i]!=NULL)
        {
            gsl_spline *spline = gsl_spline_alloc (gsl_interp_cspline, splines[i]->size);
            gsl_spline_init (spline, splines[i]->x, splines[i]->y, splines[i]->size);
            newpot->splines[i]=spline;
        }
        else newpot->splines[i]=NULL;
    return newpot;
}

/// \cond INTERNAL_DOCS
double intSpline_r(double r, void* params)
{
    if(r==0) return 0;
    int n=((CPotentialParamSH*)params)->sh_n;  // power index for s^n, either l+2 or 1-l
    double result, error;
    size_t neval;
    ((CPotentialParamSH*)params)->r=r;
    gsl_function F;
    F.function=&intSH_theta;
    F.params=params;
    gsl_integration_qng(&F, 0, ((CPotentialParamSH*)params)->theta_max, 0, EPSREL_POTENTIAL_INT, &result, &error, &neval);
    if(((CPotentialParamSH*)params)->sh_l>2 && fabs(result)<EPSABS_POTENTIAL_INT) 
        result=0;  // to avoid huge drop in performance for very small coef values
    return result * pow(r, n*1.0);
}
/// \endcond

void CPotentialSpline::prepareCoefsAnalytic(const CDensity* density, const vectord *srcradii)
{
    std::vector< vectord > coefsArray(Ncoefs_radial+1);  // SHE coefficients to pass to initspline routine
    vectord radii(Ncoefs_radial+1);  // true radii to pass to initspline routine
    for(size_t i=0; i<=Ncoefs_radial; i++)
        coefsArray[i].assign(pow_2(1+Ncoefs_angular), 0);
    bool initUserRadii = (srcradii!=NULL && srcradii->size()==Ncoefs_radial+1 && srcradii->front()==0);
    if(srcradii!=NULL && !initUserRadii)  // something went wrong with manually supplied radii, initializing by default
        my_error("Warning, invalid call to constructor of Spline potential or SHGrid Schwarzschild model");
    if(initUserRadii)
        radii= *srcradii;
    else
    {
        // find inner/outermost radius
        double totalMass=density->totalMass();
        double epsout = 0.1/sqrt(pow_2(Ncoefs_radial)+0.01*pow(Ncoefs_radial*1.0,4.0));      // how far should be the outer node (leave out this fraction of mass)
        double epsin = 5.0/pow(Ncoefs_radial*1.0,3.0);                                       // how close can we get to zero, in terms of innermost grid node
        double rout = totalMass<0 ? 10 : density->getRadiusByMass(totalMass*(1-epsout));     // totalMass<0 should not really happen! means that mass is infinite
        double rin = std::min<double>(epsin, density->getRadiusByMass(totalMass*epsin*0.1)); // somewhat arbitrary choice for min/max radii, but probably reasonable
        createNonuniformGrid(Ncoefs_radial+1, rin, rout, true, &radii); 
#ifdef DEBUGPRINT
        std::cerr << "Mtotal="<<totalMass << ", rin="<<radii[1]<<", rout="<<rout<<"\n";
#endif
    }
    gsl_integration_workspace * w = gsl_integration_workspace_alloc (1000);
    CPotentialParamSH PP;
    PP.P=density;
    PP.theta_max= mysymmetry & ST_REFLECTION ? M_PI_2 : M_PI;  // if symmetries exist, no need to integrate over whole space
    PP.phi_max= mysymmetry & ST_PLANESYM ? M_PI_2 : 2*M_PI;
    int multfactor = (mysymmetry & ST_PLANESYM ? 4 : 1) * (mysymmetry & ST_REFLECTION ? 2 : 1);  // compensates integration of only half- or 1/8-space
    gsl_function F;
    F.function=&intSpline_r;
    F.params=&PP;
    vectord coefsInner, coefsOuter;
    radii.front()=BH_SMOOTH*radii[1];  // to prevent log divergence for gamma=2 potentials
    for(int l=0; l<=lmax; l+=lstep)
    {
        PP.sh_l=l;
        for(int m=l*mmin; m<=l*mmax; m+=mstep)
        {
            PP.sh_m=m;
            // first precompute inner and outer density integrals at each radial grid point, summing contributions from each interval of radial grid
            coefsInner.assign(Ncoefs_radial+1, 0);
            coefsOuter.assign(Ncoefs_radial+1, 0);
            // loop over inner intervals
            double result, error;
            PP.sh_n = l+2;
            for(size_t c=0; c<Ncoefs_radial; c++)
            {   // the reason we give absolute error threshold is that for flat-density-core profiles the high-l,m coefs at small radii are extremely tiny and their exact calculation is impractically slow
                gsl_integration_qags(&F, radii[c], radii[c+1], EPSABS_POTENTIAL_INT, EPSREL_POTENTIAL_INT, 1000, w, &result, &error);
                coefsInner[c+1] = result + coefsInner[c];
            }
            // loop over outer intervals, starting from infinity backwards
            PP.sh_n = 1-l;
            gsl_integration_qagiu(&F, radii.back(), EPSABS_POTENTIAL_INT, EPSREL_POTENTIAL_INT, 1000, w, &result, &error);
            coefsOuter.back() = result;
            for(size_t c=Ncoefs_radial; c>static_cast<size_t>(l==0 ? 0 : 1); c--)
            {
                gsl_integration_qags(&F, radii[c-1], radii[c], EPSREL_POTENTIAL_INT*fabs(coefsOuter[c]), EPSREL_POTENTIAL_INT, 1000, w, &result, &error);
                coefsOuter[c-1] = result + coefsOuter[c];
            }
            // now compute the coefs of potential expansion themselves
            for(size_t c=0; c<=Ncoefs_radial; c++)
            {
                coefsArray[c][l*(l+1) + m] = ((c>0 ? coefsInner[c]*pow(radii[c], -l-1.0) : 0) + coefsOuter[c]*pow(radii[c], l*1.0)) *
                    multfactor* -4*M_PI/(2*l+1) * (m==0 ? 1 : M_SQRT2);
            }
#ifdef DEBUGPRINT
            std::cerr << "l="<<l<<",m="<<m<<"\n";
#endif
        }
    }
    radii.front()=0;
    gsl_integration_workspace_free (w);
    initSpline(radii, coefsArray);
}

/// \cond INTERNAL_DOCS
// auxiliary structure to sort particles in radius
struct CParticle{
    float r;   // float not double, to save memory
    size_t index;
    CParticle() : r(0), index(0) {}
    CParticle(const std::pair< CPosVelPoint<double>, double> &src, size_t srcindex) :
        r(static_cast<float>(sqrt(pow_2(src.first.Pos[0])+pow_2(src.first.Pos[1])+pow_2(src.first.Pos[2])))),
        index(srcindex) {};
    inline bool operator<(CParticle p) const
    { return (r<p.r); }
};
/// \endcond

void CPotentialSpline::computeCoefsFromPoints(const CPointMassSetDouble &srcpoints, const vectord* srcradii, vectord* outradii, std::vector< vectord > *outcoefs)
{
    assert(outcoefs!=NULL);
    bool initUserRadii = srcradii!=NULL;   // if true, then it was called in the context of SHgrid Schwarzschild model
    std::vector< vectord > *CoefsInner=NULL, *CoefsOuter=NULL;  // coefficients of expansion at each point
    try{
        if(initUserRadii) 
        {
            outcoefs->resize(srcradii->size());
            for(size_t i=0; i<outcoefs->size(); i++)
                outcoefs->at(i).assign(pow_2(Ncoefs_angular+1), 0);
        }
        double legendre_array[MAX_NCOEFS_ANGULAR][MAX_NCOEFS_ANGULAR-1];
        size_t npoints=srcpoints.size();
        if(npoints==0) return;
        std::vector<CParticle> points;
        points.reserve(npoints);
        size_t hadNegativeMass=0;
        for(size_t i=0; i<npoints; i++)
        {
            if(srcpoints[i].second>0)  // don't consider zero-mass points
                points.push_back(CParticle(srcpoints[i], i));
            if(srcpoints[i].second<0) hadNegativeMass++;
        }
        if(hadNegativeMass>0)   // don't fail entirely, but display an error message
            my_error(convertToString(hadNegativeMass)+" particle(s) with negative mass are discarded");
        std::sort(points.begin(), points.end());
        npoints=points.size();   // may be smaller than before if we throw out zero masses
        if(npoints<=Ncoefs_radial*10)
        {
            if(!initUserRadii) my_error("Number of points is too small");
            return;
        }
        // having sorted particles in radius, may now initialize coefs
        if(!initUserRadii)  // will need to output coefs at each particle's radius
        {
            outradii->resize(npoints);
            for(size_t i=0; i<npoints; i++)
                outradii->at(i)=points[i].r;
            outcoefs->resize(pow_2(Ncoefs_angular+1));   // note that array indexing is swapped w.r.t. coefsArray - to save memory by not initializing unnecessary coefs
            CoefsOuter = outcoefs;   // re-use memory allocated for target coefs array  for the intermediate calculations
        }
        else 
            CoefsOuter = new std::vector< vectord> (pow_2(Ncoefs_angular+1));
        CoefsInner = new std::vector< vectord> (pow_2(Ncoefs_angular+1));  
        for(int l=0; l<=lmax; l+=lstep)
            for(int m=l*mmin; m<=l*mmax; m+=mstep)
                CoefsOuter->at(l*(l+1)+m).assign(npoints, 0);  // reserve memory only for those coefs that will be used
        for(int l=0; l<=lmax; l+=lstep)
            for(int m=l*mmin; m<=l*mmax; m+=mstep)
                CoefsInner->at(l*(l+1)+m).assign(npoints, 0);  // yes do it separately from the above, to allow contiguous block of memory to be freed after deleting CoefsInner
        
        double addMbh=0;       // count particles at r=0; in principle, there shouldn't be any, but in case there are, add their mass to the central black hole.
        size_t addNpoints=0;   // note: this assignment of r=0 particles to the black hole will not be reflected on the computed and stored coefficients
        // initialize SH expansion coefs at each point's location
        for(size_t i=0; i<npoints; i++)
        {
            const CPosVelPoint<double> &point = srcpoints[ points[i].index ].first;
            double mass=srcpoints[ points[i].index ].second;
            double costheta = points[i].r>0 ?   // don't use points[i].r because it's float while Pos[] is double -- may end up having cos>1...
                point.Pos[2]/sqrt(pow_2(point.Pos[0])+pow_2(point.Pos[1])+pow_2(point.Pos[2])) : 0;
            if(points[i].r==0) {
                addNpoints++;
                addMbh+=mass;
                mass=0;
            }
            double phi = atan2(point.Pos[1], point.Pos[0]); 
            for(int m=0; m<=lmax; m+=mstep)
                gsl_sf_legendre_sphPlm_array(lmax, m, costheta, legendre_array[m]);
            for(int l=0; l<=lmax; l+=lstep)
                for(int m=l*mmin; m<=l*mmax; m+=mstep)
                {
                    int coefind=l*(l+1)+m;
                    int absm=abs(m);  // negative m correspond to sine, positive - to cosine
                    double mult= -sqrt(4*M_PI)/(2*l+1) * (m==0 ? 1 : M_SQRT2) * mass *
                        legendre_array[absm][l-absm] * (m>=0 ? cos(m*phi) : sin(m*phi));
                    CoefsOuter->at(coefind)[i] = points[i].r>0 ? mult * pow(points[i].r, -(1+l)) : 0;
                    CoefsInner->at(coefind)[i] = mult * pow(points[i].r, l);
                }
        }
        if(addNpoints>0) 
            my_message("Warning, "+convertToString(addNpoints)+" particles at r=0 with total mass "+convertToString(addMbh)+" were added to the central point mass");
        Mbh+=addMbh;
        // sum inner coefs interior and outer coefs exterior to each point's location
        for(int l=0; l<=lmax; l+=lstep)
            for(int m=l*mmin; m<=l*mmax; m+=mstep)
            {
                int coefind=l*(l+1)+m;
                for(size_t i=1; i<npoints; i++)
                    CoefsInner->at(coefind)[i] += CoefsInner->at(coefind)[i-1];
                for(size_t i=npoints-1; i>0; i--)
                    CoefsOuter->at(coefind)[i-1] += CoefsOuter->at(coefind)[i];
            }
        // initialize potential expansion coefs by multiplying inner and outer coefs by r^(-1-l) and r^l, correspondingly
        if(initUserRadii)   // if using manually supplied srcradii, output coefficients only at these radii
        {
            size_t indPoint=0;
            for(size_t indGrid=0; indGrid<srcradii->size(); indGrid++)
            {   // find indPoint so that points[indPoint-1].r <= srcRadii[indGrid] < points[indPoint].r
                double radGrid=srcradii->at(indGrid);
                while(indPoint<npoints && points[indPoint].r<=radGrid) indPoint++;
                for(int l=0; l<=lmax; l+=lstep)
                    for(int m=l*mmin; m<=l*mmax; m+=mstep)
                    {
                        int coefind=l*(l+1)+m;
                        outcoefs->at(indGrid)[coefind] = 
                            (indPoint>0 ? CoefsInner->at(coefind)[indPoint-1] * pow(radGrid, -(1+l)) : 0) + 
                            (indPoint<npoints ? CoefsOuter->at(coefind)[indPoint] * pow(radGrid, l) : 0);
                    }
            }
        } else {   // compute coefs at every particle radius
            for(size_t i=0; i<npoints; i++)
            {
                for(int l=0; l<=lmax; l+=lstep)
                    for(int m=l*mmin; m<=l*mmax; m+=mstep)
                    {
                        int coefind=l*(l+1)+m;
                        outcoefs->at(coefind)[i] = 
                            (i>0 ? CoefsInner->at(coefind)[i-1] * pow(points[i].r, -(1+l)) : 0) + 
                            (i<npoints-1 ? CoefsOuter->at(coefind)[i+1] * pow(points[i].r, l) : 0);
                    }
            }
        }
        if(CoefsOuter!=NULL && initUserRadii) delete CoefsOuter;   // if !initUserRadii, this array actually holds the output coefs so it should not be deleted
        if(CoefsInner!=NULL) delete CoefsInner;
    }
    catch(...)   // there could well be an out-of-memory exception, but for the sake of generality we catch everything including an asteroid hitting earth
    {
        outcoefs->clear();  // marks an error
        if(CoefsInner!=NULL && initUserRadii) delete CoefsInner;   // in any case we need to free temporary arrays
        if(CoefsOuter!=NULL) delete CoefsOuter;
    }
}

/// \cond INTERNAL_DOCS
// auxiliary functions to find outer slope and 1st order correction to inner power-law slope for potential
struct CGammaParam{
    double r1,r2,r3,K,K2,K3;
};
double findGammaOut(double y, void* params)
{
    CGammaParam* par=static_cast<CGammaParam*>(params);
    return( pow(par->r2, 3-y) - pow(par->r1, 3-y))/( pow(par->r3, 3-y) - pow(par->r2, 3-y)) - par->K;
}
double findBcorrIn(double B, void* params)
{
    CGammaParam* par=static_cast<CGammaParam*>(params);
    return (par->K2 - log( (1-B*par->r2)/(1-B*par->r1) ))*log(par->r3/par->r1) - (par->K3 - log( (1-B*par->r3)/(1-B*par->r1) ))*log(par->r2/par->r1);
}
/// \endcond

void CPotentialSpline::prepareCoefsDiscrete(const CPointMassSetDouble &srcpoints, double smoothfactor, const vectord *userradii)
{
  try{

    std::vector<double> pointRadii;      // radii of each point and coefficients of expansion at each point
    std::vector< vectord > pointCoefs;   // note that array indexing is swapped w.r.t. coefsArray - to save memory by not initializing unnecessary coefs
    computeCoefsFromPoints(srcpoints, NULL, &pointRadii, &pointCoefs);
    if(pointCoefs.size()==0) {
        my_error("Not enough memory in computing SH coefs");
        initDefault();
        return;
    }
    size_t npoints=pointRadii.size();
    vectord radii(Ncoefs_radial+1);                         // radii of grid nodes to pass to initspline routine
    std::vector< vectord > coefsArray(Ncoefs_radial+1);     // SHE coefficients to pass to initspline routine
    for(size_t i=0; i<=Ncoefs_radial; i++)
        coefsArray[i].assign(pow_2(Ncoefs_angular+1), 0);

    // choose the radial grid parameters: innermost cell contains minBinMass and outermost radial node should encompass cutoffMass
    // spline definition region extends up to outerRadius which is ~several times larger than outermost radial node, however coefficients at that radius are not used in the potential computation later
    const size_t minBinPoints=10;
    bool initUserRadii = (userradii!=NULL && userradii->size()==Ncoefs_radial+1 && userradii->front()==0 && userradii->at(1)>= pointRadii[minBinPoints]);
    if(userradii!=NULL && !initUserRadii)  // something went wrong with manually supplied radii, initializing by default
        my_error("Warning, invalid radial grid provided to constructor of SHGrid Schwarzschild model");
    if(initUserRadii)
        radii= *userradii;
    else
    {
        size_t npointsMargin    = static_cast<size_t>(sqrt(npoints*0.1));
        size_t npointsInnerGrid = std::max<size_t>(minBinPoints, npointsMargin);    // number of points within 1st grid radius
        size_t npointsOuterGrid = npoints - std::max<size_t>(minBinPoints, npointsMargin);   // number of points within outermost grid radius
        double innerBinRadius = pointRadii[npointsInnerGrid];
        double outerBinRadius = pointRadii[npointsOuterGrid];
        createNonuniformGrid(Ncoefs_radial+1, innerBinRadius, outerBinRadius, true, &radii);
    }
    // find index of the inner- and outermost points which are used in b-spline fitting
    size_t npointsInnerSpline = 0;
    while(pointRadii[npointsInnerSpline]<radii[1]) npointsInnerSpline++;
    npointsInnerSpline = std::min<size_t>(npointsInnerSpline-2, std::max<size_t>(minBinPoints, npointsInnerSpline/2));
    // index of last point used in b-spline fitting (it is beyond outer grid radius, since b-spline definition region is larger than the range of radii for which the spline approximation will eventually be constructed)
    double outerRadiusSpline = pow_2(radii[Ncoefs_radial])/radii[Ncoefs_radial-1];  // roughly equally logarithmically spaced from the last two points
    size_t npointsOuterSpline = npoints-1;
    while(pointRadii[npointsOuterSpline]>outerRadiusSpline) npointsOuterSpline--;
    //!!!FIXME: what happens if the outermost pointRadius is < outerBinRadius ?

    size_t numPointsUsed = npointsOuterSpline-npointsInnerSpline;  // outer and inner points are ignored
    size_t numBSplineKnots = Ncoefs_radial+2;  // including zero and outermost point; only interior nodes are actually used for computing best-fit coefs (except l=0 coef, for which r=0 is also used)
    vectord scaledPointRadii(numPointsUsed), scaledPointCoefs(numPointsUsed);  // transformed x- and y- values of original data points which will be approximated by a spline regression
    vectord scaledKnotRadii(numBSplineKnots), scaledSplineValues;              // transformed x- and y- values of regression spline knots

    // open block so that temp variable "appr" will be destroyed upon closing this block
    {
        // first construct spline for zero-order term (radial dependence)
        potcenter=pointCoefs[0][0];  // value of potential at origin (times 1/2\sqrt{\pi} ?)
        for(size_t p=1; p<=Ncoefs_radial; p++)
            scaledKnotRadii[p] = log(radii[p]);
        scaledKnotRadii[Ncoefs_radial+1] = log(outerRadiusSpline);
        scaledKnotRadii[0] = log(pointRadii[npointsInnerSpline]);
        for(size_t i=0; i<numPointsUsed; i++)
        {
            scaledPointRadii[i] = log(pointRadii[i+npointsInnerSpline]);
            scaledPointCoefs[i] = log(1/(1/potcenter - 1/pointCoefs[0][i+npointsInnerSpline]));
        }
        CSplineApprox appr(scaledPointRadii, scaledKnotRadii);
        if(appr.status()==CSplineApprox::AS_SINGULAR)
            my_message("Warning, in Spline potential initialization: singular matrix for least-square fitting; fallback to a slow algorithm");
        if(appr.status()==CSplineApprox::AS_NOMEM)
        {
            my_error("Not enough memory in performing spline fitting to potential coefficients");
            // in principle we could just pick up coefficients at radii corresponding to grid points, 
            // instead of creating a smoother curve, but this doesn't make much sense anyway - 
            // it's better to reduce the number of input points
            initDefault();
            return;
        }
        appr.loadyvalues(scaledPointCoefs);
        CSplineApprox::APPROXSTATUS status = appr.fitData(0, &scaledSplineValues);
        if(status != CSplineApprox::AS_OK && status != CSplineApprox::AS_SINGULAR)
            scaledSplineValues.assign(Ncoefs_radial+1, 0);   // in case of error fallback to zero values
        // now store fitted values in coefsArray to pass to initspline routine
        coefsArray[0][0] = potcenter;
        for(size_t c=1; c<=Ncoefs_radial; c++)
            coefsArray[c][0] = -1./(exp(-scaledSplineValues[c])-1/potcenter);
    }
    if(lmax>0) {  // construct splines for all l>0 spherical-harmonic terms separately
        // first estimate the asymptotic power-law slope of coefficients at r=0 and r=infinity
        double gammaInner = 2-log((coefsArray[1][0]-potcenter)/(coefsArray[2][0]-potcenter))/log(radii[1]/radii[2]);
        if(gammaInner<0) gammaInner=0; 
        if(gammaInner>2) gammaInner=2;
        // this was the estimate of density slope. Now we need to convert it to the estimate of power-law slope of l>0 coefs
        gammaInner = 2.0-gammaInner;   // the same recipe as used later in initSpline
        double gammaOuter = -1.0;      // don't freak out, assume default value
#ifdef DEBUGPRINT
        std::cerr << "Estimated slope: inner="<<gammaInner<<", outer="<<gammaOuter<<"\n";
#endif
        // init x-coordinates from scaling transformation
        for(size_t p=0; p<=Ncoefs_radial; p++)
            scaledKnotRadii[p] = log(1+radii[p]);
        scaledKnotRadii[Ncoefs_radial+1] = log(1+outerRadiusSpline);
        for(size_t i=0; i<numPointsUsed; i++)
            scaledPointRadii[i] = log(1+pointRadii[i+npointsInnerSpline]);
        CSplineApprox appr(scaledPointRadii, scaledKnotRadii);
        if(appr.status()==CSplineApprox::AS_SINGULAR)
            my_message("Warning, in Spline potential initialization: singular matrix for least-square fitting; fallback to a slow algorithm without smoothing");
        // loop over l,m
        for(int l=lstep; l<=lmax; l+=lstep)
        {
            for(int m=l*mmin; m<=l*mmax; m+=mstep)
            {
                int coefind=l*(l+1) + m;
                // init matrix of values to fit
                for(size_t i=0; i<numPointsUsed; i++)
                    scaledPointCoefs[i] = pointCoefs[coefind][i+npointsInnerSpline]/pointCoefs[0][i+npointsInnerSpline];// *
                    //pow(std::min<double>(1.0, pointRadii[i+npointsInnerSpline]/radii[2]), gammaInner);    // a trick to downplay wild fluctuations at r->0
                appr.loadyvalues(scaledPointCoefs);
                double edf=0;  // equivalent number of free parameters in the fit; if it is ~2, fit is oversmoothed to death (i.e. to a linear regression, which means we should ignore it)
                CSplineApprox::APPROXSTATUS status = appr.fitDataOversmooth(smoothfactor, &scaledSplineValues, NULL, &edf);
                if((status != CSplineApprox::AS_OK && status != CSplineApprox::AS_SINGULAR) || edf<3.0)
                    scaledSplineValues.assign(Ncoefs_radial+1, 0);   // in case of error or an oversmoothed fit fallback to zero values
#ifdef DEBUGPRINT
                std::cerr << "l="<<l <<",m="<<m<<" - EDF="<<edf<<"\n";
#endif
                // now store fitted values in coefsArray to pass to initspline routine
                coefsArray[0][coefind] = 0;  // unused
                for(size_t c=1; c<=Ncoefs_radial; c++)
                    coefsArray[c][coefind] = scaledSplineValues[c] * coefsArray[c][0];  // scale back (multiply by l=0,m=0 coefficient)
                // correction to avoid fluctuation at first and last grid radius
                if( coefsArray[1][coefind] * coefsArray[2][coefind] < 0 || coefsArray[1][coefind]/coefsArray[2][coefind] > pow(radii[1]/radii[2], gammaInner))
                    coefsArray[1][coefind] = coefsArray[2][coefind] * pow(radii[1]/radii[2], gammaInner);   // make the smooth curve drop to zero at least as fast as gammaInner'th power of radius
                if( coefsArray[Ncoefs_radial][coefind] * coefsArray[Ncoefs_radial-1][coefind] < 0 || 
                    coefsArray[Ncoefs_radial][coefind] / coefsArray[Ncoefs_radial-1][coefind] > pow(radii[Ncoefs_radial]/radii[Ncoefs_radial-1], gammaOuter))
                    coefsArray[Ncoefs_radial][coefind] = coefsArray[Ncoefs_radial-1][coefind] * pow(radii[Ncoefs_radial]/radii[Ncoefs_radial-1], gammaOuter);
            }
        }
    }
    initSpline(radii, coefsArray);
    
  }  //< try
  catch(const std::bad_alloc&) {
    my_error("Not enough memory in initializing Spline potential");
    initDefault();
  }
}

void CPotentialSpline::checkSymmetry(const std::vector< vectord > &coefsArray)
{   
    SYMMETRYTYPE sym=ST_SPHERICAL;  // too optimistic:))
    const double MINCOEF=1e-8;   // if ALL coefs of a certain subset of indices are below this value, assume some symmetry
    for(size_t n=0; n<=Ncoefs_radial; n++)
    {
        for(int l=0; l<=(int)Ncoefs_angular; l++)
            for(int m=-l; m<=l; m++)
                if(fabs(coefsArray[n][l*(l+1)+m])>MINCOEF)  
                {   // nonzero coef.: check if that breaks any symmetry
                    if(l%2==1)  sym = (SYMMETRYTYPE)(sym & ~ST_REFLECTION);
                    if(m<0 || m%2==1)  sym = (SYMMETRYTYPE)(sym & ~ST_PLANESYM);
                    if(m!=0) sym = (SYMMETRYTYPE)(sym & ~ST_ZROTSYM);
                    if(l>0) sym = (SYMMETRYTYPE)(sym & ~ST_SPHSYM);
                }
    }
    mysymmetry = sym; 
    assignlmrange();
}

void CPotentialSpline::initSpline(const vectord &radii, const std::vector< vectord > &coefsArray)
{
    if(radii[0]!=0)  my_error("Warning, in Spline potential: radii[0]!=0 (assumed to be zero)");
    if(radii.size()!=Ncoefs_radial+1 || coefsArray.size()!=Ncoefs_radial+1) my_error("Warning, in Spline potential: Ncoefs_radial+1 != array size");  // likely to get segmentation fault shortly soon
    checkSymmetry(coefsArray);   // assign nontrivial symmetry class if some of coefs are equal or close to zero
    gridradii=radii;    // copy real radii
    minr=gridradii[1];
    maxr=gridradii.back();
    std::vector<double> spnodes(Ncoefs_radial);  // scaled radii
    std::vector<double> spvalues(Ncoefs_radial+2);
    splines.assign(pow_2(Ncoefs_angular+1), (gsl_spline*)NULL);
    slopein. assign(pow_2(Ncoefs_angular+1), 1.);
    slopeout.assign(pow_2(Ncoefs_angular+1), -1.);
    potcenter=coefsArray[0][0];    // here assumed that gridradii[0]=0, i.e. innermost node is at origin

    // estimate outermost slope  (needed for accurate extrapolation beyond last grid point)
    gsl_function F;
    CGammaParam PP;
    PP.r1=radii[Ncoefs_radial];
    PP.r2=radii[Ncoefs_radial-1];
    PP.r3=radii[Ncoefs_radial-2];
    PP.K = (coefsArray[Ncoefs_radial][0]*PP.r1 - coefsArray[Ncoefs_radial-1][0]*PP.r2)/(coefsArray[Ncoefs_radial-1][0]*PP.r2 - coefsArray[Ncoefs_radial-2][0]*PP.r3);
    F.function=&findGammaOut;
    F.params=&PP;
    gammaout = findRoot(&F, 3.01, 10., 4.0);
    coefout = (1 - coefsArray[Ncoefs_radial-1][0]*PP.r2/(coefsArray[Ncoefs_radial][0]*PP.r1)) / (pow(PP.r2/PP.r1, 3-gammaout) - 1);
    if(coefout<0) coefout=0;

    // estimate innermost slope with 1st order correction for non-zero radii 
    F.function=&findBcorrIn;
    PP.r1=radii[1];
    PP.r2=radii[2];
    PP.r3=radii[3];
    PP.K2=log((coefsArray[2][0]-potcenter)/(coefsArray[1][0]-potcenter));
    PP.K3=log((coefsArray[3][0]-potcenter)/(coefsArray[1][0]-potcenter));
    double B = findRoot(&F, 0, 0.9/radii[3], 0);
    gammain = 2. - (log((coefsArray[2][0]-potcenter)/(coefsArray[1][0]-potcenter)) - log((1-B*radii[2])/(1-B*radii[1])))/log(radii[2]/radii[1]);
    double gammainuncorr = 2. - log((coefsArray[2][0]-potcenter)/(coefsArray[1][0]-potcenter)) /log(radii[2]/radii[1]);
    if(gammain>=1) gammain=gammainuncorr;
    if(gammain<0) gammain=0; 
    if(gammain>2) gammain=2;
    coefin = (1-coefsArray[1][0]/potcenter)/pow(radii[1],2-gammain);
#ifdef DEBUGPRINT
    std::cerr << "#gammain="<<gammain<<" ("<<gammainuncorr<<");  gammaout="<<gammaout<<"\n";
#endif
    potmax=coefsArray.back()[0];
    potminr=coefsArray[1][0];
    // first init l=0 spline which has radial scaling "log(r)" and nontrivial transformation 1/(1/phi-1/phi0)
    for(size_t i=0; i<Ncoefs_radial; i++)
    {
        spnodes[i] = log(gridradii[i+1]);
        spvalues[i]= log(1/ (1/potcenter - 1/coefsArray[i+1][0]));
    }
    spvalues[Ncoefs_radial] = -(2-gammain)*potcenter/coefsArray[1][0];   // derivative at leftmost node
    spvalues[Ncoefs_radial+1] = - (1+coefout*(3-gammaout))/(1 - potmax/potcenter);  // derivative at rightmost node
    gsl_spline *spline = gsl_spline_alloc (gsl_interp_ccspline, Ncoefs_radial);
    gsl_spline_init (spline, &(spnodes.front()), &(spvalues.front()), Ncoefs_radial);
    splines[0]=spline;
    coef0(maxr, NULL, NULL, &der2out);
    // next init all higher-order splines which have radial scaling log(a+r) and value scaled to l=0,m=0 coefficient
    const double ascale=1.0;
    for(size_t i=0; i<Ncoefs_radial; i++)
        spnodes[i]=log(ascale+gridradii[i+1]);
    double C00val, C00der;
    coef0(minr, &C00val, &C00der, NULL);
    for(int l=lstep; l<=lmax; l+=lstep)
    {
        for(int m=l*mmin; m<=l*mmax; m+=mstep)
        {
            int coefind=l*(l+1)+m;
            for(size_t i=0; i<Ncoefs_radial; i++)
                spvalues[i] = coefsArray[i+1][coefind]/coefsArray[i+1][0];
            slopein[coefind] = log(coefsArray[2][coefind]/coefsArray[1][coefind]) / log(gridradii[2]/gridradii[1]);   // estimate power-law slope of Clm(r) at r->0
            if(gsl_isnan(slopein[coefind])) slopein[coefind]=1.0;  // default
            slopein[coefind] = std::max<double>(slopein[coefind], std::min<double>(l, 2-gammain));  // the asymptotic power-law behaviour of the coefficient expected for power-law density profile
            spvalues[Ncoefs_radial] = spvalues[0] * (1+ascale/minr) * (slopein[coefind] - minr*C00der/C00val);   // derivative at innermost node
            slopeout[coefind] = log(coefsArray[Ncoefs_radial][coefind]/coefsArray[Ncoefs_radial-1][coefind]) / log(gridradii[Ncoefs_radial]/gridradii[Ncoefs_radial-1]) + 1;   // estimate slope of Clm(r)/C00(r) at r->infinity (+1 is added because C00(r) ~ 1/r at large r)
            if(gsl_isnan(slopeout[coefind])) slopeout[coefind]=-1.0;  // default
            slopeout[coefind] = std::min<double>(slopeout[coefind], std::max<double>(-l, 3-gammaout));
            spvalues[Ncoefs_radial+1] = spvalues[Ncoefs_radial-1] * (1+ascale/maxr) * slopeout[coefind];   // derivative at outermost node
            spline = gsl_spline_alloc (gsl_interp_ccspline, Ncoefs_radial);
            gsl_spline_init (spline, &(spnodes.front()), &(spvalues.front()), Ncoefs_radial);
            splines[coefind]=spline;
#ifdef DEBUGPRINT
            std::cerr << "#l="<<l<<",m="<<m<<" - inner="<<slopein[coefind] << ", outer="<<slopeout[coefind] <<"\n";
#endif
        }
    }
    bool densityNonzero = checkDensityNonzero();
    bool massMonotonic  = checkMassMonotonic();
    if(!massMonotonic || !densityNonzero) 
        my_message(std::string("Warning, ") + 
        (!massMonotonic ? "mass does not monotonically increase with radius" : "") +
        (!massMonotonic && !densityNonzero ? " and " : "") + 
        (!densityNonzero ? "density drops to zero at a finite radius" : "") + "!");
}

void CPotentialSpline::getCoefs(vectord *radii, std::vector< vectord > *coefsArray, bool useNodes) const
{
    if(radii==NULL || coefsArray==NULL) return;
    if(useNodes)
    {
        radii->resize(Ncoefs_radial+1);
        for(size_t i=0; i<=Ncoefs_radial; i++)
            (*radii)[i] = gridradii[i];
    }
    size_t numrad=radii->size();
    coefsArray->resize(numrad);
    for(size_t i=0; i<numrad; i++)
    {
        double rad=(*radii)[i];
        double xi=log(1+rad);
        double Coef00;
        coef0(rad, &Coef00, NULL, NULL);
        (*coefsArray)[i].assign(pow_2(Ncoefs_angular+1), 0);
        (*coefsArray)[i][0] = Coef00;
        for(int l=lstep; l<=lmax; l+=lstep)
            for(int m=l*mmin; m<=l*mmax; m+=mstep)
            {
                int coefind=l*(l+1)+m;
                coeflm(coefind, rad, xi, &((*coefsArray)[i][l*(l+1)+m]), NULL, NULL, Coef00);
            }
    }
}

void CPotentialSpline::coeflm(size_t lm, double r, double xi, double *val, double *der, double *der2, double c0val, double c0der, double c0der2) const  // works only for l2>0
{
    const double ascale=1.0;
    double cval=0, cder=0, cder2=0;   // value and derivatives of \tilde Clm = Clm(r)/C00(r)
    if(r < maxr)
    {
        if(r > minr)  // normal interpolation
        {
#ifdef NO_OPTIMIZATION
            cval= gsl_spline_eval(splines[lm], xi, NULL);
            if(der!=NULL){ cder = gsl_spline_eval_deriv(splines[lm], xi, NULL)/(r+ascale);    // to save time, evaluate these only if needed
            if(der2!=NULL) cder2=(gsl_spline_eval_deriv2(splines[lm], xi, NULL)/(r+ascale) - cder)/(r+ascale); } // obviously, if der2!=NULL then assuming that der!=NULL
#else
            // use optimized routines to compute value and derivatives at once
            if(der==NULL) {
                cspline_eval_all(splines[lm], xi, &cval, NULL, NULL);
            } else if(der2==NULL) {
                cspline_eval_all(splines[lm], xi, &cval, &cder, NULL);
                cder /= r+ascale;
            } else {
                cspline_eval_all(splines[lm], xi, &cval, &cder, &cder2);
                cder /= r+ascale;
                cder2 = (cder2/(r+ascale) - cder)/(r+ascale);
            }
#endif
        }
        else  // power-law asymptotics at r<minr
        {
            cval = splines[lm]->y[0] * potminr;
            if(val!=NULL)  *val = cval * pow(r/minr, slopein[lm]);
            if(der!=NULL){ *der = (*val) * slopein[lm]/r;
            if(der2!=NULL) *der2= (*der) * (slopein[lm]-1)/r; }
            return;   // for r<minr, Clm is not scaled by C00
        }
    }
    else  // power-law asymptotics for r>maxr
    {
        cval = splines[lm]->y[splines[lm]->interp->size-1] * pow(r/maxr, slopeout[lm]);
        cder = cval * slopeout[lm]/r;
        cder2= cder * (slopeout[lm]-1)/r;
        double ximax=splines[lm]->x[splines[lm]->interp->size-1];
        double der2left = (gsl_spline_eval_deriv2(splines[lm], ximax, NULL) - gsl_spline_eval_deriv(splines[lm], ximax, NULL))/pow_2(r+ascale);
        double der2right = splines[lm]->y[splines[lm]->interp->size-1]*slopeout[lm]*(slopeout[lm]-1)/pow_2(maxr);
        double acorr = (der2left-der2right)*0.5;
        double slopecorr = slopeout[lm]-4;
        double powcorr = pow(r/maxr, slopecorr);
        cval += acorr*powcorr*pow_2(r-maxr);
        cder += acorr*powcorr*(r-maxr)*(2 + slopecorr*(r-maxr)/r);
        cder2+= acorr*powcorr*(2 + 4*slopecorr*(r-maxr)/r + pow_2(1-maxr/r)*slopecorr*(slopecorr-1));
    }
    // scale by C00
    if(val!=NULL)  *val = cval*c0val;
    if(der!=NULL)  *der = cder*c0val + cval*c0der;
    if(der2!=NULL) *der2= cder2*c0val + 2*cder*c0der + cval*c0der2;
}

void CPotentialSpline::coef0(double r, double *val, double *der, double *der2) const  // works only for l=0
{
    if(r<=maxr)
    {
        double logr=log(r);
        double sval, sder, sder2;
        if(r<minr)
        {
            double ratio = 1-coefin*pow(r, 2-gammain);  // C00(r)/C00(0)
            sval = log(-potcenter/coefin) - (2-gammain)*logr + log(ratio);
            sder = -(2-gammain)/ratio;
            sder2= -pow_2(sder)*(1-ratio);
        }
        else
        {
#ifdef NO_OPTIMIZATION
            sval = gsl_spline_eval(splines[0], logr, NULL);
            sder = gsl_spline_eval_deriv(splines[0], logr, NULL);
            sder2= gsl_spline_eval_deriv2(splines[0], logr, NULL);
#else
            cspline_eval_all(splines[0], logr, &sval, &sder, &sder2);
#endif
        }
        double sexp = (r>0? exp(-sval) : 0);
        double vval = 1./(sexp-1/potcenter);
        if(val!=NULL)  *val = -vval;
        if(der!=NULL)  *der = -vval*vval*sexp/r * sder;  // this would not work for r=0 anyway...
        if(der2!=NULL) *der2= -pow_2(vval/r)*sexp * (sder2 - sder + sder*sder*(2*vval*sexp-1) );
    }
    else
    {
        double r_over_maxr_g=pow(r/maxr, 3-gammaout);
        double der2right = -2*potmax*maxr/pow(r,3) * (1 - coefout*(r_over_maxr_g*(gammaout-1)*(gammaout-2)/2 - 1));
        double slopecorr = -gammaout-4;
        double acorr = 0*pow(r/maxr, slopecorr) * (der2out-der2right)*0.5;   //!!! why 0?
        if(val!=NULL)  *val = -(-potmax*maxr/r * (1 - coefout*(r_over_maxr_g-1))  + acorr*pow_2(r-maxr) );
        if(der!=NULL)  *der = -( potmax*maxr/r/r * (1 - coefout*(r_over_maxr_g*(gammaout-2) - 1))  + acorr*(r-maxr)*(2 + slopecorr*(r-maxr)/r) );
        if(der2!=NULL) *der2= -(der2right  + acorr*(2 + 4*slopecorr*(r-maxr)/r + pow_2(1-maxr/r)*slopecorr*(slopecorr-1)) );
    }
}

double CPotentialSpline::Mass(const double r) const
{
    if(r<=0) return 0;
    double der;
    coef0(r, NULL, &der, NULL);
    return der * r*r;   // d Phi(r)/d r = G M(r) /r^2
}

void CPotentialSpline::computeSHCoefs(const double r, double coefsF[], double coefsdFdr[], double coefsd2Fdr2[]) const
{
    const double ascale=1.0;
    double xi = log(r+ascale);
    double val00, der00, der200;
    coef0(r, &val00, &der00, &der200);  // compute value and two derivatives of l=0,m=0 spline
    if(coefsF)    coefsF[0]    = val00;
    if(coefsdFdr) coefsdFdr[0] = der00;
    if(coefsd2Fdr2) coefsd2Fdr2[0] = der200;
    for(int l=lstep; l<=lmax; l+=lstep)
    {
        for(int m=l*mmin; m<=l*mmax; m+=mstep)
        {
            int coefind=l*(l+1)+m;
            coeflm(coefind, r, xi, 
                coefsF!=NULL ? coefsF+coefind : NULL, 
                coefsdFdr!=NULL ? coefsdFdr+coefind : NULL, 
                coefsd2Fdr2!=NULL ? coefsd2Fdr2+coefind : NULL, 
                val00, der00, der200);
        }
    }
}

//----------------------------------------------------------------------------//
// Scale-free potential 

double CPotentialScaleFree::Rho(double X, double Y, double Z, double /*t*/) const
{
    return pow(pow_2(X) + pow_2(Y/q) + pow_2(Z/p), -Gamma/2);
}

/// \cond INTERNAL_DOCS
struct CPotentialParamSF2{
    double theta, phi;
    double p, q, Gamma;
};
struct CPotentialParamSF3{
    const CPotentialScaleFree* P;
    double X, Y, Z;
    double p, q, Gamma;
};

double intPhiSF_angular(double s, void* params)
{
    double theta= ((CPotentialParamSF2*)params)->theta;
    double phi  = ((CPotentialParamSF2*)params)->phi;
    double p= ((CPotentialParamSF2*)params)->p;
    double q= ((CPotentialParamSF2*)params)->q;
    double Gamma= ((CPotentialParamSF2*)params)->Gamma;
    return 2*pow(pow_2(s)*( pow_2(sin(theta))*(pow_2(cos(phi)) + pow_2(sin(phi))/(1-(1-pow_2(q))*pow_2(s))) + pow_2(cos(theta))/(1-(1-pow_2(p))*pow_2(s)) ), 1-Gamma/2) /
        sqrt( (1-(1-pow_2(q))*pow_2(s)) * (1-(1-pow_2(p))*pow_2(s)));
}

double intForceSFx(double s, void* params)
{
    double X= ((CPotentialParamSF3*)params)->X;
    double Y= ((CPotentialParamSF3*)params)->Y;
    double Z= ((CPotentialParamSF3*)params)->Z;
    double p= ((CPotentialParamSF3*)params)->p;
    double q= ((CPotentialParamSF3*)params)->q;
    double Gamma= ((CPotentialParamSF3*)params)->Gamma;
    return pow(s, 2-Gamma) * pow( pow_2(X) + pow_2(Y)/(1-(1-pow_2(q))*pow_2(s)) + pow_2(Z)/(1-(1-pow_2(p))*pow_2(s)), -Gamma/2) /
        sqrt( (1-(1-pow_2(q))*pow_2(s)) * (1-(1-pow_2(p))*pow_2(s)));
}
double intForceSFy(double s, void* params)
{
    double X= ((CPotentialParamSF3*)params)->X;
    double Y= ((CPotentialParamSF3*)params)->Y;
    double Z= ((CPotentialParamSF3*)params)->Z;
    double p= ((CPotentialParamSF3*)params)->p;
    double q= ((CPotentialParamSF3*)params)->q;
    double Gamma= ((CPotentialParamSF3*)params)->Gamma;
    return pow(s, 2-Gamma) * pow( pow_2(X) + pow_2(Y)/(1-(1-pow_2(q))*pow_2(s)) + pow_2(Z)/(1-(1-pow_2(p))*pow_2(s)), -Gamma/2) /
        sqrt( (1-(1-pow_2(q))*pow_2(s)) * (1-(1-pow_2(p))*pow_2(s))) / (1-(1-pow_2(q))*pow_2(s));
}
double intForceSFz(double s, void* params)
{
    double X= ((CPotentialParamSF3*)params)->X;
    double Y= ((CPotentialParamSF3*)params)->Y;
    double Z= ((CPotentialParamSF3*)params)->Z;
    double p= ((CPotentialParamSF3*)params)->p;
    double q= ((CPotentialParamSF3*)params)->q;
    double Gamma= ((CPotentialParamSF3*)params)->Gamma;
    return pow(s, 2-Gamma) * pow( pow_2(X) + pow_2(Y)/(1-(1-pow_2(q))*pow_2(s)) + pow_2(Z)/(1-(1-pow_2(p))*pow_2(s)), -Gamma/2) /
        sqrt( (1-(1-pow_2(q))*pow_2(s)) * (1-(1-pow_2(p))*pow_2(s))) / (1-(1-pow_2(p))*pow_2(s));
}
/// \endcond 

double CPotentialScaleFree::Phi(double X, double Y, double Z, double /*t*/) const
{
    double result, error;
    size_t neval;
    CPotentialParamSF2 PP;
    PP.theta=atan2(sqrt(pow_2(X)+pow_2(Y)), Z); 
    PP.phi=atan2(Y, X);
    PP.p=p; PP.q=q; PP.Gamma=Gamma;
    gsl_function F;
    F.function=&intPhiSF_angular;
    F.params=&PP;
    gsl_integration_qng (&F, 0, 1, 0, EPSREL_POTENTIAL_INT, &result, &error, &neval); 
    return result * 2*M_PI/(2-Gamma)*q*p * pow(pow_2(X)+pow_2(Y)+pow_2(Z),1-Gamma/2) - ((Mbh>0)?(Mbh/sqrt(pow_2(BH_SMOOTH) + pow_2(X) + pow_2(Y) + pow_2(Z))):0);
}

void CPotentialScaleFree::DiffEq(unsigned n, double /*t*/, const double v[], double *f) const
{
    bool calcLyapunov=n>2*N_DIM;
    double X=v[0];
    double Y=v[1];
    double Z=(N_dim==3)?v[2]:0;    // for universality: if only 2 dimensions, Z=0 will add nothing
    f[0] = v[N_dim];
    f[1] = v[N_dim+1];
    if(N_dim==3)
        f[2] = v[N_dim+2];
    double mult=-4*M_PI*q*p;
    double r3=pow(pow_2(BH_SMOOTH) + pow_2(X) + pow_2(Y) + pow_2(Z), 1.5);
    double result, error; 
    size_t neval;
    CPotentialParamSF3 PP;
    PP.X=X; PP.Y=Y; PP.Z=Z;
    PP.p=p; PP.q=q; PP.Gamma=Gamma;
    gsl_function F;
    F.params=&PP;
    F.function=&intForceSFx;
    gsl_integration_qng (&F, 0, 1, 0, EPSREL_POTENTIAL_INT, &result, &error, &neval); 
    f[N_dim]   = mult * X * result - ((Mbh>0)?(Mbh*X/r3):0);
    F.function=&intForceSFy;
    gsl_integration_qng (&F, 0, 1, 0, EPSREL_POTENTIAL_INT, &result, &error, &neval); 
    f[N_dim+1] = mult * Y * result - ((Mbh>0)?(Mbh*Y/r3):0);
    if((N_dim==3))
    {
        F.function=&intForceSFz;
        gsl_integration_qng (&F, 0, 1, 0, EPSREL_POTENTIAL_INT, &result, &error, &neval); 
        f[N_dim+2] = mult * Z * result - ((Mbh>0)?(Mbh*Z/r3):0);
    }
    if(calcLyapunov)
    {   // variational equation - not implemented in the scale-free model; may use Scale-free SH for that purpose
        f[N_dim*2]   = 0;
        f[N_dim*2+1] = 0;
        if(N_dim==3) f[N_dim*2+2]=0;
        f[N_dim*3]   = 0;
        f[N_dim*3+1] = 0;
        if(N_dim==3) f[N_dim*3+2]=0;
    }
}

//----------------------------------------------------------------------------//
// scale-free potential - spherical harmonic expansion
CPotentialScaleFreeSH::CPotentialScaleFreeSH(unsigned int _N_dim, double q, double p, double _Mbh, double _Gamma, size_t _Ncoefs_angular): 
CPotentialSH(_N_dim, _Mbh, _Ncoefs_angular), Gamma(_Gamma)
{
    mysymmetry = (q==1 ? (p==1 ? ST_SPHERICAL : ST_AXISYMMETRIC) : ST_TRIAXIAL);
    assignlmrange();
    prepareCoefsAnalytic(q, p);
}

CPotentialScaleFreeSH::CPotentialScaleFreeSH(unsigned int _N_dim, double _Mbh, double _Gamma, const vectord &_coefs): 
CPotentialSH(_N_dim, _Mbh, static_cast<size_t>(sqrt(_coefs.size()*1.0)-1)), Gamma(_Gamma)
{
    assert(_coefs.size()==pow_2(Ncoefs_angular+1));
    SHcoefs=_coefs;  // no array size check performed!! 
    mysymmetry=ST_TRIAXIAL;  // no check performed if it is the case
}

/// \cond INTERNAL_DOCS
struct CPotentialParamSFSH{
    int sh_l, sh_m;
    double costheta;
    double p, q, Gamma;
    double phi_max;
};
double intSF_phi(double phi, void* params)
{
    double costheta= ((CPotentialParamSFSH*)params)->costheta;
    size_t sh_m=((CPotentialParamSFSH*)params)->sh_m;
    double q=((CPotentialParamSFSH*)params)->q;
    double p=((CPotentialParamSFSH*)params)->p;
    double Gamma=((CPotentialParamSFSH*)params)->Gamma;
    return pow((1-pow_2(costheta))*(1+(1/pow_2(q)-1)*pow_2(sin(phi))) + pow_2(costheta/p), -Gamma/2) * cos(sh_m*phi);
}
double intSF_theta(double theta, void* params)
{
    double result, error;
    double costheta=cos(theta);
    ((CPotentialParamSFSH*)params)->costheta=costheta;
    int sh_l=((CPotentialParamSFSH*)params)->sh_l;
    int sh_m=((CPotentialParamSFSH*)params)->sh_m;
    gsl_function F;
    F.function=&intSF_phi;
    F.params=params;
    gsl_integration_workspace * w = gsl_integration_workspace_alloc (1000);
    gsl_integration_qag(&F, 0, ((CPotentialParamSFSH*)params)->phi_max, 0, EPSREL_POTENTIAL_INT, 1000, 3, w, &result, &error);
    gsl_integration_workspace_free (w);
    return result * sin(theta) * gsl_sf_legendre_sphPlm(sh_l, sh_m, costheta) / (2*M_SQRTPI);
}
/// \endcond

void CPotentialScaleFreeSH::prepareCoefsAnalytic(double q, double p)
{   // this function assumes that potential in question has triaxial symmetry, unlike the more general function for BSE potential
    SHcoefs.assign(pow_2(Ncoefs_angular+1), 0);
    CPotentialParamSFSH PP;
    PP.q=q; PP.p=p;
    PP.Gamma=std::max<double>(Gamma,1e-5);  // HACK to avoid singularity in gamma=0 coefficient computation
    //gsl_integration_workspace * w = gsl_integration_workspace_alloc (1000);
    gsl_function F;
    F.function=&intSF_theta;
    F.params=&PP;
    double theta_max= mysymmetry & ST_REFLECTION ? M_PI_2 : M_PI;  // if symmetries exist, no need to integrate over whole space
    PP.phi_max= mysymmetry & ST_PLANESYM ? M_PI_2 : 2*M_PI;
    int multfactor = (mysymmetry & ST_PLANESYM ? 4 : 1) * (mysymmetry & ST_REFLECTION ? 2 : 1);  // compensates integration of only half- or 1/8-space
    for(int l=0; l<=lmax; l+=lstep)
        for(int m=l*mmin; m<=l*mmax; m+=mstep)
        {
            double result, error;
            size_t neval;
            PP.sh_l=l; PP.sh_m=m;
            gsl_integration_qng(&F, 0, theta_max, 0, EPSREL_POTENTIAL_INT, &result, &error, &neval);
            SHcoefs[l*(l+1)+m] = multfactor*result * 4*M_PI/(l+3-Gamma)/(2-l-Gamma) * (m==0 ? 1 : M_SQRT2);
        }
    //gsl_integration_workspace_free (w);
}

void CPotentialScaleFreeSH::computeSHCoefs(const double r, double coefsF[], double coefsdFdr[], double coefsd2Fdr2[]) const
{
    double multr=pow(r, 2-Gamma);
    for(int l=0; l<=lmax; l+=lstep)
    {
        for(int m=l*mmin; m<=l*mmax; m+=mstep)
        {
            int indx=l*(l+1)+m;
            double coef = SHcoefs[indx] * multr;
            if(coefsF)      coefsF     [indx] = coef;
            if(coefsdFdr)   coefsdFdr  [indx] = coef * (2-Gamma)/r;
            if(coefsd2Fdr2) coefsd2Fdr2[indx] = coef * (2-Gamma)*(1-Gamma)/(r*r);
        }
    }
}

//----------------------------------------------------------------------------//
// Frozen-N-body potential (adapted from Barnes&Hut treecode from 'hackforce')
CPotentialNB::CPotentialNB(unsigned int _N_dim, double _Mbh) :
    CPotential(_N_dim, _Mbh), eps(0), tol(0)
{ 
    maketree();
}

template<typename NumT> CPotentialNB::CPotentialNB(
unsigned int _N_dim, double _Mbh, double _eps, double _tol, const CPointMassSet<NumT> &points):
    CPotential(_N_dim, _Mbh), eps(_eps), tol(_tol)
{
    if(N_dim==N_DIM)
    {
        bodytab.resize(points.size());
        for(size_t i=0; i<points.size(); i++)
        {
            bodytab[i].type=BODY;
            bodytab[i].mass=points[i].second;
            bodytab[i].pos[0]=points[i].first.Pos[0];
            bodytab[i].pos[1]=points[i].first.Pos[1];
            bodytab[i].pos[2]=points[i].first.Pos[2];
        }
    }
    maketree();  // should be called anyway
}
// explicit instantiations of constructor for two variants of template argument
template CPotentialNB::CPotentialNB(unsigned int _N_dim, double _Mbh, double _eps, double _tol, const CPointMassSet<double> &points);
template CPotentialNB::CPotentialNB(unsigned int _N_dim, double _Mbh, double _eps, double _tol, const CPointMassSet<float > &points);

double CPotentialNB::Phi(double X, double Y, double Z, double /*t*/) const
{
    double phi=0;
    vec3 point={X,Y,Z};
    walktree(troot, pow_2(rsize), point, CALCPHI, &phi);
    return -phi - ((Mbh>0)?(Mbh/sqrt(pow_2(BH_SMOOTH) + pow_2(X) + pow_2(Y) + pow_2(Z))):0) - ((Omega>0)?pow_2(Omega)*(pow_2(X)+pow_2(Y))/2:0);
}
/// \cond INTERNAL_DOCS
bool comparepd(paird elem1, paird elem2)
{
    return(elem1.first<elem2.first);
}
/// \endcond
double CPotentialNB::Rho(double X, double Y, double Z, double /*t*/) const
{
    vectorpd vec;
    vec3 point={X,Y,Z};
    // first determine which cell does this point belong to, and its size (which will be used as an estimate of neighbour search radius)
    double searchrad=rsize;
    nodeptr *qptr=(nodeptr*)&troot;
    int xp[N_DIM];
    if(!intcoord(xp, point)) return 0;   // if point is outside root cell, density is zero
    int l = IMAX >> 1;                /* start with top bit       */
    while (*qptr != NULL && l>0) {  /* loop descending tree     */
        if ((*qptr)->type == BODY) {/*   reached a "leaf"?      */
            l=0;
        }
        else{
            qptr = &(((cellptr)(*qptr))->subp[subindex(xp, l)]);    /*   move down one level    */
            l = l >> 1;                /*   and test next bit      */
            searchrad/=2;
        }
    }
    searchrad *= 4;  // to have somewhat more than one neighbour!
    // now look for neighbours inside searchrad
    size_t numneighbour=0;
    do{ // recursively scan tree and collect the list of particles closer to our position than the search radius
        vec.clear();
        walktreedens(troot, rsize*rsize, searchrad*searchrad, point, xp, IMAX >> 1, &vec);
        numneighbour=vec.size();
        searchrad*=2;
    } while(numneighbour<TREECODE_DENSITY_KTH_NEIGHBOUR);  // if the number of particles was too low, repeat the search
    std::sort(vec.begin(), vec.end(), comparepd);
    size_t kth=std::min<size_t>(vec.size(), TREECODE_DENSITY_KTH_NEIGHBOUR);
    if(kth<3) return 0; // failure to get density
    double mass=0;
    double smoothrad=vec[kth-1].first;
    for(size_t k=0; k<vec.size(); k++)
    {
        double radrel=vec[k].first/smoothrad;  // smoothing radius is the distance to Kth neighbour
        double weight=0;  
#if TREECODE_DENSITY_SPLINE
        if(radrel<0.5) weight=8-48*radrel*radrel+48*radrel*radrel*radrel;  // use spline weighting function with compact support
        else if(radrel<1) weight=16*(1-radrel)*(1-radrel)*(1-radrel);
#else
        if(radrel<1) weight=15.0/8*(1-radrel*radrel);  // use simpler Ferrer F1 kernel
#endif
        mass += vec[k].second*weight;
        if(weight==0) k=vec.size(); // all other neighbours are farther than spline support
    }
    mass /= (M_PI*pow(smoothrad, 3.0));  // for spline
    return mass;
}

void CPotentialNB::DiffEq(unsigned n, double /*t*/, const double v[], double *f) const
{
    double X=v[0];
    double Y=v[1];
    double Z=v[2];
    f[0] = v[3];
    f[1] = v[4];
    f[2] = v[5];
    f[3]=f[4]=f[5]=0;
    f[N_DIM*2]=rsize*rsize;  // init min distance (squared)
    if(n<2*N_DIM) return;   // do not live in Flatland...
    vec3 point={X,Y,Z};
    walktree(troot, pow_2(rsize), point, CALCACC, f+N_DIM);  // recursively scan tree from the root and compute accelerations and distance to the nearest particle
    f[2*N_DIM]=sqrt(f[2*N_DIM]);  // this holds minimum distance (unsoftened) to nearest particle (un-squared:)
    if(Mbh>0 || Omega>0)
    {
        double r3=pow(pow_2(BH_SMOOTH) + pow_2(X) + pow_2(Y) + pow_2(Z), 1.5);
        f[N_DIM]   += - ((Mbh>0)?(Mbh*X/r3):0) - ((Omega>0)?  2*Omega*v[N_dim+1]+pow_2(Omega)*X : 0) ;
        f[N_DIM+1] += - ((Mbh>0)?(Mbh*Y/r3):0) - ((Omega>0)? -2*Omega*v[N_dim]  +pow_2(Omega)*Y : 0) ;
        f[N_DIM+2] += - ((Mbh>0)?(Mbh*Z/r3):0);
    }
}

double CPotentialNB::Mass(const double r) const
{
    double m=0, r2=r*r;
    for(size_t i=0; i<bodytab.size(); i++)
        if(pow_2(bodytab[i].pos[0])+pow_2(bodytab[i].pos[1])+pow_2(bodytab[i].pos[2]) < r2)
            m+=bodytab[i].mass;
    return m;
}

void CPotentialNB::maketree()    // initialize tree from set of bodies
{
    int celltabsize = (int) (100+bodytab.size()*0.52);  // should be enough, but if it was not, repeat the tree construction with larger allocated number of cells
    bool result=true; 
    do{
        troot=NULL;
        for(unsigned int d=0; d<N_DIM; d++) rmin[d]=-2.0;
        rsize=4.0;  // init box size
        celltab.clear();
        celltab.reserve(celltabsize); 
        result=true; 
        size_t b=0;
        for(; b<bodytab.size() && result; b++)
        {
            result = expandbox(b);   /* expand root to fit */
            result&= appendtree(b);  /* insert into tree   */
        }
        if(!result)
        {
            if(celltab.size()>=celltab.capacity() && b>bodytab.size()/2)  // error due to insufficient allocation of celltab
            {
                celltabsize = (int) (celltab.capacity() *1.05* bodytab.size()/b); // increase it and repeat
                my_error("Warning, restarting tree initialization because of insufficient initial allocation of celltab");
            }
            else
            {
                troot=NULL;
                celltab.clear();
                my_error("Error (unknown) in tree construction, perhaps pathological mass distribution");
                return;  // unknown error...
            }
        }
    } while(!result);

    centerofmass(troot, rmin, rsize,0);                /* find c-of-m coordinates  */
    
    double avgmass=troot->mass/bodytab.size();   // average particle mass (to assign variable softening length depending on mass)
    if(eps<0) {  // need to initialize each particle's softening length according to local density
#if TREECODE_EPS_FROM_DENSITY
        // this is more exact method, based on local density, but much more costly
        for(size_t i=0; i<bodytab.size(); i++)
        {
            double dens=Rho(bodytab[i].pos[0], bodytab[i].pos[1], bodytab[i].pos[2]);
            bodytab[i].eps = fabs(eps)*sqrt(bodytab[i].mass/avgmass) * pow(dens/avgmass, -1.0/3);
        }
#else
        // this is a more approximate method, based on just cell sizes, with iterative averaging of epsilons for nearby nodes in the tree
        // it is a lot faster (at least unless the density computation is made substantially more efficient) and gives slightly larger spread in epsilon
        // compared to the density-based method; however, as the effect of spatially adaptive softening depends only weakly on the softening parameter,
        // at least for the (recommended) Epanechnikov kernel, this approximation should be sufficient for practical purposes.
        for(int iter=0; iter<10; iter++)   // iterations to reduce 'binning' effect when eps ~ size of the smallest cell containing this one particle
            assigneps(troot, 0, rsize);    // this computes the approximation to local softening length for each point (stored is eps_i=|eps|*n^{-1/3})
        for(size_t i=0; i<bodytab.size(); i++)   // now adjust these values for variation with particle mass
            bodytab[i].eps = sqrt(bodytab[i].mass/avgmass) * bodytab[i].eps;
#endif
    } else {  // no spatial variation of softening length, but retain variation with particle mass
        for(size_t b=0; b<bodytab.size(); b++)
            bodytab[b].eps = eps*sqrt(bodytab[b].mass/avgmass);
    }
    propagateeps(troot);   // initialize cells' epsilon by averaging from underlying nodes
    
}

bool CPotentialNB::expandbox(const size_t b)  // expand bounding box if needed to accomodate for a new body
{
    int xtmp[N_DIM];
    while (! intcoord(xtmp, bodytab[b].pos)) {  /* expand box (rarely)      */
       vec3 rmid;
       for(unsigned int d=0; d<N_DIM; d++)
            rmid[d]=rmin[d]+0.5*rsize;          /*   find box midpoint      */
        for(unsigned int d=0; d<N_DIM; d++)     /*   loop over dimensions   */
            if (bodytab[b].pos[d] < rmid[d])    /*     is p left of mid?    */
                rmin[d] -= rsize;               /*       extend to left     */
        rsize = 2.0 * rsize;                    /*   double length of box   */
        if (troot != NULL) {                    /*   repot existing tree?   */
            cell c;                             /*     create new root cell */
            int xmid[N_DIM];
            if(!intcoord(xmid, rmid)) return false;   /*     locate old root cell */
            int s = subindex(xmid, IMAX >> 1);  /*     find old tree index  */
            c.subp[s] = troot;                  /*     graft old on new     */
            if(celltab.size()>=celltab.capacity()) return false;  /* cannot reallocate vector, otherwise all pointers will be invalidated */
            celltab.push_back(c);
            troot = (nodeptr) (&(celltab.back()));             /*     plant new tree       */
        }
    }
    return true;
}

bool CPotentialNB::appendtree(const size_t b)            // append a body to the tree, creating new cell if necessary
{
    int xp[N_DIM], xq[N_DIM];
    nodeptr *qptr;

    intcoord(xp, bodytab[b].pos);   /* form integer coords      */
    int l = IMAX >> 1;              /* start with top bit       */
    qptr = &troot;                  /* start with tree root     */
    while (*qptr != NULL) {         /* loop descending tree     */
        if(l==0) {                  /*   dont run out of bits   */
            (*qptr)->mass += bodytab[b].mass;  /* if two particles coincide, just sum up their mass */
            return true;
        }
        if ((*qptr)->type == BODY) {/*   reached a "leaf"?      */
            cell c;                 /*     alloc a new cell     */
            if(!intcoord(xq, (*qptr)->pos)) return false;   /*     get integer coords   */
            c.subp[subindex(xq, l)] = *qptr;    /*     put body in cell     */
            if(celltab.size()>=celltab.capacity()) return false;  /* cannot reallocate vector, otherwise all pointers will be invalidated */
            celltab.push_back(c);
            *qptr = (nodeptr) (&(celltab.back()));        /*     link cell in tree    */
        }
        qptr = &(((cellptr)(*qptr))->subp[subindex(xp, l)]);    /*   move down one level    */
        l = l >> 1;                /*   and test next bit      */
    }
    *qptr = (nodeptr) (&(bodytab[b]));            /* found place, store p     */
    return true;
}

bool CPotentialNB::intcoord(    // compute integer coords, return false if point is out of bounds for current root cell
            int xp[N_DIM],   /* integerized coordinate vec3 [0,IMAX) */
            const vec3 rp)   const   /* double coordinate vec3 (system coords) */
{
    bool inb = true;                    /* use to check bounds      */
    for (unsigned int k = 0; k < N_DIM; k++) {        /* loop over dimensions     */
        double xsc = (rp[k] - rmin[k]) / rsize; /*   scale to range [0,1)   */
        if (0.0 <= xsc && xsc < 1.0)            /*   within unit interval?  */
            xp[k] = (int) floor(IMAX * xsc);    /*     then integerize      */
        else                                    /*   out of range           */
            inb = false;                        /*     then remember that   */
    }
    return inb;
}

int CPotentialNB::subindex(   // determine the subcell that a given position belongs to
           const int x[N_DIM],        /* integerized coordinates of particle */
           const int l)     const      /* current level of tree */
{
    int i = 0;                                  /* sum index in i           */
    for (unsigned int k = 0; k < N_DIM; k++)             /* check each dimension     */
        if (x[k] & l)                           /*   if beyond midpoint     */
            i += NSUB >> (k + 1);               /*     skip over subcells   */
    return (i);
}

void CPotentialNB::centerofmass(const nodeptr q, const vec3 corner, const double cellsize, int l)      // recursive center-of-mass finder, also initializes rmax and quadrupole moment
{
    if(q==NULL) return;
    if(q->type == CELL) {                      /* is this a cell?          */
        q->mass = 0.0;                          /*   init total mass        */
        for(unsigned int d=0; d<N_DIM; d++) q->pos[d]=0; /*   and c. of m.           */
        for(unsigned int i=0; i< NSUB; i++) {            /*   loop over subcells     */
            vec3 subcellcorner;
            nodeptr r = ((cellptr)q)->subp[i];
            if(r != NULL) {                    /*     does subcell exist?  */
                if(r->type==BODY) 
                    ((cellptr)q)->numbody++;   
                else
                {
                    double halfsize=cellsize/2;
                    for(unsigned int d=0; d<N_DIM; d++)
                        subcellcorner[d]=corner[d] + halfsize*((i & (NSUB >> (d+1))) > 0);
                    centerofmass(r, subcellcorner, halfsize, l+1);  /*       find subcell cm    */
                    ((cellptr)q)->numbody+=((cellptr)r)->numbody;   /*  count total number of bodies */
                }
                q->mass += r->mass;             /*       sum total mass     */
                for(unsigned int d=0; d<N_DIM; d++)      /*   sum c.o.m. coordinates */
                    q->pos[d] += r->mass * r->pos[d];
            }
        }
        if(q->mass>0)
            for(unsigned int d=0; d<N_DIM; d++)
                q->pos[d] /= q->mass;
        // now we have center of mass coordinates, may compute quadrupole moment and other stuff
#if TREECODE_QUADRUPOLE
        for(unsigned int d1=0; d1<N_DIM; d1++)
            for(unsigned int d2=0; d2<N_DIM; d2++)
                ((cellptr)q)->quad[d1][d2]=0;
        for(unsigned int i = 0; i < NSUB; i++) {        /*   loop over subnodes     */
            nodeptr r = ((cellptr)q)->subp[i];
            if (r != NULL) {                    /*     does subnode exist?  */
                vec3 dr={r->pos[0]-q->pos[0], r->pos[1]-q->pos[1], r->pos[2]-q->pos[2]};  /* displacement vect. */
                for(unsigned int d1=0; d1<N_DIM; d1++)
                    for(unsigned int d2=0; d2<N_DIM; d2++)
                    {
                        ((cellptr)q)->quad[d1][d2] += dr[d1]*dr[d2] * r->mass;
                        if(r->type==CELL)       /*       if subnode is cell */
                            ((cellptr)q)->quad[d1][d2] += ((cellptr)r)->quad[d1][d2];  /*         use its moment   */
                    }
            }
        }
#endif
#if TREECODE_DEHNEN_MAC
        double rmax1=0, rmax2=0;
        // loop over sub-nodes
        for (unsigned int i = 0; i < NSUB; i++) {        /*   loop over subnodes     */
            nodeptr r = ((cellptr)q)->subp[i];
            if (r != NULL) {            /*     does subnode exist?  */
                vec3 dr={r->pos[0]-q->pos[0], r->pos[1]-q->pos[1], r->pos[2]-q->pos[2]};  /* displacement vect. */
                double drmag=sqrt(pow_2(dr[0])+pow_2(dr[1])+pow_2(dr[2]));
                rmax1 = std::max<double>(rmax1, (r->type==BODY ? 0 : ((cellptr)r)->rmax) + drmag );
            }
        }
        // find distance from this cell's c.o.m. to its most distant corner
        for(unsigned int c=0; c<NSUB; c++)  // loop over corners
        {
            double dist2corner=0;
            for(unsigned int d=0; d<N_DIM; d++)
                dist2corner += pow_2(corner[d] + ((c & (NSUB >> (d+1))) > 0)*cellsize - q->pos[d]);
            rmax2 = std::max<double>(rmax2, sqrt(dist2corner));
        }
        // assign cell's rmax to lower of these two values (Dehnen 2000)
        ((cellptr)q)->rmax=std::min<double>(rmax1, rmax2);
#if TREECODE_QUADRUPOLE
        // check that quadrupole terms are not too large, otherwise increase rmax (a modification to Dehnen's criterion to combat worst-case configurations)
        //double quadfact = (((cellptr)q)->quad[0][0]+((cellptr)q)->quad[1][1]+((cellptr)q)->quad[2][2])/(q->mass*pow_2(((cellptr)q)->rmax));
        //if(quadfact>0.3)  ((cellptr)q)->rmax *= sqrt(quadfact/0.3);
#endif
#else
        ((cellptr)q)->rmax=cellsize;   // use Barnes&Hut cell opening criterion based on geometric size of the cell only
#endif
    }
}

/** recursive assign of softening length: 
    compute the number density of points in a given cell and average it with the parent cell,
    repeat process for the entire tree is scanned several times to perform back and forth 
    averaging over parent/daughter nodes.
    eps_i = |eps| * n^{-1/3}, where eps is the global softening parameter and n is 
    the average number density in the vicinity of the node.
    epsparent is the softening length of the parent cell (zero on the first iteration).
*/
void CPotentialNB::assigneps(const nodeptr q, const double epsparent, const double cellsize)
{
    if(q==NULL) return;
    double epsavg=0; int numsub=0;
    // if on the first pass, compute epsilon from the number density of particles in this cell, otherwise don't re-assign it
    double epscur = q->eps>0 ? q->eps : fabs(eps)*cellsize/(q->type==BODY ? 1.0 : pow(((cellptr)q)->numbody*1.0, 1.0/3) );
    if(epsparent>0)
        epscur = sqrt(epsparent*epscur);     // geometric average with the parent cell's eps (which is an averaged value for all daughter nodes)
    q->eps=epscur;   // assign some preliminary guess
    if(q->type==BODY) 
        return;  // no descendants
    for(unsigned int i=0; i< NSUB; i++) {   // this node is a cell -- loop over daughter nodes
        nodeptr r = ((cellptr)q)->subp[i];
        if (r != NULL) {
            assigneps(r, ((cellptr)q)->eps, cellsize/2);  // descend one level deeper into the tree
            epsavg += pow(r->eps, -1./2);                 // add averaged value (weighted with some power, index chosen empirically!)
            numsub++;
        }
    }
    q->eps = pow(epsavg/numsub, -2./1);  // compute averaged value of non-empty daughter nodes' mean number density; 
    // note that for cells it is no longer the mean density within the cell, but rather the mean density averaged over its non-empty nodes
    // otherwise we will end up having the root cell with a very low average density, which will then be spread down into its daughter cells via iterative averaging.
}

void CPotentialNB::propagateeps(const nodeptr q)   // recursively compute averaged epsilons for tree cells from its child nodes
{
    if(q==NULL) return;
    double epsavg=0;
    for(unsigned int i=0; i< NSUB; i++) {    // assign eps to tree cells by averaging the values in sub-cells/leafs
        nodeptr r = ((cellptr)q)->subp[i];
        if (r != NULL) {
            if(r->type==CELL) propagateeps(r);
            epsavg += r->mass * pow(r->eps, -2.);  // add averaged value, weighted with the node mass and an empirically chosen power
        }
    }
    q->eps = pow(epsavg/q->mass, -0.5);   // the softening for a tree cell is the average over its non-empty subcells
}

void CPotentialNB::walktree(const nodeptr p,   // pointer to interacting node or cell 
           const double dsq,                   // cell size squared
           const vec3 pos0,                    // position to calculate quantity in question
           const WALKOP operation,             // operation to perform during tree-walk
           double data[])  const               // data accumulated according to the operation
{
    if(p==NULL) return;
    vec3 dr={p->pos[0]-pos0[0], p->pos[1]-pos0[1], p->pos[2]-pos0[2]};
    double distsq=pow_2(dr[0])+pow_2(dr[1])+pow_2(dr[2]);
    // test whether cell should be opened
    if(p->type==CELL && tol*tol*distsq < pow_2(((cellptr)p)->rmax) ) {
        for(unsigned int c=0; c<NSUB; c++)
            if(((cellptr)p)->subp[c] != NULL)  /* walk into sub-cells */
                walktree(((cellptr)p)->subp[c], dsq/4.0, pos0, operation, data);
    }
    else if(distsq>0) 
    {   // perform requested operation on leaf or unopened cell
        // assign softened distance, based either on Plummer softening or on F1 compact kernel;  
        // we need to compute the gravitation field (including quadrupole) of the softened particle,
        // so we compute additional quantities g1-g3 (notation is taken from Springel et al.2001)
#if TREECODE_EPS_COMPACT_KERNEL
        double eps2=p->eps*p->eps, eps3=eps2*p->eps;    /* use Epanechnikov kernel with compact support */
        double dist_sq_over_eps = p->eps==0?2 : distsq/eps2;
        double dist_inv_eps = dist_sq_over_eps>=1 ? 1/sqrt(distsq) : (15./8 - 5./4*dist_sq_over_eps + 3./8*pow_2(dist_sq_over_eps) )/p->eps;
        double g1 = dist_sq_over_eps>=1 ? dist_inv_eps/distsq : (2.5-1.5*dist_sq_over_eps)/eps3;
        double g2 = dist_sq_over_eps>=1 ? -3*g1/distsq : -3/(eps2*eps3);
        double g3 = dist_sq_over_eps>=1 ? -5*g2/distsq : 0;
#else
        double dist_sq_eps = distsq + pow_2(p->eps);    /* use Plummer softening */
        double dist_inv_eps = 1/sqrt(dist_sq_eps);
        double g1 = dist_inv_eps/dist_sq_eps;
        double g2 = -3*g1/dist_sq_eps;
        double g3 = -5*g2/dist_sq_eps;
#endif
#if TREECODE_QUADRUPOLE
        vec3 quaddr;
        double drquaddr=0, quadtrace=0;
        if(p->type == CELL){        /* if cell, add quad. term  */
            quadtrace=( ((cellptr)p)->quad[0][0]+((cellptr)p)->quad[1][1]+((cellptr)p)->quad[2][2] );
            for(unsigned int d1=0; d1<N_DIM; d1++)
            {
                quaddr[d1] = 
                    ((cellptr)p)->quad[d1][0] * dr[0] + 
                    ((cellptr)p)->quad[d1][1] * dr[1] + 
                    ((cellptr)p)->quad[d1][2] * dr[2];
                drquaddr += quaddr[d1]*dr[d1];
            }
        }
#endif
        switch(operation){
            case CALCPHI: 
                data[0] += p->mass*dist_inv_eps; 
#if TREECODE_QUADRUPOLE
                data[0] -= 0.5*(g2*drquaddr + g1*quadtrace);
#endif
                break;
            case CALCACC: 
                for(unsigned int d=0; d<N_DIM; d++)
                    data[d] += p->mass*dr[d] * g1;
#if TREECODE_QUADRUPOLE
                if(p->type == CELL) {        /* if cell, add quad. term  */
                    for(unsigned int d=0; d<N_DIM; d++)
                        data[d] += g2*quaddr[d] + 0.5*(g3*drquaddr+g2*quadtrace)*dr[d];
                }
#endif
                if(distsq<data[N_DIM]) data[N_DIM]=distsq;   // update minimum distance
                //data[N_DIM]=std::min<double>(data[N_DIM], 1/pow_2(dist_inv_eps));
                break;
        }
    }
}

void CPotentialNB::walktreedens(const nodeptr p,   // pointer to interacting node or cell 
           const double dsq,      // cell size squared
           const double searchradsq, // search radius squared
           const vec3 pos0,       // position to calculate quantity in question
           const int xp[N_DIM],   // integerized coordinates of the point in question
           const int lev,         // current grid level
           vectorpd* data)  const // array to store neighbours
{
    if(p==NULL) return;
    vec3 dr={p->pos[0]-pos0[0], p->pos[1]-pos0[1], p->pos[2]-pos0[2]};
    double distsq=pow_2(dr[0])+pow_2(dr[1])+pow_2(dr[2]);
    if(p->type==CELL) {  
        if(distsq < dsq*3 + searchradsq) {  // worst case - CoM may lie on the other side of diagonal
            for(unsigned int c=0; c<NSUB; c++)
                if(((cellptr)p)->subp[c] != NULL)  /* walk into sub-cells */
                    walktreedens(((cellptr)p)->subp[c], dsq/4, searchradsq, pos0, xp, lev>>1, data);
        }
        else {  // point is far from our CoM, but lies within one of child cells, which still needs to be opened
            int subcell=subindex(xp, lev);
            if(((cellptr)p)->subp[subcell] != NULL)  /* walk into this sub-cell */
                walktreedens(((cellptr)p)->subp[subcell], dsq/4, searchradsq, pos0, xp, lev>>1, data);
        }
    }
    else if (distsq < searchradsq)   // this is a leaf within search radius
        data->push_back(paird(sqrt(distsq),p->mass));
}

//----------------------------------------------------------------------------//
// simple density models, without potential
double CDensityPlummer::Rho(double X, double Y, double Z, double /*t*/) const
{
    double m = sqrt(pow_2(X) + pow_2(Y/q) + pow_2(Z/p));
    return 3/(4*M_PI*p*q) * pow(1+m*m, -2.5);
}

double CDensityIsochrone::Rho(double X, double Y, double Z, double /*t*/) const
{
    double m = sqrt(pow_2(X) + pow_2(Y/q) + pow_2(Z/p));
    double a=sqrt(m*m+1);
    return (3*(1+a)*a*a-m*m*(1+3*a))/(4*M_PI*p*q*pow(a*(1+a),3));
}

double CDensityPerfectEllipsoid::Rho(double X, double Y, double Z, double /*t*/) const
{
    double m = sqrt(pow_2(X) + pow_2(Y/q) + pow_2(Z/p));
    return 1/(M_PI*M_PI*p*q*pow_2(1+m*m));
}

//----------------------------------------------------------------------------//
/// \cond INTERNAL_DOCS
// some more machinery for NFW potential with a gradual cutoff in radius
double findNFWrcutoff(double r, void* params)
{
    return 4*M_PI*(r!=1 ? r*(r*log(r)-r+1)/pow_2(r-1) : 0.5) - *((double*)params);
}
/// \endcond
CDensityNFW::CDensityNFW(double _q, double _p, double c):
  q(_q), p(_p)
{
    if(c<1.0) c=1.0;
    double totalMass=(4*M_PI*(log(1+c)-c/(c+1)));
    /*  "c" is the standard NFW concentration, defined as the ratio of virial radius to scale radius, 
        and assuming that the density drops instantly to zero beyond that radius.
        We introduce another radius "rcutoff" such that (1) density drops more gradually (as r^-4) 
        beyond rcutoff, and (2) total mass is the same as for truncated NFW model with given concentration.
        Now the task is to find rcutoff. */
    gsl_function F;
    F.function=&findNFWrcutoff;
    F.params=&totalMass;
    rcutoff = findRoot(&F, 0.25, c);
#ifdef DEBUGPRINT
    std::cerr << "rcutoff="<<rcutoff<<"\n";
#endif
    norm = 1.0/(totalMass*p*q);
}

double CDensityNFW::Rho(double X, double Y, double Z, double /*t*/) const
{
    double m = sqrt(pow_2(X) + pow_2(Y/q) + pow_2(Z/p));
    return norm / (m * pow_2(1.0 + m) * (1 + m/rcutoff));
}

//----------------------------------------------------------------------------//
// even more machinery for Ellipsoidal mass model with varying axis ratios and user-specified cumulative mass profile
CDensityEllipsoidal::CDensityEllipsoidal(const vectord &_radii, const vectord &_inmass, const vectord &_axesradii, const vectord &_qvals, const vectord &_pvals)
{
    qvals=pvals=dens=NULL;
    if(_radii.size()<=2 || _radii.size()!=_inmass.size() || _axesradii.size()!=_qvals.size() || _axesradii.size()!=_pvals.size())
    {
        my_message("Error, input arrays are not equal in length");
        return;
    }
    CMassModel* massModel = new CMassModel(_radii, _inmass);
    if(massModel->getStatus()!=CMassModel::MS_OK && massModel->getStatus()!=CMassModel::MS_DF_NEGATIVE)
    { 
        my_message("Error in creating spherical mass model!");
        return; 
    }
    // get values of density at the grid in radius and construct our own spline (for efficiency)
    vectord radii, densval;
    for(size_t i=0; i<_radii.size(); i++)
        if(_radii[i]>0) {
            radii.push_back(log(_radii[i]));
            densval.push_back(log(massModel->dens(_radii[i])));
        }
    dens = gsl_spline_alloc(gsl_interp_cspline, radii.size());
    gsl_spline_init (dens, &(radii.front()), &(densval.front()), densval.size());
    delete massModel;  // not needed anymore - we have our own spline
    // check if all values of q, p are equal
    if(_axesradii.size()==0) {
        q=1; p=1; return;
    }
    q=_qvals[0]; p=_pvals[0];
    for(size_t i=1; i<_axesradii.size() && q>0; i++)
        if(_qvals[i]!=q || _pvals[i]!=p) q=0;
    if(p>0) return;  // no need to initialize splines -- constant axis ratios
    // initialize radius-dependent axis ratio
    vectord axesradii, axesq, axesp;
    for(size_t i=0; i<_axesradii.size(); i++)
        if(_axesradii[i]>0) {
            axesradii.push_back(log(_axesradii[i]));
            axesq.push_back(_qvals[i]);
            axesp.push_back(_pvals[i]);
        }
    axesq.push_back(0); axesq.push_back(0);  // two additional values of zero to specify spline endpoint derivatives
    qvals = gsl_spline_alloc(gsl_interp_ccspline, axesradii.size());
    gsl_spline_init (qvals, &(axesradii.front()), &(axesq.front()), axesradii.size());
    axesp.push_back(0); axesp.push_back(0);
    pvals = gsl_spline_alloc(gsl_interp_ccspline, axesradii.size());
    gsl_spline_init (pvals, &(axesradii.front()), &(axesp.front()), axesradii.size());
}

CDensityEllipsoidal::~CDensityEllipsoidal()
{
    if(dens!=NULL)  delete dens;
    if(qvals!=NULL) delete qvals;
    if(pvals!=NULL) delete pvals;
}

/*CDensityEllipsoidal::Mass(const double r) const
{
    return massModel->mass(r);  // spherically averaged approximation
}*/

double CDensityEllipsoidal::Rho(double X, double Y, double Z, double /*t*/) const
{
    if((X==0&&Y==0&&Z==0) || dens==NULL) return 0;
    double qs, ps;
    if(p>0) { qs=q; ps=p; }  // simplification for constant axis ratio
    else {  // find from spline interpolation
        double rslog=log(X*X+Y*Y+Z*Z)/2;  // used to find q and p
        if(rslog<=qvals->x[0]) { qs = qvals->y[0]; ps = pvals->y[0]; } else
        if(rslog>=qvals->x[qvals->size-1]) { qs = qvals->y[qvals->size-1]; ps = pvals->y[qvals->size-1]; } else
        {
            qs = gsl_spline_eval(qvals, rslog, 0);
            ps = gsl_spline_eval(pvals, rslog, 0);
        }
    }
    double qp=pow(qs*ps, -1.0/3);
    qs*=qp; ps*=qp;   // scale axis ratio so that their product is 1
    double logr = log(pow_2(X/qp)+pow_2(Y/qs)+pow_2(Z/ps))/2.0;
    return exp(gsl_spline_evalx(dens, logr, NULL));
}

//----------------------------------------------------------------------------//
// Multi-gaussian expansion of density profile
CDensityMGE::CDensityMGE(const smile::CDensityMGE::vectorg &_data) : data(_data)
{
    // check if all values of q, p are equal
    if(data.empty()) { q=p=1; return; }
    q=data[0].q; p=data[0].p;
    for(size_t i=1; i<data.size(); i++)
        if(data[i].q!=q || data[i].p!=p) q=0;
}

double CDensityMGE::Rho(double X, double Y, double Z, double /*t*/) const
{
    double result=0;
    for(size_t i=0; i<data.size(); i++)
    {
        double r2=pow_2(X) + pow_2(Y/data[i].q) + pow_2(Z/data[i].p);
        result += data[i].mass/(pow(2*M_PI*pow_2(data[i].width),1.5)*data[i].p*data[i].q)
            * exp(-r2/(2*pow_2(data[i].width)));
    }
    return result;
}

//----------------------------------------------------------------------------//
// 'class factory':)  (makes correspondence between enum potential and symmetry types and string names)
PotentialNameMapType PotentialNames;
DensityNameMapType DensityNames;
SymmetryNameMapType SymmetryNames;

/// create a correspondence between names and enum identifiers for potential, density and symmetry types
bool initPotentialAndSymmetryNameMap()
{
    PotentialNames.clear();
    PotentialNames[CDensity::PT_LOG] = CPotentialLog::myName();
    PotentialNames[CDensity::PT_HARMONIC] = CPotentialHarmonic::myName();
    PotentialNames[CDensity::PT_DEHNEN] = CPotentialDehnen::myName();
    PotentialNames[CDensity::PT_SCALEFREE] = CPotentialScaleFree::myName();
    PotentialNames[CDensity::PT_SCALEFREESH] = CPotentialScaleFreeSH::myName();
    PotentialNames[CDensity::PT_BSE] = CPotentialBSE::myName();
    PotentialNames[CDensity::PT_SPLINE] = CPotentialSpline::myName();
    PotentialNames[CDensity::PT_NB] = CPotentialNB::myName();

    // list of density models available for BSE and Spline approximation
    DensityNames.clear();
    DensityNames[CDensity::PT_NB] = CPotentialNB::myName();   // denotes not the actual tree-code potential, but rather the fact that the density model comes from discrete points in Nbody file
    DensityNames[CDensity::PT_ELLIPSOIDAL] = CDensityEllipsoidal::myName();
    DensityNames[CDensity::PT_MGE] = CDensityMGE::myName();
    DensityNames[CDensity::PT_COEFS] = "Coefs";  // denotes that potential expansion coefs are loaded from a text file rather than computed from a density model
    DensityNames[CDensity::PT_DEHNEN] = CPotentialDehnen::myName();
    DensityNames[CDensity::PT_PLUMMER] = CDensityPlummer::myName();
    DensityNames[CDensity::PT_PERFECTELLIPSOID] = CDensityPerfectEllipsoid::myName();
    DensityNames[CDensity::PT_ISOCHRONE] = CDensityIsochrone::myName();
    DensityNames[CDensity::PT_NFW] = CDensityNFW::myName();

    SymmetryNames[CDensity::ST_NONE]="None";
    SymmetryNames[CDensity::ST_REFLECTION]="Reflection";
    SymmetryNames[CDensity::ST_TRIAXIAL]="Triaxial";
    SymmetryNames[CDensity::ST_AXISYMMETRIC]="Axisymmetric";
    SymmetryNames[CDensity::ST_SPHERICAL]="Spherical";
    
    return rand()==42;
}
bool mapinitializer = initPotentialAndSymmetryNameMap();  // this weird move ensures that the function be implicitly called upon program start

CDensity::POTENTIALTYPE getPotentialTypeByName(std::string PotentialName)
{
    for(PotentialNameMapType::const_iterator iter=PotentialNames.begin(); iter!=PotentialNames.end(); iter++)
        if(iter->second==PotentialName) return iter->first;
    return CDensity::PT_UNKNOWN;
}

CDensity::POTENTIALTYPE getDensityTypeByName(std::string DensityName)
{
    for(DensityNameMapType::const_iterator iter=DensityNames.begin(); iter!=DensityNames.end(); iter++)
        if(iter->second==DensityName) return iter->first;
    return CDensity::PT_UNKNOWN;
}

CDensity::SYMMETRYTYPE getSymmetryTypeByName(std::string SymmetryName)
{
    for(SymmetryNameMapType::const_iterator iter=SymmetryNames.begin(); iter!=SymmetryNames.end(); iter++)
        if(iter->second==SymmetryName) return iter->first;
    return CDensity::ST_TRIAXIAL;
}

}