/** \file    potential.h
    \brief   Classes for computing potential, forces and density for various mass models
    \author  EV
    \date    2009-2013

    This file defines several classes of two kind:
    - density profiles (descendants of smile::CDensity class), which provide only density 
      as a function of coordinates, the corresponding potential needs not be known;
    - potential solvers (derived from smile::CPotential class, which inherits CDensity), 
      which in addition compute potential and forces, and provide several utility functions 
      like finding an intersection with equipotential surface.

    There are several flexible general-purpose potential representations, which can take an arbitrary 
    density profile as an input; moreover there are at least two flexible density parametrizations suitable 
    to represent almost anything one may need. 
    If, nevertheless, you want to add a new density model, the necessary steps are summarized below:
    - Add new identifier to CDensity::POTENTIALTYPE enum;
    - Implement a new class derived from CDensity, which provides the density profile in the virtual function 
      Rho(x,y,z), and, optionally, returns the asymptotic power-law slope at r->0 in getGamma();
      also it should return correct PotentialType and name;
    - Add item to DensityNames list in initPotentialAndSymmetryNameMap(); 
    - Add corresponding line to CSmileCore::initPotential() in core.cpp under the section PT_BSE, PT_SPLINE;
    - If the density model has additional parameters besides axis ratio, one needs also to modify 
      CSmileCore::loadSettings()/saveSettings() in core.cpp and potentialChanged() in gui.cpp to show/hide 
      and interactively modify these parameters, and probably additional visual elements in GUI (smile.ui)
**/
#pragma once
#include "common.h"
#include <gsl/gsl_spline.h>
#include <map>

namespace smile{

/** Base class defining density model (without corresponding potential). 
    May be used for initializing density for general-purpose potential approximations (BSE and Spline). 
**/
class CDensity
{
public:
    /// list of all existing types of density or density/potential models, each of them implemented in its own class
enum POTENTIALTYPE   
{
    PT_UNKNOWN,     ///< means that potential will be initialized by default, currently as PT_LOG
    PT_COEFS,       ///< not an actual density model, but a way to load pre-computed coefficients of potential expansion
    PT_NB,          ///< a set of frozen particles:  CPotentialNB
    PT_BSE,         ///< basis-set expansion:  CPotentialBSE
    PT_SPLINE,      ///< spline spherical-harmonic expansion:  CPotentialSpline
    PT_LOG,         ///< logaritmic potential:  CPotentialLog
    PT_HARMONIC,    ///< simple harmonic oscillator:  CPotentialHarmonic
    PT_SCALEFREE,   ///< single power-law density profile:  CPotentialScaleFree
    PT_SCALEFREESH, ///< spherical-harmonic approximation to a power-law density:  CPotentialScaleFreeSH
    PT_DEHNEN,      ///< Dehnen(1993) density model:  CPotentialDehnen
    PT_PLUMMER,     ///< Plummer model:  CDensityPlummer
    PT_ISOCHRONE,   ///< isochrone model:  CDensityIsochrone
    PT_PERFECTELLIPSOID,  ///< Kuzmin/de Zeeuw integrable potential:  CDensityPerfectEllipsoid
    PT_NFW,         ///< Navarro-Frenk-White profile:  CDensityNFW
    PT_ELLIPSOIDAL, ///< a generalization of spherical mass profile with arbitrary axis ratios:  CDensityEllipsoidal
    PT_MGE          ///< Multi-Gaussian expansion:  CDensityMGE
};
    /// Type of symmetry. 
    /// used mainly in spherical-harmonic expansion potentials to compute and use only 
    /// a subset of coefficients based on the symmetry properties of the underlying potential)
enum SYMMETRYTYPE{     
    ST_NONE = 0,       ///< all coefficients should be considered
    ST_REFLECTION = 1, ///< reflection about origin (change of sign of all coordinates simultaneously) - terms with odd l are zero
    ST_PLANESYM = 2,   ///< reflection about principal planes (change of sign of any coordinate) - no terms with sin(m phi) or odd m
    ST_ZROTSYM = 4,    ///< axisymmetric model (all terms with m>0 are zero)
    ST_SPHSYM = 8,     ///< spherical symmetry (all terms with l>0 are zero)
    ST_TRIAXIAL = ST_REFLECTION | ST_PLANESYM,    ///< triaxial symmetry (only even l,m terms are used)
    ST_AXISYMMETRIC = ST_TRIAXIAL | ST_ZROTSYM,   ///< axial symmetry (only m=0 terms are used for even l)
    ST_SPHERICAL = ST_AXISYMMETRIC | ST_SPHSYM    ///< spherical symmetry (only l=0 terms)
};
    CDensity() {};
    virtual ~CDensity() {};
    virtual POTENTIALTYPE PotentialType() const=0;  ///< enumerable potential type
    virtual std::string PotentialName() const=0;    ///< string representation of potential type
    virtual SYMMETRYTYPE symmetry() const=0;        ///< returns symmetry type of this potential
    virtual double Rho(double X, double Y, double Z, double t=0) const=0;    ///< returns density at given coordinates, this should obviously be overriden in derivative classes
    virtual double Mass(const double r) const;      ///< returns mass inside given radius (approximately! not necessary to integrate density over sphere, just a rough estimate used e.g. in choosing radial nodes of Schwarzschild grid). Excluding black hole!
    double totalMass() const;                       ///< returns estimated M(r=infinity) or -1 if mass is infinite
    double getRadiusByMass(const double m) const;   ///< solves for Mass(r)=m
    bool checkMassMonotonic() const;                ///< safety measure: check (roughly) that mass is increasing with radius
    bool checkDensityNonzero() const;               ///< another safety measure: check that density doesn't drop to zero along any of three axes (important to assess spherical-harmonic approximation quality)
    virtual double getGamma() const;                ///< returns inner density slope estimate (only used in BSE potential expansion for the automatic selection of shape parameter Alpha)
private:
    CDensity& operator=(const CDensity& );          ///< assignment forbidden, use clone() for CPotential
};

/** Plummer density profile, triaxial **/
class CDensityPlummer: public CDensity
{
public:
    CDensityPlummer(double _q, double _p): CDensity(), q(_q), p(_p) {};
    virtual POTENTIALTYPE PotentialType() const { return PT_PLUMMER; };
    virtual std::string PotentialName() const { return myName(); };
    static std::string myName() { return "Plummer"; };
    virtual SYMMETRYTYPE symmetry() const { return (q==1 ? (p==1 ? ST_SPHERICAL : ST_AXISYMMETRIC) : ST_TRIAXIAL); };   // returns symmetry type of this potential based on p, q being unity or not
    virtual double Rho(double X, double Y, double Z, double t=0) const;
    virtual double getGamma() const { return 0; };
private:
    const double q, p;          ///< axis ratio (y/x and z/x) of equidensity surfaces
};

/** Isochrone density profile, triaxial **/
class CDensityIsochrone: public CDensity
{
public:
    CDensityIsochrone(double _q, double _p): CDensity(), q(_q), p(_p) {};
    virtual POTENTIALTYPE PotentialType() const { return PT_ISOCHRONE; };
    virtual std::string PotentialName() const { return myName(); };
    static std::string myName() { return "Isochrone"; };
    virtual SYMMETRYTYPE symmetry() const { return (q==1 ? (p==1 ? ST_SPHERICAL : ST_AXISYMMETRIC) : ST_TRIAXIAL); };   // returns symmetry type of this potential based on p, q being unity or not
    virtual double Rho(double X, double Y, double Z, double t=0) const;
    virtual double getGamma() const { return 0; };
private:
    const double q, p;          ///< axis ratio (y/x and z/x) of equidensity surfaces
};

/** Triaxial Perfect Ellipsoid density profile **/
class CDensityPerfectEllipsoid: public CDensity
{
public:
    CDensityPerfectEllipsoid(double _q, double _p): CDensity(), q(_q), p(_p) {};
    virtual POTENTIALTYPE PotentialType() const { return PT_PERFECTELLIPSOID; };
    virtual std::string PotentialName() const { return myName(); };
    static std::string myName() { return "Perfect Ellipsoid"; };
    virtual SYMMETRYTYPE symmetry() const { return (q==1 ? (p==1 ? ST_SPHERICAL : ST_AXISYMMETRIC) : ST_TRIAXIAL); };   // returns symmetry type of this potential based on p, q being unity or not
    virtual double Rho(double X, double Y, double Z, double t=0) const;
    virtual double getGamma() const { return 0; };
private:
    const double q, p;          ///< axis ratio (y/x and z/x) of equidensity surfaces
};

/** modified Navarro-Frenk-White density profile, triaxial.
    The density is given by rho ~ 1/(r*(1+r)^2*(1+r/r_cut)), i.e. modified to have a steeper 
    r^{-4} outer slope beyond the cutoff radius, to make the total mass finite. 
    The cutoff radius is computed from the condition that the total mass is the same as for 
    a sharply cut NFW profile with a given concentration parameter, but the gentle transition 
    to the outer slope instead of a sharp cut makes it more suitable to use in the potential 
    expansions. **/
class CDensityNFW: public CDensity
{
public:
    CDensityNFW(double _q, double _p, double c);  // c == concentration parameter
    virtual POTENTIALTYPE PotentialType() const { return PT_NFW; };
    virtual std::string PotentialName() const { return myName(); };
    static std::string myName() { return "NFW"; };
    virtual SYMMETRYTYPE symmetry() const { return (q==1 ? (p==1 ? ST_SPHERICAL : ST_AXISYMMETRIC) : ST_TRIAXIAL); };   // returns symmetry type of this potential based on p, q being unity or not
    virtual double Rho(double X, double Y, double Z, double t=0) const;
    virtual double getGamma() const { return 1.0; };
private:
    const double q, p; ///< axis ratio (y/x and z/x) of equidensity surfaces
    double rcutoff;    ///< cutoff radius
    double norm;       ///< normalization factor for density
};

/** Ellipsoidal model.
    Arbitrary spline-interpolated radial density profile with radially varying axis ratios 
    which are also spline-interpolated. **/
class CDensityEllipsoidal: public CDensity
{
public:
    CDensityEllipsoidal(const vectord &_radii, const vectord &_inmass, const vectord &_axesradii, const vectord &_qvals, const vectord &_pvals);
    virtual ~CDensityEllipsoidal();
    virtual POTENTIALTYPE PotentialType() const { return PT_ELLIPSOIDAL; };
    virtual std::string PotentialName() const { return myName(); };
    static std::string myName() { return "Ellipsoidal"; };
    virtual SYMMETRYTYPE symmetry() const { return (q==1 ? (p==1 ? ST_SPHERICAL : ST_AXISYMMETRIC) : ST_TRIAXIAL); };   // returns symmetry type of this potential based on p, q being unity or not
    virtual double Rho(double X, double Y, double Z, double t=0) const;
private:
    double q, p;               ///< axis ratio (y/x and z/x) of equidensity surfaces; only make sense if they are constant with radius
    gsl_spline* dens;          ///< spline interpolated density (log-log scaling)
    gsl_spline* qvals, *pvals; ///< spline interpolated axis ratios
};

/** Multi-Gaussian expansion of a triaxial density profile **/
class CDensityMGE: public CDensity
{
public:
    /// defines one component of the multi-gaussian expansion
struct CGaussian {   
    double width;    ///< "sigma" - width of the gaussian in x-direction
    double mass;     ///< total mass in this component
    double q, p;     ///< axes ratio (widths of the gaussian in y and z directions w.r.t x)
    CGaussian(double _width, double _mass, double _q, double _p) : width(_width), mass(_mass), q(_q), p(_p) {};
};
typedef std::vector<CGaussian> vectorg;  ///< an array of gaussian components

    CDensityMGE(const vectorg &_data);
    virtual ~CDensityMGE() {};
    virtual POTENTIALTYPE PotentialType() const { return PT_MGE; };
    virtual std::string PotentialName() const { return myName(); };
    static std::string myName() { return "MGE"; };
    virtual SYMMETRYTYPE symmetry() const { return (q==1 ? (p==1 ? ST_SPHERICAL : ST_AXISYMMETRIC) : ST_TRIAXIAL); };   // returns symmetry type of this potential based on p, q being unity or not
    virtual double Rho(double X, double Y, double Z, double t=0) const;
private:
    double q, p;     ///< axis ratio (y/x and z/x) of equidensity surfaces; only make sense if they are constant with radius
    vectorg data;    ///< array of gaussians representing the density distribution
};

/** base class for potential-density model including potential derivatives (=forces)  **/
class CPotential: public CDensity
{
public:
    const unsigned int N_dim; ///< dimensions (2 or 3)
    double Mbh;               ///< mass of a central black hole
    const double Omega;       ///< rotation speed (around z axis) [rotating potentials are not yet implemented]
    
    CPotential(unsigned int _N_dim, double _Mbh): 
      CDensity(), N_dim(std::min<unsigned int>(3, std::max<unsigned int>(2,_N_dim))), Mbh(_Mbh), Omega(0) {};
    
    /** Return a pointer to a copy of this instance of potential.
        A standard copy constructor or assignment is disabled because of 
        different amount of data needed to be copied in different derived classes). */
    virtual CPotential* clone() const=0;
    
    /** Return potential at a given spatial point (possibly a time-varying one) */
    virtual double Phi(double X, double Y, double Z, double t=0) const=0;

    /** Called from the integrator of ODE:  dx[i]/dt = f[i] .
        the input array x holds coordinates and velocities (depending on N_dim, it may be 2+2 or 3+3 double values), 
        and optionally the deviation vector (another 2+2 or 3+3 values for {dr[i],dv[i]}, 
        if we compute the Lyapunov exponent in the orbit integrator using the variation equation method.
        The number of variables n distinguishes between these two cases 
        (if n>2*N_dim, then the variation equation is also needed).
        Output array dx must be of the same size as the input one. */
    virtual void DiffEq(unsigned int n, double t, const double x[], double *dx) const=0;

    /** Convenience function to find the maximal elongation of long-axis orbit of given energy.
        (i.e. solve for Phi(x,0,0)=E ). */
    double longaxisradius(double E) const;

    /** Compute the period of oscillations for long-axis orbit of given energy.
        it is given by the integral
          \f$ T_{rad} = 4*\int_0^{X_{max}(E)}  dx/\sqrt{2*(E-\Phi(x))}. \f$
        For convenience (to avoid duplicate computation), the extent of the orbit may be provided 
        in Xmax, otherwise it is computed by longaxisradius(E). */
    double longaxisperiod(double E, double Xmax=-1) const;

    /** Find the location of the equipotential surface in a given direction: 
        solve for k such that Phi(k*X, k*Y, k*Z) = E . */
    double findintersection(double E, double X, double Y, double Z) const;

    /** Inform whether the potential tends to zero at r->infinity (the mass is finite).
        all potentials used for Schwarzschild modelling must have this property, 
        but toy models like the logarithmic, harmonic or scale-free potentials do not. */
    virtual bool finite() const=0;

    /** Convenience function to return the total energy for a given position+velocity point. */
    double totalEnergy(const CPosVelPoint<double> &point) const
    { return Phi(point.Pos[0], point.Pos[1], point.Pos[2]) + (pow_2(point.Vel[0])+pow_2(point.Vel[1])+pow_2(point.Vel[2]))/2; };

    /** find corotation radius for a rotating potential [unused] */
    double corotationradius() const;
};

/** Logarithmic potential, with possible core **/
class CPotentialLog: public CPotential
{
public:
    CPotentialLog(unsigned int _N_dim, double _q, double _p, double _Mbh, double _Rc): CPotential(_N_dim, _Mbh), q(_q), p(_p), Rc(_Rc) {};
    virtual CPotential* clone() const { return new CPotentialLog(*this); };
    virtual POTENTIALTYPE PotentialType() const { return PT_LOG; };
    virtual std::string PotentialName() const  { return myName(); };
    static std::string myName() { return "Logarithmic"; };
    virtual SYMMETRYTYPE symmetry() const { return (q==1 ? (p==1 ? ST_SPHERICAL : ST_AXISYMMETRIC) : ST_TRIAXIAL); };   // returns symmetry type of this potential based on p, q being unity or not
    virtual double Rho(double X, double Y, double Z, double t=0) const;
    virtual double Phi(double X, double Y, double Z, double t=0) const;
    virtual void DiffEq(unsigned n, double t, const double x[], double *dx) const;
    virtual bool finite() const { return false; };
private:
    const double q, p;    ///< axis ratio (y/x and z/x) of equipotential surfaces
    const double Rc;      ///< core radius of log potential
};

/** Potential of a constant-density (harmonic) core with different frequencies of oscillations in three directions **/
class CPotentialHarmonic: public CPotential 
{
public:
    CPotentialHarmonic(unsigned int _N_dim, double _q, double _p, double _Mbh): CPotential(_N_dim, _Mbh), q(_q), p(_p) {};
    virtual CPotential* clone() const { return new CPotentialHarmonic(*this); };
    virtual POTENTIALTYPE PotentialType() const { return PT_HARMONIC; };
    virtual std::string PotentialName() const { return myName(); };
    static std::string myName() { return "Harmonic"; };
    virtual SYMMETRYTYPE symmetry() const { return (q==1 ? (p==1 ? ST_SPHERICAL : ST_AXISYMMETRIC) : ST_TRIAXIAL); };   // returns symmetry type of this potential based on p, q being unity or not
    virtual double Rho(double X, double Y, double Z, double t=0) const;
    virtual double Phi(double X, double Y, double Z, double t=0) const;
    virtual void DiffEq(unsigned n, double t, const double x[], double *dx) const;
    virtual bool finite() const { return false; };
private:
    const double q, p;    ///< axis ratio (y/x and z/x) of equidensity surfaces
};

/** Dehnen(1993) double power-law model **/
class CPotentialDehnen: public CPotential
{
public:
    CPotentialDehnen(unsigned int _N_dim, double _q, double _p, double _Mbh, double _Gamma);
    virtual CPotential* clone() const { return new CPotentialDehnen(*this); };
    virtual POTENTIALTYPE PotentialType() const { return PT_DEHNEN; };
    virtual std::string PotentialName() const { return myName(); };
    static std::string myName() { return "Dehnen"; };
    virtual SYMMETRYTYPE symmetry() const { return (q==1 ? (p==1 ? ST_SPHERICAL : ST_AXISYMMETRIC) : ST_TRIAXIAL); };   // returns symmetry type of this potential based on p, q being unity or not
    virtual double Rho(double X, double Y, double Z, double t=0) const;
    virtual double Phi(double X, double Y, double Z, double t=0) const;
    virtual void DiffEq(unsigned n, double t, const double x[], double *dx) const;
    virtual double getGamma() const {return Gamma; }
    virtual bool finite() const { return true; };
private:
    const double q, p;       ///< axis ratio (y/x and z/x) of equidensity surfaces
    const double Gamma;      ///< cusp exponent for Dehnen potential
};

/** parent class for all potential expansions based on spherical harmonics for angular variables **/
class CPotentialSH: public CPotential
{
public:
    CPotentialSH(unsigned int _N_dim, double _Mbh, size_t _Ncoefs_angular): 
        CPotential(_N_dim, _Mbh), Ncoefs_angular(std::min<size_t>(MAX_NCOEFS_ANGULAR-1, _Ncoefs_angular)), mysymmetry(ST_TRIAXIAL) { assignlmrange(); };
    virtual SYMMETRYTYPE symmetry() const { return mysymmetry; };
    virtual double Rho(double X, double Y, double Z, double t=0) const;
    virtual double Phi(double X, double Y, double Z, double t=0) const;
    virtual void DiffEq(unsigned n, double t, const double x[], double *dx) const;  ///< common function for all derivative classes
    virtual bool finite() const { return true; };
    /// return l_max -- the order of angular expansion
    size_t getNcoefs_angular() const { return Ncoefs_angular; }
protected:
    size_t Ncoefs_angular;               ///< l_max, the order of angular expansion (0 means spherically symmetric model)
    SYMMETRYTYPE mysymmetry;             ///< may have different type of symmetry
    int lmax, lstep, mmin, mmax, mstep;  ///< range of angular coefficients used for given symmetry
    void assignlmrange();                ///< assigns the above variables based on mysymmetry, should be called whenever mysymmetry has changed

    /** The function that computes spherical-harmonic coefficients for potential 
        and its radial (first/second) derivative at given radius.
        Must be implemented in derived classes, and is used in evaluation of potential and forces; 
        unnecessary coefficients are indicated by passing NULL for coefs** and should not be computed.
    */
    virtual void computeSHCoefs(const double r, double coefsF[], double coefsdFdr[], double coefsd2Fdr2[]) const = 0;   
};

/** basis-set expansion on the Zhao(1996) basis set (alpha models) **/
class CPotentialBSE: public CPotentialSH
{
public:
    /// init coefficients from a discrete point mass set
    template<typename NumT> CPotentialBSE(unsigned int _N_dim, double _Mbh, double _Alpha, size_t _Ncoefs_radial, size_t _Ncoefs_angular, 
        const CPointMassSet<NumT> &points, SYMMETRYTYPE _sym=ST_TRIAXIAL, bool checkMonotonic=true);
    /// load coefficients from stored values
    CPotentialBSE(unsigned int _N_dim, double _Mbh, double _Alpha, 
        const std::vector< vectord > &coefs);
    /// init potential coefficients from an analytic mass model
    CPotentialBSE(unsigned int _N_dim, double _Mbh, double _Alpha, size_t _Ncoefs_radial, size_t _Ncoefs_angular, 
        const CDensity* density);
    virtual CPotential* clone() const { return new CPotentialBSE(*this); };
    virtual POTENTIALTYPE PotentialType() const { return PT_BSE; };
    virtual std::string PotentialName() const { return myName(); };
    static std::string myName() { return "BSE"; };
    /// a faster estimate of M(r) from the l=0 harmonic only
    virtual double Mass(const double r) const;
    //  get functions:
    /// return BSE coefficients array
    void getCoefs(std::vector< vectord > *coefsArray) const
    { if(coefsArray!=NULL) *coefsArray=SHcoefs; };
    /// return the shape parameter of basis set
    double getAlpha() const { return Alpha; };
    /// return the number of radial basis functions
    size_t getNcoefs_radial() const { return Ncoefs_radial; }
private:
    size_t Ncoefs_radial;          ///< number of radial basis functions [ =SHcoefs.size() ]
    std::vector<vectord> SHcoefs;  ///< array of coefficients A_nlm of potential expansion
    double Alpha;                  ///< shape parameter controlling inner and outer slopes of basis functions

    /// compute angular expansion coefs at the given radius
    virtual void computeSHCoefs(const double r, double coefsF[], double coefsdFdr[], double coefsd2Fdr2[]) const;

    /// compute coefficients from a discrete point mass set; 
    /// if Alpha=0 then it is computed automatically from the data
    template<typename NumT> void prepareCoefsDiscrete(const CPointMassSet<NumT> &points);

    /// compute coefficients from a smooth mass profile; 
    ///if Alpha=0 then it is chosen automatically from density->getGamma()
    void prepareCoefsAnalytic(const CDensity* density);
    
    void initDefault();           ///< called as a default initialization when everything else fails
    void checkSymmetry();         ///< assigns symmetry class if some coefficients are (near-)zero
};

/** spherical-harmonic expansion of potential with coefficients being spline functions of radius **/
class CPotentialSpline: public CPotentialSH
{
public:
    /// init potential from N point masses with given assumed symmetry type, 
    /// may also provide desired grid radii (otherwise assigned automatically)
    CPotentialSpline(unsigned int _N_dim, double _Mbh, size_t _Ncoefs_radial, size_t _Ncoefs_angular, 
        const CPointMassSetDouble &points, SYMMETRYTYPE _sym=ST_TRIAXIAL, double smoothfactor=0, 
        const vectord *_gridradii=NULL);
    /// init potential from stored SHE coefficients at given radii
    CPotentialSpline(unsigned int _N_dim, double _Mbh, 
        const vectord &_gridradii, const std::vector< vectord > &_coefs);
    /// init potential from analytic mass model, 
    /// may also provide desired grid radii (if not given, assign automatically)
    CPotentialSpline(unsigned int _N_dim, double _Mbh, size_t _Ncoefs_radial, size_t _Ncoefs_angular, 
        const CDensity* density, const vectord *_gridradii=NULL);
    /** not really a constructor, but a way to compute potential expansion coefficients 
        for a set of discrete point masses at given radii.
        \param[in] points contains the source set of point masses;
        \param[in] _Ncoefs_angular is the order of angular expansion (l_max);
        \param[in] radii contains the array of radii at which the coefficients should be computed;
        \param[out] coefsArray contains the computed coefficients.
        Smoothing is not performed, splines are not initialized and the potential is not usable after this call!
    */
    CPotentialSpline(size_t _Ncoefs_angular, const CPointMassSetDouble &points, const vectord *radii, 
        std::vector< vectord > *coefsArray);
    ~CPotentialSpline();
    virtual CPotential* clone() const;
    virtual POTENTIALTYPE PotentialType() const { return PT_SPLINE; };
    virtual std::string PotentialName() const { return myName(); };
    static std::string myName() { return "Spline"; };
    virtual double Mass(const double r) const;  ///< faster estimate of M(r) using l=0 terms only
    // get functions
    size_t getNcoefs_radial() const { return Ncoefs_radial; }   ///< return the number of radial points in the spline (excluding r=0)
    /** compute SHE coefficients at given radii.
        \param[in] useNodes tells whether the radii in question are 
                   the original spline nodes (true) or user-specified radii (true);
        \param[in,out] radii: if useNodes=true, this array is filled with spline nodes [output parameter, but must point to an existing array], 
                   otherwise the coefficients are computed at the supplied radii [input parameter];
        \param[out] coefsArray is filled by the computed coefficients (must point to an existing array)
    */
    void getCoefs(vectord *radii, std::vector< vectord > *coefsArray, bool useNodes=true) const;
private:
    size_t Ncoefs_radial;                ///< number of radial coefficients (excluding the one at r=0)
    vectord gridradii;                   ///< defines nodes of radial grid in splines
    double minr, maxr;                   ///< definition range of splines; extrapolation beyond this radius 
    double gammain,  coefin;             ///< slope and coef. for extrapolating potential inside minr (spherically-symmetric part, l=0)
    double gammaout, coefout, der2out;   ///< slope and coef. for extrapolating potential outside maxr (spherically-symmetric part, l=0)
    double potcenter, potmax, potminr;   ///< (abs.value) potential in the center (for transformation of l=0 spline), at the outermost spline node, and at 1st spline node
    std::vector<gsl_spline*> splines;    ///< spline coefficients at each harmonic
    vectord slopein, slopeout;           ///< slope of coefs for l>0 for extrapolating inside rmin/outside rmax

    /// reimplemented function to compute the coefficients of angular expansion of potential at the given radius
    virtual void computeSHCoefs(const double r, double coefsF[], double coefsdFdr[], double coefsd2Fdr2[]) const;

    /// assigns symmetry class if some coefficients are (near-)zero.
    /// called on an intermediate step after computing coefs but before initializing splines
    void checkSymmetry(const std::vector< vectord > &coefsArray);

    /// called as a default initialization when everything else fails
    void initDefault();

    /// create spline objects for all non-zero spherical harmonics from the supplied radii and coefficients.
    /// radii should have "Ncoefs_radial" elements and coefsArray - "Ncoefs_radial * (Ncoefs_angular+1)^2" elements
    void initSpline(const vectord &radii, const std::vector< vectord > &coefsArray);

    /** calculate (non-smoothed) spherical harmonic coefs for a discrete point mass set.
        \param[in] srcpoints contains the positions and masses of particles that serve as the input.
        There are two modes of operation, depending on the content of "outradii" vector: 

        - in the context of initialization of potential from a discrete point set, "srcradii"=NULL;
        "outradii" will be initialized with radii of all particles sorted in ascending order, 
        and "outcoefs" will contain coefficients of SH expansion for all points, 
        "outradii" and "outcoefs" should point to existing arrays; in case of error outcoefs is set to zero length.
        \note: the order of indexes in outcoefs is swapped w.r.t. standard convention in this file:
        outcoefs[coefind][pointind] - this is done to save memory by not allocating arrays for unused coefind'ices.

        - in the context of SHgrid Schwarzschild modelling, this is called from a quasi-constructor 
        for the purpose of computing SH coefs at a given set of radii. 
        This set is passed as "srcradii" array, and coefs are returned in "outcoefs"
        in the standard way: outcoefs[pointind][coefind]. 
        "outcoefs" should point to an existing array, in case of error it is filled with zeroes.
    */
    void computeCoefsFromPoints(const CPointMassSetDouble &points, const vectord* srcradii, 
        vectord* outradii, std::vector< vectord > *outcoefs);

    /** create smoothing splines from the coefficients computed at each particle's radius.
        \param[in] userradii may contain the radial grid to construct the splines; 
                   if it is NULL then the radial grid is constructed automatically 
                   taking into account the range of radii covered by source points.
        \param[in] smoothfactor determines how much smoothing is applied to the spline 
                   (good results are obtained for smoothfactor=1-2).  
    */
    void prepareCoefsDiscrete(const CPointMassSetDouble &points, double smoothfactor, const vectord* userradii);

    /** compute expansion coefficients from an analytical mass profile.
        \param[in] srcradii may contain the radial grid; if NULL then it is assigned automatically.
    */
    void prepareCoefsAnalytic(const CDensity* density, const vectord *srcradii);

    /// evaluate value and optionally up to two derivatives of l=0 coefficient, 
    /// taking into account extrapolation beyond the grid definition range and log-scaling of splines
    void coef0(double r, double *val, double *der, double *der2) const;

    /// evaluate value, and optionally first and second derivative of l>0 coefficients 
    /// (lm is the combined index of angular harmonic >0); corresponding values for 0th coef must be known
    void coeflm(size_t lm, double r, double xi, double *val, double *der, double *der2, 
        double c0val, double c0der=0, double c0der2=0) const;
};

/** simple power-law potential/density pair: rho ~ r^-gamma **/
class CPotentialScaleFree: public CPotential
{
public:
    CPotentialScaleFree(unsigned int _N_dim, double _q, double _p, double _Mbh, double _Gamma): CPotential(_N_dim, _Mbh), q(_q), p(_p), Gamma(_Gamma) {};
    virtual CPotential* clone() const { return new CPotentialScaleFree(*this); };
    virtual POTENTIALTYPE PotentialType() const { return PT_SCALEFREE; };
    virtual std::string PotentialName() const { return myName(); };
    static std::string myName() { return "Scale-free"; };
    virtual SYMMETRYTYPE symmetry() const { return (q==1 ? (p==1 ? ST_SPHERICAL : ST_AXISYMMETRIC) : ST_TRIAXIAL); };   // returns symmetry type of this potential based on p, q being unity or not
    virtual double Rho(double X, double Y, double Z, double t=0) const;
    virtual double Phi(double X, double Y, double Z, double t=0) const;
    virtual void DiffEq(unsigned n, double t, const double x[], double *dx) const;
    virtual double getGamma() const { return Gamma; }
    virtual bool finite() const { return false; };
private:
    const double q, p;       ///< axis ratio (y/x and z/x) of equidensity surfaces
    const double Gamma;      ///< scale-free profile exponent
};

/** angular expansion of scale-free potential in spherical harmonics **/
class CPotentialScaleFreeSH: public CPotentialSH
{
public:
    /// init potential from a scale-free density profile with the power-law exponent _Gamma
    CPotentialScaleFreeSH(unsigned int _N_dim, double _q, double _p, double _Mbh, double _Gamma, size_t _Ncoefs_angular);
    /// init potential from stored SHE coefficients
    CPotentialScaleFreeSH(unsigned int _N_dim, double _Mbh, double _Gamma, const vectord &_coefs);
    virtual CPotential* clone() const { return new CPotentialScaleFreeSH(*this); };
    virtual POTENTIALTYPE PotentialType() const { return PT_SCALEFREESH; };
    virtual std::string PotentialName() const { return myName(); };
    static std::string myName() { return "Scale-free SH"; };
    virtual bool finite() const { return false; };
    //  get functions
    /// return spherical-harmonic coefficients array
    void getCoefs(vectord *coefsArray) const { if(coefsArray!=NULL) *coefsArray=SHcoefs; };
    virtual double getGamma() const { return Gamma; }
private:
    const double Gamma;      ///< scale-free profile exponent
    vectord SHcoefs;         ///< coefficients of angular expansion
    virtual void computeSHCoefs(const double r, double coefsF[], double coefsdFdr[], double coefsd2Fdr2[]) const;
    void prepareCoefsAnalytic(double q, double p);   ///< called from the constructor to compute coefs for the given density profile
};

/** frozen N-body potential calculated by tree-code algorithm (based on hackcode1.c) **/
class CPotentialNB: public CPotential   // frozen N-body potential
{
private:

/// tree contains nodes of two kinds: individual particles and child cells
enum NODETYPE{
    BODY,   ///< one particle
    CELL    ///< tree cell containing up to 8 child nodes
};

typedef double vec3[N_DIM], matrix3[N_DIM][N_DIM];
/// NODE: data common to BODY and CELL structures.
struct node {
    NODETYPE type;    ///< code for node type
    double mass;      ///< total mass of node
    vec3 pos;         ///< position of node
    double eps;       ///< softening length associated with this node
};
typedef node *nodeptr;

/// BODY: data structure used to represent particles.
typedef node body, *bodyptr;

static const int IMAX=(1 << (8 * sizeof(int) - 2));  ///< max integer coordinate
static const unsigned int NSUB=(1 << N_DIM);         ///< subcells per cell (2^3)

/// CELL: structure used to represent internal nodes of tree. 
/// First few fields coincide with the definition of body structure
struct cell{
    NODETYPE type;       ///< code for node type
    double mass;         ///< total mass of cell
    vec3 pos;            ///< center of mass position of cell
    double eps;          ///< softening length associated with this node
    double rmax;         ///< max distance between cell's center of mass and any of child nodes
    size_t numbody;      ///< total number of bodies inside this cell
    nodeptr subp[NSUB];  ///< descendents of cell (eight pointers to child nodes, possibly NULL)
#if TREECODE_QUADRUPOLE
    matrix3 quad;        ///< quadrupole moment of cell (sum of m*dx_i*dx_j)
#endif
    cell() { type=CELL; for(unsigned int s=0; s<NSUB; s++) subp[s]=NULL; mass=-1; eps=rmax=0; numbody=0; }
};
typedef cell *cellptr;

/// operation to perform during the tree-walk
enum WALKOP{
    CALCPHI,   ///< compute potential
    CALCACC    ///< compute acceleration(3d)
};  
public:
    
    CPotentialNB(unsigned int _N_dim, double _Mbh);  ///< default (empty) initialization
    /// Initialize from a list of particles; \tparam NumT = double or float
    template<typename NumT> CPotentialNB(unsigned int _N_dim, double _Mbh, double _eps, double _tol, 
        const CPointMassSet<NumT> &points);
    virtual CPotential* clone() const { return new CPotentialNB(*this); };
    virtual POTENTIALTYPE PotentialType() const { return PT_NB; };
    virtual std::string PotentialName() const { return myName(); };
    virtual SYMMETRYTYPE symmetry() const { return ST_NONE; };  // nothing assumed about the position of points, right?
    static std::string myName() { return "Nbody"; };
    virtual double Phi(double X, double Y, double Z, double t=0) const;
    virtual double Rho(double X, double Y, double Z, double t=0) const;
    virtual void DiffEq(unsigned n, double t, const double x[], double *dx) const;
    virtual double Mass(const double r) const;    ///< reimplement the mass estimate by counting bodies inside given radius
    virtual bool finite() const { return true; };
    //  get functions
    /// return number of particles
    size_t bodyNum() const { return bodytab.size(); };
    /// return the position of a given particle
    template<typename NumT> CPosPoint<NumT> bodyPos(size_t index) const
    { return CPosPoint<NumT>(static_cast<NumT>(bodytab.at(index).pos[0]), static_cast<NumT>(bodytab.at(index).pos[1]), static_cast<NumT>(bodytab.at(index).pos[2])); };
    /// return the mass of a given particle
    double bodyMass(size_t index) const { return bodytab.at(index).mass; };
private:
    const double eps;           ///< potential softening parameter; negative value means position-dependent softening proportional to local density^{-1/3}
    const double tol;           ///< accuracy parameter (opening angle): 0 means exact N^2 force calculation (don't ever try!)
    std::vector<body> bodytab;  ///< list of all particles
    std::vector<cell> celltab;  ///< list of all tree cells
    
    nodeptr troot;              ///< pointer to the root cell of the tree
    vec3 rmin;                  ///< lower-left corner of coordinate box
    double rsize;               ///< side-length of integerized coordinate box
    // tree construction
    void maketree();                                       ///< initialize tree for the constructor
    bool expandbox(const size_t b);                        ///< expand bounding box (aka root cell) if particle does not fit into existing one
    bool appendtree(const size_t b);                       ///< append particle to the tree 
    bool intcoord(int xp[N_DIM], const vec3 rp) const;     ///< calculate integerized coords
    int subindex(const int x[N_DIM], const int l) const;   ///< determine subcell that a given position belongs to (l is cell level)
    void centerofmass(const nodeptr q, const vec3 corner, const double cellsize, int l);  ///< recursive calculation of cell center-of-mass
    void assigneps(const nodeptr q, const double epsparent, const double cellsize);       ///< recursive initialization of individual softening radii of cells
    void propagateeps(const nodeptr q);                    ///< recursively assign epsilon to tree cells from underlying nodes
    /// recursive computation of potential or forces from tree node/cell p at position pos0
    void walktree(const nodeptr p, const double dsq, const vec3 pos0, const WALKOP operation, double data[]) const;
    /// recursive search of particles in node/cell p which are closer to pos0 than sqrt{dsq).
    /// the list of such particles is stored in data and then used to compute density via kernel estimate
    void walktreedens(const nodeptr p, const double dsq, const double searchradsq, const vec3 pos0, const int xp[N_DIM], const int lev, vectorpd* data) const;
};

/* Correspondence between potential/density names and corresponding classes */

/// lists all 'true' potentials, i.e. those providing a complete density-potential(-force) pair
typedef std::map<CDensity::POTENTIALTYPE, std::string> PotentialNameMapType;

/// lists all analytic density profiles (including those that don't have corresponding potential, 
/// but excluding general-purpose expansions)
typedef std::map<CDensity::POTENTIALTYPE, std::string> DensityNameMapType;

/// lists available symmetry types
typedef std::map<CDensity::SYMMETRYTYPE, std::string > SymmetryNameMapType;

extern PotentialNameMapType PotentialNames;
extern DensityNameMapType DensityNames;
extern SymmetryNameMapType SymmetryNames;
CDensity::POTENTIALTYPE getPotentialTypeByName(std::string PotentialName);  ///< perform reverse lookup in PotentialNames map
CDensity::POTENTIALTYPE getDensityTypeByName(std::string DensityName);      ///< same for DensityNames
CDensity::SYMMETRYTYPE getSymmetryTypeByName(std::string SymmetryName);     ///< same for symmetry

/// structure that contains parameters for all possible potentials
struct CConfigPotential
{
    unsigned int N_dim;                      ///< dimensions of space (2 or 3)
    double q, p, Mbh, Rc, Gamma, Alpha;      ///< these values will be used to initialize proper potential type
    double Omega;                            ///< rotation frequency of potential figure (unused)
    size_t Ncoefs_radial, Ncoefs_angular;    ///< number of radial and angular coefficients in spherical-harmonic expansion
    double nbpotEps;                         ///< treecode smooothing length (negative means adaptive smoothing based on local density, absolute value is the proportionality coefficient between eps and mean interparticle distance)
    double nbpotTol;                         ///< tree cell opening angle
    CDensity::POTENTIALTYPE PotentialType;   ///< currently selected potential type
    CDensity::POTENTIALTYPE DensityType;     ///< if pot.type == BSE or Spline, this gives the underlying density profile approximated by these expansions or flags that an Nbody file should be used
    CDensity::SYMMETRYTYPE SymmetryType;     ///< if using Nbody file with the above two potential expansions, may assume certain symmetry on the coefficients (do not compute them but just assign to zero)
    double splineSmoothFactor;               ///< for smoothing Spline potential coefs initialized from discrete point mass set
    double splineRadiusMin, splineRadiusMax; ///< if nonzero, specifies the inner- and outermost grid node radii
};

}