#include "schwarzschild.h"
#include <cmath>
#include <algorithm>
#include <gsl/gsl_math.h>
#include <cassert>
#include "potential.h"
#include "orbitlib.h"
#include "orbitan.h"
#include "utils.h"

namespace smile{

// class for recording time spent in each shell during orbit integration, 
// and corresponding radial & tangential velocity dispersion
COrbitRuntimeSchwKinemShell::COrbitRuntimeSchwKinemShell(const COrbit* _orbit, const CSchwDataKinemShell* _model): 
CBasicOrbitRuntimeFnc(_orbit), model(_model)
{
    assert(orbit->getPotential()->N_dim==N_DIM);
    shellTime.assign(model->numConstraints()+1, 0); 
    shellVr.assign(model->numConstraints()+1, 0); 
    shellVt.assign(model->numConstraints()+1, 0); 
    COrbitInitData<double> ID=orbit->getInitData();
    shellprev = model->whichShellR(sqrt(pow_2(ID.initCond.Pos[0])+pow_2(ID.initCond.Pos[1])+pow_2(ID.initCond.Pos[2])));
    timeAccuracy = 1e-6*ID.timeStep;
}

bool COrbitRuntimeSchwKinemShell::Timestep(const double told, const double t, const double y[])
{
    double r2=pow_2(y[0]) + pow_2(y[1]) + pow_2(y[2]);   // assume N_dim==3
    double Lx=y[1]*y[5]-y[2]*y[4];
    double Ly=y[2]*y[3]-y[0]*y[5];
    double Lz=y[0]*y[4]-y[1]*y[3];
    double L=sqrt(pow_2(Lx)+pow_2(Ly)+pow_2(Lz));
    double vr2= pow_2(y[0]*y[3]+y[1]*y[4]+y[2]*y[5])/r2;
    double vt2= pow_2(L)/r2;
    if(r2==0) { vr2=pow_2(y[0])+pow_2(y[1])+pow_2(y[2]); vt2=0; }
    size_t shell=model->whichShellR(sqrt(r2));
    if(shell==shellprev)
    {
        shellTime[shell] += (t-told);
        shellVr[shell] += (t-told)*vr2;
        shellVt[shell] += (t-told)*vt2;
    }
    else  // calculate fractional time spent in shell
        calcShellTime(told, t, shellprev, shell, vr2, vt2);
    shellprev=shell;
    return true;
}

void COrbitRuntimeSchwKinemShell::calcShellTime(double tlow, double tupp, size_t slow, size_t supp, double vr2, double vt2)
{
    double tl=tlow, tu=tupp;
    int niter=0;
    double tmid;
    do{
        tmid=(tu+tl)/2;
        size_t smid=model->whichShellR(sqrt(pow_2(orbit->getInterpolatedTrajectory(0,tmid)) + pow_2(orbit->getInterpolatedTrajectory(1,tmid)) + pow_2(orbit->getInterpolatedTrajectory(2,tmid))));
        if(smid==slow)
            tl=tmid;
        else if(smid==supp)
            tu=tmid;
        else
        {   // neither of the two shells - go deeper into recursion
            calcShellTime(tlow, tmid, slow, smid, vr2, vt2);
            calcShellTime(tmid, tupp, smid, supp, vr2, vt2);
            return;
        }
        niter++;
    }
    while((tu-tl>timeAccuracy) && (niter<20));
    shellTime[slow] += tmid-tlow;
    shellTime[supp] += tupp-tmid;
    shellVr[slow]   += (tmid-tlow)*vr2;
    shellVr[supp]   += (tupp-tmid)*vr2;
    shellVt[slow]   += (tmid-tlow)*vt2;
    shellVt[supp]   += (tupp-tmid)*vt2;
}

void COrbitRuntimeSchwKinemShell::Finish()
{
    double intTime=orbit->getIntTime();
    if(intTime>0)
    {
        for(size_t s=0; s<shellTime.size(); s++)
        {
            if(shellTime[s]>0)
            {
                shellVr[s] /= shellTime[s];
                shellVt[s] /= shellTime[s];
            }
            shellTime[s] /= intTime;    // normalize to unit time
        }
    }
    shellVr.pop_back();  // eliminate last shell which goes to infty
    shellVt.pop_back();  
    shellTime.pop_back();
}

// class for recording time spent in each cell during orbit integration for classic Schwarzschild modelling
COrbitRuntimeSchwDensClassic::COrbitRuntimeSchwDensClassic(const COrbit* _orbit, const CSchwDataDensClassic* _model): 
CBasicOrbitRuntimeFnc(_orbit), model(_model)
{
    cellTime.assign(model->numConstraints()+1, 0);
    COrbitInitData<double> ID=orbit->getInitData();
    cellprev = model->whichCell(ID.initCond.Pos[0],ID.initCond.Pos[1],ID.initCond.Pos[2]);
    timeAccuracy = 1e-6*ID.timeStep;
}

bool COrbitRuntimeSchwDensClassic::Timestep(const double told, const double t, const double y[])
{
    size_t cell=model->whichCell(y[0], y[1], y[2]);
    if(cell==cellprev)
        cellTime[cell] += (t-told);
    else  // calculate fractional time spent in cell
        calcCellTime(told, t, cellprev, cell);
    cellprev=cell;
    return true;
}

void COrbitRuntimeSchwDensClassic::calcCellTime(double tlow, double tupp, size_t clow, size_t cupp)
{
    double tl=tlow, tu=tupp;
    int niter=0;
    double tmid;
    do{
        tmid=(tu+tl)/2;
        size_t cmid=model->whichCell(orbit->getInterpolatedTrajectory(0,tmid), orbit->getInterpolatedTrajectory(1,tmid), orbit->getInterpolatedTrajectory(2,tmid));  // assume that N_dim=3
        if(cmid==clow)
            tl=tmid;
        else if(cmid==cupp)
            tu=tmid;
        else
        {   // neither of the two cells - go deeper into recursion
            calcCellTime(tlow, tmid, clow, cmid);
            calcCellTime(tmid, tupp, cmid, cupp);
            return;
        }
        niter++;
    }
    while((tu-tl>timeAccuracy) && (niter<20));
    cellTime[clow]  += tmid-tlow;
    cellTime[cupp]  += tupp-tmid;
}

void COrbitRuntimeSchwDensClassic::Finish()
{
    double intTime=orbit->getIntTime();
    if(intTime>0)
        for(size_t c=0; c<cellTime.size(); c++)
            cellTime[c] /= intTime;    // normalize to unit time
    cellTime.pop_back();  // eliminate last cell which goes to infty
}

// class for recording density data in terms of spherical-harmonic expansion
COrbitRuntimeSchwDensSH::COrbitRuntimeSchwDensSH(const COrbit* _orbit, const CSchwDataDensSH* _model): 
CBasicOrbitRuntimeFnc(_orbit), model(_model)
{
    timeStep=orbit->getInitData().timeStep;
}

bool COrbitRuntimeSchwDensSH::Timestep(const double told, const double t, const double /*y*/[])
{
#if 0
    // old method -- store trajectory at roughly timeStep intervals
    if(t-told<timeStep)
        traj.add(CPosVelPoint<double>(y[0], y[1], y[2], y[3], y[4], y[5]), t-told);
    else
    {
        int numIntervals=static_cast<int>(ceil((t-told)/timeStep));
        for(int i=numIntervals-1; i>=0; i--)
        {
            double t1=t-i*(t-told)/numIntervals;
            traj.add(CPosVelPoint<double>(
                orbit->getInterpolatedTrajectory(0, t1),
                orbit->getInterpolatedTrajectory(1, t1),
                orbit->getInterpolatedTrajectory(2, t1),
                0, 0, 0), (t-told)/numIntervals);
        }
    }
#else 
    // new method -- store trajectory at regular intervals of time
    int i1=(int)(told/timeStep); 
    int i2=(int)(t/timeStep);
    for (int i=i1; i<=i2; i++)
    {
        double tt=timeStep*i;
        if ((tt>=told) && (tt<t))
        {
            traj.add(CPosVelPoint<double>(
                orbit->getInterpolatedTrajectory(0, tt),
                orbit->getInterpolatedTrajectory(1, tt),
                orbit->getInterpolatedTrajectory(2, tt),
                0, 0, 0), timeStep);
        }
    }
#endif
    return true;
}

void COrbitRuntimeSchwDensSH::Finish()
{
    double intTime=orbit->getIntTime();
    if(intTime>0) 
        for(size_t i=0; i<traj.size(); i++)
            traj[i].second /= intTime;
    model->computeDensityData(traj, &SHcoefs);
    double sum=0; 
    for(size_t i=0; i<SHcoefs.size(); i++) sum+=SHcoefs[i];
    traj.data.clear();
}

// base Schw.data class, one convenience function
void CBasicSchwData::createRadialGrid(const smile::CPotential *potential, size_t numShells, double innerShellMass, double outerShellMass, smile::vectord *grid)
{
    assert(grid);
    if(outerShellMass==0) outerShellMass=1.0-1.0/(numShells+1);
    if(innerShellMass==0) innerShellMass=outerShellMass/numShells;
    double totalMass = potential->totalMass();
    if(totalMass<=0) { 
        my_error("Error, density model does not seem to have finite mass!"); 
        totalMass=1; 
    }
    double innerRadius=potential->getRadiusByMass(innerShellMass * totalMass);
    double outerRadius=potential->getRadiusByMass(outerShellMass * totalMass);
    double maxStepRad =pow(outerRadius/innerRadius, 3.0/(numShells-1));  // maximum allowed ratio between adjacent radii (empirical formula)
    vectord gridtmp;
    grid->resize(numShells);
    bool ok=true;
    do{
        // first create a (possibly non-uniform) grid in the enclosed mass (normalized to total mass)
        createNonuniformGrid(numShells, innerShellMass, outerShellMass, false, &gridtmp);
        // now convert the grid in mass to the grid in radius
        for(size_t s=0; s<numShells; s++)
            grid->at(s) = potential->getRadiusByMass(gridtmp[s] * totalMass);
        // then check that outer radii differ no more than by a factor of maxStepRad
        ok=true;
        if(numShells>2 && grid->at(numShells-2)*maxStepRad < grid->at(numShells-1)) {
            outerRadius/=maxStepRad;
            outerShellMass=potential->Mass(outerRadius);
            numShells--;
            ok=false;
        }
    } while(!ok);
}

// Schw.data class for handling kinematic information in terms of radial/tangential 
// velocity dispersions in a set of concentric spherical shells
CSchwDataKinemShell::CSchwDataKinemShell(const CPotential* potential, unsigned int numShells, double innerShellMass, double outerShellMass)
{
    createRadialGrid(potential, numShells, innerShellMass, outerShellMass, &shellRadius);
    shellEnergy.assign(numShells, 0);
    for(unsigned int s=0; s<numShells; s++)
        shellEnergy[s] = potential->Phi(shellRadius[s], 0, 0);
}

size_t CSchwDataKinemShell::whichShellR(double radius) const
{
    size_t shnum=0;
    while(shnum<shellRadius.size() && radius>=shellRadius[shnum]) shnum++;
    return shnum;
}

size_t CSchwDataKinemShell::whichShellE(double energy) const
{
    size_t shnum=0;
    while(shnum<shellRadius.size() && energy>=shellEnergy[shnum]) shnum++;
    return shnum;
}

// Class for performing classic 3d-grid-based Schwarzschild density modelling
CSchwDataDensClassic::CSchwDataDensClassic(const CPotential* potential, unsigned int numShells, double innerShellMass, double outerShellMass, unsigned int _linesPerSegment): 
  linesPerSegment(_linesPerSegment), 
  cellsPerSegment(linesPerSegment*linesPerSegment),
  cellsPerShell(3*cellsPerSegment),
  numCells(numShells*cellsPerShell)
{
    // init angular grid boundaries
    cellBoundary.resize(linesPerSegment+1);
    for(size_t i=0; i<linesPerSegment; i++)
        cellBoundary[i] = tan(M_PI/4/linesPerSegment*i);
    cellBoundary[linesPerSegment] = 1.0;  // exact value
    // init radial grid in the same way as for shell-kinematic Schw.data class
    createRadialGrid(potential, numShells, innerShellMass, outerShellMass, &shellRadius);
    double massShell=0, prevMassShell=0;
    shellMassNorm.resize(numShells);
    for(unsigned int s=0; s<numShells; s++) {
        massShell = potential->Mass(shellRadius[s]);
        // massShell<=prevMassShell shouldn't normally happen, as it indicates a badly behaved potential; 
        // however, to avoid division by zero we initialize it by some reasonable value
        shellMassNorm[s]=(massShell>prevMassShell ? massShell-prevMassShell : 1)/cellsPerShell;
        prevMassShell=massShell;
    }
    // more complex initialization of cell masses is done in the separate routine
    initCellMass(potential);
}

void CSchwDataDensClassic::initCellMass(const CPotential* potential)
{
    // calculate mass in cells
    cellMass.assign(numCells, 0);
    if(potential->PotentialType()!=CDensity::PT_NB)
    for(unsigned int c=0; c<numCells; c++)
    {
        unsigned int nshell = c / cellsPerShell;
        unsigned int indseg = c % cellsPerShell;
        unsigned int nseg = indseg / cellsPerSegment;
        unsigned int ind1 =(indseg % cellsPerSegment) % linesPerSegment;
        unsigned int ind2 =(indseg % cellsPerSegment) / linesPerSegment;
        double dr1 = cellBoundary[ind1+1] - cellBoundary[ind1];
        double dr2 = cellBoundary[ind2+1] - cellBoundary[ind2];
        double mass=0;
        for(unsigned int i1=0; i1<NUM_SUBDIV_ANGULAR; i1++)
            for(unsigned int i2=0; i2<NUM_SUBDIV_ANGULAR; i2++)
            {
                double r1l = cellBoundary[ind1] + dr1*i1/NUM_SUBDIV_ANGULAR;
                double r1h = cellBoundary[ind1] + dr1*(i1+1)/NUM_SUBDIV_ANGULAR;
                double r2l = cellBoundary[ind2] + dr2*i2/NUM_SUBDIV_ANGULAR;
                double r2h = cellBoundary[ind2] + dr2*(i2+1)/NUM_SUBDIV_ANGULAR;
                double phill = asin(r1l*r2l/sqrt((1+pow_2(r1l))*(1+pow_2(r2l))));
                double philh = asin(r1l*r2h/sqrt((1+pow_2(r1l))*(1+pow_2(r2h))));
                double phihl = asin(r1h*r2l/sqrt((1+pow_2(r1h))*(1+pow_2(r2l))));
                double phihh = asin(r1h*r2h/sqrt((1+pow_2(r1h))*(1+pow_2(r2h))));
                double area = (phill+phihh-philh-phihl);
                double ratio1 = (r1l+r1h)/2;
                double ratio2 = (r2l+r2h)/2;
                double r=0, rprev, rho=0, rhoprev;
                for(unsigned int ir=0; ir<=NUM_SUBDIV_RADIAL; ir++)
                {
                    rprev=r; rhoprev=rho;
                    double shellRprev=nshell==0 ? 0 : shellRadius[nshell-1];
                    r = shellRprev + (shellRadius[nshell] - shellRprev) * ir / NUM_SUBDIV_RADIAL;
                    if(r==0) r=shellRadius[0]*0.01/NUM_SUBDIV_RADIAL;
                    double coord0 = r/sqrt(1 + pow_2(ratio1) + pow_2(ratio2));
                    double coord1 = ratio1*coord0;
                    double coord2 = ratio2*coord0;
                    double X=0, Y=0, Z=0;
                    switch(nseg)
                    {
                    case 0:
                        X=coord0;
                        Y=coord1;
                        Z=coord2;
                        break;
                    case 1:
                        Y=coord0;
                        Z=coord1;
                        X=coord2;
                        break;
                    case 2:
                        Z=coord0;
                        X=coord1;
                        Y=coord2;
                        break;
                    }
                    rho = potential->Rho(X, Y, Z);
                    if(ir>0)
                    {
                        double gamma = - log(rho/rhoprev)/log(r/rprev);
                        if(!gsl_finite(gamma)) gamma=0; // avoid NaN
                        if(gamma<0)
                            gamma=0; 
                        double intrad = rho*pow(r,gamma) * (gamma!=3 ? (pow(r, 3-gamma)-pow(rprev, 3-gamma))/(3-gamma) : log(r/rprev));
                        if(!gsl_finite(intrad)) intrad=0;
                        mass +=  area * intrad * 8;
                    }
                }
            }
        cellMass[c] = mass;
    }
    else  // frozen Nbody potential - simply count bodies in each cell
    {
        const CPotentialNB* potentialNB=static_cast<const CPotentialNB*>(potential);
        for(size_t ind=0; ind<potentialNB->bodyNum(); ind++)
        {
            CPosPoint<double> pt=potentialNB->bodyPos<double>(ind);
            size_t cell=whichCell(pt.Pos[0], pt.Pos[1], pt.Pos[2]);
            double ptmass=potentialNB->bodyMass(ind);
            if(cell<numCells) cellMass[cell] += ptmass;
        }
    }
}

void CSchwDataDensClassic::cellCenter(const size_t C, double* X, double* Y, double* Z) const
{
    assert(X); assert(Y); assert(Z);
    size_t nshell = C / cellsPerShell;
    if(nshell>=shellRadius.size()) 
    {
        X=0; Y=0; Z=0;
        return;
    }
    size_t indseg = C % cellsPerShell;
    size_t nseg = indseg / cellsPerSegment;
    size_t ind1 =(indseg % cellsPerSegment) % linesPerSegment;
    size_t ind2 =(indseg % cellsPerSegment) / linesPerSegment;
    double ratio1 = (cellBoundary[ind1] + cellBoundary[ind1+1])/2;
    double ratio2 = (cellBoundary[ind2] + cellBoundary[ind2+1])/2;
    double r = ((nshell>0 ? shellRadius[nshell-1] : 0) + shellRadius[nshell])/2;
    double coord0 = r/sqrt(1 + pow_2(ratio1) + pow_2(ratio2));
    double coord1 = ratio1*coord0;
    double coord2 = ratio2*coord0;
    switch(nseg)
    {
    case 0:
        *X=coord0;
        *Y=coord1;
        *Z=coord2;
        break;
    case 1:
        *Y=coord0;
        *Z=coord1;
        *X=coord2;
        break;
    case 2:
        *Z=coord0;
        *X=coord1;
        *Y=coord2;
        break;
    }
    if(whichCell(*X,*Y,*Z) != C)
        my_error("Error in determining cell center, shouldn't happen!");
}

size_t CSchwDataDensClassic::whichCell(const double _X, const double _Y, const double _Z) const
{
    double X=fabs(_X), Y=fabs(_Y), Z=fabs(_Z), radius=sqrt(pow_2(X)+pow_2(Y)+pow_2(Z));
    int nseg=-1;
    size_t nshell=0;
    while(nshell<shellRadius.size() && radius>=shellRadius[nshell]) nshell++;
    double coord0=0, coord1=0, coord2=0;
    if((X>=Y)&&(X>Z))
    {
        nseg=0;
        coord0=X;
        coord1=Y;
        coord2=Z;
    } else
    if((Y>=Z)&&(Y>X))
    {
        nseg=1;
        coord0=Y;
        coord1=Z;
        coord2=X;
    } else
    if((Z>=X)&&(Z>=Y))
    {
        nseg=2;
        coord0=Z;
        coord1=X;
        coord2=Y;
    }
    if(nseg==-1)
    {
        my_error("Error in determining cell index, shouldn't happen!");
        return 0;  // should not happen, but who knows?...
    }
    double ratio1 = coord1/coord0;
    double ratio2 = coord2/coord0;
    unsigned int ind1=0, ind2=0;
    for(unsigned int i=1; i<linesPerSegment; i++)
    {
        if(ratio1>=cellBoundary[i]) ind1=i;
        if(ratio2>=cellBoundary[i]) ind2=i;
    }
    unsigned int indseg = ind1 + ind2*linesPerSegment + nseg*cellsPerSegment;
    size_t ncell = nshell*cellsPerShell + indseg;
    return (ncell<numCells)?ncell:numCells;
}

//-------------------------------------------------------------//
CSchwDataDensSHGrid::CSchwDataDensSHGrid(const smile::CPotential *potential, size_t numShells, double innerShellMass, double outerShellMass, size_t _Nang) :
  CSchwDataDensSH(_Nang), numGrid(numShells)
{
    std::vector< vectord > tmpCoefs;
    createRadialGrid(potential, numShells, innerShellMass, outerShellMass, &GridRadii);
    /*if(outerShellMass==0) outerShellMass=1.0-1.0/(numShells+1);
    if(innerShellMass==0) innerShellMass=outerShellMass/numShells;
    createNonuniformGrid(numShells, potential->getRadiusByMass(innerShellMass), potential->getRadiusByMass(outerShellMass), false, &GridRadii); */

    GridRadii.insert(GridRadii.begin(), 0);  // additional grid node inserted at origin
    if(potential->PotentialType()==CPotential::PT_SPLINE)
        static_cast<const CPotentialSpline*>(potential)->getCoefs(&GridRadii, &tmpCoefs, false);   // HACK!! to speed up initialization
    else
    {
        CPotentialSpline pot(N_DIM, 0, numGrid, Nang, potential, &GridRadii);
        pot.getCoefs(&GridRadii, &tmpCoefs, true);
    }
    assert(GridRadii.size()==numGrid+1 && tmpCoefs.size()==numGrid+1);
    // take only even l,m>=0 coefficients
    size_t nAngCoefs = getNumCoefsAngularExpansion();
    coefConstraints.assign(1 + numGrid * nAngCoefs, 0);  // one coef for 0th node, nAngCoefs coefs for each of other nodes
    coefConstraints[0] = tmpCoefs[0][0];
    for(size_t radInd=0; radInd<numGrid; radInd++)
        for(size_t l=0; l<=Nang/2; l++)
            for(size_t m=0; m<=l; m++)
            {
                size_t tmpInd=2*l*(2*l+1)+2*m;
                coefConstraints[1 + radInd*nAngCoefs + l*(l+1)/2 + m] = 
                    tmpInd<tmpCoefs[radInd+1].size() ? tmpCoefs[radInd+1][tmpInd] : 0;  // may happen because of hack above, i.e. tmpcoefs array may have less items that required by the model
            }
}

void CSchwDataDensSHGrid::computeDensityData(const smile::CPointMassSetDouble &traj, vectord* data) const
{
    assert(data);
    std::vector< vectord > tmpCoefs;
    size_t nAngCoefs = getNumCoefsAngularExpansion();
    data->assign(1 + numGrid * nAngCoefs, 0);
    if(traj.size()<=1) return;  // zerozero
    CPotentialSpline pot(Nang, traj, &GridRadii, &tmpCoefs);
    if(tmpCoefs.size()==0)  // signals error
        return;   // return array of the necessary length filled with zeroes
    assert(tmpCoefs.size()==numGrid+1);
    // take only even l,m>=0 coefficients
    data->at(0) = tmpCoefs[0][0];
    for(size_t radInd=0; radInd<numGrid; radInd++)
        for(size_t l=0; l<=Nang/2; l++)
            for(size_t m=0; m<=l; m++)
            {
                data->at(1 + radInd*nAngCoefs + l*(l+1)/2 + m) = tmpCoefs[radInd+1][2*l*(2*l+1)+2*m];
            }
}

double CSchwDataDensSHGrid::getConstraintNormFactor(size_t index) const
{
    if(index==0) return coefConstraints[0];
    size_t nAngCoefs = getNumCoefsAngularExpansion();
    size_t n=(index-1)/nAngCoefs;
    size_t lm=(index-1)%nAngCoefs;
    int l=static_cast<int>(sqrt(2.*lm-floor(sqrt(2.*lm))));
    double norm=0;
    for(int m=0; m<=l; m++)
        norm+=pow_2(coefConstraints[1 + n*nAngCoefs + l*(l+1)/2 + m]);
    if(norm>1e-20) return sqrt(norm) * pow_2(l+1.); 
    else if(coefConstraints[1+n*nAngCoefs]>1e-10) return coefConstraints[1+n*nAngCoefs];
    else return 1;
}

void CSchwDataDensSHGrid::getCoefDecomposition(size_t index, double *n, int *l, int *m) const
{
    if(index==0 || index>numConstraints()) { *l=0; *m=0; *n=0; return; }
    index--;
    size_t indRad= index/getNumCoefsAngularExpansion();
    size_t indAng= index%getNumCoefsAngularExpansion();
    *l=static_cast<int>(sqrt(2.*indAng-floor(sqrt(2.*indAng))));
    *m=static_cast<int>(indAng-(*l)*(*l+1)/2)*2;
    *l*=2;
    *n=GridRadii[indRad+1];
}

//-------------------------------------------------------------//
CSchwDataDensSHBSE::CSchwDataDensSHBSE(const smile::CPotential *potential, size_t _Nang, size_t _Nrad, double _Alpha) :
  CSchwDataDensSH(_Nang), Nrad(_Nrad), totalMass(potential->totalMass())
{
    std::vector< vectord > tmpCoefs;
    if(potential->PotentialType()==CPotential::PT_BSE && 
        (_Alpha==0 || static_cast<const CPotentialBSE*>(potential)->getAlpha()==_Alpha))
    {  // HACK: to speed up initialization, assign coefficients from existing density model instead of computing costly integrals
        static_cast<const CPotentialBSE*>(potential)->getCoefs(&tmpCoefs);
        Alpha=static_cast<const CPotentialBSE*>(potential)->getAlpha();
        // the number of coefficients in potential may be greater or less than needed in S.M.
    }
    else
    {
        CPotentialBSE pot(N_DIM, 0, _Alpha, Nrad, Nang, potential);
        Alpha=pot.getAlpha();   // if _Alpha==0, it means autodetect; here we get this autodetected value (or the original supplied value if it was >=0.5)
        pot.getCoefs(&tmpCoefs);
        assert(tmpCoefs.size()==Nrad+1);
    }
    // take only even l,m>=0 coefficients
    size_t nAngCoefs = getNumCoefsAngularExpansion();
    coefConstraints.assign((Nrad+1) * nAngCoefs, 0);  // Nang*(Nang+1)/2 coefs for each radial basis function
    for(size_t radInd=0; radInd<=Nrad; radInd++)
        for(size_t l=0; l<=Nang/2; l++)
            for(size_t m=0; m<=l; m++)
            {
                size_t tmpInd=2*l*(2*l+1)+2*m;
                coefConstraints[radInd*nAngCoefs + l*(l+1)/2 + m] = 
                    radInd<tmpCoefs.size() && tmpInd<tmpCoefs[radInd].size() ? tmpCoefs[radInd][tmpInd] : 0;  // may happen because of hack above, i.e. tmpcoefs array may have less items that required by the model
            }
}

void CSchwDataDensSHBSE::computeDensityData(const smile::CPointMassSetDouble &traj, vectord *data) const
{
    assert(data);
    std::vector< vectord > tmpCoefs;
    size_t nAngCoefs = getNumCoefsAngularExpansion();
    data->assign((Nrad+1) * nAngCoefs, 0);
    if(traj.size()<=1) return;  // zerozero
    CPotentialBSE pot(N_DIM, 0, Alpha, Nrad, Nang, traj, CPotential::ST_TRIAXIAL, false);
    pot.getCoefs(&tmpCoefs);
    assert(tmpCoefs.size()==Nrad+1);
    // take only even l,m>=0 coefficients
    for(size_t radInd=0; radInd<=Nrad; radInd++)
        for(size_t l=0; l<=Nang/2; l++)
            for(size_t m=0; m<=l; m++)
            {
                data->at(radInd*nAngCoefs + l*(l+1)/2 + m) = tmpCoefs[radInd][2*l*(2*l+1)+2*m];
            }
}

/** this function is necessary because the values of coefficients may span many orders of magnitude,
    especially for high alpha. 
    To ensure a more uniform scaling of penalties for constraint violation, the scaling function is introduced. 
    The expression below is based on a number of `test cases' and is very approximate (within a few orders of mag.)
*/
double CSchwDataDensSHBSE::getConstraintNormFactor(size_t index) const
{
    size_t n=index/getNumCoefsAngularExpansion();
    size_t lm=index%getNumCoefsAngularExpansion();
    int l=static_cast<int>(sqrt(2.*lm-floor(sqrt(2.*lm))));
#if 0
    double norm=0;
    for(int m=0; m<=l; m++)
        norm+=pow_2(densityData[n*Nang*(Nang+1)/2 + l*(l+1)/2 + m]);
    if(norm>1e-20) return sqrt(norm);
#endif
    // some arbitrary, not very good heuristical approximation. A better one is desirable.
    double coef_n0=pow(n/(Alpha+1)+1, -4-2*log(Alpha)/log(2.)) * pow_2(n+1.);
    if(l==0) return totalMass*coef_n0;
    double A=2.8*Alpha-0.6;
    double B=-0.1*Alpha-0.024;
    double C=-0.01*Alpha-0.05;
    double D=-0.11*Alpha*Alpha+0.07*Alpha-1.9;
    return totalMass*coef_n0*exp(A*l + B*n + C*l*n + D) * pow_2(l+1.);
}

void CSchwDataDensSHBSE::getCoefDecomposition(size_t index, double *n, int *l, int *m) const
{
    if(index>numConstraints()) { *l=0; *m=0; *n=0; return; }
    size_t indRad= index/getNumCoefsAngularExpansion();
    size_t indAng= index%getNumCoefsAngularExpansion();
    *l=static_cast<int>(sqrt(2.*indAng-floor(sqrt(2.*indAng))));
    *m=static_cast<int>(indAng-(*l)*(*l+1)/2)*2;
    *l*=2;
    *n=static_cast<double>(indRad);
}

// driver class for quadratic Schwarzschild modelling of density and shell-kinematic data 
// (the only variant implemented so far)
CSchwModelQuadOpt::CSchwModelQuadOpt(COrbitLibrary* _orbits, 
    const vectorSchwData &_data, 
    const CBasicQuadraticOptimizationSolver<smNumType>* _solver,
    const CBasicOrbitFilteringFnc* _orbitPenaltyFnc,
    double _lambda, double _penaltyCons,
    double _maxWeight,
    bool _constrainBeta,
    double _betaIn, double _betaOut) :
  CBasicSchwModel(_orbits, _data), 
  solver(_solver), orbitPenaltyFnc(_orbitPenaltyFnc),
  lambda(_lambda), penaltyCons(_penaltyCons), 
  maxWeight(_maxWeight), constrainBeta(_constrainBeta), numConstr(0),
  correct(true)
{
    assert(orbits);
    totalMass=orbits->getPotential()->totalMass(); 
    if(totalMass<=0) totalMass=1;  // shouldn't happen...
    // determine what kind of constraints do we have
    // can handle arbitrary density constraints (equality) and optionally shell-kinematic constraints on velocity anisotropy
    for(vectorSchwData::iterator iter=data.begin(); iter!=data.end();)
        if( ((*iter)->dataType() & CBasicSchwData::SD_DENS) > 0 ||
            ((*iter)->dataType()==CBasicSchwData::SD_KINEM_SHELL && constrainBeta) )
        {
            numConstr += (*iter)->numConstraints();
            if((*iter)->dataType()==CBasicSchwData::SD_KINEM_SHELL) {  // assign required values of vel.anisotr. in each shell
                size_t nc=(*iter)->numConstraints();
                beta.resize(nc);
                for(size_t i=0; i<nc; i++)
                    beta[i]=static_cast<smNumType>(_betaIn + (_betaOut-_betaIn) * 
                    orbits->getPotential()->Mass(static_cast<const CSchwDataKinemShell*>(*iter)->getShellRadius(i)) / totalMass);
            }
            ++iter;
        } else {
            iter=data.erase(iter);  // do not consider this kind of data (we made a copy of data array beforehand so are free to erase its elements)
        }
    // should check that all orbits have necessary kind of data with right array sizes...
    for(size_t ind_orb=0; correct && ind_orb<orbits->size(); ind_orb++)
    {
        std::vector<bool> orbHasData(data.size(), false);   // record what kinds of orbit information we have
        for(size_t ind_inf=0; ind_inf<orbits->getOrbitDesc(ind_orb)->getInfoNum(); ind_inf++)
        {
            const CBasicInformation* info=orbits->getOrbitDesc(ind_orb)->getInfo(ind_inf);
            if(info->infoType()==CBasicInformation::IT_SCHW) {
                const CSchwInformation* infoSchw=static_cast<const CSchwInformation* >(info);
                CBasicSchwData::SCHWDATATYPE dtype=infoSchw->schwDataType();
                size_t dsize=infoSchw->getSchwData().size();
                if(dtype==CBasicSchwData::SD_KINEM_SHELL) dsize/=3;  // each kinematic data array has 3*numShells values
                // find a corresponding schw.data type from what we have in the model
                for(size_t ind_data=0; ind_data<data.size(); ind_data++)
                    if(data[ind_data]->dataType() == dtype) {
                        orbHasData[ind_data]=true;
                        if(dsize!=data[ind_data]->numConstraints())  { // incorrect size
                            if(correct)  my_error("Orbit library isn't compatible with Schwarzschild data:\n"+
                                data[ind_data]->dataName()+" has size="+convertToString(data[ind_data]->numConstraints())+
                                " but orbit #"+convertToString(ind_orb)+" has size="+convertToString(dsize) );
                            correct=false;
                        }
                    }
            }
        }
        // check that orbit has all necessary data
        for(size_t ind_data=0; correct && ind_data<data.size(); ind_data++)
            if(!orbHasData[ind_data]) {
                my_error("Orbit library isn't compatible with Schwarzschild data:\n"
                    "otbit #"+convertToString(ind_orb)+" doesn't have "+data[ind_data]->dataName()+" data");
                correct=false;
            }
    }
    numConstrData=numConstr;   // store the total number of constraints in all data objects, without the total mass constraint or additional tolerance constraints in the case penaltyCons<0
    numVar = orbits->size() + (penaltyCons!=0 ? 2*numConstrData : 0);   // if have constraint penalties, then the linear system has additional `slack' variables, two per constraint        
    if(penaltyCons<0) numConstr*=2;  // if constraints must be satisfied to within a given tolerance, then the linear system has additional block of equations, one per `normal' constraint
    numConstr+=1;  // additional equality constraint for total mass
    // initialize rhs
    rhs.assign(numConstr, 0);        // the rhs of equations
    normFactor.assign(numConstr, 1); // normalization factors for constraints
    size_t ind_cons=0;
    for(vectorSchwData::iterator iter=data.begin(); iter!=data.end(); iter++)
    {
        for(size_t c=0; c<(*iter)->numConstraints(); c++, ind_cons++)
        {
            normFactor[ind_cons] = static_cast<smNumType>((*iter)->getConstraintNormFactor(c));
            rhs[ind_cons] = static_cast<smNumType>((*iter)->getConstraint(c)/normFactor[ind_cons]);
        }
    }
    if(penaltyCons<0)  // add the block of equations for constraint tolerance intervals
        for(size_t c=0; c<ind_cons; c++)
        {
            normFactor[ind_cons+c] = 1;
            rhs[ind_cons+c] = static_cast<smNumType>(fabs(penaltyCons*rhs[c]));
        }
    rhs.back()=static_cast<smNumType>(totalMass);  // the last constraint: sum of all orbit weights should be equal to total model mass
}

smNumType CSchwModelQuadOpt::getLinearMatrixElem(size_t ind_orb, size_t ind_cons) const
{
    if(!correct || ind_cons>=numConstr || ind_orb>=numVar)
        return 0;   // shouldn't happen?
    if(ind_cons == numConstr-1) { // constraint for total mass
        return static_cast<smNumType>(ind_orb<orbits->size() ? 1.0 : 0.0);
    } else if(penaltyCons!=0 && ind_orb>=orbits->size()) {  // return coef for additional slack variable
        ind_orb -= orbits->size();
        if(penaltyCons<0 && ind_cons>=numConstrData) {  // additional block of equations
            ind_cons -= numConstrData;
            if(ind_orb==ind_cons || ind_orb==ind_cons+numConstrData) return +1.0; else return 0;
        }
        if(ind_orb==ind_cons) return +1.0; else
        if(ind_orb==ind_cons+numConstrData) return -1.0; else return 0;
    } else {  // find which Schw.data does the constraint correspond to
        vectorSchwData::const_iterator iter=data.begin();   // the corresponding data object
        size_t ind_cons_this=ind_cons;                      // index of constraint in this object
        while(iter!=data.end() && ind_cons_this >= (*iter)->numConstraints()) {
            ind_cons_this -= (*iter)->numConstraints(); 
            iter++; 
        }
        if(iter==data.end()) 
            return 0;   // may only happen in the case penaltyCons<0 and ind_orb<numOrbits; this block of linear matrix is zero
        const CBasicSchwData::SCHWDATATYPE dtype=(*iter)->dataType();
        const COrbitDesc* orb=orbits->getOrbitDesc(ind_orb);
        for(size_t i=0; i<orb->getInfoNum(); i++)
        {   // find info of corresponding Schw.data type
            const CBasicInformation* info=orb->getInfo(i);
            if(info->infoType()==CBasicInformation::IT_SCHW &&
                static_cast<const CSchwInformation*>(info)->schwDataType()==dtype)
            {   // found corresponding information
                const CSchwInformation* info_s=static_cast<const CSchwInformation*>(info);
                if(dtype == CBasicSchwData::SD_KINEM_SHELL)  // need special treatment
                {
                    sdNumType shellVt = info_s->getSchwData(ind_cons_this+2*(*iter)->numConstraints());
                    sdNumType shellVr = info_s->getSchwData(ind_cons_this+1*(*iter)->numConstraints());
                    sdNumType shellTime = info_s->getSchwData(ind_cons_this);
                    return static_cast<smNumType>((shellVt-2*(1-beta[ind_cons_this])*shellVr) * shellTime);
                } else 
                    return static_cast<smNumType>(info_s->getSchwData(ind_cons_this) / normFactor[ind_cons]);
            }
        }
        return 0;   // coefficient not found, shouldn't raise an error?
    }
}

std::string CSchwModelQuadOpt::performModelling()
{
    if(!correct) 
        return "Error, model is not consistent with orbit data";
    if(orbits==NULL || orbits->size()==0 || solver==NULL || data.size()==0) 
        return "Error, empty model or orbit library!";
    // internal arrays used in optimization procedure
    CMatrixDiagonal<smNumType> *objFuncQuad=NULL;
    CMatrixBasicSchwModel* matrixInterface=new CMatrixBasicSchwModel(this);
    int result = 0;
    try{
        std::vector<smNumType> objFuncLinear(numVar);    // linear terms in the objective function
        std::vector<smNumType> orbitWeight(numVar);      // the solution to be obtained
        std::vector<smNumType> maxOrbitWeight(numVar);   // upper limits on orbit weights if necessary
        if(lambda>0)                      // quadratic terms in the objective function
            objFuncQuad = new CMatrixDiagonal<smNumType>(numVar);
        // 1st stage: preparation
        //my_message("Preparing model");
        for(size_t o=0; o<orbits->size(); o++)
        {
            if(orbitPenaltyFnc!=NULL)
                objFuncLinear[o] = static_cast<smNumType>(orbitPenaltyFnc->eval(orbits->getOrbitDesc(o))/totalMass);
            orbitWeight[o]=static_cast<smNumType>(orbits->getOrbitDesc(o)->getWeight());  // current orbit weights, may be used to warm-start the solver
            if(lambda>0)
                objFuncQuad->set(o, static_cast<smNumType>(lambda*orbits->size()/pow_2(totalMass)));
            maxOrbitWeight[o]=static_cast<smNumType>(maxWeight);
        }
        if(penaltyCons>0)   // add penalty for slack variables
            for(size_t c=0; c<2*numConstrData; c++)
                objFuncLinear[c+orbits->size()] = static_cast<smNumType>(penaltyCons);

        // 2nd stage: optimization itself
        my_message("Calling solver");
        result = solver->callSolver(matrixInterface, rhs, maxOrbitWeight, objFuncLinear, objFuncQuad, &orbitWeight);
        if(objFuncQuad!=NULL) delete objFuncQuad;
        delete matrixInterface;
        if(result<0)   // error in solver
            return solver->errorDescription(result);

        // 3rd stage: post-processing
        //my_message("Finishing model");
        for(size_t o=0; o<orbits->size(); o++)
        {
            if(orbitWeight[o] < totalMass*SCHW_MIN_ORBIT_WEIGHT_USED/orbits->size()) 
                orbitWeight[o]=0;
            orbits->setOrbitWeight(o, orbitWeight[o]);
        }
        return getStatistics();  // on success return statistical information about solution
    }
    catch(const std::bad_alloc&) {
        if(objFuncQuad!=NULL) delete objFuncQuad;
        delete matrixInterface;
        return "Not enough memory in preparing model";
    }
    catch(...) {
        if(objFuncQuad!=NULL) delete objFuncQuad;
        delete matrixInterface;
        return "Unknown error in Schwarzschild modelling";
    }
}

std::string CSchwModelQuadOpt::getStatistics() const
{
    if(!correct) 
        return "Error, model is not consistent with orbit data";
    if(orbits==NULL || orbits->size()==0) 
        return "Error in getStatistics: no orbits!";
    size_t numOrbitsNonzero=0;     // number of orbits with non-zero weight
    size_t numCoefsInfeasible=0;   // number of coefficients that did not meet requirement values
    double totalDensityPenalty=0;  // sum of densityPenalty for all infeasible coefs
    double entropy=0;
    double totalWeight=0;
    std::vector<double> weights(orbits->size(), 0);
    for(size_t o=0; o<orbits->size(); o++)
    {
        double orbitWeight=orbits->getOrbitDesc(o)->getWeight();
        totalWeight+=orbitWeight;
        if(orbitWeight>0) numOrbitsNonzero++;
        weights[o]=orbitWeight;
    }
    std::sort(weights.begin(), weights.end(), std::greater<double>());
    double avgWeight=totalWeight/orbits->size();
    double sumHighWeights=0;
    size_t numOrbitsHalfMass=0;
    if(avgWeight>0) {
        for(size_t o=0; o<orbits->size(); o++)  // compute entropy assuming flat (uniform) priors for weights
        {
            double orbitWeight=orbits->getOrbitDesc(o)->getWeight();
            if(orbitWeight>0) entropy -= orbitWeight*log(orbitWeight/avgWeight);
            sumHighWeights+=weights[o];
            if(sumHighWeights>=totalWeight*0.5 && numOrbitsHalfMass==0) numOrbitsHalfMass=o;
        }
        entropy /= totalWeight;
    }
    double maxToAvgWeight = weights.front()/avgWeight;
    double fracOrbitsHalfMass = numOrbitsHalfMass*1.0/orbits->size();

    for(size_t c=0; c<numConstr; c++)
    {
        double deviation=calcCoefDeviation(c);
        if(deviation!=0) 
        {
            numCoefsInfeasible++;
            totalDensityPenalty+=deviation;
        }
    }
    return
        convertToString(numOrbitsNonzero)+" orbits out of "+convertToString(orbits->size())+" with nonzero weights\n" +
        (totalWeight>0 ?   "Total weight: " + convertToString(totalWeight) + "\n" +
        convertToString(fracOrbitsHalfMass*100, 3) + "% orbits make up 50% total mass" +
        (fracOrbitsHalfMass>0.2 ? " (good)" : fracOrbitsHalfMass>0.1 ? " (fair)" : " (poor)") + "\n" +
        "entropy: " + convertToString(entropy, 3) + 
        (entropy>=-0.5 ? " (good)" : entropy>=-1.0 ? " (fair)" : " (poor)") + "\n" +
        "Max/avg weight: " + convertToString(maxToAvgWeight, 3) + 
        (maxToAvgWeight<=10 ? " (good)" : maxToAvgWeight<=50 ? " (fair)" : " (poor)") + "\n" : "" ) +
        (numCoefsInfeasible>0 ? "WARNING: " +
        convertToString(numCoefsInfeasible)+" coefs out of "+convertToString(numConstr)+" infeasible\n" +
        convertToString(totalDensityPenalty, 3)+" total difference" : "Model is feasible");
}

double CSchwModelQuadOpt::calcCoefDeviation(size_t coefIndex) const
{
    if(!correct || coefIndex>numConstr) return 0;
    double deviation=0;
    for(size_t o=0; o<orbits->size(); o++)
        deviation += getLinearMatrixElem(o, coefIndex) * orbits->getOrbitDesc(o)->getWeight();
    deviation -= rhs[coefIndex];
    if(fabs(deviation) <= fabs(rhs[coefIndex]*SCHW_MAX_CELLMASS_REL_DIFFERENCE) || 
        (rhs[coefIndex]==0 && fabs(deviation) <= normFactor[coefIndex]*SCHW_MAX_CELLMASS_REL_DIFFERENCE))
        deviation=0;
    return deviation;
}

//-------------------------------------------------------------//
double CChaosOrbitFilteringFnc::eval(const smile::COrbitDesc *orbit) const
{
    if(orbit==NULL) return 0;
    const COrbitInformation<float>* info=static_cast<const COrbitInformation<float>*>(orbit->getInfoByType(CBasicInformation::IT_TRAJ_ANALYSIS));
    if(info!=NULL) 
    {
        double wf=log10(info->getlfccdiff())-log10(chaoticMinFreqDiff) + 0.5;   // `soft' transition from reg to chaos, based on freq.diff.rate
        if(wf<0) wf=0; if(wf>1) wf=1;
        if(info->getlambda()>chaoticMinLambda) wf=1;
        return chaoticWeightFactor*wf;
    }
    else return 0;
}

double CCentrophilicOrbitFilteringFnc::eval(const smile::COrbitDesc *orbit) const
{
    if(orbit==NULL) return 0;
	const CPericenterInformation<float>* info=static_cast<const CPericenterInformation<float>*>(orbit->getInfoByType(CBasicInformation::IT_PERICENTER));
	if(info!=NULL) {
		float fitIntercept, fitSlope, fitScatter, fitSignificance, Lcirc2;
		info->getParams(&fitIntercept, &fitSlope, &fitScatter, &fitSignificance, &Lcirc2);
		if(fitIntercept==0) return weightFactor; else return 0;
	}
    else return 0;
}

double CShellOrbitFilteringFnc::eval(const smile::COrbitDesc *orbit) const
{
    if(orbit==NULL) return 0;
    if(model==NULL || numShell==0) return 1;
    double E=0;
    const COrbitInformation<float>* info=static_cast<const COrbitInformation<float>*>(orbit->getInfoByType(CBasicInformation::IT_TRAJ_ANALYSIS));
    if(info!=NULL) E=info->getEinit();
    // need to find which energy shell does this orbit belong to
    if(E==0)  // need to figure out orbit energy
    {
        const COrbitInitData<float>& InitData = orbit->getInitData();
        E = InitData.potential->totalEnergy(InitData.initCond);
    }
    if((numShell>model->numConstraints() && E>model->getShellEnergy(model->numConstraints()-1)) ||
       (E<=model->getShellEnergy(numShell-1) && (numShell==1 || E>model->getShellEnergy(numShell-2)))) 
       return 1;
    else return 0;
}

/** Correspondence between Schw.model names and types **/
std::string getSchwDataNameByType(CBasicSchwData::SCHWDATATYPE dataType)
{
    switch(dataType) {
    case CBasicSchwData::SD_DENS_CLASSIC: return CSchwDataDensClassic::myName();
    case CBasicSchwData::SD_DENS_SHGRID:  return CSchwDataDensSHGrid::myName();
    case CBasicSchwData::SD_DENS_SHBSE:   return CSchwDataDensSHBSE::myName();
    case CBasicSchwData::SD_KINEM_SHELL:  return CSchwDataKinemShell::myName();
    default: return "Error";
    }
};

CBasicSchwData::SCHWDATATYPE getSchwDataTypeByName(std::string dataName)
{
    if(dataName==CSchwDataDensClassic::myName()) return CBasicSchwData::SD_DENS_CLASSIC;
    if(dataName==CSchwDataDensSHGrid ::myName()) return CBasicSchwData::SD_DENS_SHGRID;
    if(dataName==CSchwDataDensSHBSE  ::myName()) return CBasicSchwData::SD_DENS_SHBSE;
    if(dataName==CSchwDataKinemShell ::myName()) return CBasicSchwData::SD_KINEM_SHELL;
    return CBasicSchwData::SD_DENS_CLASSIC;  // default
};

CBasicOrbitRuntimeFncCreator* createSchwCreator(const CBasicSchwData* schwData)
{
    switch(schwData->dataType()) {
        case CBasicSchwData::SD_KINEM_SHELL: 
            return new COrbitRuntimeSchwKinemShellCreator(static_cast<const CSchwDataKinemShell*>(schwData)); 
        case CBasicSchwData::SD_DENS_CLASSIC: 
            return new COrbitRuntimeSchwDensClassicCreator(static_cast<const CSchwDataDensClassic*>(schwData)); 
        case CBasicSchwData::SD_DENS_SHBSE: 
        case CBasicSchwData::SD_DENS_SHGRID: 
            return new COrbitRuntimeSchwDensSHCreator(static_cast<const CSchwDataDensSH*>(schwData)); 
        default:
            return NULL;  // unknown class
    }
};

}  // namespace