/** \file    schwarzschild.h
    \brief   Classes that compute and store various data for Schwarzschild modelling, 
             and the driver class for performing modelling
    \author  EV
    \date    2009-2013

    Class hierachy for Schwarzschild modelling is split into two connected class lineages.

    The Data objects, descendants of smile::CBasicSchwData, are used to represent various density and kinematic data.
    They must be initialized prior to orbit library construction. 
    Each data class implements the computation of constraints for a given potential model 
    and for each orbit, and has a corresponding orbit runtime function which records the necessary data 
    for each orbit (and its associated runtime function creator class). 
    All data classes share the same information class (smile::CSchwInformation), which stores 
    the data as a plain 1d array and has an additional flag specifying the data class.    
    In the present version, three variants of density models (Classic, BSE and Spline) 
    and one fairly simple kinematic data model (which allows to specify velocity anisotropy coefficient 
    as a function of radius) are implemented.

    The Modelling class is constructed only temporary for solving the problem of finding orbit weights 
    which reproduce the set of constraints specified by one or more individual data objects, by solving 
    some kind of optimization problem (i.e. quadratic optimization, non-negative least squares, etc.).
    Presently there is only one implementation, CSchwModelQuadratic.
    It combines several data classes (e.g. one for density model and one for kinematic data) 
    into one set of linear equations with some objective(penalty) function to be minimized, 
    and calls an instance of LP/QP solver to obtain the solution. 
    Another function implemented in the  is the statistical information about the quality 
    of the model (i.e. the number of infeasible constraints, the spread in orbit weights, etc.)
    This statistics may be obtained without invoking the optimization itself.
*/

#pragma once
#include "common.h"

namespace smile{

/*------- Definition of base (abstract) classes --------*/

/** Base class for Schwarzschild Data objects.

    Each object stores the information about some property of the model, 
    like the mass in each cell of the classic grid-based Schwarzschild model. 
    The information is stored in array(s) of abstractly defined constraints, 
    which are needed to be satisfied in some way (exactly or approximately, 
    or within a given tolerance) in the course of modelling.

    The instances of descendant classes are constructed for a given potential, 
    which is not used thereafter, and represent the relevant data for the entire model. 

    They also provide the way of computing the values of constraints for each orbit.
    This is done by corresponding orbit runtime functions, derived from CBasicOrbitRuntimeFnc;
    each variant of Schwarzschild data object has its associated runtime function 
    which uses specific methods of this class to compute the constraints for a given orbit.
*/
class CBasicSchwData
{
public:
    /** Lists all possible variants of data objects */
enum SCHWDATATYPE {
    SD_DENS=0x100,       ///< all variants of density information
    SD_KINEM=0x200,      ///< all variants of kinematic information
    SD_DENS_CLASSIC=SD_DENS | 1,       ///< the classic grid-based Schwarzschild density model: CSchwDataDensClassic
    SD_DENS_SPHERICAL_HARMONIC=0x80,   ///< all density models based on angular spherical-harmonic expansion
    SD_DENS_SHBSE  =SD_DENS | SD_DENS_SPHERICAL_HARMONIC | 2,  ///< model based on BSE potential expansion: CSchwDataDensSHBSE
    SD_DENS_SHGRID =SD_DENS | SD_DENS_SPHERICAL_HARMONIC | 3,  ///< model based on spherical-harmonic expansion on a radial grid: CSchwDataDensSHGrid
    SD_KINEM_SHELL =SD_KINEM| 1        ///< kinematic information about radial and tangential velocity dispersion as a function of radius: CSchwDataKinemShell
};
    virtual ~CBasicSchwData() {};
    virtual SCHWDATATYPE dataType() const = 0;    ///< descendant classes return their type
    virtual std::string dataName() const = 0;     ///< string identificator of the class
    virtual size_t numConstraints() const = 0;    ///< number of constraints that this kind of data object handles
    virtual double getConstraint(size_t index) const = 0;     ///< return the value of the given constraing
    /// return a scaling coefficient which is used to normalize the penalty for constraint deviation in the optimization problem.
    /// Introduced to make the optimization problem more uniformly scaled (optional, by default returns 1)
    virtual double getConstraintNormFactor(size_t /*index*/) const { return 1; };
    /// convenience function that creates a "non-uniform grid" in mass and returns array of corresponding radii. 
    /// Static function shared among several descendant classes.
    static void createRadialGrid(const CPotential* potential, size_t numShells, double innerShellMass, double outerShellMass, vectord* grid);  
};

/** array of heterogeneous data objects to be used simultaneously in the modelling */
typedef std::vector<const CBasicSchwData*> vectorSchwData;

/** Base class for Schwarzschild Modelling driver objects.

    It takes the set of constraints to be satisfied in some way 
    (exactly or approximately, with suitably defined penalties for violating them) 
    from an array of Schwarschild Data objects, and an orbit library in which each orbit must have 
    the same set of Schwarzschild information objects corresponding to data objects.

    The performModelling() method then creates an optimization problem in some way (e.g. a quadratic optimization one)
    and passes it to the instance of a solver supplied when the model is constructed.
*/
class CBasicSchwModel
{
public:
    /// constructors of descendant classes would also take a pointer to some kind of solver as a parameter
    CBasicSchwModel(COrbitLibrary* _orbits, const vectorSchwData &_data):
      orbits(_orbits), data(_data) {};           // store a pointer to orbits and make a copy of array of pointers to data
    virtual ~CBasicSchwModel() {};

    /// perform the actual work, and return statistical information on success or error description on failure
    virtual std::string performModelling()=0;

    /// computes and returns statistical information.
    /// The information returned may describe, e.g. how much the model differs from data, etc.
    /// Uses existing Schwarzschild data and orbit weights, without re-doing modelling.
    virtual std::string getStatistics() const=0;
    
    virtual size_t numVariables() const=0;    ///< total number of variables in the linear matrix (number of orbits plus slack variables, etc)
    virtual size_t numConstraints() const=0;  ///< total number of constraints in the linear matrix (in all data objects plus anything else)

    /// return the elements of linear constraint matrix to pass to the solver 
    /// through a separately defined matrix interface object
    virtual smNumType getLinearMatrixElem(size_t i1, size_t i2) const=0;
protected:
    COrbitLibrary* orbits;   ///< pointer to the orbit library
    vectorSchwData data;     ///< array of pointers to Schwarzschild data objects (which contain various constraints in the model)
};

/** matrix interface class that passes the linear matrix elements from Schwarzschild model object to the solver */
class CMatrixBasicSchwModel: public CMatrix<smNumType>
{
public:
    CMatrixBasicSchwModel(const CBasicSchwModel* _model): model(_model) {};
    virtual size_t size1() const { return model->numVariables(); };
    virtual size_t size2() const { return model->numConstraints(); };
    virtual smNumType at(size_t i1, size_t i2) const { return model->getLinearMatrixElem(i1, i2); };
private:
    const CBasicSchwModel* model;
};

/* -------- Descendant classes that provide actual Scharzschild data computation ------- */

/** Schwarzschild data object to handle (rudimentary) kinematic data, in the form of 
    radial and tangential velocity dispersions in a number of radial shells.
    For each orbit, the corresponding runtime function records the fraction of time that the orbit spends 
    inside every shell, as well as the radial and tangential velocity dispersion in that shell.
    The class that does the actual Schwarzschild modelling takes this information and 
    tries to achieve a given velocity anisotropy profile throughout the model;
    the actual value of anisotropy in a given shell is not specified at the time of orbit integration 
    but rather is set at the modelling stage, therefore the "required" constraints are not used literally.
*/
class CSchwDataKinemShell: public CBasicSchwData
{
public:
    /// create shell-kinematic Schwarzschild model for a given potential model 
    /// (which is not used after constructor call)
    CSchwDataKinemShell(const CPotential* potential, unsigned int _numShells, double innerShellMass, double outerShellMass);
    virtual SCHWDATATYPE dataType() const { return SD_KINEM_SHELL; };
    virtual std::string dataName() const { return myName(); };
    static std::string myName() { return "KinematicShell"; };
    /// total number of constraints is the number of shells
    virtual size_t numConstraints() const { return shellRadius.size(); };
    /// actual constraint value is set internally in CSchwModelQuadOpt
    virtual double getConstraint(size_t /*index*/) const { return 0; };
    
    //  get functions to be called from the corresponding orbit runtime function
    /// return index of shell in which a given radius lies, or shellRadius.size() if it is beyond the last shell
    size_t whichShellR(double radius) const;

    /// return index of shell containing a given value of energy, or shellEnergy.size() if it is beyond the last shell
    size_t whichShellE(double energy) const;

    /// return the (outer) radius of a given shell
    double getShellRadius(size_t index) const { return shellRadius.at(index); }
    /// return the energy at the outer radius of a given shell
    double getShellEnergy(size_t index) const { return shellEnergy.at(index); }
private:
    vectord shellRadius;   ///< radii of concentric shells for storing kinematic data
    vectord shellEnergy;   ///< values of potential along the x axis at the radius of each shell
};

/** Schwarzschild data object for the classic grid-based density model. 
    The constraints are the masses contained in cells of a three-dimensional grid
    covering radii containing mass from innerShellMass to outerShellMass and using the angular division 
    of the positive octant of space into three segments and their further division into NxN cells.
*/
class CSchwDataDensClassic: public CBasicSchwData
{
public:
    /// create classic 3d-grid density Schwarzschild model for a given potential model (it is not used after constructor call)
    CSchwDataDensClassic(const CPotential* potential, unsigned int _numShells, double innerShellMass, double outerShellMass, unsigned int _linesPerSegment);
    virtual SCHWDATATYPE dataType() const { return SD_DENS_CLASSIC; };
    virtual std::string dataName() const { return myName(); };
    static std::string myName() { return "Classic"; };
    virtual size_t numConstraints() const { return cellMass.size(); };
    virtual double getConstraint(size_t index) const { return cellMass.at(index); };
    virtual double getConstraintNormFactor(size_t index) const { return shellMassNorm.at(index/cellsPerShell); };
    //  get functions to be called from the corresponding orbit runtime function
    /// return index of cell containing given coordinates, called from COrbitRuntimeSchwClassic to compute time spent in each cell
    size_t whichCell(const double X, const double Y, const double Z) const;
    /// return coordinates of geometric center of a cell
    void cellCenter(const size_t cell, double* X, double* Y, double* Z) const;
    /// return number of cells in one radial shell (3*cellsPerSegment)
    size_t getCellsPerShell() const { return cellsPerShell; }
private:
    const unsigned int linesPerSegment;    ///< Schwarzschild's scheme for dividing octant into three segments, each of them divided into linesPerSegment^2 squares
    const unsigned int cellsPerSegment;
    const unsigned int cellsPerShell;
    const unsigned int numCells;           // all these constants are stored just for convenience; they can be deduced from the sizes of arrays below
    vectord shellRadius;    ///< radii of concentric shells
    vectord cellBoundary;   ///< angles dividing segment into linesPerSegment parts
    vectord cellMass;       ///< integral of potential->rho() over the volume of each cell
    vectord shellMassNorm;  ///< approximate normalization for cell masses (order-of-magnitude expected mass of each cell, the same throughout each shell)
    void initCellMass(const CPotential* potential);  ///< called from constructor to compute the mass of each cell
};

/** Base class for variants of Schwarzschild density data objects based on 
    spherical-harmonic expansion of density of both target model and orbits.
    Two derived classes share several virtual methods, and the runtime function 
    computing coefs of expansion for an orbit is the same for both variants.
*/
class CSchwDataDensSH: public CBasicSchwData
{
public:
    CSchwDataDensSH(size_t _Nang): Nang(_Nang) {};
    virtual size_t numConstraints() const { return coefConstraints.size(); };
    virtual double getConstraint(size_t index) const { return coefConstraints.at(index); };
    //  get functions to be called from the corresponding orbit runtime function
    /// evaluate coefs for orbit trajectory data, called from COrbitRuntimeSchwSH, implemented differently in derived classes
    virtual void computeDensityData(const CPointMassSetDouble& traj, vectord* data) const=0;
    /// return number of coefficients in angular spherical-harmonic expansion
    size_t getNumCoefsAngularExpansion() const { return (Nang/2+1)*(Nang/2+2)/2; }
    /// convenience function returning n,l,m for a given coef.index
    virtual void getCoefDecomposition(size_t index, double* n, int* l, int* m) const=0;
protected:
    const size_t Nang;       ///< order of expansion in angular harmonics (lmax)
    vectord coefConstraints; ///< required constraints of the model (specified by descendant classes)
};

/** Variant of Schwarzschild density data object based on spherical-harmonic expansion of 
    potential coefficients at radial grid points.
    Only even non-negative harmonics in l,m are used because of triaxial symmetry. 
    Nang is the expansion order (lmax), and the number of coefficients at each radius is 
      (Nang/2+1)*(Nang/2+2)/2
    Note: the potential in which the model is evolved, needs not to be itself a spline expansion,
    but if it is, then the model coefs are initialized from the potential rather than 
    by computing integrals over density (which are expensive). 
*/
class CSchwDataDensSHGrid: public CSchwDataDensSH
{
public:
    CSchwDataDensSHGrid(const CPotential* potential, size_t _numShells, double innerShellMass, double outerShellMass, size_t _Nang);
    virtual SCHWDATATYPE dataType() const { return SD_DENS_SHGRID; };
    virtual std::string dataName() const { return myName(); };
    static std::string myName() { return "SHGrid"; };
    // get functions
    virtual void computeDensityData(const CPointMassSetDouble& traj, vectord* data) const;
    virtual double getConstraintNormFactor(size_t index) const;
    virtual void getCoefDecomposition(size_t index, double* n, int* l, int* m) const;
private:
    const size_t numGrid;  ///< number of grid points for computing expansion coefficients stored for convenience (equal to GridRadii.size())
    vectord GridRadii;     ///< expansion coefficients are evaluated at given radii
};

/** Variant of Schwarzschild density data object based on basis-set expansion in terms of 
    Zhao(1995) basis set for density of both required profile and of orbits.
    Nrad+1 is the number of radial functions and Nang is the order of angular expansion 
    in spherical harmonics, the number of angular coefs for each radial function is 
      (Nang/2+1)*(Nang/2+2)/2. 
    Only even non-negative harmonics in l,m are used because of triaxial symmetry. 
    The density profile is thus fitted in 'nonlocal' sense.
    Alpha is the parameter controlling the family of basis functions, same as in CPotentialBSE.
    Note: the potential in which the model is evolved, needs not to be itself a BSE expansion,
    but if it is, then the model coefs are initialized from the potential rather than 
    by computing integrals over density (which are expensive). 
*/
class CSchwDataDensSHBSE: public CSchwDataDensSH
{
public:
    CSchwDataDensSHBSE(const CPotential* potential, size_t _Nang, size_t _Nrad, double _Alpha);
    virtual SCHWDATATYPE dataType() const { return SD_DENS_SHBSE; };
    virtual std::string dataName() const { return myName(); };
    static std::string myName() { return "BSE"; };
    // get functions
    virtual void computeDensityData(const CPointMassSetDouble& traj, vectord* data) const;
    virtual double getConstraintNormFactor(size_t index) const;
    virtual void getCoefDecomposition(size_t index, double* n, int* l, int* m) const;
private:
    const size_t Nrad;     ///< number of radial basis functions
    double Alpha;          ///< parameter controlling the form of Zhao(1995) basis functions
    double totalMass;      ///< multiplier for constraint normalization
};

/*--------- Descendant information container classes that keep Schwarzschild data for each orbit --------*/

/** Information object that keeps density data for all variants of Schwarzschild data objects.
    Model-specific density or kinematic or any other data is kept in a single array "schwData", 
    whose meaning depends on the variant of model.
    The data type is returned by another dedicated function.
*/
class CSchwInformation: public CBasicInformation
{
public:
    /// create the data object from a one-dimensional array
    template<typename NumT> CSchwInformation(CBasicSchwData::SCHWDATATYPE _dataType, const std::vector<NumT> &_schwData) :
      dataType(_dataType) 
    {
        schwData.reserve(_schwData.size());
        for(size_t c=0; c<_schwData.size(); c++) 
            schwData.push_back(static_cast<sdNumType>(_schwData[c]));
    };
    /// all Schwarzschild information objects have the same information type
    virtual INFOTYPE infoType() const { return IT_SCHW; };
    /// return the type of Schwarzschild data object for which this information is applicable
    CBasicSchwData::SCHWDATATYPE schwDataType() const { return dataType; }
    /// no textual representation available
    virtual std::string toString() const { return std::string(); };
    // get functions
    /// return the entire data serialized into a one-dimensional array
    const std::vector<sdNumType>& getSchwData() const { return schwData; }
    /// overloaded function for convenience access to data elements
    sdNumType getSchwData(size_t i) const { return schwData.at(i); }
private:
    const CBasicSchwData::SCHWDATATYPE dataType;  ///< the type of Schwarzschild data stored in this object
    std::vector<sdNumType> schwData;              ///< one-dimensional data array
};

/*---------- orbit runtime functions and their creator classes for each Schwarzschild data object ---------*/

/** Runtime function recording shell-based kinematic data
    (time spent in each shell and radial/tangential velocity dispersion in a shell).
*/
class COrbitRuntimeSchwKinemShell: public CBasicOrbitRuntimeFnc
{
public:
    COrbitRuntimeSchwKinemShell(const COrbit* _orbit, const CSchwDataKinemShell* _model);
    virtual FNCTYPE FncType() const { return FT_SCHW; };
    virtual bool Timestep(const double told, const double t, const double y[]);
    virtual void Finish();
    virtual CBasicInformation* getData() const   /// serialize three arrays into one
    {
        vectord tempData;
        tempData.insert(tempData.end(), shellTime.begin(), shellTime.end());
        tempData.insert(tempData.end(), shellVr.begin(), shellVr.end());
        tempData.insert(tempData.end(), shellVt.begin(), shellVt.end());
        return new CSchwInformation(model->dataType(), tempData); 
    }
private:
    const CSchwDataKinemShell* model;    ///< pointer to Schwarzschild kinem.data object
    double timeAccuracy;                 ///< accuracy of computation of time spent in each shell
    vectord shellTime, shellVr, shellVt; ///< time spent in each shell and corresponding radial/tangential velocity dispersions
    size_t shellprev;                    ///< index of shell on the previous timestep
    /// record time-in-shell with fraction-of-period accuracy
    void calcShellTime(double tlow, double tupp, size_t slow, size_t supp, double vr2, double vt2);
};

/** corresponding creator class for shell-kinematic Schw runtime function */
class COrbitRuntimeSchwKinemShellCreator: public CBasicOrbitRuntimeFncCreator
{
public:
    COrbitRuntimeSchwKinemShellCreator(const CSchwDataKinemShell* _model) : model(_model) {};
    virtual CBasicOrbitRuntimeFnc* createRuntimeFnc(const COrbit* orbit) const { return new COrbitRuntimeSchwKinemShell(orbit, model); };
private:
    const CSchwDataKinemShell* model;
};

/** Runtime function corresponding to classic 3d-grid Schwarzschild density data object */
class COrbitRuntimeSchwDensClassic: public CBasicOrbitRuntimeFnc
{
public:
    COrbitRuntimeSchwDensClassic(const COrbit* _orbit, const CSchwDataDensClassic* _model);
    virtual FNCTYPE FncType() const { return FT_SCHW; };
    /// find the cell(s) which the orbit traversed during this timestep
    virtual bool Timestep(const double told, const double t, const double y[]);
    /// normalize the time spent in each cell by the total integration time
    virtual void Finish();
    virtual CBasicInformation* getData() const 
    { return new CSchwInformation(model->dataType(), cellTime); }
private:
    const CSchwDataDensClassic* model; ///< pointer to the Schwarzschild data object which handles the partitioning of space into cells
    vectord cellTime;                  ///< time spent in each cell of Schwarzschild's model
    size_t cellprev;                   ///< cell index at previous timestep (for recording time-spent-in-cell)
    double timeAccuracy;               ///< compute the time in each cell with this accuracy
    void calcCellTime(double tlow, double tupp, size_t clow, size_t cupp);  ///< record time-in-cell with fraction-of-period accuracy
};

/** corresponding creator class for classic Schw runtime function */
class COrbitRuntimeSchwDensClassicCreator: public CBasicOrbitRuntimeFncCreator
{
public:
    COrbitRuntimeSchwDensClassicCreator(const CSchwDataDensClassic* _model) : model(_model) {};
    virtual CBasicOrbitRuntimeFnc* createRuntimeFnc(const COrbit* orbit) const { return new COrbitRuntimeSchwDensClassic(orbit, model); };
private:
    const CSchwDataDensClassic* model;
};

/** Runtime function corresponding to Schwatzschild density data object 
    based on spherical-harmonic expansion of potential coefficients 
    at radial grid points or for some radial basis functions (serves both variants)
*/
class COrbitRuntimeSchwDensSH: public CBasicOrbitRuntimeFnc
{
public:
    COrbitRuntimeSchwDensSH(const COrbit* _orbit, const CSchwDataDensSH* _model);
    virtual FNCTYPE FncType() const { return FT_SCHW; };
    virtual bool Timestep(const double told, const double t, const double y[]);  ///< record trajectory
    virtual void Finish();     ///< the actual coefs computation is done here
    virtual CBasicInformation* getData() const
    { return new CSchwInformation(model->dataType(), SHcoefs); }
private:
    const CSchwDataDensSH* model;    ///< pointer to one of the two spherical-harmonic based Schwarzschild data objects used to compute expansion coefficients for each orbit
    CPointMassSetDouble traj;        ///< keeps trajectory data to compute SH coefficients when finished
    vectord SHcoefs;                 ///< coefficients of spherical-harmonic expansion
    double timeStep;                 ///< defined maximum interval between stored trajectory points (need not be stored equidistantly in time)
};

/** corresponding creator class for SH Schw runtime function */
class COrbitRuntimeSchwDensSHCreator: public CBasicOrbitRuntimeFncCreator
{
public:
    COrbitRuntimeSchwDensSHCreator(const CSchwDataDensSH* _model) : model(_model) {};
    virtual CBasicOrbitRuntimeFnc* createRuntimeFnc(const COrbit* orbit) const { return new COrbitRuntimeSchwDensSH(orbit, model); };
private:
    const CSchwDataDensSH* model;
};

/*--------- Descendant classes that provide actual Scharzschild modelling algorithms --------*/

/** The class for performing Schwarzschild modelling of density and kinematic data 
    from the theoretical point of view (without "observational errors").

    The problem is solved by linear or quadratic optimization, in which the cost function takes 
    the contribution to penalty from orbitPenaltyFnc if it is supplied (i.e. one may 
    favour or disfavour some kinds of orbits, such as long-axis tubes or all regular orbits), 
    and -- in the case of quadratic programming -- also the contribution from squared sum 
    of orbit weights normalized to some priors, with a given proportionality constant lambda 
    which has the meaning of regularization (smoothing). 
    In other words, in the case without any penalty functions, it tries to achieve 
    the distribution of orbit weights as close as possible to the priors (most often, uniform weights).
    
    There are also two option to allow from departures from exact solution (i.e. not all constraints 
    may be satisfied to machine precision). We denote deviation of a constraint as 
      \f$  dev[c] = abs| \sum_o M[o][c] sol[o] - rhs[c] |  \f$.

    In the first approach, when penaltyCons>0, a linear term is added into the cost function 
    which is just  \sum_c penaltyCons * dev[c].

    In the second approach, when penaltyCons<0, another set of N_c equations is introduced to keep 
    the deviation within a given tolerance interval (as in van den Bosch et al.2008):
      \f$  |dev[c]| < |penaltyCons * rhs[c]|  \f$.

    The entire cost function is written as 
      \f[  \mathcal{L} = \lambda/N_o * \sum_o (sol[o]/prior[o])^2 
                       + \sum_o penaltyOrb[o] * sol[o]/M_{total}
                       + \sum_c penaltyCons[c] * dev[c] .
      \f]

    If the problem is solved by linear programming, the first terms is omitted.
    If penaltyCons<0, the last term is absent and the tolerance intervals are absorbed 
    by the additional set of non-negative variables.

    Presently, some simplifications are in effect: priors for orbit weights are taken to be uniform 
      prior[o] = M_total/N_o,
    and penalty terms for constraint violation are also the same for all constraints, but normalized 
    to the intrinsic "normalization factors" provided by corresponding data objects.
*/
class CSchwModelQuadOpt: public CBasicSchwModel
{
public:
    CSchwModelQuadOpt(COrbitLibrary* _orbits,  ///< pointer to an existing orbit library
        const vectorSchwData &_data,           ///< pointers to the array of Schwarzschild data objects (density+kinematic data)
        const CBasicQuadraticOptimizationSolver<smNumType>* _solver, ///< pointer to an instance of quad.opt. solver. Will be deleted when the object is destroyed! If NULL then no modelling could be performed, but statistics may be computed anyway
        const CBasicOrbitFilteringFnc* _orbitPenaltyFnc,             ///< pointer to an instance of a function evaluating penalties for orbits, or NULL for no such penalty. Will be deleted when the object is destroyed!
        double _lambda,                        ///< coefficients in the penalty function for quadratic optimization; if 0 then a linear optimization is performed
        double _penaltyCons,                   ///< penalties for constraint violation; if 0 then constraints must be satisfied exactly or the model is infeasible
        double sm_maxWeight,                   ///< upper bound on orbit weight
        bool sm_constrainBeta,                 ///< whether to constrain anisotropy
        double sm_betaIn, double sm_betaOut    ///< anisotropy coefficient requested (values for the inner and outer shells, linearly interpolated in between)
    );
    virtual ~CSchwModelQuadOpt() {             ///  destroys solver and penaltyfnc, but doesn't touch orbits or schwData
        if(solver!=NULL) delete solver; 
        if(orbitPenaltyFnc!=NULL) delete orbitPenaltyFnc;
    };
    virtual std::string performModelling();
    virtual std::string getStatistics() const;
    virtual size_t numVariables() const { return numVar; };
    virtual size_t numConstraints() const { return numConstr; };
    virtual smNumType getLinearMatrixElem(size_t i1, size_t i2) const;
    double calcCoefDeviation(size_t coefIndex) const;   ///< compute deviations of a model constraint from its required value, normalized by coefNormFactor
    bool isCorrect() const { return correct; }          ///< whether the orbit library is consistent with data objects
private:
    const CBasicQuadraticOptimizationSolver<smNumType>* solver;   ///< pointer to an existing instance of linear/quadratic solver
    const CBasicOrbitFilteringFnc* orbitPenaltyFnc;               ///< pointer to a function which assigns penalties to some kinds of orbits, may be NULL
    const double lambda;               ///< regularization term in quadratic optimization function, may be zero (then a linear optimization is performed) 
    const double penaltyCons;          ///< penalty coefficient for constraint violation; if zero, then it means infinity (i.e. constraints must be satisfied exactly, but the problem may turn out to be infeasible)
    const double maxWeight;            ///< if nonzero, puts upper limit to orbit weights
    const bool constrainBeta;          ///< whether to add kinematic data (in terms of velocity anisotropy beta) to optimization constraints
    size_t numVar, numConstr;          ///< dimensions of linear matrix, set up in the constructor
    size_t numConstrData;              ///< total number of constraints in all data objects, without the total mass constraint or additional tolerance constraints in the case penaltyCons<0
    std::vector<smNumType> beta;       ///< required values of vel.anisotr.coef in radial shells
    std::vector<smNumType> rhs;        ///< required r.h.s. values of linear eq. system
    std::vector<smNumType> normFactor; ///< normalization factors for r.h.s. values
    double  totalMass;                 ///< total model mass, stored for convenience 
    bool correct;                      ///< flag set in the constructor to ensure that orbit library is consistent with Schwarzschild data
};

/* ------- A couple of examples for orbit filtering and penalty functions ------ */

/** Orbit filtering function that evaluates chaotic properties of an orbit, based on threshold in 
    frequency diffusion rate and in Lyapunov exponent value, multiplied by some predefined factor.
    May be used for telling regular from chaotic orbits in e.g. computing orbit population of a model, 
    or for assigning positive or negative bias for penalizing or favouring orbits of particular type 
    in Schwarzschild model.
    Positive factor favours regular orbits, negative - chaotic.
*/
class CChaosOrbitFilteringFnc: public CBasicOrbitFilteringFnc
{
public:
    CChaosOrbitFilteringFnc(double _chaoticWeightFactor, double _chaoticMinLambda, double _chaoticMinFreqDiff) :
      chaoticWeightFactor(_chaoticWeightFactor), chaoticMinLambda(_chaoticMinLambda), chaoticMinFreqDiff(_chaoticMinFreqDiff) {};
    virtual double eval(const COrbitDesc* orbit) const;
private:
    double chaoticWeightFactor;  ///< (maximum) returned value for chaotic orbits (with a soft transition to zero for regular orbits)
    double chaoticMinLambda;     ///< threshold in Lyapunov exponent to designate the orbit as a chaotic (typically zero)
    double chaoticMinFreqDiff;   ///< threshold in FDR (actually use a soft transition spanning +-0.5 in log(FDR)
};

/** Orbit filtering function that selects centrophilic orbits.
    May be used for favouring (weightFactor<0) or penalizing (weightFactor>0) orbits with 
    estimated minimum angular momentum equal to zero in the Schwarzschild model.
*/
class CCentrophilicOrbitFilteringFnc: public CBasicOrbitFilteringFnc
{
public:
    CCentrophilicOrbitFilteringFnc(double _weightFactor) :
      weightFactor(_weightFactor) {};
    virtual double eval(const COrbitDesc* orbit) const;
private:
    double weightFactor;   ///< returned value for centrophilic orbits
};

/** Orbit filtering function based on whether an orbit belongs to a given energy shell.
    Energy shells are taken from an existing shell-kinematic Schwarzschild data object.
    if model==NULL or numShell==0, any orbit is acceptable, otherwise 
    only if orbit energy is within numShell-1 energy shell.
*/
class CShellOrbitFilteringFnc: public CBasicOrbitFilteringFnc
{
public:
    CShellOrbitFilteringFnc(const CSchwDataKinemShell* _model, unsigned int _numShell) :
      model(_model), numShell(_numShell) {};
    virtual double eval(const COrbitDesc* orbit) const; ///< return 1 if the orbit is inside a given energy shell, 0 otherwise
private:
    const CSchwDataKinemShell* model;  ///< pointer to a shell-kinematic Schwarzschild data object from which the shell index for an orbit is determined
    unsigned int numShell;             ///< required shell index
};

//  Correspondence between Schwarzschild model names and types 
/// return the name for a given Schwarzschild data type
std::string getSchwDataNameByType(CBasicSchwData::SCHWDATATYPE DataType);
/// return Schwarzschild data type associated with the text name (Classic by default)
CBasicSchwData::SCHWDATATYPE getSchwDataTypeByName(std::string DataName);

/// convenience function to return an orbit runtime function creator associated with a given Schwarzschild data class
CBasicOrbitRuntimeFncCreator* createSchwCreator(const CBasicSchwData* schwData);

}