/** \file    energydiff.cpp
    \brief   Tool: Measure the rate of energy diffusion in N-body simulation
    \author  EV
    \date    2010-2013

This program is used to determine the rate of energy diffusion of particles in an N-body simulation 
of a system that is assumed to be in self-consistent equilibrium, and compare it against analytical 
prediction from the theory of two-body relaxation.

Input file in the NEMO format contains multiple time snapshots.
The program computes the squared change of particle energy, averaged over particles in a number of 
energy bins, as a function of time. For a diffusion process, this change should grow linearly with 
time, so a best-fit slope estimate is taken to be the energy diffusion rate. 
Output consists of the diffusion rates and their uncertainties, computed for each energy bin.
It can be compared with the analytical estimate of diffusion coefficient from a smooth mass model, 
scaled to the appropriate number of particles.

*/
#include <stdlib.h>
#include <stdinc.h>
#include <getparam.h>
#include <io_nemo.h>
#include <vector>
#include <algorithm>
#include <iostream>
#include <fstream>
#include <gsl/gsl_errno.h>
#include <gsl/gsl_fit.h>
#include "massmodel.h"
#include "utils.h"
#define DensBit (1<<15)
#define PotentialBit (1<<3)

/// \cond INTERNAL_DOCS
struct particle{
    int index;
    float mass, etot, l2;
    int bin;
    particle(): 
        index(0), mass(0), etot(0), l2(0), bin(-1) {}
    particle(int _index, float _mass, float _etot, float _l2):
        index(_index), mass(_mass), etot(_etot), l2(_l2), bin(-1) {}
};
bool compareindex(particle p1, particle p2)
{ return p1.index<p2.index; }
bool compareetot(particle p1, particle p2)
{ return p1.etot<p2.etot; }

bool comparepair(std::pair<double,double> x1, std::pair<double,double> x2)
{ return x1.first<x2.first; }
/// \endcond

extern "C" {
string defv[]=
{
    "in=???\n         input snapshot",
    "nbins=100\n      number of bins for which to calculate energy diffusion",
    "minbin=0\n       fraction of mass in the innermost bin (most bound particles) [default 1/nbins]",
    "rel=f\n          output relative energy change (dE/E)^2 instead of (dE)^2 and rel.ang.mom.",
    "tab=\n           spherical mass model (for calculation of relative ang.mom.), if not given compute from first snapshot",
    "mbh=0\n          additional point mass (black hole) at origin (which is not in snapshot)",
    "outdelta=\n      output file for the time evolution of deltaE,L for each bin (default $in.delta)",
    "outcoef=\n       output file for fitted diffusion coefficients deltaE,L/dt(E) (default $in.difcoef)",
    "VERSION=1.02\n   "__DATE__" eugvas",
    NULL
};
string usage="plot average energy and angular momentum change vs. time as a function of energy bin (measure relaxation rate)";

/// a modification of linear regression fit with zero intercept, which is more robust against outliers
int my_fit_mul(const double *x, const size_t, const double *y, const size_t, const size_t num, double *coef1, double *, double *sumsq)
{
    std::vector<double> yoverx(num);
    for(size_t i=0; i<num; i++)
        yoverx[i]=y[i]/x[i];
    size_t newnum=num;
    double median=0;
    size_t imin, imax;
    std::sort(yoverx.begin(), yoverx.end());
    // iteratively fit a straight line with a median slope, throw away outliers for which the 
    // difference in slope is more than a factor of two, until no such points left or less than 
    // a half of original points remained.
    do{
        median=yoverx[newnum/2];
        imin=0; imax=newnum;
        for(size_t i=0; i<newnum; i++)
        {
            if(yoverx[i]<median/2) imin=i;
            if(yoverx[i]>median*2 && imax==newnum) imax=i;
        }
        for(size_t i=imin; i<imax; i++) yoverx[i-imin]=yoverx[i];
        newnum=imax-imin;
    } while(newnum>num/2 && (imin>0 || imax<newnum));
    median=yoverx[newnum/2];
    *coef1=median;
    *sumsq=0;
    for(size_t i=0; i<num; i++) *sumsq += gsl_pow_2(y[i]-median*x[i]);
    return GSL_SUCCESS;
}

void nemo_main()
{
    gsl_set_error_handler(&smile::gsl_verbose_error_handler);

    int Nbody;
    float * pos=NULL, * vel=NULL, * mass=NULL, * pot=NULL;
    float * time=NULL, currenttime, inittime;
    int   * nbody=NULL, *bits=NULL;

    string in = getparam("in");
    int nbins = getiparam("nbins");
    double minbin=getdparam(const_cast<char*>("minbin"));
    if(minbin==0) minbin=1.0/nbins;
    bool relative = getbparam("rel");
    double bhmass=getdparam("mbh");
    std::string tabname=std::string(getparam("tab"));
    bool firstsnapshot=true;
    std::string outname=std::string(getparam("outdelta")); if(outname=="") outname=std::string(in)+std::string(".delta");
    std::string totname=std::string(getparam("outcoef")); if(totname=="") totname=std::string(in)+std::string(".difcoef");
    std::ofstream out;
    out.open(outname.c_str());
    std::cout.precision(5);

    /* This reads the data from the input snapshot */
    std::vector<particle> pointsinit;
    std::vector<double> avgenergy(nbins), maxenergy(nbins), binmass(nbins);
    std::vector<double> binboundaries;
    std::vector<int> binnum(nbins);
    std::vector< std::vector<double> > deltae(nbins), deltal2(nbins);
    std::vector<double> times, einit(nbins);
    smile::createNonuniformGrid(nbins, minbin, 1, false, &binboundaries);

    smile::CMassModel *model=NULL;

    while (io_nemo(in,"float,n,t,m,x,v,p,b,read",&nbody,&time,&mass,&pos,&vel,&pot,&bits)!=0) {
        binmass.assign(nbins,0);
        if((*bits & PotentialBit) == 0) { std::cerr << "Error! No data for potential\n"; return; }
        Nbody = *nbody;
        currenttime=*time;

        if(relative && model==NULL)  // need to initialize mass profile
        {
            std::vector<double> radii, inmass;
            if(!tabname.empty())  // read profile from supplied file
            {
                // read input table
                std::ifstream in(tabname.c_str());
                if(!in) {
                    std::cerr << "Invalid tab filename: " << tabname << "\n";
                    return;
                }
                while(in)
                {
                    std::string str;
                    std::getline(in, str);
                    if(in && str.size()>2 && str[0]!='#' && str[0]!='/' && str[0]!='%') 
                    {
                        std::istringstream stm(str);
                        double r=0, m=0;
                        stm >> r >> m;
                        radii.push_back(r);
                        inmass.push_back(m);
                    }
                }
            }
            else   // create profile from the 1st snapshot itself
            {
                // in this application, no smoothing of mass model is necessary, because it is 
                // only used for computing the radius of circular orbit as a function of energy,
                // not for the distribution function, unlike ``mkspherical''.
                std::cerr << "Creating spherical model at time "<<currenttime<<"\n";
                std::vector< std::pair<double,double> > radmass;
                double innerMass=0, totalMass=0;
                for(int i=0; i<Nbody; i++)
                {
                    radmass.push_back(std::pair<double,double>(sqrt(gsl_pow_2(pos[i*3])+gsl_pow_2(pos[i*3+1])+gsl_pow_2(pos[i*3+2])), mass[i]));
                    totalMass+=mass[i];
                }
                std::sort(radmass.begin(), radmass.end(), comparepair);
                std::vector< std::pair<double,double> >::iterator point=radmass.begin();
                int prevbin=0;
                radii.push_back(0);
                inmass.push_back(bhmass);
                while(point!=radmass.end())
                {
                    int bin = 0;
                    while(innerMass/totalMass>binboundaries[bin]) bin++;
                    if(bin!=prevbin && point!=radmass.begin())
                    {
                        radii.push_back((point-1)->first);
                        inmass.push_back(innerMass+bhmass);
                    }
                    innerMass += point->second;
                    prevbin=bin;
                    point++;
                }
                if(inmass.back()<innerMass)
                {
                    radii.push_back(radmass.back().first);
                    inmass.push_back(totalMass+0.1);
                }
            }
            model = new smile::CMassModel(radii, inmass);
            switch(model->getStatus())
            {
            case smile::CMassModel::MS_ARRAY_NOT_EQ:  std::cerr << "Invalid format of input file\n"; return;
            case smile::CMassModel::MS_ARRAY_SHORT:   std::cerr << "Too few points\n"; return; 
            case smile::CMassModel::MS_RADII_NOT_ASC: std::cerr << "Radii not in ascending order!\n"; return;
            case smile::CMassModel::MS_MASS_NOT_ASC:  std::cerr << "Mass not increasing with radius!\n"; return;
            case smile::CMassModel::MS_DF_NEGATIVE:   std::cerr << "Warning, f(E) not positive!\n"; break; // no fail
            }
        }
        std::vector<particle> points(Nbody);
        double totalMass=0;
        for(int i=0; i<Nbody; i++)
        {
            float 
                lx=pos[i*3+1]*vel[i*3+2]-pos[i*3+2]*vel[i*3+1], 
                ly=pos[i*3+2]*vel[i*3]-pos[i*3]*vel[i*3+2], 
                lz=pos[i*3]*vel[i*3+1]-pos[i*3+1]*vel[i*3], 
                l2=sqrt(lx*lx+ly*ly+lz*lz),
                E=( (vel[i*3]*vel[i*3]+vel[i*3+1]*vel[i*3+1]+vel[i*3+2]*vel[i*3+2])/2 + pot[i]);
            if(relative && model)
            {
                float Lcirc=E<0 ? static_cast<float>(model->lcirc(E)) : l2;  // if E>=0 this point will not be used anyway
                if(gsl_isnan(Lcirc)) { Lcirc=l2; mass[i]=0; }   // energy is below the minimum value for the initial snapshot, discard this point
                l2 = l2*l2/Lcirc/Lcirc;
            }
            if(E>0) { mass[i]=0; }   // exclude this point from averaging
            points[i]=particle(i, mass[i], pot[i] + (vel[i*3]*vel[i*3]+vel[i*3+1]*vel[i*3+1]+vel[i*3+2]*vel[i*3+2])/2, l2);
            totalMass += mass[i];
        }
        if(!firstsnapshot)
        {
            std::vector<double> energydiff(nbins), l2diff(nbins);
            for(int i=0; i<Nbody; i++)
            {
                binmass[pointsinit[i].bin] += points[i].mass;
                energydiff[pointsinit[i].bin] += gsl_pow_2(pointsinit[i].etot-points[i].etot) * points[i].mass
                    * (relative ? 1/gsl_pow_2(pointsinit[i].etot) : 1);
                l2diff[pointsinit[i].bin] += gsl_pow_2(pointsinit[i].l2-points[i].l2) * points[i].mass;
            }

            out << currenttime;
            times.push_back(currenttime-inittime);
            for(int b=0; b<nbins; b++)
            {
                if(binmass[b])
                {
                    energydiff[b] /= binmass[b];
                    l2diff[b] /= binmass[b];
                }
                if(b<nbins) 
                {
                    deltae [b].push_back(energydiff[b]);
                    deltal2[b].push_back(l2diff[b]);
                    out << "\t" << energydiff[b] << "\t" << l2diff[b];
                }
            }
            out << "\n" << std::flush;
        } else {  // first snapshot
            inittime=currenttime;
            std::sort(points.begin(), points.end(), compareetot); // sort in energy
            std::vector<particle>::iterator point=points.begin();
            double innerMass=0;
            while(point!=points.end())
            {
                innerMass += point->mass;
                point->bin = 0;
                while(innerMass/totalMass>binboundaries[point->bin]) point->bin++;
                binmass[point->bin] += point->mass;
                binnum[point->bin]++;
                avgenergy[point->bin] += point->etot*point->mass;
                maxenergy[point->bin] = point->etot;
                point++;
            }
            std::sort(points.begin(), points.end(), compareindex); // restore order
            pointsinit=points;
            std::cerr << "Nbody="<<Nbody<<"\tMass="<<totalMass<<"\n";
            out << "#init.E,Eb";
            for(int b=0; b<nbins; b++) 
            {
                if(binmass[b])
                {
                    avgenergy[b] /= binmass[b];
                }
                if(b<nbins) {
                    einit[b]=avgenergy[b];
                    out << "\t"<<avgenergy[b] << "\t" << maxenergy[b];
                }
            }
            out << "\n";
        }
        firstsnapshot=false;
    }
    out.close();
    // fit energy and ang.mom. diffusion rates in each bin
    out.open(totname.c_str());  
    size_t nump=times.size();
    for(int b=0; b<nbins; b++)
    {
        out << einit[b] << "\t";
        double coef, cov, chisq, disp;
        {
            my_fit_mul(&(times.front()), 1, &(deltae[b].front()), 1, nump, &coef, &cov, &chisq);
            disp=sqrt(chisq/nump)/(coef*times.back());
            out << coef << "\t" << disp << "\t";
            my_fit_mul(&(times.front()), 1, &(deltal2[b].front()), 1, nump, &coef, &cov, &chisq);
            disp=sqrt(chisq/nump)/(coef*times.back());
            out << coef << "\t" << disp << "\n";
        }
    }
    return;
}

}
