/** \file   mkspherical.cpp
    \brief  Tool: Create spherical mass models from a text table or from Nbody snapshots
    \author EV
    \date   2010-2013

This program uses spherical mass models for two different purposes and created from two possible 
sources:  either from a supplied table with r, m(r) values specifying enclosed mass as a function 
of radius, or an N-body snapshot, in which case it first fits a penalized spline to compute M(r) 
from the snapshot. 

The spherical model, in turn, may be used to compute a number of parameters as spline-interpolated 
functions of radius (potential, radial/circular period, distribution function via Eddington 
inversion formula, diffusion coefficients, etc.), or to generate an N-body snapshot, in which 
particles are distributed according to the given density profile and isotropically in velocities.

In short, this is a generalization of tools such as `halogen` or `spherICs` for creating 
a spherical isotropic model with a given arbitrary density profile, and at the same time a useful 
tool to study dynamical properties of a given \Nbody system (or, rather, its spherically-symmetric 
isotropic counterpart).

Two compilation options are available: either to use NEMO machinery for parsing input parameters 
from the command-line, or use a separate lightweight class in config.h
*/

#ifdef HAVE_UNSIO
#define HAVE_NEMO
#endif

extern "C" { 
/// default values of command-line parameters
char* defv[]=
{
  "intab=\n        input table with r and M(r) values",
  "outtab=\n       output table with r,M,Phi,f(E) and other parameters",
  "rmin=0\n        output table inner nonzero radius [0 for auto-detect]",
  "rmax=0\n        output table outer radius [0 for auto-detect]",
  "npoints=100\n   output table number of points in radius [-1 to use input points]",
  "insnap=\n       input snapshot name (to construct spherical model from a snapshot instead of a table)",
  "smoothing=1\n   smoothing coefficient for the input snapshot M(r) profile",
  "mbh=0\n         additional point mass (black hole) at origin",
  "outsnap=\n      output snapshot file name (to create an N-body representation of the model)",
  "outformat=Nemo\noutput snapshot format [Text/Nemo/Gadget]",
  "nbody=0\n       number of points to create",
  "quiet=0\n       0=noisy 1=somewhat quiet 2=more quiet", 
  "elow=0\n        lowest(negative) energy of emitted particles", 
  "VERSION=2.0\n   "__DATE__" eugvas",
  0
};
/// short description
const char* usage_my="Create spherical isotropic models for a given mass profile using Eddington inversion.\n"
  "Either an input table with M(r) values or an existing N-body snapshot should be given to create a model.\n"
  "output may consist of an extended table with many parameters as functions of radius and energy, and/or "
  "an N-body snapshot corresponding to the spherical isotropic model (or both a table and a snapshot)";
}

#include <cmath>
#include <vector>
#include <algorithm>
#include <iostream>
#include <fstream>
#include <sstream>
#include <time.h>
#include <gsl/gsl_rng.h>
#include <gsl/gsl_fit.h>
#include "common.h"
#include "iosnapshot.h"
#include "massmodel.h"
#include "utils.h"
#ifdef HAVE_NEMO
#include <stdinc.h>
#include <getparam.h>
#ifndef HAVE_UNSIO
extern "C" { char* usage=const_cast<char*>(usage_my); }
#endif
#else
#include "config.h"
#endif

/// sort pairs of floating-point values by the first element
bool sortfnc(const std::pair<float, float> &elem1, const std::pair<float, float> &elem2)
{   return elem1.first<elem2.first; }

/// \cond INTERNAL_DOCS
typedef struct{
    double* xdata;
    double* ydata;
    size_t  npoints;
} fitdata;
/// \endcond
/// helper routine to find extrapolated mass of an Nbody snapshot at infinity Mtot 
/// by requiring that log(M(r)/(Mtot-M(r))) vs log(r) is best described by a linear relation close to outer radius
double getfitquality(double totalmass, void* params)
{
    size_t npoints=static_cast<fitdata*>(params)->npoints;
    double* xdata=static_cast<fitdata*>(params)->xdata;
    double* ydata=static_cast<fitdata*>(params)->ydata;
    std::vector<double> y1data(npoints);
    for(size_t i=0; i<npoints; i++)
        y1data[i]=log(ydata[i]/(totalmass-ydata[i]));
    double cov00, cov01, cov11, sumsq, a, b;
    gsl_fit_linear(xdata, 1, &(y1data.front()), 1, npoints, &a, &b, &cov00, &cov01, &cov11, &sumsq);
    double norm=b*(xdata[npoints-1]-xdata[0]);  // normalization factor: difference between first and last point according to linear fit
    return sumsq/(norm*norm);
}

/// replacement for GSL error handler that doesn't crash program
void gsl_quiet_error_handler (const char *reason, const char *file, int line, int er)
{
    if(er == GSL_ETOL || er == GSL_EROUND || er == GSL_ESING || er == GSL_EDIVERGE || er == GSL_EMAXITER) return;  // do nothing
    std::cerr << "GSL error " << er << " in " << file << " at line " << line << ": " << reason << "\n";
}

/// main program begins here
int main(int argc, char* argv[])
{
    for(int arg=1; arg<argc; arg++) 
        if(std::string(argv[arg]).substr(0,4) == "help") {  // display help and return
            std::cerr << usage_my << "\n";
            char** par=defv;
            while(*par!=0) {
                std::string parstr(*par);
                size_t pp=parstr.find('\n');
                if(pp!=std::string::npos) parstr[pp]=' ';
                std::cerr << parstr << "\n";
                par++;
            }
            return 0;
        }
    initparam(argv, defv);
    gsl_set_error_handler(&gsl_quiet_error_handler);
    std::string inputtab = getparam("intab");
    std::string outputtab= getparam("outtab");
    std::string inputsnap= getparam("insnap");
    std::string outputsnap=getparam("outsnap");
    std::string outputformat=getparam("outformat");
    double gridrmin = getdparam("rmin");
    double gridrmax = getdparam("rmax");
    int numNodes=getiparam("npoints");
    double smoothing = getdparam("smoothing"); 
    double bhmass = getdparam("mbh"); 
    if(smoothing==0) smoothing=1; //?! always use some smoothing, even if the user did not ask...
    int nbody=getiparam("nbody");
    int quiet=getiparam("quiet");
    double elow = getdparam("elow");
    // r, M(r) table to be initialized either directly from a text file or from an N-body snapshot
    std::vector<double> radii, inmass;
    if(inputtab!="" && inputsnap!="")
    {
        std::cerr << "Must give either an input snapshot or M(r) table file name, not both\n";
        return 0;
    }
    if(outputsnap!="" && nbody==0 || outputsnap=="" && nbody>0) {
        std::cerr << "Must give both output snapshot filename and number of bodies\n";
        return 0;
    }
    if(inputsnap!="")
    {
        size_t nbody_in=0;
        smile::vectorpf radmass;
        smile::CBasicIOSnapshot* snap_in = smile::createIOSnapshotRead(inputsnap);
        smile::CPointMassSet<float>* points_in=NULL;
        if(snap_in!=NULL && snap_in->readSnapshot(points_in))
        {
            nbody_in = points_in->size();
            if(nbody_in<100) {
                std::cerr << "Too few points for a meaningful smooth model\n";
                return 0;
            }
            radmass.reserve(nbody_in);
            for(size_t i=0; i<nbody_in; i++)
            {
                radmass.push_back( std::pair<float, float> (
                    sqrt(smile::pow_2(points_in->at(i).first.Pos[0]) + 
                         smile::pow_2(points_in->at(i).first.Pos[1]) + 
                         smile::pow_2(points_in->at(i).first.Pos[2]) ),
                    points_in->at(i).second ) );
            }
            delete points_in;
        } else {
            std::cerr << "Can't read input snapshot file: " << inputsnap << "\n";
            return 0;
        }
        // sort particules in radius and initialize M(r) grid by fitting a smoothing spline to 
        // log(r), log[M(r)]
        std::sort(radmass.begin(), radmass.end(), sortfnc);
        double totalmass=0;
        std::vector<double> logradii, logmassrel;
        logradii.reserve(nbody_in);
        logmassrel.reserve(nbody_in);
        for(size_t i=0; i<nbody_in; i++)
        {
            if(radmass[i].first<=1e-10)
                bhmass+=radmass[i].second;
            else 
            {
                totalmass+=radmass[i].second;
                logradii.push_back(log(radmass[i].first));
                logmassrel.push_back(totalmass);  // not log-mass yet; will require one more step to find extrapolated Mtot 
            }
        }
        size_t nbody1=logradii.size();
        size_t minbin=static_cast<size_t>(10 + 5*log10(nbody1/1000.));  if(minbin<5) minbin=5;
        size_t numgrid=static_cast<size_t>(10*pow(nbody1/1000., 1./3)); if(numgrid*minbin>nbody1) numgrid=nbody1/minbin;
        // assign knots in radial grid trying to achieve equal log-r intervals 
        // but requiring that each interval holds at least minbin points
        double innerrad=logradii.front(), outerrad=logradii.back();
        std::vector<double> knots(numgrid);   // equidistant grid in log radius
        std::vector<double> gridlogmassrel(numgrid); // values of log-scaled mass returned by the fitting routine
        smile::createAlmostUniformGrid(logradii, minbin, &numgrid, &knots);
        // find "totalmassfit" such that log(m[i]/(totalmassfit-m[i])) is best described by a linear function in log(r)
        gsl_function Fmin;
        fitdata Fdata;
        size_t nfit=1; 
        while(nfit<nbody1 && logradii[nbody1-nfit]>knots[knots.size()-3]) nfit++;
        Fdata.xdata=&(logradii[nbody1-nfit]);
        Fdata.ydata=&(logmassrel[nbody1-nfit]);
        Fdata.npoints=nfit;
        Fmin.function=&getfitquality;
        Fmin.params=&Fdata;
        double totalmassfit=smile::findMinGuess(&Fmin, totalmass*1.0000001, totalmass*2);
        for(size_t i=0; i<nbody1; i++)
            logmassrel[i]=log( logmassrel[i]/(totalmassfit-logmassrel[i]));  // scale to M(r)/(Mtot-M(r))
        // init penalized spline fitting routine
        smile::CSplineApprox fit(logradii, knots);
        if(fit.status()!=smile::CSplineApprox::AS_OK && fit.status()!=smile::CSplineApprox::AS_SINGULAR) {
            std::cerr << "Error in creating smoothing spline for M(r)\n";
            return 0;
        }
        fit.loadyvalues(logmassrel);
        double rmserr, edf, lambda;
        smile::CSplineApprox::APPROXSTATUS status = fit.fitDataOversmooth(smoothing, &gridlogmassrel, &rmserr, &edf, &lambda);
        if(status != smile::CSplineApprox::AS_OK && status != smile::CSplineApprox::AS_SINGULAR) {
            std::cerr << "Error in fitting smoothing spline for M(r)\n";
            return 0;
        }
        else 
            std::cerr << "nbody mass="<<(totalmass+bhmass)<<
            ", fit mass="<<(bhmass+totalmassfit/(1+exp(-gridlogmassrel[knots.size()-1])))<<
            ", extrapolated mass="<<(totalmassfit+bhmass)<<
            "\nrmserr="<<rmserr<<", edf(# independent radial pts)="<<edf<<", lambda="<<lambda <<
            "\ngrid in radii from "<<exp(knots.front()) << " to " << exp(knots.back()) << "\n";
        // scale back from log-scaling
        if(bhmass>0) {
            radii.push_back(0);
            inmass.push_back(bhmass);
        }
        for(size_t k=0; k<knots.size(); k++)
        {
            double masscur=bhmass + totalmassfit/(1+exp(-gridlogmassrel[k]));
            if(inmass.size()==0 || masscur>inmass.back()) {   // enforce strictly increasing mass with radius
                radii.push_back(exp(knots[k]));
                inmass.push_back(masscur);
            }
        }   // model will be initialized from these two arrays
#if 0
        std::vector<double> logmassfit;
        fit.computeRegressionAtPoints(logradii, &logmassfit);
        std::ofstream tabout((inputsnap+"t").c_str(), std::ios_base::out);
        for(size_t i=0; i<logradii.size(); i++)
            tabout << logradii[i] << "\t" << logmassrel[i] << "\t" << logmassfit[i] << "\n";
        tabout << "\n";
        for(size_t k=0; k<radii.size(); k++)
            tabout << radii[k] << "\t" << inmass[k] << "\n";
#endif
    }
    else if(inputtab!="")
    {
        // read input table
        std::ifstream in(inputtab.c_str());
        if(!in) {
            std::cerr << "Invalid input table filename: " << inputtab << "\n";
            return 0;
        }
        while(in)   // parse input file
        {
            std::string str;
            std::getline(in, str);
            if(in && str.size()>2 && str[0]!='#' && str[0]!='/' && str[0]!='%') 
            {
                std::istringstream stm(str);
                double r=0, m=0;
                stm >> r >> m;
                radii.push_back(r);
                inmass.push_back(m+bhmass);
            }
        }
    }
    else {
        std::cerr << "Must give either an input snapshot or M(r) table file name\n";
        return 0;
    }
    smile::CMassModel model(radii, inmass);
    if( model.getStatus()==smile::CMassModel::MS_ARRAY_SHORT ||
        model.getStatus()==smile::CMassModel::MS_RADII_NOT_ASC ||
        model.getStatus()==smile::CMassModel::MS_MASS_NOT_ASC ) 
    { // before reporting error, dump out r-m(r) table
        std::cerr << "radius\tmass\n";
        for(size_t i=0; i<radii.size(); i++) std::cerr << radii[i]<<"\t"<<inmass[i]<<"\n";
    }
    switch(model.getStatus())
    {
    case smile::CMassModel::MS_ARRAY_NOT_EQ:  std::cerr << "Invalid format of input table file (columns of unequal length)\n"; return 0;
    case smile::CMassModel::MS_ARRAY_SHORT:   std::cerr << "Too few points\n"; return 0; 
    case smile::CMassModel::MS_RADII_NOT_ASC: std::cerr << "Radii not in ascending order!\n"; return 0;
    case smile::CMassModel::MS_MASS_NOT_ASC:  std::cerr << "Mass not increasing with radius!\n"; return 0;
    case smile::CMassModel::MS_DF_NEGATIVE:   std::cerr << "Warning, f(E) not positive!\n"; break; // no fail
    }
    double rho0, drhodr, alpha;
    if(model.hasDensityCore(&rho0, &drhodr, &alpha))
        std::cerr << "Constant-density core determined: rho(r) = " << rho0 << "*(1-" << drhodr << "*r^" << alpha << ")\n";

    if(outputtab!="")   // output table with all computed quantities vs radius
    {
        std::ofstream out(outputtab.c_str(), std::ios_base::out);
        if(!out) {
            std::cerr << "Cannot write output table: " << outputtab << "\n";
            return 0;
        }
        std::vector<double> gridRadius;
        if(numNodes<0) {  // use input points, excluding r=0
            std::vector<double>::iterator beg=radii.begin();
            while(beg!=radii.end() && *beg<=0) beg++;
            gridRadius.assign(beg, radii.end());
            numNodes=static_cast<int>(gridRadius.size());
        } else {  // create a log-grid with reasonable inner and outer radii
            double zeroMass=model.mass(0), totalMass=model.getTotalMass()-zeroMass;
            if(gridrmin==0) {
                gridrmin=0.125;
                while(gridrmin>1e-6 && model.mass(gridrmin)-zeroMass>totalMass*1e-6) gridrmin/=2;  // a quick default algo
            }
            if(gridrmax==0) {
                gridrmax=8;
                while(gridrmax<1e6 && model.mass(gridrmax)-zeroMass<totalMass*(1-1e-3)) gridrmax*=2;
            }
            if(numNodes==0) numNodes=100;
            smile::createNonuniformGrid(numNodes, gridrmin, gridrmax, false, &gridRadius);
        }
        out << "#r\tM(r)\tPhi=E\trho\tf(E)\tg(E)\tM(E)\tperiod(E)\tr_circ(E)\tL_circ(E)\tsigma\tsigma_los\tdens_proj\tD_EE\tD_RR\n";
        out << 0 << "\t" << model.mass(0) << "\t" << model.pot(0) << "\n";
        for(int i=0; i<numNodes; i++)
        {
            if(i%10==0) std::cerr << (i*100/numNodes) << "%\r";
            double rad = gridRadius[i];
            double pot = model.pot(rad);
            double DEE, DRR;
            model.DifCoefAvE(pot, &DEE, &DRR, NULL, NULL);
            out << rad << "\t" << smile::convertToString(model.mass(rad),12) << "\t" 
                << smile::convertToString(pot,10) << "\t" << smile::convertToString(model.dens(rad),10) << "\t" 
                << model.distrf(pot) << "\t" << model.ded(pot) << "\t" << model.masse(pot) << "\t"
                << model.trad(pot) << "\t" << model.rcirc(pot) << "\t" << model.lcirc(pot) << "\t"
                << model.veldisp(rad) << "\t" << model.losvd(rad) << "\t" << model.densproj(rad) << "\t"
                << DEE << "\t" << DRR;
            out << "\n";
        }
        std::cerr << "100%\n";
        out.close();
    }

    if(outputsnap!="" && nbody>0)
    {
        // generate initial conditions using derived d.f.
        double totalmasswithoutbh=(inmass.back()-inmass.front());
        double maxrad=1, maxmass=totalmasswithoutbh*(1-1.0/nbody);  
        if(inmass[0]>0)  nbody++; //bh
        smile::CPointMassSet<float>* points_out=new smile::CPointMassSet<float>();
        points_out->data.reserve(nbody);
        if(inmass[0]>0)  // central black hole
            points_out->add(smile::CPosVelPoint<float>(), static_cast<float>(inmass[0]));
        while(maxrad<1e6 && model.mass(maxrad) < inmass[0]+maxmass) maxrad*=2;
        if(model.mass(maxrad)-inmass[0]<maxmass) maxmass=model.mass(maxrad)-inmass[0];
        gsl_rng* randn=gsl_rng_alloc(gsl_rng_default);
        gsl_rng_set(randn, (unsigned)time( NULL ));
        for(int ind=(inmass[0]>0); ind<nbody; ind++)  // start from 0 or 1
        {   // sample accessible configuration space with appropriate weight
            double radius=0, veloc, poten, encmass;
            // first pick up position not too close to BH
            if (quiet==1) 
                encmass=(ind+gsl_rng_uniform(randn))/nbody;
            else if (quiet==2) 
                encmass = (ind+0.5)/nbody;
            else
                encmass=gsl_rng_uniform(randn);
            encmass=maxmass*encmass + inmass[0];
            radius=model.rad(encmass);
            double costh=gsl_rng_uniform(randn)*2.0-1;
            double sinth=sqrt(1-costh*costh);
            double phi=gsl_rng_uniform(randn)*2*M_PI;
            smile::CPosVelPoint<float> pt(
                static_cast<float>(radius*sinth*cos(phi)),
                static_cast<float>(radius*sinth*sin(phi)), 
                static_cast<float>(radius*costh), 0, 0, 0);
            poten=model.pot(radius);
            do{
                veloc=model.samplev(radius);
            } while(elow!=0 && poten+veloc*veloc/2<elow);
            costh=gsl_rng_uniform(randn)*2.0-1;
            sinth=sqrt(1-costh*costh);
            phi=gsl_rng_uniform(randn)*2*M_PI;
            pt.Vel[0]=static_cast<float>(veloc*sinth*cos(phi)); 
            pt.Vel[1]=static_cast<float>(veloc*sinth*sin(phi)); 
            pt.Vel[2]=static_cast<float>(veloc*costh);
            points_out->add(pt, static_cast<float>(totalmasswithoutbh/(nbody - (inmass[0]>0 ? 1 : 0))) );
            //pot[ind]=static_cast<float>(poten);
        }
        gsl_rng_free(randn);
        std::string cmdlinefull="mkspherical";
        char** argv0=argv+1;
        for(char *arg=*argv0; (arg = *argv0); argv0 ++ ){
            cmdlinefull += " " + std::string(arg);
        }
        smile::CBasicIOSnapshot* snap_out = smile::createIOSnapshotWrite(outputformat, outputsnap, cmdlinefull);
        if(snap_out!=NULL) {
            if(!snap_out->writeSnapshot(points_out)) 
                std::cerr << "Can't write output snapshot: " << outputsnap << "\n";
        } else {
            std::cerr << "Unknown output format: " << outputformat << "\n";
        }
        delete snap_out;
        delete points_out;
    }

    finiparam();
    return 0;
}
