/** \file   renderdensity.cpp
    \brief  Tool: Create a visual representation of a non-spherical density model specified by a BSE/Spline potential
    \author EV
    \date   2011-2013

Construct an Nbody snapshot with particles distributed according to the density profile of 
a BSE or Spline potential expansion specified by a coefficients file. 
Particles are given zero velocity and thus the snapshot is not an equilibrium model.
*/

#ifdef HAVE_UNSIO
#define HAVE_NEMO
#endif

extern "C" {
/// default values of command-line parameters
char* defv[]=
{
  "in=???\n         input coefs text-file",
  "out=???\n        output snapshot",
  "nbody=???\n      number of points in snapshot",
  "outformat=Nemo\n output format (Text/Nemo/Gadget)",
  "VERSION=1.3\n    Jan 2013 EV",
  0
};
/// short description
const char* usage_my="create nbody snapshot with density from basis-set or spline expansion supplied as an input text file";
}

#include <cmath>
#include <vector>
#include <iostream>
#include <time.h>
#include <gsl/gsl_rng.h>
#include "iopotential.h"
#include "iosnapshot.h"
#include "potential.h"
#include "massmodel.h"
#include "utils.h"
#ifdef HAVE_NEMO
#include <stdinc.h>
#include <getparam.h>
#ifndef HAVE_UNSIO
extern "C" { char* usage=const_cast<char*>(usage_my); }
#endif
#else
#include "config.h"
#endif

int main(int argc, char* argv[])
{
    initparam(argv, defv);
    std::string infilename  = getparam("in");
    std::string outfilename = getparam("out");
    std::string outformat   = getparam("outformat");
    int nbody = getiparam("nbody");
    gsl_set_error_handler(&smile::gsl_verbose_error_handler);

    if(infilename.empty() || outfilename.empty() || nbody<=0) 
    {
        std::cerr << "Usage: renderdensity in=(coef_file) out=(snapshot_file) nbody=??? [outformat=Text/Nemo/Gadget]\n";
        return 0;
    }
    smile::CConfigPotential configPotential;
    configPotential.N_dim=3;
    configPotential.PotentialType=smile::CPotential::PT_UNKNOWN;
    configPotential.Mbh=0;
    const smile::CPotential* potential=smile::readPotential(infilename, &configPotential);
    if(potential==NULL)
    {
        std::cerr << "Error loading potential from file " << infilename << "\n";
        return 0;
    }
    // create a spherical mass model corresponding to the potential in question
    double totalMass=potential->totalMass();
    double rin  = potential->getRadiusByMass(totalMass*0.001);
    double rout = potential->getRadiusByMass(totalMass*0.999);
    smile::CMassModel* massModel=NULL;
    if(totalMass>0 && rout>0 && rin>0 && rout>rin)
    {
        std::cerr << "Creating mass model for "<<rin<<"<r<"<<rout<<", Mtotal="<<totalMass<<"\n";
        int numNodes=100;
        std::vector<double> radii(numNodes), inmass(numNodes);
        smile::createNonuniformGrid(numNodes-1, rin, rout, true, &radii); 
        radii.push_back(rout*10);
        for(int i=1; i<numNodes; i++)
            inmass[i]=potential->Mass(radii[i]);
        // ensure that mass strictly increases with radius (who knows what a heck could happen...)
        for(int i=1; i<numNodes; i++)
            if(inmass[i]<=inmass[i-1])
            {
                radii.erase(radii.begin()+i);
                inmass.erase(inmass.begin()+i);
                numNodes--;
                i--;
            }
        massModel=new smile::CMassModel(radii, inmass);
        if( massModel->getStatus()!=smile::CMassModel::MS_OK && 
            massModel->getStatus()!=smile::CMassModel::MS_DF_NEGATIVE) 
        { delete massModel; massModel=NULL; }
    }
    if(massModel==NULL) {
        std::cerr << "Error creating mass model\n";
        return 0;
    }
    // sample points according to M(r) from the spherical model and uniformly in angles
    double totalMassSpher=massModel->getTotalMass()*0.999;
    int nSampling = std::max(nbody*100, 10000);
    // set of trial points from which the resulting points will be drawn
    std::vector< smile::CPosPoint<float> > sampPoints(nSampling);
    std::vector<double> weight(nSampling);  // weights assigned to trial points
    double totalWeight=0, prevradius=0;
    gsl_rng* randn=gsl_rng_alloc(gsl_rng_default);
    gsl_rng_set(randn, (unsigned)time( NULL ));
    for(int ind=0; ind<nSampling; ind++)
    {
        double radius=massModel->rad( (ind+1.0)/nSampling * totalMassSpher);
        double costh=gsl_rng_uniform(randn)*2.0-1;
        double sinth=sqrt(1-costh*costh);
        double phi=gsl_rng_uniform(randn)*2*M_PI;
        sampPoints[ind].Pos[0] = static_cast<float>(radius*sinth*cos(phi));
        sampPoints[ind].Pos[1] = static_cast<float>(radius*sinth*sin(phi));
        sampPoints[ind].Pos[2] = static_cast<float>(radius*costh);
        weight[ind] = potential->Rho(sampPoints[ind].Pos[0], sampPoints[ind].Pos[1], sampPoints[ind].Pos[2]) 
            * 4*M_PI*radius*radius*(radius-prevradius);
        totalWeight += weight[ind];
        prevradius=radius;
        if(ind%1000==0) std::cerr << static_cast<int>(ind*100.0/nSampling) << "%\r";
    }
    gsl_rng_free(randn);
    std::cerr << "      \r";
    // select points from sampled
    smile::CPointMassSet<float>* points=new smile::CPointMassSet<float>();
    points->data.reserve(nbody);
    float pointMass=static_cast<float>(potential->totalMass() / nbody);
    int ns=0, nd=0; double ws=0;
    // if a sampling point has too large weight, it may be duplicated with possible reflection 
    // in up to 8 directions, depending on the potential symmetry
    int mult[8][3]={ {1,1,1}, {-1,-1,-1}, {1,-1,-1}, {-1,1,-1}, {-1,-1,1}, {-1,1,1}, {1,-1,1}, {1,1,-1} };
    int maxmult= potential->symmetry() == smile::CPotential::ST_NONE ? 1 : 
                 potential->symmetry() == smile::CPotential::ST_REFLECTION ? 2 : 8;
    for(int np=0; np<nbody; np++)
    {
        int inc=0;
        while(ws+weight[ns]<(np+1)*totalWeight/nbody && ns<nSampling-1) { ws += weight[ns]; ns++; inc++; }
        if(inc==0) nd++; else nd=0;
        float Xs=sampPoints[ns].Pos[0];
        float Ys=sampPoints[ns].Pos[1];
        float Zs=sampPoints[ns].Pos[2];
        if(nd>=maxmult) { 
            nd-=maxmult; 
            std::cerr << "duplicate point #"<<np<<"/"<<ns<<" at "<<Xs<<","<<Ys<<","<<Zs<<std::endl; 
        }
        points->add(smile::CPosVelPoint<float>(Xs*mult[nd][0], Ys*mult[nd][1], Zs*mult[nd][2], 0, 0, 0), pointMass );
    }
    std::string cmdlinefull="renderdensity";
    char** argv0=argv+1;
    for(char *arg=*argv0; (arg = *argv0); argv0 ++ ){
        cmdlinefull += " " + std::string(arg);
    }
    smile::CBasicIOSnapshot* snap_out = smile::createIOSnapshotWrite(outformat, outfilename, cmdlinefull);
    if(snap_out!=NULL) {
        if(!snap_out->writeSnapshot(points)) 
            std::cerr << "Can't write output snapshot: " << outfilename << "\n";
    } else {
        std::cerr << "Unknown output format: " << outformat << "\n";
    }
    delete snap_out;
    delete points;
    finiparam();
    return 0;
}
