/** \file    snaporbits.cpp
    \brief   Tool: Perform orbit analysis on trajectories of particles from an N-body snapshot
    \author  EV
    \date    2012-2013
    
This program demonstrates how to use SMILE orbit analysis machinery on trajectories from an N-body 
simulation. It loads NEMO snapshot file containing particles at multiple moments of time and 
treats trajectories of each particle as a separate orbit, performing the same kind of orbit 
analysis as for loading an orbit from a text file in SMILE GUI. 

To conduct orbit analysis, one needs to provide the potential. It may be done either by specifying 
a potential coefs file or simply by using the first snapshot in the input file as the source data 
for potential approximation (using BSE or Spline potential expansions). 
In this latter case, several additional parameters may be given for initializing this potential 
approximation (number of radial and angular coefficients, symmetry and potential type), but they 
are not very important, since the potential is only used to compute the orbit energy and the 
natural orbital period, which serves as the unit of frequency. The potential coefficients are 
stored in a text file to simplify latter use (for instance, when loading the orbit library file 
to SMILE, one may provide this text file for the potential model). In addition, a point mass 
(central black hole) may be added to the potential. Beware that initializing the potential from 
the 1st snapshot itself or from another snapshot file will only account for the central point mass 
if it is precisely at origin; otherwise the potential will not be initialized properly. 
On the other hand, when loading the potential coefs from a text file, this additional central mass 
may need to be provided manually since it is not stored in the coefs file.

Once the entire snapshot file has been read, the orbit analysis is performed on each particle 
trajectory, and the resulting orbit library is saved in the output text file. 

Note that all particles at all times are loaded to memory; to reduce memory demand, one may need 
to filter the input snapshot by standard NEMO tools such as 
`snapmask input_file - select=0:1000 | snaporbits - out=output_library pot=input_file`
(note that in this case we must provide the original input file containing all particles that 
contribute to the potential, not only the subset of particles to be processed).

**/
#include <vector>
#include <iostream>
#include <stdinc.h>      // from NEMO
#include <getparam.h>    // from NEMO
#include <io_nemo.h>     // from NEMO
#include "potential.h"   // from SMILE
#include "orbitan.h"     // from SMILE
#include "orbitlib.h"    // from SMILE
#include "iopotential.h" // from SMILE
#include "iodata.h"      // from SMILE
#include "utils.h"       // from SMILE
#include <gsl/gsl_errno.h>

extern "C" {
/// default values for command-line options
string defv[]=
{
  const_cast<char*>("in=???\n         input snapshot containing trajectories to be analyzed (test particles)"),
  const_cast<char*>("out=???\n        output text file with orbit library"),
  const_cast<char*>("pot=\n           text file with potential coefs, or snapshot containing field particles that create potential (if not given, use 1st snapshot from 'in')"),
  const_cast<char*>("nrad=20\n        number of radial grid points in potential approximation"),
  const_cast<char*>("nang=6\n         number of angular harmonics"),
  const_cast<char*>("sym=t\n          assumed symmetry type: [n]one, [r]eflection, [t]riaxial, [a]xisymmetric, [s]pherical"),
  const_cast<char*>("pottype=s\n      potential approximation type: [b]se or [s]pline"),  
  const_cast<char*>("mbh=0\n          additional point mass (black hole) at origin"),
  const_cast<char*>("VERSION=1.02\n   "__DATE__" eugvas"),
  NULL
};
/// short description
string usage_my=const_cast<char*>("perform orbit analysis for orbits from Nbody snapshots");
#ifndef HAVE_UNSIO
char* usage=const_cast<char*>(usage_my); 
#endif

/// main program
void nemo_main()
{
  gsl_set_error_handler(&smile::gsl_verbose_error_handler);

  unsigned int Nbody;
  float * pos=NULL, * vel=NULL, * mass=NULL;
  float * time=NULL, currenttime;
  int   * nbody=NULL;

  std::string infile    = std::string(getparam(const_cast<char*>("in")));
  std::string outfile   = std::string(getparam(const_cast<char*>("out")));
  std::string potfile   = std::string(getparam(const_cast<char*>("pot")));
  int nrad = getiparam(const_cast<char*>("nrad"));
  int nang = getiparam(const_cast<char*>("nang"));
  double mbh = getdparam(const_cast<char*>("mbh"));
  const smile::CPotential* potential=NULL;
  if(!potfile.empty())  // try to load potential coefs from a text file
  {
    smile::CConfigPotential configPotential;
    configPotential.N_dim=3;
    configPotential.PotentialType=smile::CPotential::PT_UNKNOWN;  // means that we will try to read text file with potential coefs but no other kind of file
    configPotential.Mbh=mbh;
    potential=smile::readPotential(potfile, &configPotential);  // NULL if unsuccessful
  }
  if(potential==NULL)  // try to init from either NEMO snapshot in file "potfile" or from the first snapshot itself
  {
    std::cerr << "Creating density/potential model"<<std::endl;
    double * fieldpos=NULL, * fieldmass=NULL;
    if(potfile.empty()) potfile=infile;
    if(!io_nemo((char*)potfile.c_str(),const_cast<char*>("double,n,x,m,read"),&nbody,&fieldpos,&fieldmass) || *nbody<=0)
    {
      std::cerr << "Error, can't read field particles from " << potfile << std::endl; 
      return;
    }
    io_nemo((char*)potfile.c_str(),const_cast<char*>("close"));
    Nbody=*nbody;
    smile::CPointMassSetDouble points;
    for(unsigned int i=0; i<Nbody; i++)
    {
      points.data.push_back(std::pair< smile::CPosVelPoint<double>, double>
        (smile::CPosVelPoint<double>(fieldpos[i*3], fieldpos[i*3+1], fieldpos[i*3+2], 0, 0, 0), fieldmass[i]) );
    }
    std::string symmetry = std::string(getparam(const_cast<char*>("sym")));
    smile::CPotential::SYMMETRYTYPE symtype=smile::CPotential::ST_TRIAXIAL;
    if(!symmetry.empty())
    {
      if(symmetry.at(0)=='a') symtype=smile::CPotential::ST_AXISYMMETRIC;
      if(symmetry.at(0)=='s') symtype=smile::CPotential::ST_SPHERICAL;
      if(symmetry.at(0)=='r') symtype=smile::CPotential::ST_REFLECTION;
      if(symmetry.at(0)=='n') symtype=smile::CPotential::ST_NONE;
      if(symmetry.at(0)=='t') symtype=smile::CPotential::ST_TRIAXIAL;
    }
    std::string pottype = std::string(getparam(const_cast<char*>("pottype")));
    bool usespline=true;
    if(!pottype.empty() && pottype.at(0)=='b') usespline=false; 
    if(usespline)
      potential=new smile::CPotentialSpline(3, mbh, nrad, nang, points, symtype);
    else
      potential=new smile::CPotentialBSE(3, mbh, 0, nrad, nang, points, symtype);
    smile::writePotential(infile+".coef_"+(usespline?"spl":"bse"), potential);  // store potential coefs for later use
  }
  std::vector< smile::CPosVelDataFloat > trj;
  std::vector< float > masses;
  int numTimesteps=0;
  std::cerr << "Loading orbits"<<std::endl;
  try{
    while (io_nemo((char*)infile.c_str(),const_cast<char*>("float,n,t,x,v,m,read"),&nbody,&time,&pos,&vel,&mass)) 
    {
      currenttime=*time;
      if(numTimesteps>0 && *nbody!=Nbody) { std::cerr<<"Error, number of particles is different between snapshots at time="<<currenttime<<std::endl; return; }
      if(pos==NULL || vel==NULL) { std::cerr<<"Error, no position/velocity data in snapshot"<<std::endl; return; }
      Nbody = *nbody;
      if(trj.size()==0)
      {
        trj.resize(Nbody);
        masses.assign(Nbody,0);
        if(mass!=NULL) 
          for(unsigned int o=0; o<Nbody; o++)
            masses[o]=mass[o];
      }
      for(unsigned int o=0; o<Nbody; o++)
        trj[o].push_back(smile::CPosVelPoint<float>(pos[o*3],pos[o*3+1],pos[o*3+2],vel[o*3],vel[o*3+1],vel[o*3+2]));
      numTimesteps++;
    }
  }
  catch(const std::bad_alloc&)
  {
    std::cerr << "Error, not enough memory at time=" << currenttime << std::endl;
    return;
  }
  if(numTimesteps<=1) { std::cerr << "Error, only a single snapshot time is not enough"<<std::endl; return; }
  std::cerr << "Performing orbit analysis"<<std::endl;
  double timeStep = currenttime / (numTimesteps-1);
  try{
    smile::CConfigOrbit configOrbit;   // several parameters used for orbit integration [unused] and classification
    configOrbit.freqStrictOrder=false; // actually the only parameter needed to be initialized
    smile::COrbitLibrary* orbits=new smile::COrbitLibrary(&configOrbit, potential, NULL);
    for(unsigned int o=0; o<Nbody; o++)
    {
      smile::COrbit* orb=new smile::COrbit(&configOrbit, potential, timeStep, 0, trj[o]);
      smile::vectorInformation vi;
      for(size_t ni=0; ni<orb->getInfoNum(); ni++)
        vi.push_back(orb->getNewInfo(ni));
      orbits->addOrbitDesc(new smile::COrbitDesc(&configOrbit, orb->getInitData(), currenttime, /*maxtime*/0, /*weight*/masses[o], vi));
      delete orb;
      if(o%1000==0) { std::cerr << (o*100/Nbody) << "% done\r"; }
    }
    std::cerr << "            \r";
    smile::CFileOutputText out(outfile);
    out.writeOrbitLibrary(orbits, false);
    delete orbits;
    if(!out.ok()) std::cerr << "Error writing orbit library to "<<outfile<< std::endl;
  }
  catch(const std::bad_alloc&)
  {
    std::cerr << "Error, not enough memory" << std::endl;
    return;
  }
}

}
